#include "adaptor.h"
#include <linux/delay.h>
#include <linux/smp_lock.h>
#include <linux/vmalloc.h>
#include <linux/slab.h>
#include <linux/proc_fs.h>

MODULE_AUTHOR("CERCS - GaTech");
MODULE_DESCRIPTION("Adaptor Module");
MODULE_LICENSE("GPL");

static adaptor_node *adaptor_list_head = NULL;
static int total_adaptors = 0;

static int a_hooks_read(char *page, char **start, off_t off, int count, int *eof,
	void *data)
{
	int len = 0;
	static char my_buffer[4096];
	adaptor_node *link;

	if (off > 0) return 0;
	MOD_INC_USE_COUNT;
	len = sprintf(my_buffer, "Available adaptors:\n");
	link = adaptor_list_head;
	while (link != NULL) {
		len += sprintf(my_buffer+len, "%s: %d\n", link->name, 
		link->id);
		link = link->next;
	} 
	*start = my_buffer;
	MOD_DEC_USE_COUNT;
	return len;
}

/* prev_id = -1 means return first entry, else look for next */
int get_next_adaptor(int prev_id) 
{
	adaptor_node *link;
    
	link = adaptor_list_head;
	if (link == NULL) return -2;
	if (prev_id == -1) return link->id;
	while ((link != NULL) && (link->id != prev_id)) 
		link = link->next;
	if (link != NULL) {
		link = link->next;
		if (link == NULL) return -2;
		else return link->id;
	}
	return -2;
}

int call_adaptor_by_id(int id, int parameter1, int parameter2, 
	int parameter3,
	int parameter4)
{
	adaptor_node *link;
	int (*func)(int, int, int, int);

	link = adaptor_list_head;
	while ((link != NULL) && (link->id != id)) 
		link = link->next;

	if (link != NULL) {
		func = (int(*)()) (int) link->callback;
		return func(parameter1, parameter2, parameter3, parameter4);
	}
	return -1;
}

int call_adaptor_by_name(char *adaptor_name, int parameter1, int parameter2, 
	int parameter3,
	int parameter4)
{
	adaptor_node *link;
	int (*func)(int, int, int, int);

	link = adaptor_list_head;
	while ((link != NULL) && (strcmp(adaptor_name, link->name) != 0)) 
		link = link->next;

	if (link != NULL) {
		func = (int(*)()) (int) link->callback;
		return func(parameter1, parameter2, parameter3, parameter4);
	}
	return -1;
}

int register_adaptor(char *adaptor_name, int id, void *callback) 
{
	adaptor_node *new_link;

	new_link = (adaptor_node *)kmalloc(sizeof(adaptor_node), GFP_KERNEL);
	new_link->next = adaptor_list_head;
	adaptor_list_head = new_link;
	strncpy(new_link->name, adaptor_name, strlen(adaptor_name));
	new_link->name[strlen(adaptor_name)] = '\0';	
	new_link->callback = callback;
	new_link->id = id;
	total_adaptors++;
	return 0;
}

int unregister_adaptor(char *adaptor_name) {
	adaptor_node *link, *plink;

	plink = link = adaptor_list_head;

	while ((link != NULL) && (strcmp(adaptor_name, link->name) != 0)) {
		plink = link;
		link = link->next;
	}
	if (link != NULL) {
		if (plink != link) 
			plink->next = link->next;
		else
			adaptor_list_head = link->next;
		kfree(link);
		total_adaptors--;
	}
	return 0;
}

int init_module(void)
{
	struct proc_dir_entry *t;

	t = create_proc_read_entry("adaptors", S_IFREG | S_IRUGO, NULL,
		a_hooks_read, NULL);
	t->owner = THIS_MODULE;

	return 0;
}

void cleanup_module(void)
{
	remove_proc_entry("adaptors", NULL);
}
