#ifndef MODULE
#define MODULE
#endif
#ifndef __KERNEL__
#define __KERNEL__
#endif

#include <linux/version.h>
#include <linux/module.h> 
#if defined(MODVERSIONS)
#include <linux/modversions.h>
#endif

#include <linux/kernel.h>
#include <linux/autoconf.h>
#include <linux/mm.h>
#include <linux/swap.h>
#include <linux/sched.h>
#include <linux/file.h>
#include <linux/netdevice.h>
#include <linux/netlink.h>
#include <net/ip.h>
#include <net/sock.h>
#include <net/tcp.h>
#include <net/udp.h>
#include <net/scm.h>
#include <net/pkt_sched.h>
#include <net/protocol.h>
#include <linux/sch_dwcs.h>
#include <linux/qsocket.h>
#include <net/utils.h>
#include <net/dwcs.h>
#include "adaptor.h"

MODULE_AUTHOR("CERCS - GaTech");
MODULE_DESCRIPTION("Control Module (DWCS NET)");
MODULE_LICENSE("GPL");

/* parameter1 = pid, parameter2 = fd, parameter3 = type (0=ipg,1=own,2=owd,3=flags,
4=droppable), parameter4= value */
int dwcs_net_callback(int parameter1, int parameter2, int parameter3, int parameter4)
{
	int result = -1;
#if defined(CONFIG_NET_SCH_DWCS) || defined (CONFIG_NET_SCH_DWCS_MODULE)
	struct task_struct *task;
	struct file *fl = NULL;
	int i_ino = 0;
	struct Qdisc *qdisc;
	struct Qdisc *leaf;
	struct net_device *dev = dev_base;
	struct socket_filter *fh = NULL;
	struct socket_head *head = NULL;
	struct tcf_proto *tp = NULL;
	struct dwcs_sched_data *sch_dwcs = NULL;
	struct dwcs_class *cl, *link;

	task = find_task_by_pid(parameter1);
	if (task && (fl=fcheck_files(task->files, parameter2))) {
		if (fl->f_dentry && fl->f_dentry->d_inode) {
			i_ino = fl->f_dentry->d_inode->i_ino;
		}
	}
	if (i_ino == 0) return result;

	for (; dev != NULL; dev = dev->next) {
		if (dev != NULL) {	
			qdisc = dev->qdisc;
			if (qdisc != NULL) {
				leaf = qdisc;
				if (strcmp("dwcs", leaf->ops->id) == 0) {
					sch_dwcs = (struct dwcs_sched_data *)
						leaf->data;
					link = sch_dwcs->link;
					for (tp = link->filter_list; tp; tp=tp->next) {
						if (tp->ops) {
							if (!strcmp("socket", 
								tp->ops->kind)) {
								head = tp->root;
								if (head)
								  fh = head->ht[socket_hash(0)];
							  	if (fh == NULL) 
									continue;
								for (; fh; fh=fh->next)
								  if (fh->i_ino==i_ino)
								    goto out;
							}
						}
					}
				}
			}
		}
	}
out: 
	if (fh == NULL)
		return result;
	if ((cl = (void *) fh->res.class ) == NULL) {
		if (TC_H_MAJ (fh->res.classid))
			cl = dwcs_class_lookup(sch_dwcs, fh->res.classid);
	}

	if (cl == NULL)
		return result;

	if (parameter3 == 0) {
		cl->attr.ipg = parameter4;
		result = 0;
	} else if (parameter3 == 1) {
		cl->attr.oln = parameter4;
		result = 0;
	} else if (parameter3 == 2) {
		cl->attr.old = parameter4;
		result = 0;
	} else if (parameter3 == 3) {
		cl->attr.flags = parameter4;
		result = 0;
	} else if (parameter3 == 4) {
		cl->attr.droppable = parameter4;
		result = 0;
	}
#endif

	return result;
}

int init_module(void)
{
	register_adaptor("DWCS Net Scheduler", 3, (void *)dwcs_net_callback);
	return 0;
}

void cleanup_module(void)
{
	unregister_adaptor("DWCS Net Scheduler");
	return;
}
