// The purpose of this module is to show information which is related to 
// disk i/o monitoring, result is output throug proc interface at
// /proc/diskmon/content

#ifndef __KERNEL__
#define __KERNEL__
#endif
#ifndef MODULE
#define MODULE
#endif

//#define EXPORT_SYMTAB
#include <linux/kernel.h>
#include <linux/module.h>
#include <linux/init.h>
#include <linux/errno.h>
#include <linux/sched.h>
#include <linux/interrupt.h>
#include <linux/timer.h>
#include <linux/proc_fs.h>
#include <asm/uaccess.h>
#include <linux/kernel_stat.h>
#include "kthread.h"


#define DISKMON_PERIOD	100
#define DECAY_FACTOR	0.3

MODULE_AUTHOR("Sandip Agarwala");
MODULE_DESCRIPTION("Loadavg Module(Dproc)");
MODULE_LICENSE("GPL");

static kthread_t diskmon_thread_data;
static int disk_rate;

static int proc_calc_metrics(char *page, char **start, off_t off,
		int count, int *eof, int len){
	if (len <= off+count) *eof = 1;
	*start = page + off;
	len -= off;
	if (len>count) len = count;
	if (len<0) len = 0;
	return len;
}

int diskmon_read_proc(char *page, char **start, off_t off, int count, int *eof, void *data) 
{
	int len;
	char buf[32];
	sprintf(buf, "%d\n",disk_rate);
	len = strlen(buf)+1;
	memcpy(page, buf, len);
	return proc_calc_metrics(page, start, off, count, eof, len);
}

static void diskmon_thread(kthread_t *kthread){
	int major, disk,tmp;
	unsigned long curr_sum, last_sum;
	disk_rate = last_sum = 0;
	init_kthread(kthread, "diskmon thread");
	for (major = 0; major < DK_MAX_MAJOR; major++)
		for (disk = 0; disk < DK_MAX_DISK; disk++)
			last_sum += kstat.dk_drive_rblk[major][disk] +
				kstat.dk_drive_wblk[major][disk];
	while(1){
                interruptible_sleep_on_timeout(&kthread->queue, DISKMON_PERIOD);
		mb();
		if (kthread->terminate)
			break;
		curr_sum = 0;
		for (major = 0; major < DK_MAX_MAJOR; major++)
			for (disk = 0; disk < DK_MAX_DISK; disk++)
				curr_sum += kstat.dk_drive_rblk[major][disk] +
					kstat.dk_drive_wblk[major][disk];
		tmp = (int)(curr_sum-last_sum);
		printk(KERN_CRIT "%d %d\n",tmp,disk_rate);
		disk_rate = DECAY_FACTOR*(disk_rate - tmp)+ tmp;
		printk(KERN_CRIT "%d %d\n",tmp,disk_rate);
		last_sum = curr_sum;
	}
	exit_kthread(kthread);
}

static int start_diskmon(void)
{
    int err = -ENOMEM;
    struct proc_dir_entry *res;
    printk("Welcome diskmon\n");
    res = create_proc_read_entry("diskmon", 0, NULL, diskmon_read_proc,NULL);
    if (!res) {
        err = -EINVAL;
        return -EINVAL;
    }
    start_kthread(diskmon_thread, &diskmon_thread_data);
    return 0;
}

static void stop_diskmon(void)
{
	stop_kthread(&diskmon_thread_data);
	remove_proc_entry ("diskmon", NULL);
	printk("Exit diskmon\n");
}

module_init(start_diskmon);
module_exit(stop_diskmon);

