/*
 * dproc.h
 *
 * Copyright (C) 2001, Neil Bright
 *
 * This source is covered by the GNU GPL.  See the file COPYING.
 */

#ifndef _DPROC_H
#define _DPROC_H

#include <linux/proc_fs.h>
#include <linux/limits.h>	/* for NAME_MAX */
#include <asm/page.h>		/* for PAGE_SIZE */


#define DPROC_DATA_SIZE		PAGE_SIZE	/* Max size of a placeholder */
#define DPROC_NAME_MAX		NAME_MAX	/* # chars in a file name */


/* Exported API
 *
 * FIXTHIS?:
 *
 * These routines share the same locking semantics when called from kernel
 * space as they do when called from user space.  This _may_ mean that the kernel
 * could block if user space has the lock...  Will this cause the machine to
 * hang, or will the user space program be rescheduled?
 */

typedef void (kdproc_handler_t)(const char *buf, unsigned long length, void *node_ptr);

extern int kdproc_create_placeholder(const char *name,
				     struct proc_dir_entry *parent, 
				     kdproc_handler_t *handler, void *node_ptr);

extern int kdproc_delete_placeholder(const char *name, int length,
				     struct proc_dir_entry *parent);

extern int kdproc_create_node(const char *name, struct proc_dir_entry **ret);

extern int kdproc_delete_node(const char *name, int length);

extern int kdproc_write(int descriptor, const char *buffer, unsigned long length);

extern int kdproc_read(int descriptor, char *buffer, int length);

#endif	// _DPROC_H
