/*
 * dproc_int.h
 *
 * Copyright (C) 2001, Neil Bright
 *
 * This source is covered by the GNU GPL.  See the file COPYING.
 */

#ifndef _DPROC_INT_H
#define _DPROC_INT_H

#include <linux/init.h>
#include <linux/config.h>
#include <linux/module.h>
#include <linux/spinlock.h>
#include <asm/uaccess.h>
#include "dproc.h"


#define DPROC_ROOTDIR		"cluster"

#define DPROC_STATSLIST		"statslist"
#define DPROC_STATSLIST_MODE	S_IWUSR

#define DPROC_NODELIST		"nodelist"
#define DPROC_NODELIST_MODE	S_IWUSR

#define DPROC_PLACEHOLDER_MODE	(S_IWUSR | S_IRUGO)

#ifdef DEBUG_LOCKING
# undef read_lock
# undef write_lock
# undef read_unlock
# undef write_unlock
# define read_lock(x)		((void)x)
# define write_lock(x)		((void)x)
# define read_unlock(x)		((void)x)
# define write_unlock(x)	((void)x)
#endif /* DEBUG_LOCKING */

#ifdef DEBUG
#define dprintk(args...)        printk(KERN_INFO args)
#else
#define dprintk(args...)
#endif /* DEBUG */



#ifndef MIN
#define MIN(a, b)		(((a) < (b)) ? (a) : (b))
#endif


int proc_match(int, const char *, struct proc_dir_entry *);

static int proc_calc_metrics(char *, char **, off_t, int, int *, int);
static struct proc_dir_entry *find_child(const char *, int,
					 struct proc_dir_entry *);
static char parse_command(char *, char **, unsigned long *);

static int _list_write(const char *, unsigned long, struct proc_dir_entry *,
		       int (*)(const char *, struct proc_dir_entry*, kdproc_handler_t, void *),
		       int (*)(const char *, int, struct proc_dir_entry*),
		       char *);
static int _create_node(const char *, struct proc_dir_entry *,
			struct proc_dir_entry **);
static int _write(int, const char *, unsigned long, int);

static int statslist_write(struct file *, const char *, unsigned long, void *);
/*Not required.. sandip*/
// static int nodelist_write(struct file *, const char *, unsigned long, void *);
static int generic_write(struct file *, const char *, unsigned long, void *);
static int generic_read(char *, char **, off_t, int, int *, void *);
static int dproc_create_placeholder(const char *, struct proc_dir_entry *, kdproc_handler_t, void *);
static int dproc_delete_placeholder(const char *, int, struct proc_dir_entry *);
/*Not required.. sandip*/
//static int dproc_create_node(const char *, struct proc_dir_entry *);
static int dproc_delete_node(const char *, int, struct proc_dir_entry *);




struct dproc_placeholder {
  struct list_head hash_list;	/* hash table (currently linked list) of
				   all placeholders */
  void *buffer;			/* pre-allocated buffer of DPROC_DATA_SIZE
				   bytes */
  rwlock_t lock;		/* protects access to this placeholder */
  int descriptor;		/* implies max of 2^31 active placeholders */
  int length;			/* current length of valid data in @buffer */
  /*Following three attributes added by Sandip*/
  
  char name[20];		/*Name of the placeholder*/
  kdproc_handler_t *kdproc_handler;
  				/* Called when something is written to this 
				 * placeholder*/
  void *node_ptr;		/* passed into the kdproc_handler*/
};

#endif	// _DPROC_INT_H
