#include "monitor_thread.h"
#include <linux/delay.h>
#include <linux/smp_lock.h>
#include <linux/vmalloc.h>
// #include <asm/msr.h>  */

#undef print_progress
#define print_progress()

//#define print_progress()	printk(KERN_CRIT "In %s at %s.%d time=%ld\n",__FUNCTION__,__FILE__,__LINE__,jiffies);
MODULE_AUTHOR("Sandip Agarwala");
MODULE_DESCRIPTION("Monitor Module(Dproc)");
MODULE_LICENSE("GPL");

static PNode_Info node_list_head=NULL, node_list_tail=NULL;
static PServiceLink service_list_head=NULL, service_list_tail=NULL;
static rwlock_t servicelist_lock = RW_LOCK_UNLOCKED;
// Signify arrival of new service
static int new_service_flag = MON_OFF;
static char *service_list_packet = NULL;
static int total_services = 0;

// Used to do synchronization for interrupt from dproc
static rwlock_t kdproc_control_lock = RW_LOCK_UNLOCKED;
static Service_Control_Info kdproc_control_info;
static PNode_Info kdproc_new_control_ptr = NULL;

// Time in jiffies that monitor loop will sleep before 
// starting another iteration
static int mon_sleep_period = 3*HZ;

/* the variable that contains the thread data */
static kthread_t monitor_thread_data;

static ECSourceHandle ecsourcehandle[MAX_CHANNELS];
static CManager cmanager;

static char this_node_name[MON_MAXHOSTNAMELEN];

// Array to be returned by out_filter function to block events
// "1" means block, "0" otherwise
static int *block_event = NULL;

static local_resources *l_res = NULL;

// Array to be returned by in_filter function to block events
// "1" means block, "0" otherwise
/*We don't need this as we are doing source filtering*/
//static int *block_incoming_event = NULL;

static PNode_Info locate_node(char *nodename){
	PNode_Info ptr = node_list_head;
	while (ptr != NULL){
		if (!strcmp (ptr->node_name, nodename))
			return ptr;
		ptr = ptr->next_node;
	}
	return NULL;
}

/* Allocate a unique id to the new service that wishes to register with monitor*/
static int allocate_service_id(){
	static int count=0;
	return count++;
}

/* Return the structure of the service with ID=id */
/* static PServiceLink locate_service_byid(int id){
 	PServiceLink tmp;
	read_lock(&servicelist_lock);
	tmp = service_list_head;
	while(tmp != NULL){
		if(tmp->id == id)
			break;
		tmp = tmp->next;
	}
	read_unlock(&servicelist_lock);
	return tmp;
}*/

/* Return the structure of the service with ID=id */
static PServiceLink locate_service_byname(char *name){
	PServiceLink tmp;
	read_lock(&servicelist_lock);
	tmp = service_list_head;
	while(tmp != NULL){
		if(!strcmp(tmp->name,name))
			break;
		tmp = tmp->next;
	}
	read_unlock(&servicelist_lock);
	return tmp;
}

/* Add those symbols which are not automatically exported by the kernel*/
/* Shamelessly copied from Symbol De-Moronifier 
* By Erik Arjan Hendriks <hendriks@lanl.gov>*/

/*
static struct module_symbol *find_symbol(const char *name){
	int i;
	for (i=0; i < THIS_MODULE->nsyms; i++)
		if (strcmp(name, THIS_MODULE->syms[i].name) == 0)
			return &(THIS_MODULE->syms[i]);
	return 0;
}

static int add_symbol(const char *name, unsigned long addr){
	int i;
	char *str;
	struct module_symbol *newsyms;
	
	lock_kernel();
	if (!(str = kmalloc(strlen(name)+1, GFP_KERNEL))){
		unlock_kernel();
		return -ENOMEM;
	}
	strcpy(str, name);
	newsyms = vmalloc((THIS_MODULE->nsyms + 1) * sizeof(struct module_symbol));
	if (!newsyms) {
		kfree(str);
		unlock_kernel();
		return -ENOMEM;
	}
	if (THIS_MODULE->syms) {
		for (i=0; i < THIS_MODULE->nsyms; i++) {
			newsyms[i].value = THIS_MODULE->syms[i].value;
			newsyms[i].name  = THIS_MODULE->syms[i].name;
		}
		//vfree(THIS_MODULE->syms);
	}
	THIS_MODULE->syms = newsyms;
	
	THIS_MODULE->syms[THIS_MODULE->nsyms].value = addr;
	THIS_MODULE->syms[THIS_MODULE->nsyms].name  = str;
	THIS_MODULE->nsyms++;
	unlock_kernel();
	return 0;
}
*/

static PNode_Info register_node(char *nodename){
	int i;
	PNode_Info new_node = (PNode_Info)kmalloc(sizeof(Node_Info),GFP_KERNEL);
	
	print_progress();
	if(new_node == NULL)
		return NULL;
	memcpy (new_node->node_name, nodename, MON_MAXHOSTNAMELEN);
	for(i=0; i<MAX_SERVICES; i++)
		new_node->service_handle[i].handle = MON_UNDEFINED;
	new_node->next_node = NULL;
	print_progress();
	if(kdproc_create_node (new_node->node_name, &(new_node->node_dproc)) != 0){
		printk(KERN_CRIT "Unable to register node %s\n",new_node->node_name);
		return NULL;
	}
	print_progress();
	new_node->total_available_services = 0;
	new_node->available_service_handle = 
		kdproc_create_placeholder (AVAILABLE_SERVICE_ENTRY, new_node->node_dproc, kdproc_control_handler, new_node);
	print_progress();
	new_node->available_services_lock = RW_LOCK_UNLOCKED;
	print_progress();
	while(ecsourcehandle[DATA_CHANNEL_INDEX] == NULL){
		CMtask_sleep(HZ/2);	// Sleep for half second
		printk(KERN_CRIT "Waiting at line %d for Data Channel to be created\n",__LINE__);
	}
	if(strncmp(nodename, this_node_name, MON_MAXHOSTNAMELEN))
		new_node->block_event_index = ECget_link_index(ecsourcehandle[DATA_CHANNEL_INDEX],1,nodename,1);
	else
		new_node->block_event_index = ECget_link_index(ecsourcehandle[DATA_CHANNEL_INDEX],1,nodename,3);

	print_progress();
	if (node_list_head == NULL)
		node_list_head = node_list_tail = new_node;
	else{
		node_list_tail->next_node = new_node;
		node_list_tail = new_node;
	}
	print_progress();
	return new_node;
}

/*Called at the time of exit*/
static void clean_monitor(){
	PNode_Info ptr = node_list_head,tmp;
	while (ptr != NULL){
/*		This Part shouldn't be necessary probably 	*/
/*
		int i;
		for(i=0; i<MAX_SERVICES; i++)
			if(ptr->service_handle[i].handle != MON_UNDEFINED)
				kdproc_delete_placeholder(ptr->service_handle[i].service_name,
						SERVICE_NAME_SIZE, ptr->node_dproc);
			else
				break;
 */
		kdproc_delete_node(ptr->node_name, MON_MAXHOSTNAMELEN);
		tmp = ptr;
		ptr = ptr->next_node;
		kfree(tmp);
		tmp = NULL;
	}
	node_list_head = NULL;
}

static inline void delete_subscription(char *service_name, PNode_Info node){
	PServiceLink serv = locate_service_byname(service_name);
	PSubscribedNode subs_node = serv->subscribed_node, subs_tmp=NULL;
	if(subs_node->node == node){
		serv->subscribed_node = subs_node->next;
		kfree(subs_node);
	}
	while(subs_node != NULL){
		if(subs_node->node == node){
			subs_tmp->next = subs_node->next;
			kfree(subs_node);
			break;
		}
		else{
			subs_tmp = subs_node;
			subs_node = subs_node->next;
		}
	}
}

static inline void modify_subscription(char *query, PNode_Info node){
	PServiceLink serv;
	PSubscribedNode subs_node, subs_tail=NULL;
	char *op;
	int val, oper = 0;
	PService_Control_Info tmp = (PService_Control_Info)query;
	op = tmp->op;
	val = atoi(tmp->val);
	serv = locate_service_byname(tmp->name);
	if(serv == NULL){
		printk(KERN_WARNING "Invalid service name %s\n",tmp->name);
		return;
	}
	if(!strcmp(op,"<"))
		oper = LT;
	else if(!strcmp(op,">"))
		oper = GT;
	else if(!strcmp(op,"="))
		oper = EQ;
	else if(!strcmp(op,"<="))
		oper = LEQ;
	else if(!strcmp(op,">="))
		oper = GEQ;
	else if(!strcmp(op,"interval")){
		oper = TIME_INTERVAL;
	}
	else if(!strcmp(op,"diff_%"))
		oper = DIFF_PERCENT;
	else if(!strcmp(op,"diff_val"))
		oper = DIFF_VAL;
	if(!oper){
		printk(KERN_WARNING "Invalid operator %s\n",op);
		return;
	}
	// Code for handling floating val
/*
	{
		char *s = strchr(query,'.');
		int j,k=1;
		if(s != NULL){
			// Floating number exist
			j = atoi(s+1);
			while(k<=j)
				k*=10;
			fval = val + ((float)j)/k;
			val = -1;	// Reset val
		}
	}
*/
	subs_node = serv->subscribed_node;
	while(subs_node != NULL){
		subs_tail = subs_node;
		if(subs_node->node == node)
			break;
		else
			subs_node = subs_node->next;
	}
	if(subs_node == NULL){
		subs_node = (PSubscribedNode)kmalloc(sizeof(SubscribedNode),GFP_KERNEL);
		subs_node->node = node;
		subs_node->last_event_sent = 0;
		subs_node->last_event_sent_time = 0;
		subs_node->next = NULL;
		if(serv->subscribed_node == NULL)
			serv->subscribed_node = subs_node;
		else{
			subs_tail->next = subs_node;
		}
	}
	subs_node->op = oper;
	subs_node->limit_val = val;
}

/**
 * Update available services(entry in the dproc) from a particular 
 * subscriber at the subscriber side
 * Called when control message of type "AVAILABLE_SERVICES" is received
 * */
void update_dproc_available_services(PNode_Info pubs, char *services, int count){
	char *s;
	int i,j;
/*Format the content of "services" to be readable from /proc */
	s = services;
	for(i=0; i< count; i++){
		print_progress();
		j = strlen(pubs->available_services[i].name);
//		printk(KERN_CRIT "Value of j is %d at %d name=%s\n",j,__LINE__,pubs->available_services[i].name);
//		mdelay(5000);
		memcpy(s, pubs->available_services[i].name, j);
		memset(s+j, ' ', SERVICE_NAME_SIZE - j);
		s += SERVICE_NAME_SIZE;
		j = strlen(pubs->available_services[i].op);
//		printk(KERN_CRIT "Value of j is %d at %d op=%s\n",j,__LINE__,pubs->available_services[i].op);
//		mdelay(5000);
		memcpy(s, pubs->available_services[i].op, j);
		memset(s+j, ' ', OPERATOR_SIZE - j);
		s += OPERATOR_SIZE;
		j = strlen(pubs->available_services[i].val);
//		printk(KERN_CRIT "Value of j is %d at %d val=%s\n",j,__LINE__,pubs->available_services[i].val);
//		mdelay(5000);
		memcpy(s, pubs->available_services[i].val, j);
		memset(s+j, ' ', SERVICE_DATA_SIZE - j);
		s += SERVICE_DATA_SIZE;
	}
	*(s-2) = '\n';
	*(s-1)='\0';
	kdproc_write (pubs->available_service_handle, 
			services, count * SERVICE_CONTROL_SIZE);
}


/**
 * Update available services from a particular subscriber at the subscriber side
 * Called when control message of type "AVAILABLE_SERVICES" is received
 * */
void update_available_services(PNode_Info pubs, char *services, int count){
	int i,j=0;
	PService_Control_Info stmp;
	char *s = services;
	char *s1;
	print_progress();
//	mdelay(5000);
//	printk(KERN_CRIT "count=%d\n",count);
//	mdelay(5000);
	for(i=0; i<count; i++){
		print_progress();
		stmp = (PService_Control_Info)s;
		if(i < pubs->total_available_services){
			s1 = pubs->available_services[i].name;
			j = i+1;
			while(strcmp(stmp->name, s1)!=0){
				// This service has been discontinued.. Remove the proc entry
				kdproc_delete_placeholder(s1, strlen(s1)+1, pubs->node_dproc);
				if(j < pubs->total_available_services){
					memcpy(&(pubs->available_services[i]), &(pubs->available_services[j]), SERVICE_CONTROL_SIZE);
					s1 = pubs->available_services[i].name;
					j++;
				}
				else
					break;
			}
			if(strcmp(stmp->name, s1) == 0)
				s += SERVICE_CONTROL_SIZE;
			else
				break;
		}
		else
			break;
	}
	while(i < count){
		print_progress();
		memcpy(&(pubs->available_services[i++]), s, SERVICE_CONTROL_SIZE);
		s += SERVICE_CONTROL_SIZE;
	}
	pubs->total_available_services = count;
	print_progress();
//	mdelay(5000);
	s = (char *)kmalloc(pubs->total_available_services*SERVICE_CONTROL_SIZE, GFP_KERNEL);
	update_dproc_available_services(pubs, s, count);
	kfree(s);
	print_progress();
}

static void process_control_message(char *control_message, int length){
	PServiceHeader header = (PServiceHeader)control_message;
	PNode_Info this_node = locate_node(header->nodename);
	char *control_body = control_message + sizeof(ServiceHeader);
	char subscription_query[SERVICE_CONTROL_SIZE];
	//int i;
	print_progress();
	if(this_node ==  NULL)
		this_node = register_node(header->nodename);
	if(this_node == NULL)
		return;
	length -= sizeof(ServiceHeader);
	
	switch(header->service_data_type){
		case AVAILABLE_SERVICES:
			write_lock(&this_node->available_services_lock);
			print_progress();
//			printk(KERN_CRIT "control_msg_length = %d\n", length);
			update_available_services(this_node, control_body, 
					length / (SERVICE_CONTROL_SIZE));
			/*for(i=0; i<10000;i++)
				print_progress();*/
			write_unlock(&this_node->available_services_lock);
			print_progress();
			break;
		case SUBSCRIBE_SERVICES:
		case TUNE_SERVICES:
			while(length > 0){
				memcpy(subscription_query, control_body, SERVICE_CONTROL_SIZE);
				modify_subscription(subscription_query, this_node);
				length -= SERVICE_CONTROL_SIZE;
				control_body += SERVICE_CONTROL_SIZE;
			}
			break;
		case UNSUBSCRIBE_SERVICES:
			while(length > 0){
				delete_subscription(control_body, this_node);
				control_body += SERVICE_NAME_SIZE;
				length -= SERVICE_NAME_SIZE;
			}
			break;
		default:
			printk(KERN_WARNING "Unrecognized Control message from %s\n", header->nodename);
	}
}

int register_service(char *service_name, data_t data, 
		int data_type, PServiceLink *slink){
	PServiceLink new_link;
	PSubscribedNode subs_node;
	int id;
	read_lock(&servicelist_lock);
	new_link = service_list_head;
	while(new_link != NULL){
		if(!strcmp(service_name, new_link->name)){
			printk(KERN_CRIT "Service %s already exist\n", service_name);
			read_unlock(&servicelist_lock);
			return MON_ERR;
		}
		else
			new_link = new_link->next;
	}
	read_unlock(&servicelist_lock);
	*slink = new_link = (PServiceLink)kmalloc(sizeof(ServiceLink), GFP_KERNEL);
	id = allocate_service_id();
	strcpy(new_link->name, service_name);
	new_link->id = id;
	new_link->data = data;
	new_link->data_type = data_type;
	new_link->polling_time = mon_sleep_period;
	new_link->last_polled_time = 0;
	// Create a default subscriber
	subs_node = new_link->subscribed_node = 
		(PSubscribedNode)kmalloc(sizeof(SubscribedNode),GFP_KERNEL);
	subs_node->node = NULL;
	subs_node->op = DEFAULT_OPERATOR;
	subs_node->limit_val = DEFAULT_CONTROL_VAL;
	subs_node->last_event_sent = 0;
	subs_node->last_event_sent_time = 0;
	subs_node->next = NULL;
	
	new_link->next = NULL;
new_service_flag = MON_OFF;
	while(new_service_flag == MON_ON){
		CMtask_sleep(HZ/2);	// Sleep for half second
		print_progress();
	}
	write_lock(&servicelist_lock);
	if(service_list_head == NULL)
		service_list_head = service_list_tail = new_link;
	else{
		service_list_tail->next = new_link;
		service_list_tail = new_link;
	}
	new_service_flag = MON_ON;
	total_services++;
	if(service_list_packet == NULL){
		PService_Control_Info ctmp;
		service_list_packet = (char *)kmalloc(sizeof(ServiceHeader)+SERVICE_CONTROL_SIZE, GFP_KERNEL);
		strncpy(service_list_packet, this_node_name, MON_MAXHOSTNAMELEN);
		((PServiceHeader)service_list_packet)->service_data_type = AVAILABLE_SERVICES;
		ctmp = (PService_Control_Info)(service_list_packet+sizeof(ServiceHeader));
		strncpy(ctmp->name,service_name, SERVICE_NAME_SIZE);
		strcpy(ctmp->op, DEFAULT_OPERATOR_STR);
		strcpy(ctmp->val, DEFAULT_CONTROL_VAL_STR);	// Default.. Send event 
		print_progress();
//		printk(KERN_CRIT "name=%s, op=%s val=%s\n",ctmp->name,ctmp->op, ctmp->val);
							// every 5 seconds
	} else{
		int size = sizeof(ServiceHeader)+SERVICE_CONTROL_SIZE*total_services;
		char *tmp = (char *)kmalloc(size, GFP_KERNEL);
		PService_Control_Info ctmp;
		PServiceHeader htmp;
		htmp = (PServiceHeader)service_list_packet;
		ctmp = (PService_Control_Info)(service_list_packet+sizeof(ServiceHeader));
//		printk(KERN_CRIT "service_list_packet:nodename=%s\n",htmp->nodename);
//		printk(KERN_CRIT "name=%s, op=%s val=%s total_services=%d at line=%d\n",ctmp->name,ctmp->op, ctmp->val,total_services,__LINE__);
		size -= SERVICE_CONTROL_SIZE;
		memcpy(tmp, service_list_packet, size);
		ctmp = (PService_Control_Info)(tmp+size);
		strncpy(ctmp->name,service_name, SERVICE_NAME_SIZE);
		strcpy(ctmp->op, DEFAULT_OPERATOR_STR);
		strcpy(ctmp->val, DEFAULT_CONTROL_VAL_STR);	// Default.. Send event 
							// every 5 seconds
		kfree(service_list_packet);
		service_list_packet = tmp;
		print_progress();
//		printk(KERN_CRIT "name=%s, op=%s val=%s total_services=%d\n",ctmp->name,ctmp->op, ctmp->val,total_services);
//		mdelay(5000);
	}
	write_unlock(&servicelist_lock);
//	printk(KERN_CRIT "Service %s successfully registered\n",service_name);
	return id;
}

int deregister_service(int service_id){
	PServiceLink tmp, tmp_free;
	PSubscribedNode tmp_node, tmp_node1;
	write_lock(&servicelist_lock);
	tmp = tmp_free=service_list_head;
	if(tmp == NULL){
		printk(KERN_CRIT "No Service registered\n");
		write_unlock(&servicelist_lock);
		return MON_ERR;
	} else if(service_list_head->id == service_id){
		if(service_list_head == service_list_tail)
			service_list_head = NULL;
		else
			service_list_head = service_list_head->next;
	} else {
		while(tmp_free != NULL){
			if(tmp_free->id == service_id){
				tmp->next = tmp_free->next;
				break;
			}
			tmp = tmp_free;
			tmp_free = tmp_free->next;
		}
		if(tmp_free == service_list_tail)
			service_list_tail = tmp;
		if(tmp_free == NULL){
			printk(KERN_CRIT "Service %d not registered\n", service_id);
			write_unlock(&servicelist_lock);
			return MON_ERR;
		}
	}
	total_services--;
	tmp_node = tmp_free->subscribed_node;
	write_unlock(&servicelist_lock);
	while(tmp_node != NULL){
		tmp_node1 = tmp_node;
		tmp_node = tmp_node->next;
		kfree(tmp_node1);
	}
	kfree(tmp_free);
	return MON_OK;
}

void advertise_new_service(char *node_name){
//	if(node_name==NULL){	// Broadcast
		write_lock(&servicelist_lock);
		new_service_flag = MON_OFF;
		write_unlock(&servicelist_lock);
//		printk(KERN_CRIT "Trying to broadcast advertisement\n");
//		ECsubmit_event(ecsourcehandle[CONTROL_CHANNEL_INDEX], (void *)service_list_packet, (sizeof(ServiceHeader) + SERVICE_CONTROL_SIZE*total_services));
//		printk(KERN_CRIT "Service advertisement broadcasted from %s\n",((PServiceHeader)service_list_packet)->nodename);
//		mdelay(5000);
//	} else{
		// Selectively send the advertisement
//	}
}

void send_control_info(){
	char *control_pkt = (char *)kmalloc(sizeof(ServiceHeader)+SERVICE_CONTROL_SIZE, GFP_KERNEL);
	strncpy(control_pkt, this_node_name, MON_MAXHOSTNAMELEN);
	((PServiceHeader)control_pkt)->service_data_type = TUNE_SERVICES;
	memcpy(control_pkt+sizeof(ServiceHeader), &kdproc_control_info, SERVICE_CONTROL_SIZE);
//	ECsubmit_event(ecsourcehandle[CONTROL_CHANNEL_INDEX], (void *)control_pkt, (sizeof(ServiceHeader) + SERVICE_CONTROL_SIZE));
}

/*
static float atof(char *s){
	int i=1, j=0;
	float f;
	f = (float)atoi(s);
	s = strchr(s,'.');
	if(s!=NULL)
		j = atoi(s+1);
	while(i<=j)
		i *= 10;
	f += ((float)j/i);
	return f;
}
*/


static int monitor_gethostname(char *name, int len)
{
	if (len < 0)
		return -EINVAL;
	down_read(&uts_sem);
	memset(name,'\0',MON_MAXHOSTNAMELEN-1);
	memcpy (name, system_utsname.nodename, strlen(system_utsname.nodename));
	if ( (memchr(name, '.', MON_MAXHOSTNAMELEN) == NULL) && 
             (strcmp(system_utsname.domainname, "(none)") != 0)       ){
		strcat (name, ".");
		strcat (name, system_utsname.domainname);
	}
	up_read(&uts_sem);
	return 1;
}

/*Called by dproc when a user try to change control settings*/
void kdproc_control_handler(const char *buf, unsigned long length, void *node_ptr){
	PNode_Info remote_node = (PNode_Info)node_ptr;
	int i=0;
//	char tmp_buf[SERVICE_CONTROL_SIZE];
	while(kdproc_new_control_ptr != NULL){
//		printk(KERN_CRIT "Waiting... ");
		print_progress();
		CMtask_sleep(HZ/2);	// Sleep for half second
	}
/*	while (i < length && i < (SERVICE_CONTROL_SIZE-2) && buf[i] != '\n'){
		tmp_buf[i+1] = buffer[i];
		i++;
	}
	tmp[i] = '\0';
	*/
	write_lock(&kdproc_control_lock);
	sscanf(buf,"%s %s %s", kdproc_control_info.name, 
			kdproc_control_info.op, kdproc_control_info.val);
//	printk(KERN_CRIT "Received from dproc: name=%s op=%s val=%s\n",kdproc_control_info.name, kdproc_control_info.op, kdproc_control_info.val);
	write_lock(&(remote_node->available_services_lock));
	for(i=0; i<remote_node->total_available_services; i++){
		if(!strcmp(kdproc_control_info.name, remote_node->available_services[i].name)){
			char *tmp = (char *)kmalloc(remote_node->total_available_services*SERVICE_CONTROL_SIZE, GFP_KERNEL);
			strncpy(remote_node->available_services[i].op, 
					kdproc_control_info.op, OPERATOR_SIZE);
			strncpy(remote_node->available_services[i].val, 
					kdproc_control_info.val, SERVICE_DATA_SIZE);
			kdproc_new_control_ptr = remote_node;
			update_dproc_available_services(remote_node, tmp, remote_node->total_available_services);
			kfree(tmp);
			break;
		}
	}
	write_unlock(&(remote_node->available_services_lock));
	write_unlock(&kdproc_control_lock);
}


static void control_channel_handler(void *event,int length, void *client_data, 
		attr_list event_attrs){
	process_control_message(event,length);
}

/* Get called when data arrives in the data channel
 * Writes the data in appropriate "placeholder"
 * */
static void data_channel_handler(void *event,int length, void *client_data, 
		attr_list event_attrs){
	PServiceHeader header = (PServiceHeader)event;
	PNode_Info this_node = locate_node(header->nodename);
	char *data = event + sizeof(ServiceHeader);
	if(this_node == NULL)
		this_node = register_node(header->nodename);
	if(this_node == NULL)
		return;
	length -= sizeof(ServiceHeader);
//	printk(KERN_CRIT "line:%d Received packet of size=%d\n",__LINE__, length);
	while(length > 0){
		PDataBody databody = (PDataBody)data;
		int i;
		print_progress();
//		printk(KERN_CRIT "Received %s from %s of size=%d at %ld\n", databody->service_name, header->nodename, length, jiffies);
		for(i=0; i<MAX_SERVICES; i++){
			if(!strcmp(this_node->service_handle[i].service_name, 
				databody->service_name) && this_node->service_handle[i].handle != MON_UNDEFINED){
				kdproc_write(this_node->service_handle[i].handle, databody->service_data, MIN(SERVICE_DATA_SIZE,strlen(databody->service_data)));
				break;
			}
			if(this_node->service_handle[i].handle == MON_UNDEFINED){
			// Handle for this service has not yet been created
				print_progress();
				this_node->service_handle[i].handle = 
					kdproc_create_placeholder(databody->service_name, this_node->node_dproc, NULL, NULL);
				strncpy(this_node->service_handle[i].service_name, databody->service_name, SERVICE_NAME_SIZE);
				print_progress();
				kdproc_write(this_node->service_handle[i].handle, databody->service_data, MIN(SERVICE_DATA_SIZE,strlen(databody->service_data)));
				print_progress();
				break;
			}
		}
		length -= sizeof(DataBody);
		data   += sizeof(DataBody);
	}
	print_progress();
}

/* Filter function for outgoing events. */
int* mon_out_filter(EChannel chan, attr_list attrs, void *event){
	static int last_subscriber_count = -1;
	PServiceHeader header = (PServiceHeader)event;
	PDataBody data = (PDataBody)(event + sizeof(ServiceHeader));
	PServiceLink service = locate_service_byname(data->service_name);
	int data_type,block_flag;
	int curr_val = 0;
	int i,total_subscribers = -1;
	int last_val, limit_val;
	PSubscribedNode subs_node;
	// The constant "1" is a magic number for rem_sink_list
	// See $(KECHO_ROOT)/kecho/channel.c
	data_type = header->service_data_type;
	if(service == NULL){
//		printk(KERN_CRIT "service: %s not found in function:%s at %s:%d\n",data->service_name,__FUNCTION__,__FILE__,__LINE__);
		return NULL;
	}
	subs_node = service->subscribed_node;
//	printk(KERN_CRIT "ECget_num_members: ");
	for(i=1;i<5;i++){
		total_subscribers = MAX(total_subscribers, ECget_num_members(chan,i));
//		printk(" %d ",ECget_num_members(chan,i));
	}
//	printk("\n");
//	printk(KERN_CRIT " total_subscribers=%d\n",total_subscribers);
	if(last_subscriber_count != total_subscribers){
		last_subscriber_count = total_subscribers;
		if(block_event != NULL)
			kfree(block_event);
		block_event = (int *)kmalloc(sizeof(int)*total_subscribers, GFP_KERNEL);
	}
//#define NO_FILTER
	for(i=0; i<total_subscribers; i++){
#ifdef NO_FILTER
		block_event[i] = 0;	// All are unblocked
#else
		block_event[i] = 1;	// All are blocked at the beginning
#endif	// NO_FILTER
	}
#ifdef NO_FILTER
		return block_event;
#endif
	switch(data_type){
		case AVAILABLE_SERVICES:
			for(i=0; i<total_subscribers; i++)
				block_event[i] = 0;	// All are unblocked
			print_progress();
			break;
		case TUNE_SERVICES:
			write_lock(&kdproc_control_lock);
//			printk(KERN_CRIT "tune_services: index=%d\n", kdproc_new_control_ptr->block_event_index);
			block_event[kdproc_new_control_ptr->block_event_index] = 0;
			kdproc_new_control_ptr = NULL;
			write_unlock(&kdproc_control_lock);
			break;
		case INT_DATA:
		case FLOAT_DATA:
		print_progress();
//		printk(KERN_CRIT "data=%s\n",data->service_data);
		curr_val = atoi(data->service_data);
//		printk(KERN_CRIT "val=%d\n",curr_val);
		
		while(subs_node != NULL){
			
			block_flag = 1;
			last_val = subs_node->last_event_sent;
			limit_val = subs_node->limit_val; 
			switch(subs_node->op){
				case LT:
					if(curr_val < limit_val)
						block_flag = 0;
					break;
				case GT:
					if(curr_val > limit_val)
						block_flag = 0;
					break;
				case EQ:
					if(curr_val == limit_val)
						block_flag = 0;
					break;
				case LEQ:
					if(curr_val <= limit_val)
						block_flag = 0;
					break;
				case GEQ:
					if(curr_val >= limit_val)
						block_flag = 0;
					break;
				case DIFF_PERCENT:
					{
						int t1;
						t1 = last_val - curr_val;
						if(t1 < 0)
							t1 *= (-1);
						t1 *= 100;
						if(last_val != 0){
							t1 /= last_val;
							if(t1 > limit_val)
								block_flag = 0;
						}
						else
							block_flag = 0;
					}
					break;
				case DIFF_VAL:
					{
						int t1;
						t1 = last_val - curr_val;
						if(t1 < 0)
							t1 *= (-1);
						if(t1 > limit_val)
							block_flag = 0;
					}
					break;
				case TIME_INTERVAL:
					print_progress();
					if(jiffies - subs_node->last_event_sent_time >= limit_val)
						block_flag = 0;
/*					if(subs_node->node == NULL)
						printk("default:");
					else
						printk("%s:",subs_node->node->node_name);
					printk("%s currtime=%ld lasttime=%d interval=%d\n", data->service_name, jiffies, subs_node->last_event_sent_time, limit_val);
*/
					break;
				default:
					printk(KERN_WARNING "Unknown operator type %d\n", subs_node->op);
			}
			print_progress();
			if(subs_node->node == NULL){// Default Case
				if(!block_flag){
					for(i=0; i<total_subscribers; i++)
						block_event[i] = 0;
				}
			}
			else
				block_event[subs_node->node->block_event_index] = block_flag;
			if(!block_flag){
				subs_node->last_event_sent = curr_val;
				subs_node->last_event_sent_time = jiffies;
			}
			subs_node = subs_node->next;
		}
		break;
		
/*		case FLOAT_DATA:
		curr_fval = atof(data->service_data);
		while(subs_node != NULL){
			float last_val, limit_val;
			last_val =  subs_node->last_event_sent.float_val;
			limit_val = subs_node->limit_val.float_val;
			block_flag = 1;
			switch(subs_node->op){
				case LT:
					if(curr_fval < limit_val)
						block_flag = 0;
					break;
					break;
				case GT:
					if(curr_fval > limit_val)
						block_flag = 0;
					break;
				case EQ:
					if(curr_fval == limit_val)
						block_flag = 0;
					break;
				case LEQ:
					if(curr_fval <= limit_val)
						block_flag = 0;
					break;
				case GEQ:
					if(curr_fval >= limit_val)
						block_flag = 0;
					break;
				case DIFF_PERCENT:
					{
						float t1;
						t1 = last_val - curr_fval;
						if(t1 < 0)
							t1 *= (-1);
						t1 *= 100;
						t1 /= last_val;
						if(t1 > limit_val)
							block_flag = 0;
					}
					break;
				case DIFF_VAL:
					{
						float t1;
						t1 = last_val - curr_fval;
						if(t1 < 0)
							t1 *= (-1);
						if(t1 > limit_val)
							block_flag = 0;
					}
					break;
				case TIME_INTERVAL:
					if(jiffies - subs_node->last_event_sent_time >= limit_val)
						block_flag = 0;
					break;
				default:
					printk(KERN_WARNING "Unknown operator type %d\n", subs_node->op);
			}
			if(!block_flag){
				block_event[subs_node->node->block_event_index] = 0;
				subs_node->last_event_sent.float_val = curr_fval;
				subs_node->last_event_sent_time = jiffies;
			}
		}
		break;
		*/
		case STRING_DATA:
		// Will be implemented later
		break;
		default:
			printk(KERN_WARNING "Unknown data type %d\n", data_type);
	}
	print_progress();
	return block_event;
}

/* Filter function for incoming events. */
/*We don't need this as we are doing source filtering*/
/*int * mon_in_filter(EChannel chan, attr_list attrs, void *event){
	static int last_local_sink_list_len = -1;
	int t1 = ECget_num_members(chan, 3);	// local_sink_list_len
	int i;
	if(last_local_sink_list_len != t1){
		last_local_sink_list_len = t1;
		kfree(block_incoming_event);
		block_incoming_event = (int *)kmalloc(sizeof(int)*t1, GFP_KERNEL);
	}
	for(i=0;i<t1; i++)
		block_incoming_event[i] = 0;
	return block_incoming_event;
}
*/

static inline int init_channel(void){

	EChannel rcv_channels[MAX_CHANNELS];
//	CManager cmanager;
	EControlContext econtrolcontext;
	comm_group_return groups;
	EChannel echannel_control,echannel_data;
	char conn_id[MAXHOSTNAMELEN];
	char *temp_channel_id, *temp_group_id;
	
	cmanager = CManager_create();
	print_progress();
//	mdelay(4000);
	mon_assert(cmanager);
	econtrolcontext = ECho_CM_init (cmanager);
	mon_assert(econtrolcontext);
	print_progress();
//	mdelay(4000);
	monitor_gethostname(this_node_name, MON_MAXHOSTNAMELEN);
//	printk("Hostname: %s\n",this_node_name);
//	mdelay(4000);
	/*Registering Filter*/
	/*We don't need in_filter as we are doing source filtering*/
	ecsourcehandle[CONTROL_CHANNEL_INDEX] = NULL;
	ecsourcehandle[DATA_CHANNEL_INDEX] = NULL;
	ECregister_channel_filter(FILTER_NAME, mon_out_filter, NULL);
	print_progress();
//	mdelay(4000);
	
	groups = matching_comm_groups (CONTROL_CHANNEL_NAME, CONTROL_CHANNEL_NAME);
	print_progress();
	mon_assert(groups);
	strncpy(conn_id, this_node_name, MON_MAXHOSTNAMELEN);
//	mdelay(4000);
	if (groups->count < 1){	//Create control Channel
		echannel_control = EChannel_create_id(econtrolcontext, conn_id);
		print_progress();
//		mdelay(4000);
		mon_assert(echannel_control);
		temp_channel_id = ECglobal_id(echannel_control);
		mon_assert(temp_channel_id);
		temp_group_id = setup_comm_group(NULL, CONTROL_CHANNEL_NAME, CONTROL_CHANNEL_NAME, 
				SETUP_TIMEOUT, this_node_name, CONTROL_CHANNEL_PORT);
		print_progress();
//		mdelay(4000);
		mon_assert(temp_group_id);
		assoc_group_info(temp_group_id, temp_channel_id, strlen(temp_channel_id) + 1);
		ecsourcehandle[CONTROL_CHANNEL_INDEX] = ECsource_subscribe_filter(echannel_control,FILTER_NAME);
		print_progress();
//		mdelay(4000);
		mon_assert(ecsourcehandle[CONTROL_CHANNEL_INDEX]);
		(void) ECsink_subscribe(echannel_control, (void *)control_channel_handler, NULL);
//		printk(KERN_CRIT "Control Channel suucessfully created\n");
//		mdelay(4000);
	}else{	//Subscribe to the Control Channel
//		printk (KERN_CRIT "Control channel already exists: %s!\n", (char *)groups->list[0].info);
		rcv_channels[CONTROL_CHANNEL_INDEX] = EChannel_open_id (econtrolcontext, (char *)groups->list[0].info, conn_id);
		print_progress();
		mon_assert(rcv_channels[CONTROL_CHANNEL_INDEX]);
		ecsourcehandle[CONTROL_CHANNEL_INDEX] = ECsource_subscribe_filter (rcv_channels[CONTROL_CHANNEL_INDEX],FILTER_NAME);
		print_progress();
		mon_assert(ecsourcehandle[CONTROL_CHANNEL_INDEX]);
		(void) ECsink_subscribe (rcv_channels[CONTROL_CHANNEL_INDEX], (void *)control_channel_handler, NULL);
//		printk(KERN_CRIT "Control Channel suucessfully subscribed\n");
	}


	groups = matching_comm_groups (DATA_CHANNEL_NAME, DATA_CHANNEL_NAME);
	mon_assert(groups);
	if (groups->count < 1){	//Create data Channel
		echannel_data = EChannel_create_id(econtrolcontext, conn_id);
		mon_assert(echannel_data);
		temp_channel_id = ECglobal_id(echannel_data);
		mon_assert(temp_channel_id);
		temp_group_id = setup_comm_group(NULL, DATA_CHANNEL_NAME, DATA_CHANNEL_NAME, 
				SETUP_TIMEOUT, this_node_name, DATA_CHANNEL_PORT);
		mon_assert(temp_group_id);
		assoc_group_info(temp_group_id, temp_channel_id, strlen(temp_channel_id) + 1);
		ecsourcehandle[DATA_CHANNEL_INDEX] = ECsource_subscribe_filter(echannel_data,FILTER_NAME);
		mon_assert(ecsourcehandle[DATA_CHANNEL_INDEX]);
		(void) ECsink_subscribe(echannel_data, (void *)data_channel_handler, NULL);
//		printk(KERN_CRIT "Data Channel suucessfully created\n");
	}else{	//Subscribe to the data Channel
//		printk ("Data channel already exists: %s!\n", (char *)groups->list[0].info);
		rcv_channels[DATA_CHANNEL_INDEX] = EChannel_open_id (econtrolcontext, (char *)groups->list[0].info, conn_id);
		print_progress();
		mon_assert(rcv_channels[DATA_CHANNEL_INDEX]);
		ecsourcehandle[DATA_CHANNEL_INDEX] = ECsource_subscribe_filter (rcv_channels[DATA_CHANNEL_INDEX],FILTER_NAME);
		mon_assert(ecsourcehandle[DATA_CHANNEL_INDEX]);
		print_progress();
		(void) ECsink_subscribe (rcv_channels[DATA_CHANNEL_INDEX], (void *)data_channel_handler, NULL);
//		printk(KERN_CRIT "Data Channel successfully subscribed\n");
	}
	return MON_OK;
}

extern struct _local_resources *poll_resources(void) {
	PServiceLink serv_link;
	void *data_buf;
	int data_len;

	read_lock(&servicelist_lock);
	serv_link = service_list_head;
	if (l_res == NULL) 
		l_res = (struct _local_resources*)kmalloc(sizeof(struct _local_resources),
			GFP_KERNEL);
	while(serv_link != NULL){
		serv_link->data(&data_buf, &data_len);
		if (strcmp(serv_link->name, "loadavg") == 0) {
			l_res->loadavg = atoi(data_buf);
		} else if (strcmp(serv_link->name, "freemem") == 0) {
			l_res->freemem = atoi(data_buf);
		} else if (strcmp(serv_link->name, "netmon") == 0) {
			l_res->netmon = atoi(data_buf);
		}
		serv_link = serv_link->next;
	}
	read_unlock(&servicelist_lock);
	return l_res;
}

static void monitor_thread(kthread_t *kthread){
	PServiceLink serv_link;
	int data_len;
	void *data_buf;
	char *data_packet;
//	unsigned long long loop_time;
	/* setup the thread environment */
	init_kthread(kthread, "monitor thread");
//	mdelay(4000);
//	if(init_channel() == MON_ERR)
//		goto EXIT_KTHREAD;
//	mdelay(4000);
	data_packet = (char *)kmalloc(sizeof(ServiceHeader) + sizeof(DataBody), GFP_KERNEL);
	if(data_packet == NULL){
//		printk("Unable to handle data of size = %d.. Exiting...\n", SERVICE_DATA_SIZE);
		goto EXIT_KTHREAD;
	}
	monitor_gethostname(this_node_name, MON_MAXHOSTNAMELEN);
	strcpy(((PServiceHeader)data_packet)->nodename, this_node_name);
	while(1){
//		unsigned int start_low, start_high, end_low, end_high;
//		print_progress();
		/* fall asleep for one second */
                interruptible_sleep_on_timeout(&kthread->queue, mon_sleep_period);
		/* We need to do a memory barrier here to be sure that 
		 * the flags are visible on all CPUs. */
		mb();
		/* here we are back from sleep, either due to the timeout 
		 * (one second), or because we caught a signal.*/
		if (kthread->terminate){
			/* we received a request to terminate ourself */
			break;
		}
		/* this is monitor functions */
//		print_progress();
//		rdtsc(start_low, start_high);
//		CMpoll_network(cmanager);
//		rdtsc(end_low, end_high);
//		loop_time = ((unsigned long long)end_high<<32 | end_low)-
//			((unsigned long long)start_high<<32 | start_low);
//		printk("loop_time %Lu\n",loop_time);
		if(new_service_flag == MON_ON){	// Some interrupt came from new service
			print_progress();
			print_progress();
			advertise_new_service(NULL);
			print_progress();
		}
		if(kdproc_new_control_ptr != NULL){	// Some control info came from clusterfs
			send_control_info();
		}
		print_progress();
		read_lock(&servicelist_lock);
		serv_link = service_list_head;
		while(serv_link != NULL){
			print_progress();
			if(jiffies - serv_link->last_polled_time >= serv_link->polling_time){
				serv_link->last_polled_time = jiffies;
				((PServiceHeader)data_packet)->service_data_type = serv_link->data_type;
				serv_link->data(&data_buf, &data_len);
				strncpy(((PDataBody)(data_packet + sizeof(ServiceHeader)))->service_name, serv_link->name, SERVICE_NAME_SIZE);
				memcpy(((PDataBody)(data_packet + sizeof(ServiceHeader)))->service_data, data_buf, data_len);
//				printk(KERN_CRIT "Submitting Data: service=%s data=%s\n",serv_link->name, (char *)data_buf);
//				ECsubmit_event(ecsourcehandle[DATA_CHANNEL_INDEX], (void *)data_packet, (sizeof(ServiceHeader) + sizeof(DataBody)));
			}
			serv_link = serv_link->next;
		}
		read_unlock(&servicelist_lock);
		print_progress();
	}
	/* here we go only in case of termination of the thread */
EXIT_KTHREAD:
        /*Clean cluster fs*/
	clean_monitor();
	/* cleanup the thread, leave */
	exit_kthread(kthread);
//	printk(KERN_CRIT "Monitor thread exiting....\n");

        /* returning from the thread here calls the exit functions */
}

/* load the module */
int init_module(void)
{
/*
	struct module_symbol *sym;
	sym = find_symbol("avenrun");
	if (sym)
		//sym->value = 0xc02d4ab8;
		sym->value = 0xc02ecef8;
	else
		add_symbol("avenrun", 0xc02ecef8);
	sym = find_symbol("nr_free_pages");
	if (sym)
		//sym->value = 0xc0136a30;
		sym->value = 0xc013ba70;
	else
		add_symbol("nr_free_pages", 0xc013ba70);
*/
	/* create new kernel threads */
	start_kthread(monitor_thread, &monitor_thread_data);
	return(0);
}

/* remove the module */
void cleanup_module(void)
{
	/* terminate the kernel threads */
	stop_kthread(&monitor_thread_data);
//	printk(KERN_CRIT "Monitor module exiting...\n");
	return;
}
