#ifndef _MONITOR_THREAD_H_
#define _MONITOR_THREAD_H_

#ifndef MODULE
#define MODULE
#endif
#ifndef __KERNEL__
#define __KERNEL__
#endif

#include <linux/version.h>
#include <linux/module.h>
#if defined(MODVERSIONS)
#include <linux/modversions.h>
#endif

#include <linux/kernel.h>
#include <linux/string.h>
#include <linux/errno.h>
#include <linux/sched.h>
#include <linux/utsname.h>
#include <linux/spinlock.h>

#include "monitor.h"
#include "kthread.h"
#include "io.h"
#include "atl.h"
#include "cm.h"
#include "echo.h"
#include "comm_group2.h"
#include "dproc.h"

// setup timeout in seconds.. Used by setup_comm_group to instruct the group_server 
// to keep the group information for the SETUP_TIMEOUT amount of time
// All the subscriber must subscribe within this period or else they won't be 
// able to join the group
#define SETUP_TIMEOUT 		1800
#define FILTER_NAME 		"monitor_filter"
#define MAX_CHANNELS		2	// Data and Control
#define CONTROL_CHANNEL_NAME 	"CONTROL_CHANNEL"
#define CONTROL_CHANNEL_PORT	5013
#define DATA_CHANNEL_NAME	"DATA_CHANNEL"
#define DATA_CHANNEL_PORT	5012
// All service name should be constrained to be of length 
// smaller than SERVICE_NAME_SIZE
#define OPERATOR_SIZE		12
#define SERVICE_CONTROL_SIZE	sizeof(Service_Control_Info)

// The default filter set to the following... Event will be sent every 2 sec

#define DEFAULT_OPERATOR_STR	"interval"
#define DEFAULT_CONTROL_VAL_STR	"200"
#define DEFAULT_OPERATOR	TIME_INTERVAL
#define DEFAULT_CONTROL_VAL	200

/*
#define DEFAULT_OPERATOR_STR	"diff_%"
#define DEFAULT_CONTROL_VAL_STR	"15"
#define DEFAULT_OPERATOR	DIFF_PERCENT
#define DEFAULT_CONTROL_VAL	15
*/
// Intentionally decreased the size of MAXHOSTNAMELEN to keep size of
// data/control packets small. Make sure that the length of fully 
// qualified hostname is within MON_MAXHOSTNAMELEN
#define MON_MAXHOSTNAMELEN	32
// Maximum number of services that can be allowed to register to 
// the monitoring module
#define MAX_SERVICES		16
//#define MAX_SERVICES		16
// Name of the available service entry
// /proc/cluster/<node_name>/AVAILABLE_SERVICE_ENTRY
#define AVAILABLE_SERVICE_ENTRY	"services"

#define MON_ON			1
#define MON_OFF			0
#define MON_UNDEFINED		-1

// Different types of Control messages
enum {AVAILABLE_SERVICES = 1, SUBSCRIBE_SERVICES, 
	TUNE_SERVICES, UNSUBSCRIBE_SERVICES};

enum {CONTROL_CHANNEL_INDEX, DATA_CHANNEL_INDEX};

// Different types of operator in subscription query
enum {LT = 1, GT, EQ, LEQ, GEQ, DIFF_PERCENT, DIFF_VAL, TIME_INTERVAL};

typedef struct _service{
	char service_name[SERVICE_NAME_SIZE];
	int handle;
}ServiceHandle, *PServiceHandle;

typedef struct _service_header{
	char nodename[MON_MAXHOSTNAMELEN];
	int service_data_type;
}ServiceHeader, *PServiceHeader;

typedef struct _data_body{
	char service_name[SERVICE_NAME_SIZE];
	char service_data[SERVICE_DATA_SIZE];
} DataBody, *PDataBody;

typedef struct _service_control_info{
	char name[SERVICE_NAME_SIZE];
	char op[OPERATOR_SIZE];
	char val[SERVICE_DATA_SIZE];
} Service_Control_Info, *PService_Control_Info;

typedef struct _node_info{
	char node_name [MON_MAXHOSTNAMELEN];
	struct proc_dir_entry *node_dproc;
	int total_available_services;
	Service_Control_Info available_services[MAX_SERVICES];
	rwlock_t available_services_lock;
	int available_service_handle;
	ServiceHandle service_handle[MAX_SERVICES];
	int block_event_index;
	struct _node_info *next_node;
}Node_Info, *PNode_Info;

typedef struct _local_resources {
	int loadavg;
	int freemem;
	int netmon;
} local_resources, *plocal_resources;

/*
union event_type{
		int int_val;
		float float_val;
	};
*/

typedef struct _subscribed_node{
	PNode_Info node;
	int op;
	int limit_val;
	int last_event_sent;
	int last_event_sent_time;
	struct _subscribed_node *next;
}SubscribedNode, *PSubscribedNode;

typedef struct _service_link{
	char name[SERVICE_NAME_SIZE];	/* Name of the service */
	int id;		/* Unique number identifying the module. Returned
			   to the module after registration */
	data_t data;	/* Function to be called to access data from 
			   service module */
	int data_type;
	int polling_time; /* in jiffies */
	int last_polled_time;	/* in jiffies */
	PSubscribedNode subscribed_node;
	struct _service_link *next;   /* link to next module */
}ServiceLink;


kdproc_handler_t kdproc_control_handler;
#ifndef MIN
#define MIN(a,b) (((a) < (b)) ? (a) : (b))
#endif

#ifndef MAX
#define MAX(a,b) (((a) > (b)) ? (a) : (b))
#endif

#endif	// _MONITOR_THREAD_H_
