#ifndef __KERNEL__
#define __KERNEL__
#endif
#ifndef MODULE
#define MODULE
#endif

#include <linux/version.h>
#include <linux/module.h> 
#if defined(MODVERSIONS)
#include <linux/modversions.h>
#endif

#include <linux/kernel.h>
#include <linux/proc_fs.h>
#include <linux/errno.h>
#include <linux/netdevice.h>
#include "kthread.h"
#include "monitor.h"

//#define print_progress()	printk(KERN_CRIT "In %s at %s.%d time=%ld\n",__FUNCTION__,__FILE__,__LINE__,jiffies);

#define NETMON_PERIOD	100
#define DECAY_FACTOR	0.3
#define NETMON_DATA_SIZE	SERVICE_DATA_SIZE
//#define NETMON_DATA_SIZE	16

MODULE_AUTHOR("Sandip Agarwala");
MODULE_DESCRIPTION("Loadavg Module(Dproc)");
MODULE_LICENSE("GPL");

static kthread_t netmon_thread_data;
static int netmon_bw;
static char netmon_string[NETMON_DATA_SIZE];
static int netmon_service_id = -1;
static PServiceLink netmon_link;

static int proc_calc_metrics(char *page, char **start, off_t off,
		int count, int *eof, int len){
	if (len <= off+count) *eof = 1;
	*start = page + off;
	len -= off;
	if (len>count) len = count;
	if (len<0) len = 0;
	return len;
}

int netmon_read_proc(char *page, char **start, off_t off, int count, int *eof, void *data){
	int len;
	char buf[32];
	sprintf(buf, "%d\n",((netmon_bw/1000)*(8*HZ))/NETMON_PERIOD);
	len = strlen(buf)+1;
	memcpy(page, buf, len);
	return proc_calc_metrics(page, start, off, count, eof, len);
}

int netmon_data(void **data_buf,int *data_len){
	sprintf(netmon_string,"%d\n",((netmon_bw/1000)*(8*HZ))/NETMON_PERIOD);
	*data_buf = (void *)netmon_string;
	*data_len = NETMON_DATA_SIZE;
	return MON_OK;
}

static void netmon_thread(kthread_t *kthread){
	struct net_device_stats *stats;
	struct net_device *dev;
	int curr_val, last_val=0;
	init_kthread(kthread, "netmon thread");
	netmon_bw = 0;
	for (dev = dev_base; dev != NULL; dev = dev->next) {
		if(!strcmp(dev->name, "eth0")){
			stats = (dev->get_stats ? dev->get_stats(dev): NULL);
			// Currently not handling overflow!! -sandip
			last_val = stats->tx_bytes + stats->rx_bytes;
		}
	}
	netmon_service_id = register_service("netmon", netmon_data, INT_DATA , &netmon_link);
	while(1){
                interruptible_sleep_on_timeout(&kthread->queue, NETMON_PERIOD);
		mb();
		if (kthread->terminate)
			break;
//		print_progress();
		read_lock(&dev_base_lock);
		for (dev = dev_base; dev != NULL; dev = dev->next) {
			if(!strcmp(dev->name, "eth0")){
				int diff;
				stats = (dev->get_stats ? dev->get_stats(dev): NULL);
				curr_val = stats->tx_bytes + stats->rx_bytes;
				diff = curr_val - last_val;
				netmon_bw = DECAY_FACTOR*(netmon_bw - diff) + diff;
				last_val = curr_val;
			}
		}
		read_unlock(&dev_base_lock);
	}
	exit_kthread(kthread);
}

/* load the module */
int init_module(void)
{
	if(!create_proc_read_entry("netmon", 0, NULL, netmon_read_proc,NULL))
		return -EINVAL;
	start_kthread(netmon_thread, &netmon_thread_data);
	return 0;
}

/* remove the module */
void cleanup_module(void)
{
	deregister_service(netmon_service_id);
	stop_kthread(&netmon_thread_data);
	remove_proc_entry ("netmon", NULL);
//	printk("Exit netmon\n");
}
