#ifndef MODULE
#define MODULE
#endif
#ifndef __KERNEL__
#define __KERNEL__
#endif

#include <linux/version.h>
#include <linux/module.h> 
#if defined(MODVERSIONS)
#include <linux/modversions.h>
#endif

#include <linux/sched.h>
#include <linux/kernel.h> 
#include <linux/delay.h>
#include <asm/xscale-pmu.h>
#include "kthread.h"
#include "power.h"
/* #include "80200_mod.h"  */

#define print_progress()	printk(KERN_CRIT "In %s at %s.%d\n",__FUNCTION__,__FILE__,__LINE__);

MODULE_AUTHOR("Hailemelekot Seifu");
MODULE_DESCRIPTION("Power Monitoring Module(Dproc)");
MODULE_LICENSE("GPL");

kthread_t power_thread_data;
char power_monitor_string[10];
unsigned int current_session = 0;
unsigned long pmu_claim_id = 0;
unsigned int cnt_sessions = 5;
unsigned int perf_cntrs[MAX_COUNT_SESSIONS*2] = {7, 7, 7, 7, 7, 7, 7, 7, 7, 7};
unsigned long long power_estimate = 0;
static PServiceLink power_servicelink;
static int power_serviceid = -1;

MODULE_PARM(cnt_sessions, "i");
MODULE_PARM(perf_cntrs, "0-10i");


int data(void **data_buf,int *data_len){
	int r = sprintf(power_monitor_string,"%llu\n",power_estimate);
	power_monitor_string[r] = '\0';
	*data_buf=(void *)power_monitor_string;
	*data_len=r;
	return MON_OK;
}

static int calc_power_stats(void) {
	unsigned int i = 0, ret = 0;
	struct pmu_results cntr_results;

	ret = pmu_stop(&cntr_results);
	if ( ret < 0 )
		printk(KERN_ERR "pmu_module:  pmu_stop error = %d\n", ret);

	if ( ++current_session < cnt_sessions ) {
		pmu_start(perf_cntrs[2*current_session], perf_cntrs[(2*current_session)+1]);
		return 1;
	}

	/* TO DO:  calculate power from usage counts */;
	power_estimate = (cntr_results.pmn0_of*(~i)) + cntr_results.pmn0;

	current_session = 0;
	pmu_start(perf_cntrs[2*current_session], perf_cntrs[(2*current_session)+1]);

        return 1; 
}


void power_thread(kthread_t *kthread)
{
	init_kthread(kthread, "power monitoring thread");
	power_serviceid = register_service("power", data, POWER_DATA_TYPE, &power_servicelink);
	while(1){
                interruptible_sleep_on_timeout(&kthread->queue, POWER_FREQ);
                mb();
                if (kthread->terminate) 
                        break;    
		calc_power_stats();
	}
        exit_kthread(kthread);
	return;
}


/* load the module */
int init_module(void)
{
	/* create new kernel threads */
	pmu_claim_id = pmu_claim();
	if ( pmu_claim_id < 0 ) {
		printk(KERN_ERR "unable to claim PMU: %lu\n", pmu_claim_id);
		return pmu_claim_id;	
	}	
	start_kthread(power_thread, &power_thread_data);
	return(0);
}

/* remove the module */
void cleanup_module(void)
{
	int ret = 0;
		
	/* terminate the kernel threads */
	deregister_service(power_serviceid);
	stop_kthread(&power_thread_data);

	ret = pmu_release(pmu_claim_id);
	if ( ret < 0 )
		printk(KERN_ERR "unable to release PMU:  %d\n", ret);

	printk(KERN_CRIT "power monitoring module exiting.. \n");
	return;
}
