#include "imgHandler.h"

/*------------------------------------------------------------------------------*/

/*
 * edgeImage: Edge-detection on a grey one, remember to convert it first
 *
 * Parameters: inImage in gray
 *
 * Returns: outImage of edge detection
 *
 */
int 
edgeImage (Raw_data* inImage, Raw_data* outImage)
{
  int col, row, indx, outIndx, row0Indx, row1Indx, row2Indx, tmprowIndx;
  double sum1, sum2, sum, abssum1, abssum2;

  if (inImage->ppm2 != '5') return 0;		/* RPGM_MAGIC2 */
  outImage->ppm1 = inImage->ppm1;
  outImage->ppm2 = inImage->ppm2;
  outImage->width = inImage->width;
  outImage->height = inImage->height;
  outImage->maxval = 255;

  /* Get the first two rows. */
  row0Indx = 0;
  row1Indx = inImage->width;

  
  /* Write out the first row, all zeros. */
  for (col = 0; col < inImage->width; ++col )
    outImage->buff[col] = 0;

  /* Now the rest of the image -- read in the next row, and write
  ** write out the current row.
  */
  for (row = 1; row < (inImage->height - 1); row++) {
    indx = (row - 1) * inImage->width;
    outIndx = (row - 1) * inImage->width;
    row2Indx = indx;

    for ( col = 1; col < inImage->width - 1; col++ ) {
      sum1 = (double) inImage->buff[row0Indx+col+1] - 
	(double) inImage->buff[row0Indx+col-1] +
	2.0 * ( (double) inImage->buff[row1Indx+col+1] - 
		(double) inImage->buff[row1Indx+col-1] ) +
	(double) inImage->buff[row2Indx+col+1] - 
	(double) inImage->buff[row2Indx+col-1];

      sum2 = ( (double) inImage->buff[row2Indx+col-1] + 
	       2.0 * (double) inImage->buff[row2Indx+col] +
	       (double) inImage->buff[row2Indx+col+1] ) -
	( (double) inImage->buff[row0Indx+col-1] + 
	  2.0 * (double) inImage->buff[row0Indx+col] +
	  (double) inImage->buff[row0Indx+col+1] );

      /* 
       * The following 8 lines of code are just to avoid using sqrt since
       * we cannot 'move' it into E-code.
       * Now, "why 0.2 of the smallest one?" you would ask: because I say
       * so.
       *
       * sum = sqrt( sum1 * sum1 + sum2 * sum2 ); 
       *
       */

      if (sum1 < 0.0) abssum1 = -sum1;
      else abssum1 = sum1;
      if (sum2 < 0.0) abssum2 = -sum2;
      else abssum2 = sum2;
      
      if (abssum1 == 0.0) sum = abssum2;
      if (abssum2 == 0.0) sum = abssum1;
      if (abssum1 > abssum2) sum = abssum1 + 0.2 * abssum2;
      else sum = 0.2 * abssum1 + abssum2;

      sum /= 1.8;		/* arbitrary scaling factor */
      if ( sum > inImage->maxval ) sum = inImage->maxval;
      outImage->buff[outIndx + col] = sum;
    }
    
    /* Rotate rows. */
    tmprowIndx = row0Indx;
    row0Indx = row1Indx;
    row1Indx = row2Indx;
    row2Indx = tmprowIndx;
  }
  
  /* And write the last row, zeros again. */
  outIndx = (inImage->height - 2) * inImage->width;
  for (col = 0; col < inImage->width; ++col)
    outImage->buff[outIndx + col] = 0;
  
  return 0;
} /* end edgeImage */
