#ifndef IMGHANDLER_H
#define IMGHANDLER_H 1

#include <stdlib.h>
#include <stdio.h>
#include <sys/types.h>

#define SERVER_PORT 5555
#define MAX_ACTIONS 10

// Filter constants
#define AND_TYPE 1
#define OR_TYPE  2


/* 
 * PPM stuff
 *
 */

typedef unsigned char pixval;

typedef struct {
  pixval r, g, b;
} pixel;

#define PPM_OVERALLMAXVAL 65535
#define PPM_MAXMAXVAL 255

#define PPM_MAGIC1  'P'
#define PPM_MAGIC2  '3'
#define RPGM_MAGIC2 '5' 
#define RPPM_MAGIC2 '6' 
#define PPM_FORMAT (PPM_MAGIC1 * 256 + PPM_MAGIC2)
#define RPPM_FORMAT (PPM_MAGIC1 * 256 + RPPM_MAGIC2)
#define PPM_TYPE PPM_FORMAT

#define PPM_ASSIGN(p,red,grn,blu) \
   do { (p).r = (red); (p).g = (grn); (p).b = (blu); } while ( 0 )

#define PPM_LUMIN(r,g,b) ( 0.299 * r + 0.587 * g + 0.114 * b )

#define IMGSIZECOLOR(width, height) width * height * 3 * sizeof(u_char); 
#define IMGSIZEGREY(width, height) width * height * sizeof(u_char); 

typedef struct {
  int tag;
  char ppm1;
  char ppm2;
  int maxval;
  int size;
  int width;
  int height;
  char *buff; 
} Raw_data, *Raw_data_ptr;

struct ppm_request {
	struct ppm_request *Next;
	int sockfd;
	struct file *filp;
	int FileLength;
	char FileName[256];
	int FileNameLength;
	int action[MAX_ACTIONS];
	int parameter1[MAX_ACTIONS];
	int parameter2[MAX_ACTIONS];
};

// static struct ppm_request *AcceptQueue;

struct stream_request {
	int filter_id;
	int rate;
	int loop;
	char FileName1[256];	
	char FileName2[256];	
	int FileNameLength1;
	int FileNameLength2;
};

struct stream_response {
	int len;
};

struct filter_packet_t {
	int type;
	char data[32];
};

extern int 
cropImage (Raw_data* inImage, Raw_data* outImage, int inRow, int inCol, 
	   int outWidth, int outHeight);

extern int 
catImage (Raw_data* inImage1, Raw_data* inImage2, Raw_data* outImage, int order);

extern int 
edgeImage (Raw_data* inImage, Raw_data* outImage);

extern int 
greyImage (Raw_data* inImage, Raw_data* outImage);

extern int 
mixImage (Raw_data*inImage1, Raw_data*inImage2, Raw_data*outImage, float fade);

extern int 
enlargeImage (Raw_data* inImage, Raw_data* outImage, int n);

extern int 
reduceImage (Raw_data* inImage, Raw_data* outImage, int n);

#endif /* !IMGHANDLER_H */
