/*******************************************************************************
*
* File:         imgServer.c
* RCS:          $Id: $
* Description:  Test for image handlers that will become streamlets
* Authors:      original: Fabian E. Bustamante
*               socket interface: Christian Poellabauer
*               Systems Research Group
*               College of Computing
*               Georgia Institute of Technology
* Language:     C
* Package:      N/A
* Status:       Experimental (Do Not Distribute)
*
* (C) Copyright 2001, Georgia Institute of Technology, all rights reserved.
*
*******************************************************************************
*/

#ifndef KECHO
#define KECHO
#endif

#include <math.h>
#include <stdlib.h>
#include <stdio.h>
#include <time.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <unistd.h>
#include <sys/time.h>
#include <sys/socket.h>
#include <netinet/in.h>
#include <netdb.h>
#include <string.h>
#include <sched.h>
#include <asm/param.h>
#include <linux/qfabric.h>
#include <linux/ecalls.h>

#include "imgHandler.h"
#include "bitio.h"
#include "errhand.h"
#include "main.h"
#include "../../../kecho/include/io.h"
#include "../../../kecho/include/atl.h"
#include "../../../kecho/include/cm.h"
#include "../../../kecho/include/echo.h"
#include "../../../kecho/include/comm_group2.h"

#define CLIENT_CROP    1
#define CLIENT_GREY    2
#define CLIENT_EDGE    3
#define CLIENT_ENLARGE 4
#define CLIENT_REDUCE  5
#define SERVER_CROP    6
#define SERVER_GREY    7
#define SERVER_EDGE    8
#define SERVER_ENLARGE 9
#define SERVER_MIX     10
#define SERVER_CONCAT  11
#define SERVER_REDUCE  12
#define HUFFMAN        13
#define LZ77           14

#define APP_ID		"PPMSTREAM"
#define GROUP_ID	"VIDEO"

//#define DEBUG 1

//struct ecall_down downcall;
//struct ecall_up upcall;

void handler(void *, int, void *, attr_list);

struct qfabric_event {
	int type;
	char hostname[MAXHOSTNAMELEN];
	int runqueuelen;
	int loadavg;
	int freemem;
	int netmon;
	int uinfo1;
	int uinfo2;
	int uinfo3;
};

typedef struct node_t {
	struct node_t *next;
	void *data;
} Node, *PNode;

typedef struct list_t {
	PNode head;
	PNode tail;
	int size;
} List, *PList;

//static char* program_name;
int first_request = 0;

int service_request(int sfd);

/*----------------------------------------------------------------------------*/

static PList init_list (void)
{
	PList new_list;

	if ((new_list = malloc (sizeof (List))) == NULL) {
		fprintf (stderr, "List allocation error!\n");
		exit(1);
	}
	new_list->head = new_list->tail = NULL;
	new_list->size = 0;

	return new_list;
}


static void add_to_list (PList list, void *data, int size)
{
	PNode new_node;

	if ((new_node = malloc (sizeof (Node))) == NULL) {
		fprintf (stderr, "Filter node allocation error!\n");
		exit (1);
	}

	new_node->next = NULL;
	if ((new_node->data = malloc (sizeof (char) * size)) == NULL) {
		fprintf (stderr, "Filter data allocation error!\n");
		exit(1);
	}
	memcpy (new_node->data, data, size);

	if (list->head == NULL) {
		list->head = list->tail = new_node;
	}
	else {
		list->tail->next = new_node;
		list->tail = new_node;
	}
	list->size++;
}


static void *get_list_node (PList list, int position)
{
	int i = 0;
	PNode helper = list->head;
	
	while (i < list->size && i != position) {
		helper = helper->next;
		i++;
	}
	if (helper != NULL) {
		return helper->data;
	}
	else {
		return NULL;
	}
}


static void free_list_helper (PNode node)
{
	if (node->next != NULL) {
		free_list_helper (node->next);
		free (node);
	}
}


static void free_list (PList list)
{
	free_list_helper (list->head);
	free (list);
}


/*------------------------------------------------------------------------------*/


static void receive_filters (PList list, int socket)
{
	int bytes_read;
	struct stream_response server_response;
	struct stream_response filter_length;
	struct filter_packet_t filter_data;

	#ifdef DEBUG
		fprintf (stdout, "Reading from socket: %d\n", socket);
	#endif
	recv (socket, &filter_length, sizeof(struct stream_response), 0);
	
	server_response.len = 11223344;
	server_response.len = htonl (server_response.len);
	while ((int) ntohl(filter_length.len) > 0) {	
		#ifdef DEBUG
			fprintf (stdout, "Filter Length: %d\n", ntohl (filter_length.len));
			fflush (stdout);
		#endif
//		send (socket, &server_response, sizeof (struct stream_response), 0);
		bytes_read = recv (socket, &filter_data, sizeof (struct filter_packet_t), 0);
		add_to_list (list, (void *) (filter_data.data), ntohl (filter_length.len));
		#ifdef DEBUG
			fprintf (stdout, "Filter AND/OR: %d\n", ntohl (filter_data.type));
			fprintf (stdout, "Adding Filter: %s\n", filter_data.data);
			fflush (stdout);
		#endif
//		send (socket, &server_response, sizeof (struct stream_response), 0);
		recv (socket, &filter_length, sizeof(struct stream_response), 0);
		
		#ifdef DEBUG
			fprintf (stdout, "Entry completely received!\n");
			fflush (stdout);
		#endif
	}
}



/*------------------------------------------------------------------------------*/

int
main (int argc, char* argv[])
{
  int listenfd;
  struct sockaddr_in server;
  struct sockaddr_in client;
  int clientlen;
  int clientsock;
  EChannel chan;
  CManager cm;
  EControlContext cc;
  ECSinkHandle handle;
  char *channel_id;
  char *group_id;
  char *thishost;
  comm_group_return groups;

  thishost = malloc(MAXHOSTNAMELEN);
  gethostname(thishost, MAXHOSTNAMELEN);
  printf("myname=%s\n", thishost);

  /* sample download code */
  /*
  char code_string[] = "\
	{\n\
		return 1;\n\
	}";
  */

  /* Create listening socket */
  listenfd = socket(PF_INET, SOCK_STREAM, 0);
  if (listenfd == -1) {
    perror("Failed opening socket.\n");    
    exit(1);
  }
  bzero((char *)&server, sizeof(server));
  server.sin_family = PF_INET;
  server.sin_addr.s_addr = INADDR_ANY;
  server.sin_port = htons(SERVER_PORT);
  bind(listenfd, (struct sockaddr *)&server, sizeof(server));
  
  listen(listenfd, 30);

  fcntl(listenfd, F_SETFL, O_NONBLOCK);

  //ECalls_start(0);

/*
  printf("Subscribe to channel...\n");
  cm = CManager_create();
  printf("cm ok\n");
  cc = ECho_CM_init(cm);
  printf("init ok\n");
  chan = EChannel_create(cc);
  printf("create ok\n");
  channel_id = ECglobal_id(chan);
  group_id = setup_comm_group(NULL, APP_ID, APP_ID, 360, thishost, 5050);
  printf("setup ok\n");
  assoc_group_info(group_id, channel_id, strlen(channel_id)+1);
  printf("assoc ok\n");
  handle = ECsink_subscribe(chan, handler, NULL);
  printf("Subscribed to channel!\n");
*/

  printf("Call qos_configure!\n");
  qos_configure(APP_ID, GROUP_ID);

  sleep(10);

  printf("Continue...\n");
  printf("Subscribe to channel...\n");
  cm = CManager_create();
  printf("cm ok\n");
  cc = ECho_CM_init(cm);
  printf("init ok\n");
  groups = matching_comm_groups(APP_ID, APP_ID);
  printf("setup ok\n");
  chan = EChannel_open(cc, (char *)groups->list[0].info);
  printf("assoc ok\n");
  handle = ECsink_subscribe(chan, handler, NULL);


  /*downcall.size1 = 0;
  downcall.size2 = 0;
  if (ECalls_subscribe("QoS", &downcall, NULL, &upcall, NULL) < 0) {
	printf("ECalls Error!\n");
	return -1;
  }*/

  //ECalls_download("q_com", APP_ID, 0, 0, code_string);
 
  /* Loop forever */
  while (1) {
    clientsock = accept(listenfd, (struct sockaddr *) &client, 
	(socklen_t *) &clientlen);
    if (clientsock < 0) {
	sleep(1);
printf("poll...\n");
	CMpoll_network(cm);
printf("poll over...\n");
    } else {
        if (fork() == 0) {
	    service_request(clientsock);
	    exit(0);
	}
printf("poll2...\n");
	CMpoll_network(cm);
printf("poll over2...\n");
    }
  }
  close(listenfd);
}

void handler(void *event, int length, void *client_data, attr_list event_attrs)
{
  struct qfabric_event qos_event;

  memcpy((void *)&qos_event, event, length);
  printf("%d bytes received from %s!!!\n", length, qos_event.hostname);
//  printf("Event Received!!!\n");
}

int service_request(int sfd) 
{
  int i;
  int compressed;
  int header_len;
  struct stream_request request;
  int filterid;
  int inCol = 0; 
  int inRow = 0; 
  int outWidth = 100; 
  int outHeight = 100;
  int enlargeFactor = 1;
  int reduceFactor = 1;
  float mixFactor = 0.5;
  int catOrder = 1;
  int nread;
  struct stream_response response;
  char header[20];
  Raw_data inImage1, inImage2, outImage, tmpImage;
  int inImagefd1, inImagefd2;
  int maxppmval;
  struct timeval t1, t2, t3;
  int loop = 1;
  int rate = 10;
  long sleeptime;
  long period;
  struct timespec time1, time2;
  pid_t my_pid;
  char filename1[20];
  char filename2[20];
  int fd_in;
  FILE *fd_in2;
  BIT_FILE *fd_out;
  struct stat stat_buf;
  PList filter_list;
  struct sched_param *params;
  int spid = getpid();

  params = malloc(sizeof(struct sched_param));
  params->sched_priority = 50;
  sched_setscheduler(0, SCHED_RR, params);

  filter_list = init_list();
  receive_filters (filter_list, sfd);
  
  if (read(sfd, &request, sizeof(struct stream_request)) <=0) {
    free(params);
    exit(1);
  }
  filterid = ntohl (request.filter_id);
#ifdef DEBUG
  printf("Filter ID: %d\n", filterid);
#endif
  loop = ntohl (request.loop);
  rate = ntohl (request.rate);
  period = 1000000/rate;

  write(sfd, &spid, sizeof(int));

  gettimeofday(&t1, NULL);
  while (loop >= 0) {
    char *send_buffer;
    struct stream_response ack;
	
    loop--;
#ifdef DEBUG
  printf("Loop=%d\n", loop);
#endif
    if (loop < 0) {
#ifdef DEBUG
  printf("Write EOF\n");
#endif
  	response.len = -999;
        response.len = htonl (response.len);	
	write(sfd, &response, sizeof(struct stream_response));
	close(sfd);
        free(params);
	exit(0); 
    }

    /* Open files */
    if (ntohl (request.FileNameLength1) > 0) 
#ifdef DEBUG
      printf("Open file: %s.\n", request.FileName1);
#endif
      if ((inImagefd1 = open(request.FileName1, O_RDONLY)) < 0) {
        perror("Failed opening file 1.\n");    
        free(params);
        exit(1);
      }

    if (ntohl (request.FileNameLength2) > 0) {
      if ((inImagefd2 = open(request.FileName2, O_RDONLY)) < 0) {
        perror("Failed opening file 2.\n");    
        free(params);
        exit(1);
      }
    } else inImagefd2 = -1;
    nread = read(inImagefd1, header, sizeof(header));   
#ifdef DEBUG
      printf("Read %d bytes.\n", nread);
#endif
if (nread < sizeof(header)) {
#ifdef DEBUG
        printf("EOF.\n");
#endif
	response.len = -999;
        response.len = htonl (response.len);	
	write(sfd, &response, sizeof(struct stream_response));
	close(sfd);
        free(params);
	exit(0); 
    }
    sscanf(header, "%c%c\n%d %d\n%d\n", 
	 &(inImage1.ppm1), &(inImage1.ppm2),
	 &(inImage1.width), &(inImage1.height), 
	 &maxppmval);

 // Cross-platform support added by RP
//      inImage2.width = ntohl (inImage2.width);
//      inImage2.height = ntohl (inImage2.height);
//      inImage1.maxval = ntohl (maxppmval);
      
    inImage1.size = inImage1.width * inImage1.height * 3 * sizeof(u_char);
    inImage1.buff = (u_char*) malloc(inImage1.size);
    if (inImage1.buff == NULL) {
      perror("Out of mem\n");
      free(params);
      exit(1);
    }
    nread = read(inImagefd1, inImage1.buff, inImage1.size);
#ifdef DEBUG
    printf("Read %d bytes.\n", nread);
#endif
    if (nread < inImage1.size) exit(0);

    if (inImagefd2 != -1) {
      nread = read(inImagefd2, header, sizeof(header));   
      if (nread < sizeof(header)) {
	      fprintf (stderr, "Error! Misread header info!\n");
	      response.len = -999;
              response.len = htonl (response.len);	
	      write(sfd, &response, sizeof(struct stream_response));
      	      free(params);
	      exit(0); 
      }
      sscanf(header, "%c%c\n%d %d\n%d\n", 
	 &(inImage2.ppm1), &(inImage2.ppm2),
	 &(inImage2.width), &(inImage2.height), 
	 &maxppmval);


 // Cross-platform support added by RP
//      inImage2.width = ntohl (inImage2.width);
//      inImage2.height = ntohl (inImage2.height);

      
      inImage2.size = inImage2.width * inImage2.height * 
        3 * sizeof(u_char);
      inImage2.buff = (u_char*) malloc(inImage2.size);
      if (inImage2.buff == NULL) {
        perror("Out of mem\n");
      	free(params);
        exit(1);
      }
      nread = read(inImagefd2, inImage2.buff, inImage2.size);
      if (nread < inImage2.size) exit(0);
    }


    compressed = 0;
    outImage.buff = NULL;
    for (i = 0; i < filter_list->size; i++) {
	    int filter_type;
	    char filter_params[32];
	    char *filter_info;

	    filter_info = (char *) get_list_node (filter_list, i);
	    sscanf (filter_info, "%d:%s", &filter_type, filter_params);
	    #ifdef DEBUG
	    	fprintf (stdout, "Filter Type: %d\nList Size: %d\n", filter_type, filter_list->size);
	    	fflush (stdout);
	    #endif

	    switch (filter_type) {
		    case CLIENT_CROP    : 
#ifdef DEBUG
			printf ("In CLIENT_CROP\n"); 
#endif
			break;
		    case CLIENT_GREY    : 
#ifdef DEBUG
			printf ("In CLIENT_GREY\n");
#endif
			break;
		    case CLIENT_EDGE    : 
#ifdef DEBUG
			printf ("In CLIENT_EDGE\n");
#endif
			break;
		    case CLIENT_ENLARGE : 
#ifdef DEBUG
			printf ("In CLIENT_ENLARGE\n");
#endif
			break;
		    case CLIENT_REDUCE  : 
#ifdef DEBUG
			printf ("In CLIENT_REDUCE\n");
#endif
			break;
					  
		    case SERVER_CROP    : 
#ifdef DEBUG
			printf ("In SERVER_CROP - params : %s\n", filter_params);
#endif
					  sscanf (filter_params, "%d:%d:%d:%d", 
							  &inCol, &inRow, &outWidth, &outHeight);
					  if (inCol < 0 || inCol > inImage1.width) inCol = 0; 
        				  if (inRow < 0 || inRow > inImage1.height) inRow = 0; 
					  if (outWidth < 0 || outWidth > (inImage1.width - inCol)) 
            					outWidth = inImage1.width - inCol; 
        				  if (outHeight < 0 || outHeight > (inImage1.height - inRow)) 
					        outHeight = inImage1.height - inRow; 
					  outImage.size = IMGSIZECOLOR(outWidth, outHeight);
					  if (outImage.buff == NULL) {
						outImage.buff = (u_char*) malloc(outImage.size);
					  }
					  else {
						outImage.buff = realloc (outImage.buff, outImage.size);
					  }
					  if (outImage.buff == NULL) {
						perror ("Memory allocation error.");
						exit(1);
        				  }
        				  if (cropImage (&inImage1, &outImage, inRow, inCol, 
								        outWidth, outHeight) < 0) {
						perror ("Error filtering image.");
						exit(1);
        				  }
					  break;
					  
		    case SERVER_GREY    : 
#ifdef DEBUG
			printf ("In SERVER_GREY\n");
#endif
					  outImage.size = IMGSIZEGREY(inImage1.width, inImage1.height);
        				  if (outImage.buff == NULL) {
						outImage.buff = (u_char*) malloc(outImage.size);
					  }
					  else {
						outImage.buff = realloc (outImage.buff, outImage.size);
					  }
					  if (outImage.buff == NULL) {
						fprintf (stderr, "Requested %d bytes...\n", outImage.size);
						perror ("Memory allocation error.");
						exit(1);
					  }
    			                  if (greyImage(&inImage1, &outImage) < 0) {
					  	perror ("Error filtering image.");
					        exit(1);
        				  }
					  break;
					  
		    case SERVER_EDGE    : 
#ifdef DEBUG
			printf ("In SERVER_EDGE\n");
#endif
					  tmpImage.size = IMGSIZEGREY(inImage1.width, inImage1.height);
					  tmpImage.buff = (u_char*) malloc(tmpImage.size);
				          if (tmpImage.buff == NULL) {
					        perror ("Memory allocation error.");
					        exit(1);
				          }
    				          if (greyImage(&inImage1, &tmpImage) < 0) {
					        perror ("Error filtering image.");
					        exit(1);
					  }

				          outImage.size = IMGSIZEGREY(tmpImage.width, tmpImage.height);
					  if (outImage.buff == NULL) {
				          	outImage.buff = (u_char*) malloc(outImage.size);
					  }
					  else {
						outImage.buff = realloc (outImage.buff, outImage.size);
					  }
				          if (outImage.buff == NULL) {
					  	perror ("Memory allocation error.");
					        exit(1);
					  }
    				          if (edgeImage(&tmpImage, &outImage) < 0) {
					          perror ("Error filtering image.");
            					  exit(1);
					  }
					  break;
					  
		    case SERVER_ENLARGE : 
#ifdef DEBUG
			printf ("In SERVER_ENLARGE - params: %s\n", filter_params);
#endif
					  sscanf (filter_params, "%d", &enlargeFactor);
					  if (enlargeFactor < 0) enlargeFactor = 1;
        				  outImage.size = IMGSIZECOLOR((inImage1.width * enlargeFactor), 
	  			      				       (inImage1.height * enlargeFactor));
					  if (outImage.buff == NULL) {
					  	outImage.buff = (u_char*) malloc(outImage.size);
					  }
					  else {
						outImage.buff = realloc (outImage.buff, outImage.size);
					  }
      
					  if (outImage.buff == NULL) {
						perror ("Memory allocation error.");
						exit(1);
					  }
					  if (enlargeImage(&inImage1, &outImage, enlargeFactor) < 0) {
						perror ("Error filtering image.");
						exit(1);
					  }
					  break;
					  
		    case SERVER_MIX     : 
#ifdef DEBUG
			printf ("In SERVER_MIX - params : %s\n", filter_params);
#endif
					  sscanf (filter_params, "%f", &mixFactor);
					  if (mixFactor < 0.0 || mixFactor > 1.0) mixFactor = 0.5;
					  outImage.size = IMGSIZECOLOR(inImage1.width,inImage1.height);    
					  if (outImage.buff == NULL) {
					  	outImage.buff = (u_char*) malloc(outImage.size);
					  }
					  else {
						outImage.buff = realloc (outImage.buff, outImage.size);
					  }
					  if (outImage.buff == NULL) {
						perror ("Memory allocation error.");
						exit(1);
					  }
					  if (mixImage(&inImage1, &inImage2, &outImage, mixFactor) < 0) {
						perror ("Error filtering image.");
						exit(1);
					  }
					  break;
					  
		    case SERVER_CONCAT  : 
#ifdef DEBUG
			printf ("In SERVER_CONCAT - params : %s\n", filter_params);
#endif
					  sscanf (filter_params, "%d", &catOrder);
					  if (catOrder != 0 && catOrder != 1) catOrder = 1;
					  if (catOrder == 0) {
						outWidth = inImage1.width * 2;
						outHeight = inImage1.height;
					  } 
					  else {
						outWidth = inImage1.width;
						outHeight = inImage1.height * 2;
					  }
					  outImage.size = IMGSIZECOLOR(outWidth, outHeight);
					  outImage.maxval = 255;
					  outImage.maxval = htonl (outImage.maxval);
					  if (outImage.buff == NULL) {
						outImage.buff = (u_char*) malloc(outImage.size);
					  }
					  else {
						outImage.buff = realloc (outImage.buff, outImage.size);
					  }
					  if (outImage.buff == NULL) {
						perror ("Memory allocation error.");
						exit(1);
					  }
					  if (catImage(&inImage1, &inImage2, &outImage, catOrder) < 0) {
						perror ("Error filtering image.");
						exit(1);
					  }
					  break;
					  
		    case SERVER_REDUCE  : 
#ifdef DEBUG
			printf ("In SERVER_REDUCE - params : %s\n", filter_params);
#endif
					  sscanf (filter_params, "%d", &reduceFactor);
					  if (reduceFactor < 0) reduceFactor = 1;
					  outImage.size = IMGSIZECOLOR((inImage1.width / reduceFactor), 
								       (inImage1.height / reduceFactor));
					  if (outImage.buff == NULL) {
					  	outImage.buff = (u_char*) malloc(outImage.size);
					  }
					  else {
						outImage.buff = realloc (outImage.buff, outImage.size);
					  }
					  if (outImage.buff == NULL) {
					  	perror ("Memory allocation error.");
						exit(1);
					  }
					  if (reduceImage(&inImage1, &outImage, reduceFactor) < 0) {
						perror ("Error filtering image.");
						exit(1);
					  }
					  break;
					  
		    case HUFFMAN        : 
#ifdef DEBUG
			printf ("In HUFFMAN\n");
#endif
					  my_pid = getpid();
					  sprintf(filename1, "%s-%d-1", "temp", my_pid);
					  sprintf(filename2, "%s-%d-2", "temp", my_pid);
					  fd_in = open(filename1, O_CREAT|O_WRONLY, S_IRWXU|S_IRGRP|S_IROTH);
					  fd_out = OpenOutputBitFile(filename2);
					  sprintf(header, "%c%c\n%d %d\n%d\n", inImage1.ppm1, inImage1.ppm2, 
						htonl (inImage1.width), 
						htonl (inImage1.height), 
						htonl (inImage1.maxval));  
					  write(fd_in, header, sizeof(header)); 
					  write(fd_in, inImage1.buff, inImage1.size); 
					  close(fd_in);
					  fd_in2 = fopen(filename1, "rb");
					  CompressFile(fd_in2, fd_out, 0, NULL);
					  CloseOutputBitFile(fd_out);
					  fclose(fd_in2);
					  inImagefd1 = open(filename2, O_RDONLY);
					  fstat(inImagefd1, &stat_buf);
					  inImage1.buff = (u_char *) malloc(stat_buf.st_size);
					  nread = read(inImagefd1, inImage1.buff, stat_buf.st_size);   
					  response.len = htonl (stat_buf.st_size);
					  write(sfd, &response, sizeof(struct stream_response));
				          write(sfd, inImage1.buff, response.len);
					  close(inImagefd1);
					  unlink(filename1);
					  unlink(filename2);
					  compressed = 1;
					  break;
					  
		    case LZ77           : 
#ifdef DEBUG
			printf ("In LZ77\n");
#endif
					  my_pid = getpid();
					  sprintf(filename1, "%s-%d-1", "temp", my_pid);
					  sprintf(filename2, "%s-%d-2", "temp", my_pid);
					  fd_in = open(filename1, O_CREAT|O_WRONLY, S_IRWXU|S_IRGRP|S_IROTH);
					  fd_out = OpenOutputBitFile(filename2);
					  sprintf(header, "%c%c\n%d %d\n%d\n", inImage1.ppm1, inImage1.ppm2, 
							  inImage1.width, inImage1.height, inImage1.maxval);  
					  write(fd_in, header, sizeof(header)); 
					  write(fd_in, inImage1.buff, inImage1.size); 
					  close(fd_in);
					  fd_in2 = fopen(filename1, "rb");
					  CompressFile2(fd_in2, fd_out, 0, NULL);
					  CloseOutputBitFile(fd_out);
					  fclose(fd_in2);
					  inImagefd1 = open(filename2, O_RDONLY);
					  fstat(inImagefd1, &stat_buf);
					  inImage1.buff = (u_char *) malloc(stat_buf.st_size);
					  nread = read(inImagefd1, inImage1.buff, stat_buf.st_size);   
					  response.len = htonl (stat_buf.st_size);
					  write(sfd, &response, sizeof(struct stream_response));
					  write(sfd, inImage1.buff, response.len);
					  close(inImagefd1);
					  unlink(filename1);
					  unlink(filename2);	
					  compressed = 1;
					  break;

		    default             : fprintf (stderr, "Unknown filter code!\n");
					  exit (1);
					  break;
	    }

	    if ((i < (filter_list->size - 1)) && (filter_type >= SERVER_CROP)) {
		    free (inImage1.buff);
		    inImage1.size = outImage.size;
		    if ((inImage1.buff = malloc (outImage.size)) == NULL) {
			    fprintf (stderr, "Unable to reallocate images!\n");
      			    free(params);
			    exit (1);
		    }
		    memcpy (inImage1.buff, outImage.buff, outImage.size);
		    free (outImage.buff);
		    outImage.buff = NULL;
	    }
    }   

    if (!compressed) {
	    int bytes_sent;
	    struct stream_response server_response;
	    
	    header_len = sprintf(header, "%c%c\n%d %d\n%d\n", outImage.ppm1, outImage.ppm2, 
		    			outImage.width, outImage.height, outImage.maxval);  
    	    header[header_len] = '\0';
	    #ifdef DEBUG
	    	fprintf (stdout, "Header = %s\n", header);
		fprintf (stdout, "Header Len = %d\n", header_len);
	    	fprintf (stdout, "New Image Size: %d bytes\n", outImage.size);
	    #endif
	    if (outImage.buff != NULL) {
	            server_response.len = htonl (outImage.size + sizeof (header));
#ifdef DEBUG
		    fprintf (stdout, "Image Length: %d\n", ntohl (server_response.len));
#endif
	            send_buffer = malloc (ntohl(server_response.len));
	            memcpy (send_buffer, header, sizeof (header));
		    memcpy (send_buffer + sizeof (header), outImage.buff, outImage.size);
	    }
	    else {
	    	    server_response.len = htonl (inImage1.size + sizeof (header));
#ifdef DEBUG
		    fprintf (stdout, "Image Length: %d\n", ntohl (server_response.len));
#endif
	            send_buffer = malloc (ntohl(server_response.len));
	            memcpy (send_buffer, header, sizeof (header));
		    memcpy (send_buffer + sizeof (header), inImage1.buff, inImage1.size);
	    }
	    send (sfd, &server_response, sizeof(struct stream_response), 0);
	    if ((bytes_sent = send (sfd, send_buffer, ntohl (server_response.len), 0)) < 0) {
		    perror ("Image transmission error");
      		    free(params);
		    exit (1);
    	    }
	    #ifdef DEBUG
	    	fprintf (stdout, "Awaiting client acknowledgement...just sent %d bytes\n", bytes_sent);
		fflush (stdout);
	    #endif
	    recv (sfd, &ack, sizeof (struct stream_response), MSG_WAITALL);
	    #ifdef DEBUG
	    	fprintf (stdout, "Received client acknowledgement: %d\n", ntohl (ack.len));
		fflush (stdout);
	    #endif
	    free (send_buffer);		
    }
 
    if (outImage.buff != NULL)
	    free(outImage.buff);
	
    close(inImagefd1);
    close(inImagefd2);

    #ifdef DEBUG
    	fprintf (stdout, "Calculating rate control...\n");
    #endif
    
    /* Rate control */
    gettimeofday(&t2, NULL);
#ifdef DEBUG
    printf("old: %ld, %ld  new: %ld, %ld\n", t1.tv_sec, t1.tv_usec, t2.tv_sec, t2.tv_usec);
#endif
    t3.tv_sec = t1.tv_sec;
    t3.tv_usec = t1.tv_usec;
    t2.tv_sec -= t3.tv_sec;
    t3.tv_sec = 0;
    if (t2.tv_sec > 0) {
      sleeptime = 1000000 - t3.tv_usec;
      sleeptime += t2.tv_usec;
      sleeptime += (t2.tv_sec - 1) * 1000000;    
    } else 
      sleeptime = t2.tv_usec - t3.tv_usec;
    if (sleeptime < period) {
      sleeptime = period - sleeptime;
      time1.tv_sec = 0;
      while (sleeptime > 999999) {
        time1.tv_sec++;
        sleeptime -= 1000000;
      }
      time1.tv_nsec = sleeptime * 1000;
      t1.tv_usec += period;
      while (t1.tv_usec > 999999) {
        t1.tv_sec++;
	t1.tv_usec -= 1000000;
      }
#ifdef DEBUG
      printf("sleep: %ld sec and %ld nsec\n", time1.tv_sec, time1.tv_nsec);
#endif
      while (nanosleep(&time1, &time2) == -1) {
        time1.tv_sec = time2.tv_sec;
        time1.tv_nsec = time2.tv_nsec;
      }
    }
  }
  close (sfd);
  free_list (filter_list);
  free(params);
  return 0;
}

