/* -*-C-*-
*******************************************************************************
*
* File:         imgServer.c
* RCS:          $Id: $
* Description:  Test for image handlers that will become streamlets
* Authors:      original: Fabian E. Bustamante
*               socket interface: Christian Poellabauer
*               Systems Research Group
*               College of Computing
*               Georgia Institute of Technology
* Language:     C
* Package:      N/A
* Status:       Experimental (Do Not Distribute)
*
* (C) Copyright 2001, Georgia Institute of Technology, all rights reserved.
*
*******************************************************************************
*/

#include <math.h>
#include <stdlib.h>
#include <stdio.h>
#include <time.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <unistd.h>
#include <sys/time.h>

#include "imgHandler.h"

static char* program_name;

/*----------------------------------------------------------------------------*/

/*
 * usage: Prints out 'usage' string
 *
 * Parameters: target 'stream' and 'status' to quit with
 *
 * Returns: -
 *
 */
 
static void
usage (FILE* stream, int status)
{
  fprintf(stream, "Usage: %s [iterations] \n"
	  program_name);
  exit(status);
} /* end usage */

/*------------------------------------------------------------------------------*/

int
main (int argc, char* argv[])
{ 
  char snapshotFile[40];
  char header[20];
  struct tm *now;
  time_t nowt;
  Raw_data inImage1, inImage2, outImage, tmpImage;
  int inCol, inRow, outWidth, outHeight;
  int sfd, inImagefd1, inImagefd2;
  int enlargeFactor, reduceFactor;
  float mixFactor;
  int catOrder;
  int streamletIndx;
  int maxppmval;
  struct timeval t1, t2;
  int listenfd;
  struct sockaddr_in server;
  struct sockaddr_in client;
  int flags;
  struct ppm_request *NewRequest, *CurrentRequest, **Prev;
  int clientlen;
  int clientsock;

  AcceptQueue = NULL;

  /* Create listening socket */
  listenfd = socket(PF_INET, SOCK_STREAM, 0);
  if (listenfd == -1) {
    perror("Failed opening socket.\n");    
    exit(1);
  }
  bzero((char *)&server, sizeof(server));
  server.sin_family = PF_INET;
  server.sin_addr.s_addr = INADDR_ANY;
  server.sin_port = htons(SERVER_PORT);
  if (bind(listenfd, (struct sock_addr *)&server, sizeof(server)) == -1)
    perror("Failed binding socket.\n");    
    exit(1);
  }
  listen(listenfd, 30);
  
  /* Loop forever */
  while () {
    if (clientsock = accept(listenfd, (struct sock_addr *) &client, clientlen) < 0) {
      NewRequest = malloc(sizeof(struct ppm_request));
      memset(NewRequest, 0, sizeof(struct ppm_request));
      NewRequest->sockfd = clientsock;
      NewRequest->Next = AcceptQueue;
      AcceptQueue = NewRequest;
    }
    CurrentRequest = AcceptQueue;
    Prev = &AcceptQueue;
    while (CurrentRequest != NULL) {
      if 

    }
  }


  /* Get first image from file */
  fprintf(stdout, "Image file 1: \t%s\n", argv[2]);
  if ((inImagefd1 = open(argv[2], O_RDONLY)) < 0) {
    perror("Failed opening image file 1.\n");    
    exit(1);
  }
  read(inImagefd1, header, sizeof(header));   
  sscanf(header, "%c%c\n%d %d\n%d\n", 
	 &(inImage1.ppm1), &(inImage1.ppm2),
	 &(inImage1.width), &(inImage1.height), 
	 &maxppmval);
  fprintf(stdout, "PPM %c%c\n", 
	  inImage1.ppm1, inImage1.ppm2);
  fprintf(stdout, "Width %d Height %d\n", 
	  inImage1.width, inImage1.height);
  fprintf(stdout, "Max PPM val %d\n", maxppmval);
  inImage1.size = inImage1.width * inImage1.height * 
    3 * sizeof(u_char);
  inImage1.buff = (u_char*) malloc(inImage1.size);
  if (inImage1.buff == NULL) {
    perror("Out of mem\n");
    exit(1);
  }
  fprintf(stdout, "Size %d\n", inImage1.size);
  read(inImagefd1, inImage1.buff, inImage1.size);

  /* Get second image from file */
  fprintf(stdout, "Image file 2: \t%s\n", argv[2]);
  if ((inImagefd2 = open(argv[3], O_RDONLY)) < 0) {
    perror("Failed opening image file 1.\n");    
    exit(1);
  }
  read(inImagefd2, header, sizeof(header));   
  sscanf(header, "%c%c\n%d %d\n%d\n", 
	 &(inImage2.ppm1), &(inImage2.ppm2),
	 &(inImage2.width), &(inImage2.height), 
	 &maxppmval);
  fprintf(stdout, "PPM %c%c\n", 
	  inImage2.ppm1, inImage2.ppm2);
  fprintf(stdout, "Width %d Height %d\n", 
	  inImage2.width, inImage2.height);
  fprintf(stdout, "Max PPM val %d\n", maxppmval);
  inImage2.size = inImage2.width * inImage2.height * 
    3 * sizeof(u_char);
  inImage2.buff = (u_char*) malloc(inImage2.size);
  if (inImage2.buff == NULL) {
    perror("Out of mem\n");
    exit(1);
  }
  fprintf(stdout, "Size %d\n", inImage2.size);
  read(inImagefd2, inImage2.buff, inImage2.size);

  time(&nowt);
  now = localtime(&nowt);

  /*
   * Now let's test the soon-to-be-streamlets
   *
   */

  switch(streamletIndx) {
  case 1:			/* Sample a part of the image */
    fprintf(stdout, "\nEnter column and row of starting point: ");
    scanf("%d %d", &inCol, &inRow);
    if (inCol < 0 || inCol > inImage1.width) inCol = 10; 
    if (inRow < 0 || inRow > inImage1.height) inRow = 10; 

    fprintf(stdout, "\nEnter width and height of new image: ");
    scanf("%d %d", &outWidth, &outHeight);
    if (outWidth < 0 || outWidth > (inImage1.width - inCol)) 
      outWidth = inImage1.width - inCol; 
    if (outHeight < 0 || outHeight > (inImage1.height - inRow)) 
      outHeight = inImage1.height - inRow; 
    
    outImage.size = IMGSIZECOLOR(outWidth, outHeight);
    outImage.buff = (u_char*) malloc(outImage.size);
    if (outImage.buff == NULL) {
      perror ("Darn!");
      exit(1);
    }
   
    gettimeofday(&t1,NULL); 
    if (cropImage(&inImage1, &outImage, inRow, inCol, outWidth, outHeight) < 0) {
      perror ("Darn!");
      exit(1);
    }

    sprintf(snapshotFile, "snapshot1-crop-%d-%d-%d.ppm", 
	    now->tm_hour, now->tm_min, now->tm_sec);
    sfd = open(snapshotFile, O_CREAT|O_WRONLY, S_IRWXU|S_IRGRP|S_IROTH);
    
    sprintf(header, "%c%c\n%d %d\n%d\n", outImage.ppm1, outImage.ppm2, 
	    outImage.width, outImage.height, outImage.maxval);  
    write(sfd, header, sizeof(header));
    write(sfd, outImage.buff, outImage.size);
    close(sfd);
    
    free(outImage.buff);
    gettimeofday(&t2,NULL); 
    printf("total time: %ld %ld - %ld %ld\n",t1.tv_sec,t1.tv_usec,t2.tv_sec,t2.tv_usec);
    break;

  case 2:			/* Get grey image -> compression as a side-effect */
    outImage.size = IMGSIZEGREY(inImage1.width, inImage1.height);
    outImage.buff = (u_char*) malloc(outImage.size);
    if (outImage.buff == NULL) {
      perror ("Darn!");
      exit(1);
    }
    
    gettimeofday(&t1, NULL); 
    if (greyImage(&inImage1, &outImage) < 0) {
      perror ("Darn!");
      exit(1);
    }

    sprintf(snapshotFile, "snapshot1-grey-%d-%d-%d.pgm", 
	    now->tm_hour, now->tm_min, now->tm_sec);
    sfd = open(snapshotFile, O_CREAT|O_WRONLY, S_IRWXU|S_IRGRP|S_IROTH);
    
    sprintf(header, "%c%c\n%d %d\n%d\n", outImage.ppm1, outImage.ppm2, 
	    outImage.width, outImage.height, outImage.maxval);  
    
    write(sfd, header, sizeof(header));
    write(sfd, outImage.buff, outImage.size);
    close(sfd);
    
    free(outImage.buff);
    gettimeofday(&t2,NULL); 
    printf("total time: %ld %ld - %ld %ld\n",t1.tv_sec,t1.tv_usec,t2.tv_sec,t2.tv_usec);
    break;

  case 3:			/* Get edge image -> compression as a side-effect */
    tmpImage.size = IMGSIZEGREY(inImage1.width, inImage1.height);
    tmpImage.buff = (u_char*) malloc(tmpImage.size);
    if (tmpImage.buff == NULL) {
      perror ("Darn!");
      exit(1);
    }
    
    gettimeofday(&t1,NULL); 
    if (greyImage(&inImage1, &tmpImage) < 0) {
      perror ("Darn!");
      exit(1);
    }

    outImage.size = IMGSIZEGREY(tmpImage.width, tmpImage.height);
    outImage.buff = (u_char*) malloc(outImage.size);
    if (outImage.buff == NULL) {
      perror ("Darn!");
      exit(1);
    }
    
    if (edgeImage(&tmpImage, &outImage) < 0) {
      perror ("Darn!");
      exit(1);
    }

    sprintf(snapshotFile, "snapshot1-edge-%d-%d-%d.pgm", 
	    now->tm_hour, now->tm_min, now->tm_sec);
    sfd = open(snapshotFile, O_CREAT|O_WRONLY, S_IRWXU|S_IRGRP|S_IROTH);
    
    sprintf(header, "%c%c\n%d %d\n%d\n", outImage.ppm1, outImage.ppm2, 
	    outImage.width, outImage.height, outImage.maxval);  
    
    write(sfd, header, sizeof(header));
    write(sfd, outImage.buff, outImage.size);
    close(sfd);
    
    free(tmpImage.buff);
    free(outImage.buff);
    gettimeofday(&t2,NULL); 
    printf("total time: %ld %ld - %ld %ld\n",t1.tv_sec,t1.tv_usec,t2.tv_sec,t2.tv_usec);
    break;

  case 4:			/* Enlarge */
    fprintf(stdout, "\nEnter enlarge factor (> 0): ");
    scanf("%d", &enlargeFactor);
    if (enlargeFactor < 0) enlargeFactor = 1;

    outImage.size = IMGSIZECOLOR((inImage1.width * enlargeFactor), 
				      (inImage1.height * enlargeFactor));
    outImage.buff = (u_char*) malloc(outImage.size);
    gettimeofday(&t1,NULL); 
    if (outImage.buff == NULL) {
      perror ("Darn!");
      exit(1);
    }
    
    if (enlargeImage(&inImage1, &outImage, enlargeFactor) < 0) {
      perror ("Darn!");
      exit(1);
    }

    sprintf(snapshotFile, "snapshot1-enlarge%d-%d-%d-%d.ppm", enlargeFactor, 
	    now->tm_hour, now->tm_min, now->tm_sec);
    sfd = open(snapshotFile, O_CREAT|O_WRONLY, S_IRWXU|S_IRGRP|S_IROTH);
    
    sprintf(header, "%c%c\n%d %d\n%d\n", outImage.ppm1, outImage.ppm2, 
	    outImage.width, outImage.height, outImage.maxval);  
    
    write(sfd, header, sizeof(header));
    write(sfd, outImage.buff, outImage.size);
    close(sfd);
    
    free(outImage.buff);
    gettimeofday(&t2,NULL); 
    printf("total time: %ld %ld - %ld %ld\n",t1.tv_sec,t1.tv_usec,t2.tv_sec,t2.tv_usec);
    break;

  case 5:			/* Mix */
    fprintf(stdout, "\nEnter mix factor (0.0 <= factor <= 1.0): ");
    scanf("%f", &mixFactor);
    if (mixFactor < 0.0 || mixFactor > 1.0) mixFactor = 0.5;

    gettimeofday(&t1,NULL); 
    outImage.size = IMGSIZECOLOR(inImage1.width,inImage1.height);    
    outImage.buff = (u_char*) malloc(outImage.size);
    if (outImage.buff == NULL) {
      perror ("Darn!");
      exit(1);
    }
    
    if (mixImage(&inImage1, &inImage2, &outImage, mixFactor) < 0) {
      perror ("Darn!");
      exit(1);
    }

    sprintf(snapshotFile, "snapshot-mix%d-%d-%d-%d.ppm", 
	    (int) (mixFactor * 10), now->tm_hour, now->tm_min, now->tm_sec);
    sfd = open(snapshotFile, O_CREAT|O_WRONLY, S_IRWXU|S_IRGRP|S_IROTH);
    
    sprintf(header, "%c%c\n%d %d\n%d\n", outImage.ppm1, outImage.ppm2, 
	    outImage.width, outImage.height, outImage.maxval);  
    
    write(sfd, header, sizeof(header));
    write(sfd, outImage.buff, outImage.size);
    close(sfd);
    
    free(outImage.buff);
    gettimeofday(&t2,NULL); 
    printf("total time: %ld %ld - %ld %ld\n",t1.tv_sec,t1.tv_usec,t2.tv_sec,t2.tv_usec);
    break;

  case 6:			/* Cat */
    fprintf(stdout, "\nEnter cat order (0|1): ");
    scanf("%d", &catOrder);
    if (catOrder != 0 && catOrder != 1) catOrder = 1;

    gettimeofday(&t1,NULL); 
    if (catOrder == 0) {
      outWidth = inImage1.width * 2;
      outHeight = inImage1.height;
    } else {
      outWidth = inImage1.width;
      outHeight = inImage1.height * 2;
    }
    outImage.size = IMGSIZECOLOR(outWidth, outHeight);
    outImage.maxval = 255;
        
    outImage.buff = (u_char*) malloc(outImage.size);
    if (outImage.buff == NULL) {
      perror ("Darn!");
      exit(1);
    }
    
    if (catImage(&inImage1, &inImage2, &outImage, catOrder) < 0) {
      perror ("Darn!");
      exit(1);
    }

    sprintf(snapshotFile, "snapshot-cat-%d-%d-%d.ppm",
	    now->tm_hour, now->tm_min, now->tm_sec);
    sfd = open(snapshotFile, O_CREAT|O_WRONLY, S_IRWXU|S_IRGRP|S_IROTH);
    
    sprintf(header, "%c%c\n%d %d\n%d\n", outImage.ppm1, outImage.ppm2, 
	    outImage.width, outImage.height, outImage.maxval);  
    write(sfd, header, sizeof(header));
    write(sfd, outImage.buff, outImage.size);
    close(sfd);
    
    free(outImage.buff);
    gettimeofday(&t2,NULL); 
    printf("total time: %ld %ld - %ld %ld\n",t1.tv_sec,t1.tv_usec,t2.tv_sec,t2.tv_usec);
    break;
  case 7:			/* Reduce */
    fprintf(stdout, "\nEnter reduce factor (> 0): ");
    scanf("%d", &reduceFactor);
    if (reduceFactor < 0) reduceFactor = 1;

    gettimeofday(&t1,NULL); 
    outImage.size = IMGSIZECOLOR((inImage1.width / reduceFactor), 
				      (inImage1.height / reduceFactor));
    outImage.buff = (u_char*) malloc(outImage.size);
    if (outImage.buff == NULL) {
      perror ("Darn!");
      exit(1);
    }
    
    if (reduceImage(&inImage1, &outImage, reduceFactor) < 0) {
      perror ("Darn!");
      exit(1);
    }

    sprintf(snapshotFile, "snapshot1-reduce%d-%d-%d-%d.ppm", reduceFactor, 
	    now->tm_hour, now->tm_min, now->tm_sec);
    sfd = open(snapshotFile, O_CREAT|O_WRONLY, S_IRWXU|S_IRGRP|S_IROTH);
    
    sprintf(header, "%c%c\n%d %d\n%d\n", outImage.ppm1, outImage.ppm2, 
	    outImage.width, outImage.height, outImage.maxval);  
    
    write(sfd, header, sizeof(header));
    write(sfd, outImage.buff, outImage.size);
    close(sfd);
    
    free(outImage.buff);
    gettimeofday(&t2,NULL); 
    printf("total time: %ld %ld - %ld %ld\n",t1.tv_sec,t1.tv_usec,t2.tv_sec,t2.tv_usec);
    break;
  } /* end switch */
  
  exit(0);    
} /* end main */
