
/*
 *   sender.c
 *
 *   Copyright (C) 2000 Christian Poellabauer
 *
 */

/* 
 *   Kernel-loadable module for event channel support
 */

/* Required definitions. */
#ifndef MODULE
#define MODULE
#endif
#ifndef __KERNEL__
#define __KERNEL__
#endif

/* Various required header files. */
#include <linux/module.h> 
#include <linux/kernel.h> 
#include <linux/tty.h>
#include <linux/unistd.h>
#include <linux/socket.h>
#include <linux/kmod.h>
#include <linux/sched.h>
#include <linux/timer.h>
#include <linux/wait.h>
#include <linux/init.h>
#include <linux/types.h>
#include <linux/stat.h>
#include <linux/utsname.h>
#include <linux/mm.h>
#include <linux/fcntl.h>
#include <linux/slab.h>
#include <linux/vmalloc.h>
#include <asm/current.h>
#include <asm/uaccess.h>
#include "sender.h"
#include "../../include/io.h"
#include "../../include/atl.h"
#include "../../include/cm.h"
#include "../../include/echo.h"
#include "../../include/comm_group2.h"
#include "../../include/utils.h"

MODULE_AUTHOR("Christian Poellabauer");
MODULE_DESCRIPTION("Example 1 for KECho");
MODULE_LICENSE("GPL");

#define DATASIZE 30000

#define APP_NAME1  "TEST1"
#define GRP_TYPE1  "TEST1"

#define my_rdtsc(x)		__asm__ __volatile__ (			\
					"rdtsc \n\t" 			\
					"movl %%eax,%0 \n\t"		\
					: "=r" (x)			\
					:				\
					: "%eax", "%edx");		

char *channel_id;

/* Prototypes. */
int init_module(void);		       
void cleanup_module(void);

/* Global variables. */
static EChannel chan1;
static CManager cm1;
static EControlContext cc1;
static ECSourceHandle handle1;
static int loop_flag = 1;
kthread_t sender_thread;

struct test_struct {
  int type;
  char hostname[64];
  int v1;
  int v2;
  int v3;
  int v4;
  int v5;
  int v6;
  int v7;
};
  
void shandler1(void *event, int length,void *client_data, attr_list event_attrs)
{
//printk("event received!\n");
  printk("in shandler1:%ld...\n", *(long *) event);
}

/* Sender function. */
static int send_data()
{
  char *group_id;
  char *channel_id;
  char *buf1;
  signed long sleep_time = 0;
  char *host_name;
  long a = 1;
  struct test_struct my_struct;

  init_kthread(&sender_thread, "sender");

  buf1 = (char *)kmalloc (DATASIZE,GFP_KERNEL);
  memset(buf1, 'A', DATASIZE);
  buf1[DATASIZE-1]='\0';

  printk("Create CManager: %d\n", current->pid);
  cm1 = CManager_create();

  printk("Init CM\n");
  cc1 = ECho_CM_init(cm1);

  printk("Create Channel\n");
  chan1 = EChannel_create(cc1);

  channel_id = ECglobal_id(chan1);
  printk("channel id is %s\n", channel_id);

  host_name = (char *) kmalloc (MAXHOSTNAMELEN, GFP_KERNEL);
  getthishostname(host_name, MAXHOSTNAMELEN);

  printk("Setup Comm Group\n");
  if ((group_id = setup_comm_group(NULL, APP_NAME1, GRP_TYPE1, 100,
	host_name, 5010)) == NULL) 
	  return -1;
  printk("Assoc Comm Group\n");
  assoc_group_info(group_id, channel_id, strlen(channel_id) + 1);

  printk("Subscribe as Source\n");
  handle1 = ECsource_subscribe(chan1);

  printk("Subscribe as Sink\n");
  (void) ECsink_subscribe(chan1, shandler1, NULL);

  printk("Start Work\n");
  while (loop_flag == 1) {
    CMpoll_network(cm1);
    if (sleep_time == 0) {
      a++;
      printk("transmit data(%ld)\n", a);
my_struct.type = a;
      ECsubmit_event(handle1, &my_struct, sizeof(struct test_struct));
      //ECsubmit_event(handle1, buf1, DATASIZE);
//      ECsubmit_event(handle1, &a, sizeof(long));
      sleep_time = HZ; 
      //if ((a > 110) && (a < 120))
//	sleep_time = HZ;
    }  
    sleep_time = CMtask_sleep(sleep_time);
  }

  exit_kthread(&sender_thread);

  return 0;
}

/* Initialize module. */
int init_module(void) 
{
  if (start_kthread(send_data, &sender_thread) <0) {
    printk("Thread creation failed!\n");
    return -1;
  }
  return 0;
}

/* Remove module. */
void cleanup_module(void) 
{
  loop_flag = 0;
  stop_kthread(&sender_thread);
}

