
/*
 *   reader.c
 *
 *   Copyright (C) 2001 Christian Poellabauer
 *
 */

/* 
 *   Kernel-loadable module for event channel support
 */

/* Required definitions. */
#ifndef MODULE
#define MODULE
#endif
#ifndef __KERNEL__
#define __KERNEL__
#endif

/* Various required header files. */
#include <linux/module.h> 
#include <linux/kernel.h> 
#include <linux/tty.h>
#include <linux/unistd.h>
#include <linux/socket.h>
#include <linux/kmod.h>
#include <linux/sched.h>
#include <linux/timer.h>
#include <linux/wait.h>
#include <asm/current.h>
#include "reader.h"
#include "../../include/io.h"
#include "../../include/atl.h"
#include "../../include/cm.h"
#include "../../include/echo.h"
#include "../../include/comm_group2.h"

MODULE_AUTHOR("Christian Poellabauer");
MODULE_DESCRIPTION("Example 1 for KECho");
MODULE_LICENSE("GPL");

#define APP_NAME1 "TEST1"
#define GRP_TYPE1 "TEST1"

//#define rdtsc(x)	__asm__ __volatile__("rdtsc \n\t" : "=A" (*(x)))

/* Prototypes. */
int init_module(void);		       
void cleanup_module(void);

/* Global variables. */
static int reader_pid;
static int loop_flag2 = 1;

/* Handler function. */
void rhandler1(void *event,int length, void *client_data, attr_list event_attrs)
{
int result;
//  printk("event data1: %ld\n", *(long *)event);
    printk("%d bytes received\n", length);
    if (event_attrs != NULL) printk("attrs received too\n");
    query_attr(event_attrs, attr_atom_from_string("QOS_DEADLINE"), NULL,
	(attr_value *)&result);
    //printk("result=%d\n", result);
}

/* Reader function. */
static int read_data()
{
  static comm_group_return groups;
  static signed long sleep_time = 0; 
  static long a = 200;
  static EChannel rchan1;
  static CManager rcm1;
  static EControlContext rcc1;
  static ECSourceHandle rhandle1;

  rcm1 = CManager_create();

  rcc1 = ECho_CM_init(rcm1);

  groups = matching_comm_groups(APP_NAME1, GRP_TYPE1);
  rchan1 = EChannel_open(rcc1, groups->list[0].info);

  rhandle1 = ECsource_subscribe(rchan1);

  (void) ECsink_subscribe(rchan1, rhandler1, NULL);

  while (loop_flag2 == 1) {
    CMpoll_network(rcm1);
    if (sleep_time == 0) {
//      ECsubmit_event(rhandle1, &a, sizeof(a));
      a++;
      sleep_time = 2 * HZ;
    }
    sleep_time = CMtask_sleep(sleep_time);
  }

  return 0;
}

/* Initialize module. */
int init_module(void) 
{
  reader_pid = kernel_thread(read_data, NULL, 0);
  return 0;
}

/* Remove module. */
void cleanup_module(void) 
{
  loop_flag2 = 0;
  CMtask_kill(reader_pid);
}
