
/*
 *   Copyright (C) 2002 Christian Poellabauer
 *
 */

/* 
 *   Kernel-loadable module for event channel support
 */

/* Required definitions. */
#ifndef MODULE
#define MODULE
#endif
#ifndef __KERNEL__
#define __KERNEL__
#endif

/* Various required header files. */
#include <linux/module.h> 
#include <linux/kernel.h> 
#include <linux/tty.h>
#include <linux/unistd.h>
#include <linux/socket.h>
#include <linux/kmod.h>
#include <linux/sched.h>
#include <linux/timer.h>
#include <linux/wait.h>
#include <linux/init.h>
#include <linux/types.h>
#include <linux/stat.h>
#include <linux/utsname.h>
#include <linux/mm.h>
#include <linux/fcntl.h>
#include <linux/malloc.h>
#include <linux/vmalloc.h>
#include <asm/current.h>
#include <asm/uaccess.h>
#include "sender.h"
#include "../../include/io.h"
#include "../../include/atl.h"
#include "../../include/cm.h"
#include "../../include/echo.h"
#include "../../include/comm_group2.h"
#include "../../include/utils.h"
#include "../../include/ecl.h"

MODULE_AUTHOR("Christian Poellabauer");
MODULE_DESCRIPTION("Example 1 for ECL");
MODULE_LICENSE("GPL");

#define APP_NAME1  "TEST1"
#define GRP_TYPE1  "TEST1"

char *channel_id;

/* Prototypes. */
int init_module(void);		       
void cleanup_module(void);

/* Global variables. */
static EChannel chan1;
static CManager cm1;
static EControlContext cc1;
static ECSourceHandle handle1;
static int loop_flag = 1;
kthread_t sender_thread;

int empty(void) {
	return 3;
}

void shandler1(void *event, int length,void *client_data, attr_list event_attrs)
{
  printk("in shandler1:%ld...\n", *(long *) event);
}

extern void *ecl_code_gen2(char *code, ecl_parse_context context);

/* Sender function. */
static int send_data()
{
  char *group_id;
  char *channel_id;
  char *buf1;
  signed long sleep_time = 0;
  char *host_name;
  long a = 100;
  int test;

  init_kthread(&sender_thread, "ecltest");

  printk("begin...\n");
  CMtask_sleep(5*HZ);
/*
  {
  char code_string[] = "\
{\n\
    int j = 4;\n\
    long k = 10;\n\
    short l = 3;\n\
\n\
    return l * (j + k);\n\
}";
*/
  {
  char code_string[] = "\
{\n\
    long k = 5;\n\
    return k;\n\
}";
	ecl_parse_context context = new_ecl_parse_context();
	ecl_code gen_code;
	long (*func)();
	long result;

printk("parsing ok\n");
CMtask_sleep(200);
	gen_code = ecl_code_gen(code_string, context);
printk("code gen ok\n");
CMtask_sleep(200);
	func = (long(*)()) (long) gen_code->func;
printk("func ptr\n");
CMtask_sleep(200);
	if (func == NULL) printk("func is null\n");
	test = empty();
	result = func();
	printk("result:%ld\n", result);
CMtask_sleep(200);
	//if (result != 42) printk("ERROR\n");
	ecl_code_free(gen_code);
	ecl_free_parse_context(context);
	printk("done...\n");
        while (loop_flag == 1) {
  		CMtask_sleep(HZ);
        }
}

  /*
    {
	char code_string[] = "\
{\n\
    int j = 4;\n\
    long k = 10;\n\
    short l = 3;\n\
\n\
    return l * (j + k + i);\n\
}";

	ecl_parse_context context = new_ecl_parse_context();
	ecl_code gen_code;
    	long (*func)(int i);

	ecl_add_param("i", "int", 0, context);
	gen_code = ecl_code_gen(code_string, context);
	func = (long(*)(int)) (long) gen_code->func;
        if (func(15) != 87) printk("ERROR!\n");
	ecl_code_free(gen_code);
	ecl_free_parse_context(context);
    }
*/
  /*
    {
	char code_string[] = "\
{\n\
    return input.l * (input.j + input.k + input.i);\n\
}";

	typedef struct test {
	    int i;
	    int j;
	    long k;
	    short l;
	} test_struct, *test_struct_p;

	IOField struct_fields[] = {
	    {"i", "integer", sizeof(int), IOOffset(test_struct_p, i)},
	    {"j", "integer", sizeof(int), IOOffset(test_struct_p, j)},
	    {"k", "integer", sizeof(long), IOOffset(test_struct_p, k)},
	    {"l", "integer", sizeof(short), IOOffset(test_struct_p, l)},
	    {(void*)0, (void*)0, 0, 0}};

	ecl_parse_context context = new_ecl_parse_context();
	test_struct str;
	ecl_code gen_code;
	long (*func)(test_struct_p s);

	ecl_add_struct_type("struct_type", struct_fields, context);
	ecl_add_param("input", "struct_type", 0, context);

	gen_code = ecl_code_gen(code_string, context);
	func = (long(*)(test_struct_p)) (long) gen_code->func;

	str.i = 15;
	str.j = 4;
	str.k = 10;
	str.l = 3;
	if (func(&str) != 87) printk("ERROR\n");
	ecl_code_free(gen_code);
	ecl_free_parse_context(context);
    }
*/

/*
  printk("Start Work\n");
  while (loop_flag == 1) {
    if (sleep_time == 0) {
      sleep_time = 2 * HZ; 
      a++;
    }
    sleep_time = CMtask_sleep(sleep_time);
  }
*/

/*
    {
	char code_string[] = "\
{\n\
    return input.l * (input.j + input.k + input.i);\n\
}";

	typedef struct test {
	    int i;
	    int j;
	    long k;
	    short l;
	} test_struct, *test_struct_p;

	IOField struct_fields[] = {
	    {"i", "integer", sizeof(int), IOOffset(test_struct_p, i)},
	    {"j", "integer", sizeof(int), IOOffset(test_struct_p, j)},
	    {"k", "integer", sizeof(long), IOOffset(test_struct_p, k)},
	    {"l", "integer", sizeof(short), IOOffset(test_struct_p, l)},
	    {(void*)0, (void*)0, 0, 0}};

	ecl_parse_context context = new_ecl_parse_context();
	test_struct str;
	ecl_code gen_code;
	long (*func)(test_struct_p s);

	ecl_add_struct_type("struct_type", struct_fields, context);
	ecl_add_param("input", "struct_type", 0, context);

	gen_code = ecl_code_gen(code_string, context);
	func = (long(*)(test_struct_p)) (long) gen_code->func;

	str.i = 15;
	str.j = 4;
	str.k = 10;
	str.l = 3;
	if (func(&str) != 87) printk("ERROR\n");
	ecl_code_free(gen_code);
	ecl_free_parse_context(context);
    }
*/
/*
  buf1 = (char *)kmalloc (DATASIZE,GFP_KERNEL);
  memset(buf1, 'A', DATASIZE);
  buf1[DATASIZE-1]='\0';

  printk("Create CManager\n");
  cm1 = CManager_create();

  printk("Init CM\n");
  cc1 = ECho_CM_init(cm1);

  printk("Create Channel\n");
  chan1 = EChannel_create(cc1);

  channel_id = ECglobal_id(chan1);
  printk("channel id is %s\n", channel_id);

  host_name = (char *) kmalloc (MAXHOSTNAMELEN, GFP_KERNEL);
  getthishostname(host_name, MAXHOSTNAMELEN);

  printk("Setup Comm Group\n");
  if ((group_id = setup_comm_group(NULL, APP_NAME1, GRP_TYPE1, 40,
	host_name, 5010)) == NULL) 
	  return -1;
  printk("Assoc Comm Group\n");
  assoc_group_info(group_id, channel_id, strlen(channel_id) + 1);

  printk("Subscribe as Source\n");
  handle1 = ECsource_subscribe(chan1);

  //printk("Subscribe as Sink\n");
  //(void) ECsink_subscribe(chan1, shandler1, NULL);

  printk("Start Work\n");
  while (loop_flag == 1) {
    CMpoll_network(cm1);
    if (sleep_time == 0) {
      printk("transmit data\n");
      ECsubmit_event(handle1, &a, sizeof(a));
      sleep_time = 2 * HZ; 
      a++;
    }
    sleep_time = CMtask_sleep(sleep_time);
  }
*/

  exit_kthread(&sender_thread);

  return 0;
}

/* Initialize module. */
int init_module(void) 
{
  if (start_kthread(send_data, &sender_thread) <0) {
    printk("Thread creation failed!\n");
    return -1;
  }
  return 0;
}

/* Remove module. */
void cleanup_module(void) 
{
  loop_flag = 0;
  stop_kthread(&sender_thread);
}

