#ifdef MODULE
#ifndef __KERNEL__
#define __KERNEL__
#endif
#include <linux/kernel.h>
#include <linux/module.h>
#endif
#include "ecl.h"
#include "ecl_internal.h"
#include "structs.h"
#ifndef MODULE
#include <malloc.h>
#include <string.h>
#else
#include <linux/string.h>
#include "kecl.h"
#define malloc (void *)DAllocMM
#define free(a) DFreeMM((addrs_t)a)
#define fprintf(fmt, args...) printk(args)
#define printf printk
#endif
#ifndef NULL
#define NULL 0
#endif
extern sm_ref
ecl_new_selection_statement()
{
    sm_ref tmp = malloc(sizeof(*tmp));
    memset(tmp, 0, sizeof(*tmp));
    tmp->visited = 0;
    tmp->node_type = ecl_selection_statement;
    return tmp;
}

extern sm_ref
ecl_new_assignment_expression()
{
    sm_ref tmp = malloc(sizeof(*tmp));
    memset(tmp, 0, sizeof(*tmp));
    tmp->visited = 0;
    tmp->node_type = ecl_assignment_expression;
    return tmp;
}

extern sm_ref
ecl_new_type_specifier()
{
    sm_ref tmp = malloc(sizeof(*tmp));
    memset(tmp, 0, sizeof(*tmp));
    tmp->visited = 0;
    tmp->node_type = ecl_type_specifier;
    return tmp;
}

extern sm_ref
ecl_new_constant()
{
    sm_ref tmp = malloc(sizeof(*tmp));
    memset(tmp, 0, sizeof(*tmp));
    tmp->visited = 0;
    tmp->node_type = ecl_constant;
    return tmp;
}

extern sm_ref
ecl_new_compound_statement()
{
    sm_ref tmp = malloc(sizeof(*tmp));
    memset(tmp, 0, sizeof(*tmp));
    tmp->visited = 0;
    tmp->node_type = ecl_compound_statement;
    return tmp;
}

extern sm_ref
ecl_new_field()
{
    sm_ref tmp = malloc(sizeof(*tmp));
    memset(tmp, 0, sizeof(*tmp));
    tmp->visited = 0;
    tmp->node_type = ecl_field;
    return tmp;
}

extern sm_ref
ecl_new_struct_type_decl()
{
    sm_ref tmp = malloc(sizeof(*tmp));
    memset(tmp, 0, sizeof(*tmp));
    tmp->visited = 0;
    tmp->node_type = ecl_struct_type_decl;
    return tmp;
}

extern sm_ref
ecl_new_element_ref()
{
    sm_ref tmp = malloc(sizeof(*tmp));
    memset(tmp, 0, sizeof(*tmp));
    tmp->visited = 0;
    tmp->node_type = ecl_element_ref;
    return tmp;
}

extern sm_ref
ecl_new_declaration()
{
    sm_ref tmp = malloc(sizeof(*tmp));
    memset(tmp, 0, sizeof(*tmp));
    tmp->visited = 0;
    tmp->node_type = ecl_declaration;
    return tmp;
}

extern sm_ref
ecl_new_identifier()
{
    sm_ref tmp = malloc(sizeof(*tmp));
    memset(tmp, 0, sizeof(*tmp));
    tmp->visited = 0;
    tmp->node_type = ecl_identifier;
    return tmp;
}

extern sm_ref
ecl_new_operator()
{
    sm_ref tmp = malloc(sizeof(*tmp));
    memset(tmp, 0, sizeof(*tmp));
    tmp->visited = 0;
    tmp->node_type = ecl_operator;
    return tmp;
}

extern sm_ref
ecl_new_return_statement()
{
    sm_ref tmp = malloc(sizeof(*tmp));
    memset(tmp, 0, sizeof(*tmp));
    tmp->visited = 0;
    tmp->node_type = ecl_return_statement;
    return tmp;
}

extern sm_ref
ecl_new_field_ref()
{
    sm_ref tmp = malloc(sizeof(*tmp));
    memset(tmp, 0, sizeof(*tmp));
    tmp->visited = 0;
    tmp->node_type = ecl_field_ref;
    return tmp;
}

extern sm_ref
ecl_new_subroutine_call()
{
    sm_ref tmp = malloc(sizeof(*tmp));
    memset(tmp, 0, sizeof(*tmp));
    tmp->visited = 0;
    tmp->node_type = ecl_subroutine_call;
    return tmp;
}

extern sm_ref
ecl_new_expression_statement()
{
    sm_ref tmp = malloc(sizeof(*tmp));
    memset(tmp, 0, sizeof(*tmp));
    tmp->visited = 0;
    tmp->node_type = ecl_expression_statement;
    return tmp;
}

extern sm_ref
ecl_new_cast()
{
    sm_ref tmp = malloc(sizeof(*tmp));
    memset(tmp, 0, sizeof(*tmp));
    tmp->visited = 0;
    tmp->node_type = ecl_cast;
    return tmp;
}

extern sm_ref
ecl_new_iteration_statement()
{
    sm_ref tmp = malloc(sizeof(*tmp));
    memset(tmp, 0, sizeof(*tmp));
    tmp->visited = 0;
    tmp->node_type = ecl_iteration_statement;
    return tmp;
}

static void ecl_apply_list(sm_list node, ecl_apply_func pre_func, ecl_apply_func post_func, ecl_apply_list_func list_func, void *data)
{
    sm_list orig = node;
    while (node != NULL) {
        ecl_apply(node->node, pre_func, post_func, list_func, data);
        node = node->next;
    }
    if (list_func) (list_func)(orig, data);
}

extern void ecl_apply(sm_ref node, ecl_apply_func pre_func, ecl_apply_func post_func, ecl_apply_list_func list_func, void *data)
{
    if (node == NULL) return;
    if (node->visited) return;
    node->visited++;
    if(pre_func) (pre_func)(node, data);
    switch(node->node_type) {
      case ecl_selection_statement: {
          ecl_apply(node->node.selection_statement.conditional, pre_func, post_func, list_func, data);
          ecl_apply(node->node.selection_statement.then_part, pre_func, post_func, list_func, data);
          ecl_apply(node->node.selection_statement.else_part, pre_func, post_func, list_func, data);
          break;
      }
      case ecl_assignment_expression: {
          ecl_apply(node->node.assignment_expression.right, pre_func, post_func, list_func, data);
          ecl_apply(node->node.assignment_expression.left, pre_func, post_func, list_func, data);
          break;
      }
      case ecl_type_specifier: {
          break;
      }
      case ecl_constant: {
          break;
      }
      case ecl_compound_statement: {
          ecl_apply_list(node->node.compound_statement.decls, pre_func, post_func, list_func, data);
          ecl_apply_list(node->node.compound_statement.statements, pre_func, post_func, list_func, data);
          break;
      }
      case ecl_field: {
          ecl_apply_list(node->node.field.type_spec, pre_func, post_func, list_func, data);
          break;
      }
      case ecl_struct_type_decl: {
          ecl_apply_list(node->node.struct_type_decl.fields, pre_func, post_func, list_func, data);
          break;
      }
      case ecl_element_ref: {
          ecl_apply(node->node.element_ref.expression, pre_func, post_func, list_func, data);
          ecl_apply(node->node.element_ref.array_ref, pre_func, post_func, list_func, data);
          break;
      }
      case ecl_declaration: {
          ecl_apply_list(node->node.declaration.params, pre_func, post_func, list_func, data);
          ecl_apply(node->node.declaration.init_value, pre_func, post_func, list_func, data);
          ecl_apply_list(node->node.declaration.type_spec, pre_func, post_func, list_func, data);
          break;
      }
      case ecl_identifier: {
          break;
      }
      case ecl_operator: {
          ecl_apply(node->node.operator.right, pre_func, post_func, list_func, data);
          ecl_apply(node->node.operator.left, pre_func, post_func, list_func, data);
          break;
      }
      case ecl_return_statement: {
          ecl_apply(node->node.return_statement.expression, pre_func, post_func, list_func, data);
          break;
      }
      case ecl_field_ref: {
          ecl_apply(node->node.field_ref.struct_ref, pre_func, post_func, list_func, data);
          break;
      }
      case ecl_subroutine_call: {
          ecl_apply_list(node->node.subroutine_call.arguments, pre_func, post_func, list_func, data);
          break;
      }
      case ecl_expression_statement: {
          ecl_apply(node->node.expression_statement.expression, pre_func, post_func, list_func, data);
          break;
      }
      case ecl_cast: {
          ecl_apply_list(node->node.cast.type_spec, pre_func, post_func, list_func, data);
          ecl_apply(node->node.cast.expression, pre_func, post_func, list_func, data);
          break;
      }
      case ecl_iteration_statement: {
          ecl_apply(node->node.iteration_statement.init_expr, pre_func, post_func, list_func, data);
          ecl_apply(node->node.iteration_statement.test_expr, pre_func, post_func, list_func, data);
          ecl_apply(node->node.iteration_statement.statement, pre_func, post_func, list_func, data);
          ecl_apply(node->node.iteration_statement.iter_expr, pre_func, post_func, list_func, data);
          break;
      }
    }
    node->visited--;
    if(post_func) (post_func)(node, data);
}

static void ecl_print_sm_list(sm_list list)
{
    while (list != NULL) {
        printf(" %lx", (long) list->node);
        list = list->next;
    }
}

extern void ecl_print(sm_ref node)
{
    switch(node->node_type) {
      case ecl_selection_statement: {
          printf("0x%lx  --  selection_statement ->\n", (long)node);
          printf("	conditional : %lx\n", (long) node->node.selection_statement.conditional);
          printf("	then_part : %lx\n", (long) node->node.selection_statement.then_part);
          printf("	else_part : %lx\n", (long) node->node.selection_statement.else_part);
          printf("	lx_srcpos : ");
          ecl_print_srcpos(node->node.selection_statement.lx_srcpos);
          printf("\n");
          break;
      }
      case ecl_assignment_expression: {
          printf("0x%lx  --  assignment_expression ->\n", (long)node);
          printf("	cg_type : %d\n", node->node.assignment_expression.cg_type);
          printf("	right : %lx\n", (long) node->node.assignment_expression.right);
          printf("	lx_srcpos : ");
          ecl_print_srcpos(node->node.assignment_expression.lx_srcpos);
          printf("\n");
          printf("	left : %lx\n", (long) node->node.assignment_expression.left);
          break;
      }
      case ecl_type_specifier: {
          printf("0x%lx  --  type_specifier ->\n", (long)node);
          printf("	token : %d\n", node->node.type_specifier.token);
          printf("	lx_srcpos : ");
          ecl_print_srcpos(node->node.type_specifier.lx_srcpos);
          printf("\n");
          break;
      }
      case ecl_constant: {
          printf("0x%lx  --  constant ->\n", (long)node);
          printf("	token : %d\n", node->node.constant.token);
          printf("	const_val : %s\n", node->node.constant.const_val);
          printf("	lx_srcpos : ");
          ecl_print_srcpos(node->node.constant.lx_srcpos);
          printf("\n");
          break;
      }
      case ecl_compound_statement: {
          printf("0x%lx  --  compound_statement ->\n", (long)node);
          printf("	decls : ");
          ecl_print_sm_list(node->node.compound_statement.decls);
          printf("\n");
          printf("	statements : ");
          ecl_print_sm_list(node->node.compound_statement.statements);
          printf("\n");
          break;
      }
      case ecl_field: {
          printf("0x%lx  --  field ->\n", (long)node);
          printf("	sm_complex_type : %lx\n", (long) node->node.field.sm_complex_type);
          printf("	cg_offset : %d\n", node->node.field.cg_offset);
          printf("	cg_size : %d\n", node->node.field.cg_size);
          printf("	dimen2 : %d\n", node->node.field.dimen2);
          printf("	sm_var_control_field : %lx\n", (long) node->node.field.sm_var_control_field);
          printf("	string_type : %s\n", node->node.field.string_type);
          printf("	name : %s\n", node->node.field.name);
          printf("	cg_type : %d\n", node->node.field.cg_type);
          printf("	type_spec : ");
          ecl_print_sm_list(node->node.field.type_spec);
          printf("\n");
          printf("	element_size : %d\n", node->node.field.element_size);
          printf("	dimen1 : %d\n", node->node.field.dimen1);
          break;
      }
      case ecl_struct_type_decl: {
          printf("0x%lx  --  struct_type_decl ->\n", (long)node);
          printf("	id : %s\n", node->node.struct_type_decl.id);
          printf("	fields : ");
          ecl_print_sm_list(node->node.struct_type_decl.fields);
          printf("\n");
          printf("	cg_size : %d\n", node->node.struct_type_decl.cg_size);
          printf("	lx_srcpos : ");
          ecl_print_srcpos(node->node.struct_type_decl.lx_srcpos);
          printf("\n");
          break;
      }
      case ecl_element_ref: {
          printf("0x%lx  --  element_ref ->\n", (long)node);
          printf("	expression : %lx\n", (long) node->node.element_ref.expression);
          printf("	lx_srcpos : ");
          ecl_print_srcpos(node->node.element_ref.lx_srcpos);
          printf("\n");
          printf("	array_ref : %lx\n", (long) node->node.element_ref.array_ref);
          break;
      }
      case ecl_declaration: {
          printf("0x%lx  --  declaration ->\n", (long)node);
          printf("	id : %s\n", node->node.declaration.id);
          printf("	sm_complex_type : %lx\n", (long) node->node.declaration.sm_complex_type);
          printf("	cg_oprnd : %d\n", node->node.declaration.cg_oprnd);
          printf("	params : ");
          ecl_print_sm_list(node->node.declaration.params);
          printf("\n");
          printf("	init_value : %lx\n", (long) node->node.declaration.init_value);
          printf("	cg_address : %lx\n", (long)node->node.declaration.cg_address);
          printf("	cg_type : %d\n", node->node.declaration.cg_type);
          printf("	type_spec : ");
          ecl_print_sm_list(node->node.declaration.type_spec);
          printf("\n");
          printf("	static_var : %d\n", node->node.declaration.static_var);
          printf("	lx_srcpos : ");
          ecl_print_srcpos(node->node.declaration.lx_srcpos);
          printf("\n");
          printf("	param_num : %d\n", node->node.declaration.param_num);
          printf("	is_subroutine : %d\n", node->node.declaration.is_subroutine);
          break;
      }
      case ecl_identifier: {
          printf("0x%lx  --  identifier ->\n", (long)node);
          printf("	sm_declaration : %lx\n", (long) node->node.identifier.sm_declaration);
          printf("	id : %s\n", node->node.identifier.id);
          printf("	cg_type : %d\n", node->node.identifier.cg_type);
          printf("	lx_srcpos : ");
          ecl_print_srcpos(node->node.identifier.lx_srcpos);
          printf("\n");
          break;
      }
      case ecl_operator: {
          printf("0x%lx  --  operator ->\n", (long)node);
          printf("	op : ");
          ecl_print_operator_t(node->node.operator.op);
          printf("\n");
          printf("	operation_type : %d\n", node->node.operator.operation_type);
          printf("	result_type : %d\n", node->node.operator.result_type);
          printf("	right : %lx\n", (long) node->node.operator.right);
          printf("	lx_srcpos : ");
          ecl_print_srcpos(node->node.operator.lx_srcpos);
          printf("\n");
          printf("	left : %lx\n", (long) node->node.operator.left);
          break;
      }
      case ecl_return_statement: {
          printf("0x%lx  --  return_statement ->\n", (long)node);
          printf("	expression : %lx\n", (long) node->node.return_statement.expression);
          printf("	lx_srcpos : ");
          ecl_print_srcpos(node->node.return_statement.lx_srcpos);
          printf("\n");
          printf("	cg_func_type : %d\n", node->node.return_statement.cg_func_type);
          break;
      }
      case ecl_field_ref: {
          printf("0x%lx  --  field_ref ->\n", (long)node);
          printf("	lx_field : %s\n", node->node.field_ref.lx_field);
          printf("	sm_field_ref : %lx\n", (long) node->node.field_ref.sm_field_ref);
          printf("	lx_srcpos : ");
          ecl_print_srcpos(node->node.field_ref.lx_srcpos);
          printf("\n");
          printf("	struct_ref : %lx\n", (long) node->node.field_ref.struct_ref);
          break;
      }
      case ecl_subroutine_call: {
          printf("0x%lx  --  subroutine_call ->\n", (long)node);
          printf("	arguments : ");
          ecl_print_sm_list(node->node.subroutine_call.arguments);
          printf("\n");
          printf("	lx_srcpos : ");
          ecl_print_srcpos(node->node.subroutine_call.lx_srcpos);
          printf("\n");
          printf("	sm_func_ref : %lx\n", (long) node->node.subroutine_call.sm_func_ref);
          break;
      }
      case ecl_expression_statement: {
          printf("0x%lx  --  expression_statement ->\n", (long)node);
          printf("	expression : %lx\n", (long) node->node.expression_statement.expression);
          break;
      }
      case ecl_cast: {
          printf("0x%lx  --  cast ->\n", (long)node);
          printf("	cg_type : %d\n", node->node.cast.cg_type);
          printf("	type_spec : ");
          ecl_print_sm_list(node->node.cast.type_spec);
          printf("\n");
          printf("	expression : %lx\n", (long) node->node.cast.expression);
          printf("	lx_srcpos : ");
          ecl_print_srcpos(node->node.cast.lx_srcpos);
          printf("\n");
          break;
      }
      case ecl_iteration_statement: {
          printf("0x%lx  --  iteration_statement ->\n", (long)node);
          printf("	init_expr : %lx\n", (long) node->node.iteration_statement.init_expr);
          printf("	test_expr : %lx\n", (long) node->node.iteration_statement.test_expr);
          printf("	statement : %lx\n", (long) node->node.iteration_statement.statement);
          printf("	lx_srcpos : ");
          ecl_print_srcpos(node->node.iteration_statement.lx_srcpos);
          printf("\n");
          printf("	iter_expr : %lx\n", (long) node->node.iteration_statement.iter_expr);
          break;
      }
    }
    printf("\n");
}

extern void ecl_free_list(sm_list list, void *junk)
{
    while (list != NULL) {
        sm_list next = list->next;
        free(list);
        list = next;
    }
}

extern void ecl_free(sm_ref node)
{
    switch(node->node_type) {
      case ecl_selection_statement: {
          break;
      }
      case ecl_assignment_expression: {
          break;
      }
      case ecl_type_specifier: {
          break;
      }
      case ecl_constant: {
	    free(node->node.constant.const_val);
          break;
      }
      case ecl_compound_statement: {
          break;
      }
      case ecl_field: {
	    free(node->node.field.string_type);
	    free(node->node.field.name);
          break;
      }
      case ecl_struct_type_decl: {
	    free(node->node.struct_type_decl.id);
          break;
      }
      case ecl_element_ref: {
          break;
      }
      case ecl_declaration: {
	    free(node->node.declaration.id);
          break;
      }
      case ecl_identifier: {
	    free(node->node.identifier.id);
          break;
      }
      case ecl_operator: {
          break;
      }
      case ecl_return_statement: {
          break;
      }
      case ecl_field_ref: {
	    free(node->node.field_ref.lx_field);
          break;
      }
      case ecl_subroutine_call: {
          break;
      }
      case ecl_expression_statement: {
          break;
      }
      case ecl_cast: {
          break;
      }
      case ecl_iteration_statement: {
          break;
      }
    }
    free(node);
}

static sm_list free_list = NULL;
extern void ecl_make_free(sm_ref node, void *junk)
{
    sm_list new_free = malloc(sizeof(*new_free));
    new_free->next = free_list;
    new_free->node = node;
    free_list = new_free;
    switch(node->node_type) {
      case ecl_selection_statement: {
          break;
      }
      case ecl_assignment_expression: {
          break;
      }
      case ecl_type_specifier: {
          break;
      }
      case ecl_constant: {
          break;
      }
      case ecl_compound_statement: {
          node->node.compound_statement.decls = NULL;
          node->node.compound_statement.statements = NULL;
          break;
      }
      case ecl_field: {
          node->node.field.type_spec = NULL;
          break;
      }
      case ecl_struct_type_decl: {
          node->node.struct_type_decl.fields = NULL;
          break;
      }
      case ecl_element_ref: {
          break;
      }
      case ecl_declaration: {
          node->node.declaration.params = NULL;
          node->node.declaration.type_spec = NULL;
          break;
      }
      case ecl_identifier: {
          break;
      }
      case ecl_operator: {
          break;
      }
      case ecl_return_statement: {
          break;
      }
      case ecl_field_ref: {
          break;
      }
      case ecl_subroutine_call: {
          node->node.subroutine_call.arguments = NULL;
          break;
      }
      case ecl_expression_statement: {
          break;
      }
      case ecl_cast: {
          node->node.cast.type_spec = NULL;
          break;
      }
      case ecl_iteration_statement: {
          break;
      }
    }
}

extern void ecl_rfree(sm_ref node) {
    free_list = NULL;
    ecl_apply(node, NULL, ecl_make_free, ecl_free_list, NULL);
    while(free_list != NULL) {
        sm_list next = free_list->next;
        ecl_free(free_list->node);
        free(free_list);
        free_list = next;
    }
}
extern void ecl_rfree_list(sm_list list, void *junk) {
    free_list = NULL;
    ecl_apply_list(list, NULL, ecl_make_free, ecl_free_list, junk);
    while(free_list != NULL) {
        sm_list next = free_list->next;
        ecl_free(free_list->node);
        free(free_list);
        free_list = next;
    }
}
extern sm_list ecl_copy_list(sm_list list)
{
    sm_list new_list = NULL;
    if (list != NULL) {
        new_list = malloc(sizeof(*new_list));
        new_list->node = ecl_copy(list->node);
        new_list->next = ecl_copy_list(list->next);
    }
    return new_list;
}

extern sm_ref ecl_copy(sm_ref node)
{
    sm_ref new_node = NULL;
    if (node == NULL) return NULL;

    switch(node->node_type) {
      case ecl_selection_statement: {
	    new_node = ecl_new_selection_statement();
	    new_node->node.selection_statement = node->node.selection_statement;
	    new_node->node.selection_statement.conditional = ecl_copy(node->node.selection_statement.conditional);
	    new_node->node.selection_statement.then_part = ecl_copy(node->node.selection_statement.then_part);
	    new_node->node.selection_statement.else_part = ecl_copy(node->node.selection_statement.else_part);
          break;
      }
      case ecl_assignment_expression: {
	    new_node = ecl_new_assignment_expression();
	    new_node->node.assignment_expression = node->node.assignment_expression;
	    new_node->node.assignment_expression.right = ecl_copy(node->node.assignment_expression.right);
	    new_node->node.assignment_expression.left = ecl_copy(node->node.assignment_expression.left);
          break;
      }
      case ecl_type_specifier: {
	    new_node = ecl_new_type_specifier();
	    new_node->node.type_specifier = node->node.type_specifier;
          break;
      }
      case ecl_constant: {
	    new_node = ecl_new_constant();
	    new_node->node.constant = node->node.constant;
	    new_node->node.constant.const_val = strdup(node->node.constant.const_val);
          break;
      }
      case ecl_compound_statement: {
	    new_node = ecl_new_compound_statement();
	    new_node->node.compound_statement = node->node.compound_statement;
	    new_node->node.compound_statement.decls = ecl_copy_list(node->node.compound_statement.decls);
	    new_node->node.compound_statement.statements = ecl_copy_list(node->node.compound_statement.statements);
          break;
      }
      case ecl_field: {
	    new_node = ecl_new_field();
	    new_node->node.field = node->node.field;
	    new_node->node.field.string_type = strdup(node->node.field.string_type);
	    new_node->node.field.name = strdup(node->node.field.name);
	    new_node->node.field.type_spec = ecl_copy_list(node->node.field.type_spec);
          break;
      }
      case ecl_struct_type_decl: {
	    new_node = ecl_new_struct_type_decl();
	    new_node->node.struct_type_decl = node->node.struct_type_decl;
	    new_node->node.struct_type_decl.id = strdup(node->node.struct_type_decl.id);
	    new_node->node.struct_type_decl.fields = ecl_copy_list(node->node.struct_type_decl.fields);
          break;
      }
      case ecl_element_ref: {
	    new_node = ecl_new_element_ref();
	    new_node->node.element_ref = node->node.element_ref;
	    new_node->node.element_ref.expression = ecl_copy(node->node.element_ref.expression);
	    new_node->node.element_ref.array_ref = ecl_copy(node->node.element_ref.array_ref);
          break;
      }
      case ecl_declaration: {
	    new_node = ecl_new_declaration();
	    new_node->node.declaration = node->node.declaration;
	    new_node->node.declaration.id = strdup(node->node.declaration.id);
	    new_node->node.declaration.params = ecl_copy_list(node->node.declaration.params);
	    new_node->node.declaration.init_value = ecl_copy(node->node.declaration.init_value);
	    new_node->node.declaration.type_spec = ecl_copy_list(node->node.declaration.type_spec);
          break;
      }
      case ecl_identifier: {
	    new_node = ecl_new_identifier();
	    new_node->node.identifier = node->node.identifier;
	    new_node->node.identifier.id = strdup(node->node.identifier.id);
          break;
      }
      case ecl_operator: {
	    new_node = ecl_new_operator();
	    new_node->node.operator = node->node.operator;
	    new_node->node.operator.right = ecl_copy(node->node.operator.right);
	    new_node->node.operator.left = ecl_copy(node->node.operator.left);
          break;
      }
      case ecl_return_statement: {
	    new_node = ecl_new_return_statement();
	    new_node->node.return_statement = node->node.return_statement;
	    new_node->node.return_statement.expression = ecl_copy(node->node.return_statement.expression);
          break;
      }
      case ecl_field_ref: {
	    new_node = ecl_new_field_ref();
	    new_node->node.field_ref = node->node.field_ref;
	    new_node->node.field_ref.lx_field = strdup(node->node.field_ref.lx_field);
	    new_node->node.field_ref.struct_ref = ecl_copy(node->node.field_ref.struct_ref);
          break;
      }
      case ecl_subroutine_call: {
	    new_node = ecl_new_subroutine_call();
	    new_node->node.subroutine_call = node->node.subroutine_call;
	    new_node->node.subroutine_call.arguments = ecl_copy_list(node->node.subroutine_call.arguments);
          break;
      }
      case ecl_expression_statement: {
	    new_node = ecl_new_expression_statement();
	    new_node->node.expression_statement = node->node.expression_statement;
	    new_node->node.expression_statement.expression = ecl_copy(node->node.expression_statement.expression);
          break;
      }
      case ecl_cast: {
	    new_node = ecl_new_cast();
	    new_node->node.cast = node->node.cast;
	    new_node->node.cast.type_spec = ecl_copy_list(node->node.cast.type_spec);
	    new_node->node.cast.expression = ecl_copy(node->node.cast.expression);
          break;
      }
      case ecl_iteration_statement: {
	    new_node = ecl_new_iteration_statement();
	    new_node->node.iteration_statement = node->node.iteration_statement;
	    new_node->node.iteration_statement.init_expr = ecl_copy(node->node.iteration_statement.init_expr);
	    new_node->node.iteration_statement.test_expr = ecl_copy(node->node.iteration_statement.test_expr);
	    new_node->node.iteration_statement.statement = ecl_copy(node->node.iteration_statement.statement);
	    new_node->node.iteration_statement.iter_expr = ecl_copy(node->node.iteration_statement.iter_expr);
          break;
      }
    }
    return new_node;
}

extern srcpos ecl_get_srcpos(expr)
sm_ref expr;
{
    switch(expr->node_type) {
      case ecl_selection_statement: return expr->node.selection_statement.lx_srcpos;
      case ecl_assignment_expression: return expr->node.assignment_expression.lx_srcpos;
      case ecl_type_specifier: return expr->node.type_specifier.lx_srcpos;
      case ecl_constant: return expr->node.constant.lx_srcpos;
      case ecl_struct_type_decl: return expr->node.struct_type_decl.lx_srcpos;
      case ecl_element_ref: return expr->node.element_ref.lx_srcpos;
      case ecl_declaration: return expr->node.declaration.lx_srcpos;
      case ecl_identifier: return expr->node.identifier.lx_srcpos;
      case ecl_operator: return expr->node.operator.lx_srcpos;
      case ecl_return_statement: return expr->node.return_statement.lx_srcpos;
      case ecl_field_ref: return expr->node.field_ref.lx_srcpos;
      case ecl_subroutine_call: return expr->node.subroutine_call.lx_srcpos;
      case ecl_cast: return expr->node.cast.lx_srcpos;
      case ecl_iteration_statement: return expr->node.iteration_statement.lx_srcpos;
      default: {
          srcpos tmp;
          tmp.line = 0;
          tmp.character = 0;
          return tmp;
       };
    };
}
