#ifndef __STRUCTS_H
#define __STRUCTS_H
typedef struct list_struct {
    sm_ref node;
    struct list_struct *next;
} *sm_list;
typedef enum {
    ecl_selection_statement,
    ecl_assignment_expression,
    ecl_type_specifier,
    ecl_constant,
    ecl_compound_statement,
    ecl_field,
    ecl_struct_type_decl,
    ecl_element_ref,
    ecl_declaration,
    ecl_identifier,
    ecl_operator,
    ecl_return_statement,
    ecl_field_ref,
    ecl_subroutine_call,
    ecl_expression_statement,
    ecl_cast,
    ecl_iteration_statement,
    ecl_last_node_type
} ecl_node_type;

typedef struct {
    sm_ref conditional;
    sm_ref then_part;
    sm_ref else_part;
    srcpos lx_srcpos;
} selection_statement;

typedef struct {
    int cg_type;
    sm_ref right;
    srcpos lx_srcpos;
    sm_ref left;
} assignment_expression;

typedef struct {
    int token;
    srcpos lx_srcpos;
} type_specifier;

typedef struct {
    int token;
    char* const_val;
    srcpos lx_srcpos;
} constant;

typedef struct {
    sm_list decls;
    sm_list statements;
} compound_statement;

typedef struct {
    sm_ref sm_complex_type;
    int cg_offset;
    int cg_size;
    int dimen2;
    sm_ref sm_var_control_field;
    char* string_type;
    char* name;
    int cg_type;
    sm_list type_spec;
    int element_size;
    int dimen1;
} field;

typedef struct {
    char* id;
    sm_list fields;
    int cg_size;
    srcpos lx_srcpos;
} struct_type_decl;

typedef struct {
    sm_ref expression;
    srcpos lx_srcpos;
    sm_ref array_ref;
} element_ref;

typedef struct {
    char* id;
    sm_ref sm_complex_type;
    int cg_oprnd;
    sm_list params;
    sm_ref init_value;
    void* cg_address;
    int cg_type;
    sm_list type_spec;
    int static_var;
    srcpos lx_srcpos;
    int param_num;
    int is_subroutine;
} declaration;

typedef struct {
    sm_ref sm_declaration;
    char* id;
    int cg_type;
    srcpos lx_srcpos;
} identifier;

typedef struct {
    operator_t op;
    int operation_type;
    int result_type;
    sm_ref right;
    srcpos lx_srcpos;
    sm_ref left;
} operator;

typedef struct {
    sm_ref expression;
    srcpos lx_srcpos;
    int cg_func_type;
} return_statement;

typedef struct {
    char* lx_field;
    sm_ref sm_field_ref;
    srcpos lx_srcpos;
    sm_ref struct_ref;
} field_ref;

typedef struct {
    sm_list arguments;
    srcpos lx_srcpos;
    sm_ref sm_func_ref;
} subroutine_call;

typedef struct {
    sm_ref expression;
} expression_statement;

typedef struct {
    int cg_type;
    sm_list type_spec;
    sm_ref expression;
    srcpos lx_srcpos;
} cast;

typedef struct {
    sm_ref init_expr;
    sm_ref test_expr;
    sm_ref statement;
    srcpos lx_srcpos;
    sm_ref iter_expr;
} iteration_statement;

typedef union {
   selection_statement selection_statement;
   assignment_expression assignment_expression;
   type_specifier type_specifier;
   constant constant;
   compound_statement compound_statement;
   field field;
   struct_type_decl struct_type_decl;
   element_ref element_ref;
   declaration declaration;
   identifier identifier;
   operator operator;
   return_statement return_statement;
   field_ref field_ref;
   subroutine_call subroutine_call;
   expression_statement expression_statement;
   cast cast;
   iteration_statement iteration_statement;
} sm_union;

struct sm_struct {
    ecl_node_type node_type;
    int visited;
    sm_union node;
};
#endif
extern sm_ref ecl_new_selection_statement();
extern sm_ref ecl_new_assignment_expression();
extern sm_ref ecl_new_type_specifier();
extern sm_ref ecl_new_constant();
extern sm_ref ecl_new_compound_statement();
extern sm_ref ecl_new_field();
extern sm_ref ecl_new_struct_type_decl();
extern sm_ref ecl_new_element_ref();
extern sm_ref ecl_new_declaration();
extern sm_ref ecl_new_identifier();
extern sm_ref ecl_new_operator();
extern sm_ref ecl_new_return_statement();
extern sm_ref ecl_new_field_ref();
extern sm_ref ecl_new_subroutine_call();
extern sm_ref ecl_new_expression_statement();
extern sm_ref ecl_new_cast();
extern sm_ref ecl_new_iteration_statement();
typedef void (*ecl_apply_func)(sm_ref node, void *data);
typedef void (*ecl_apply_list_func)(sm_list list, void *data);
extern void ecl_apply(sm_ref node, ecl_apply_func pre_func, ecl_apply_func post_func, ecl_apply_list_func list_func, void *data);
extern void ecl_print(sm_ref node);
extern void ecl_free(sm_ref node);
extern void ecl_free_list(sm_list list, void *junk);
extern void ecl_rfree(sm_ref node);
extern void ecl_rfree_list(sm_list list, void *junk);
extern sm_ref ecl_copy(sm_ref node);
extern sm_list ecl_copy_list(sm_list list);
extern srcpos ecl_get_srcpos(sm_ref expr);
