
#ifdef MODULE
#ifndef __KERNEL__
#define __KERNEL__
#endif
#endif

#ifndef MODULE
#include "config.h"
#include "io.h"
#include "gen_thread.h"
#include <atl.h>
#include "cm.h"
#include "cm_internal.h"
#include "assert.h"
#include <stdio.h>
#else
#include "config.h"
#include "../include/io.h"
#include "../include/atl.h"
#include "cm.h"
#include "cm_internal.h"
#include "kcm.h"
#include "library.h"
#include "assert.h"
#endif

extern void
IntCManager_lock(cm, file, line)
CManager cm;
char *file;
int line;
{
    CMtrace_out(cm, CMLowLevelVerbose, "CManager Lock at \"%s\" line %d",
		file, line);
#ifdef MODULE
    down(&cm->exchange_lock);
#else
    if (gen_thr_initialized()) {
	if (cm->exchange_lock == NULL) {
	    cm->exchange_lock = thr_mutex_alloc();
	}
	thr_mutex_lock(cm->exchange_lock);
    }
#endif
    cm->locked++;
    assert(cm->locked == 1);
}

extern void
IntCManager_unlock(cm, file, line)
CManager cm;
char *file;
int line;
{
    CMtrace_out(cm, CMLowLevelVerbose, "CManager Unlock at \"%s\" line %d",
		file, line);
    cm->locked--;
    assert(cm->locked == 0);
#ifdef MODULE
    up(&cm->exchange_lock);
#else
    if (gen_thr_initialized() && cm->exchange_lock) {
	thr_mutex_unlock(cm->exchange_lock);
    }
#endif
}

extern int
CManager_locked(cm)
CManager cm;
{
    return cm->locked;
}

extern void
CMConn_write_lock(conn)
CMConnection conn;
{
#ifdef MODULE
    down(&conn->write_lock);
#else
    if (gen_thr_initialized()) {
	if (conn->write_lock == NULL) {
	    conn->write_lock = thr_mutex_alloc();
	}
	thr_mutex_lock(conn->write_lock);
    }
#endif
}

extern void
CMConn_write_unlock(conn)
CMConnection conn;
{
#ifdef MODULE
    up(&conn->write_lock);
#else
    if (gen_thr_initialized() && conn->write_lock) {
	thr_mutex_unlock(conn->write_lock);
    }
#endif
}

extern void
CMControlList_lock(cl)
CMControlList cl;
{
#ifdef MODULE
    down(&cl->list_mutex);
#else
    if (gen_thr_initialized()) {
	if (cl->list_mutex == NULL) {
	    cl->list_mutex = thr_mutex_alloc();
	}
	thr_mutex_lock(cl->list_mutex);
    }
#endif
    cl->locked++;
}

extern void
CMControlList_unlock(cl)
CMControlList cl;
{
    cl->locked--;
#ifdef MODULE
    up(&cl->list_mutex);
#else
    if (gen_thr_initialized() && cl->list_mutex) {
	thr_mutex_unlock(cl->list_mutex);
    }
#endif
}

extern int
CMControlList_locked(cl)
CMControlList cl;
{
    return cl->locked;
}

#ifdef MODULE
DECLARE_MUTEX(CMglobal_data_mutex);
#else
static thr_mutex_t CMglobal_data_mutex = NULL;
#endif
static int CMglobal_data_lock_val = 0;

extern void
CMglobal_data_lock()
{
#ifdef MODULE
    down(&CMglobal_data_mutex);
#else
    if (gen_thr_initialized()) {
	if (CMglobal_data_mutex == NULL) {
	    CMglobal_data_mutex = thr_mutex_alloc();
	}
	thr_mutex_lock(CMglobal_data_mutex);
    }
#endif
    CMglobal_data_lock_val++;
}

extern void
CMglobal_data_unlock()
{
    CMglobal_data_lock_val--;
#ifdef MODULE
    up(&CMglobal_data_mutex);
#else
    if (CMglobal_data_mutex != NULL) {
	thr_mutex_unlock(CMglobal_data_mutex);
    }
#endif
}

extern int
CMglobal_data_locked()
{
    return CMglobal_data_lock_val;
}


