
#ifdef MODULE
#ifndef __KERNEL__
#define __KERNEL__
#endif
#endif

#ifndef MODULE
#include "config.h"
#include "ltdl.h"
#include <stdio.h>
#include <string.h>
#include <atl.h>
#include "cm.h"
#include "cm_internal.h"
#include "cm_transport.h"
#include "assert.h"
#else
#include "../include/atl.h"
#include "cm.h"
#include "cm_internal.h"
#include "cm_transport.h"
#include "kcm.h"
#include "library.h"
#include "assert.h"

#define printf         printk
#define perror         printk
#define fprintf(fmt, args...) printk (args) 
#define malloc         (void *)DAllocMM
#define free(a)        DFreeMM((addrs_t)a)
#define realloc(a,b)   (void *)DReallocMM((addrs_t)a,b)
#define exit           sys_exit
#define qsort          _quicksort
#define getpid()       current->pid
#define signal(a,b)    
#define gethostname    sys_gethostname
#define gethostbyname  lookup_name
#define select         sys_select

typedef struct socket_client_data {
    CManager	cm;
    char        *hostname;
    int		listen_port;
    CMtrans_services svc;
} *socket_client_data_ptr;

typedef struct socket_connection_data {
    char *remote_host;
    int   remote_IP;
    int   remote_contact_port;
    int   fd;
    void *read_buffer;
    int   read_buffer_len;
    socket_client_data_ptr sd;
    struct task_struct *parent;
} *socket_conn_data_ptr;
#endif

extern struct CMtrans_services CMstatic_trans_svcs;
/*const lt_dlsymlist lt_preloaded_symbols[1] = { { 0, 0 } };*/

#ifdef MODULE
extern void *libcmsockets_LTX_initialize(CManager cm, CMtrans_services svc);
extern attr_list libcmsockets_LTX_non_blocking_listen(CManager cm, 
	CMtrans_services svc, transport_entry trans, attr_list listen_info);
extern CMConnection libcmsockets_LTX_initiate_conn(CManager cm, 
	CMtrans_services svc, transport_entry trans, attr_list attrs);
extern int libcmsockets_LTX_self_check(CManager cm, CMtrans_services svc,
	transport_entry trans, attr_list attrs);
extern int libcmsockets_LTX_connection_eq(CManager cm, CMtrans_services svc,
	transport_entry trans, attr_list attrs, socket_conn_data_ptr scd);
extern void libcmsockets_LTX_shutdown_conn(CMtrans_services svc, 
	socket_conn_data_ptr scd);
extern int libcmsockets_LTX_read_to_buffer_func(CMtrans_services svc,
	socket_conn_data_ptr scd, void *buffer, int requested_len);  
extern void *libcmsockets_LTX_read_func(CMtrans_services svc,  
	socket_conn_data_ptr scd, int requested_len, int *actual_len);  
extern int libcmsockets_LTX_write_func(CMtrans_services svc,  
	socket_conn_data_ptr scd, void *buffer, int length);  
extern int libcmsockets_LTX_writev_func(CMtrans_services svc,
	socket_conn_data_ptr scd, struct iovec *iov, int iovcnt);  
extern int libcmsockets_LTX_writev_attr_func(CMtrans_services svc,
	socket_conn_data_ptr scd, struct iovec *iov, int iovcnt, 
	attr_list attrs);
extern void *libcmrudp_LTX_initialize(CManager cm, CMtrans_services svc);
extern attr_list libcmrudp_LTX_non_blocking_listen(CManager cm, 
	CMtrans_services svc, transport_entry trans, attr_list listen_info);
extern CMConnection libcmrudp_LTX_initiate_conn(CManager cm, 
	CMtrans_services svc, transport_entry trans, attr_list attrs);
extern int libcmrudp_LTX_self_check(CManager cm, CMtrans_services svc,
	transport_entry trans, attr_list attrs);
extern int libcmrudp_LTX_connection_eq(CManager cm, CMtrans_services svc,
	transport_entry trans, attr_list attrs, socket_conn_data_ptr scd);
extern void libcmrudp_LTX_shutdown_conn(CMtrans_services svc, 
	socket_conn_data_ptr scd);
extern int libcmrudp_LTX_read_to_buffer_func(CMtrans_services svc,
	socket_conn_data_ptr scd, void *buffer, int requested_len);  
extern void *libcmrudp_LTX_read_func(CMtrans_services svc,  
	socket_conn_data_ptr scd, int requested_len, int *actual_len);  
extern int libcmrudp_LTX_write_func(CMtrans_services svc,  
	socket_conn_data_ptr scd, void *buffer, int length);  
extern int libcmrudp_LTX_writev_func(CMtrans_services svc,
	socket_conn_data_ptr scd, struct iovec *iov, int iovcnt);  
extern int libcmrudp_LTX_writev_attr_func(CMtrans_services svc,
	socket_conn_data_ptr scd, struct iovec *iov, int iovcnt, 
	attr_list attrs);
#endif

static transport_entry *global_transports = NULL;

static transport_entry
add_transport_to_cm(cm, transport)
CManager cm;
transport_entry transport;
{
    int num_trans;
    if (cm->transports == NULL) {
	cm->transports = CMmalloc(sizeof(transport_entry) * 2);
	num_trans = 0;
    } else {
	num_trans = 0;
	while(cm->transports[num_trans] != NULL) num_trans++;
	cm->transports = CMrealloc(cm->transports,
				   sizeof(transport_entry) * (num_trans +2));
    }
    cm->transports[num_trans] = CMmalloc(sizeof(struct _transport_item));
    *(cm->transports[num_trans]) = *transport;
    cm->transports[num_trans + 1] = NULL;
    transport = cm->transports[num_trans];
    transport->cm = cm;
    return transport;
}

int
load_transport(cm, trans_name)
CManager cm;
const char *trans_name;
{
    transport_entry *trans_list = global_transports;
    transport_entry transport;
    int i = 0;
#ifndef MODULE
    char *libname;
    lt_dlhandle handle;	

    assert(CMglobal_data_locked());  /* must be already locked */
#endif

    while ((trans_list != NULL) && (*trans_list != NULL)) {
	if (strcmp((*trans_list)->trans_name, trans_name) == 0) {
	    transport_entry trans = add_transport_to_cm(cm, *trans_list);
	    if (trans->transport_init) {
		trans->trans_data = 
		    trans->transport_init(cm, &CMstatic_trans_svcs);
	    }
	    return 1;
	}
	i++;
    }
#ifdef MODULE
    if (global_transports != NULL)
#endif
    global_transports = CMrealloc(global_transports, 
				  sizeof(global_transports) * (i + 2));
#ifdef MODULE
    else
        global_transports = CMmalloc(sizeof(global_transports) * (i+2));
#endif
    global_transports[i] = 
	transport = CMmalloc(sizeof(struct _transport_item));
    global_transports[i+1] = NULL;

#ifndef MODULE
    libname = CMmalloc(strlen(trans_name) + strlen("libcm") + strlen(".la") 
		       + 1);
    
    strcpy(libname, "libcm");
    strcat(libname, trans_name);
    strcat(libname, ".la");

    if (lt_dlinit() != 0) {
	fprintf (stderr, "error during initialization: %s\n", lt_dlerror());
	return 0;
    }

    lt_dladdsearchdir(CM_LIBRARY_BUILD_DIR);
    lt_dladdsearchdir(CM_LIBRARY_INSTALL_DIR);
    handle = lt_dlopen(libname);
    if (!handle) {
	fprintf(stderr, "Failed to load required '%s' dll.  Error \"%s\".\n",
		trans_name, lt_dlerror());
	fprintf(stderr, "Search path includes '.', '%s', '%s' and any default search paths supported by ld.so\n", CM_LIBRARY_BUILD_DIR, 
		CM_LIBRARY_INSTALL_DIR);

    } else {
	CMtrace_out(cm, CMTransportVerbose, "Loading local or staticly linked version of \"%s\" transport",
		    trans_name);
    }
    if (!handle) {
	return 0;
    }
    CMfree(libname);
    CMtrace_out(cm, CMTransportVerbose, "Loaded transport.");
    transport->trans_name = strdup(trans_name);
    transport->cm = cm;
    transport->data_available = CMDataAvailable;  /* callback pointer */
    transport->transport_init = (CMTransport_func)
	lt_dlsym(handle, "initialize");  
    transport->listen = (CMTransport_listen_func)
	lt_dlsym(handle, "non_blocking_listen");  
    transport->initiate_conn = (CMConnection(*)())
	lt_dlsym(handle, "initiate_conn");  
    transport->self_check = (int(*)())lt_dlsym(handle, "self_check");
    transport->connection_eq = (int(*)())lt_dlsym(handle, "connection_eq");
    transport->shutdown_conn = (CMTransport_shutdown_conn_func)
	lt_dlsym(handle, "shutdown_conn");  
    transport->read_to_buffer_func = (CMTransport_read_to_buffer_func)
	lt_dlsym(handle, "read_to_buffer_func");  
    transport->read_func = (CMTransport_read_func)
	lt_dlsym(handle, "read_func");  
    transport->write_func = (CMTransport_write_func)
	lt_dlsym(handle, "write_func");  
    transport->writev_func = (CMTransport_writev_func)
	lt_dlsym(handle, "writev_func");  
    transport->writev_attr_func = (CMTransport_writev_attr_func)
	lt_dlsym(handle, "writev_attr_func");  
#else
    transport->trans_name = strdup((char *)trans_name);
    transport->cm = cm;
    transport->data_available = CMDataAvailable; 
    if (strcmp(transport->trans_name, "sockets") == 0) {
      transport->transport_init = 
	(CMTransport_func)libcmsockets_LTX_initialize;  
      transport->listen = 
	(CMTransport_listen_func)libcmsockets_LTX_non_blocking_listen;  
      transport->initiate_conn = 
	(CMConnection(*)())libcmsockets_LTX_initiate_conn;  
      transport->self_check = 
	(int(*)())libcmsockets_LTX_self_check;
      transport->connection_eq = 
	(int(*)())libcmsockets_LTX_connection_eq;
      transport->shutdown_conn = 
	(CMTransport_shutdown_conn_func)libcmsockets_LTX_shutdown_conn;  
      transport->read_to_buffer_func = 
	(CMTransport_read_to_buffer_func)libcmsockets_LTX_read_to_buffer_func;  
      transport->read_func = 
	(CMTransport_read_func)libcmsockets_LTX_read_func;  
      transport->write_func = 
	(CMTransport_write_func)libcmsockets_LTX_write_func;  
      transport->writev_func = 
	(CMTransport_writev_func)libcmsockets_LTX_writev_func;  
      transport->writev_attr_func = 
	(CMTransport_writev_attr_func)libcmsockets_LTX_writev_attr_func; 
    } else if (strcmp(transport->trans_name, "rudp") == 0) {
      transport->transport_init = 
	(CMTransport_func)libcmrudp_LTX_initialize;  
      transport->listen = 
	(CMTransport_listen_func)libcmrudp_LTX_non_blocking_listen;  
      transport->initiate_conn = 
	(CMConnection(*)())libcmrudp_LTX_initiate_conn;  
      transport->self_check = 
	(int(*)())libcmrudp_LTX_self_check;
      transport->connection_eq = 
	(int(*)())libcmrudp_LTX_connection_eq;
      transport->shutdown_conn = 
	(CMTransport_shutdown_conn_func)libcmrudp_LTX_shutdown_conn;  
      transport->read_to_buffer_func = 
	(CMTransport_read_to_buffer_func)libcmrudp_LTX_read_to_buffer_func;  
      transport->read_func = NULL;
      transport->write_func = NULL;
      transport->writev_func = 
	(CMTransport_writev_func)libcmrudp_LTX_writev_func;  
      transport->writev_attr_func = 
	(CMTransport_writev_attr_func)libcmrudp_LTX_writev_attr_func; 
    }
#endif
    CMtrace_out(cm, CMTransportVerbose, "Listen is %lx", transport->listen);
    if (transport->transport_init) {
	transport->trans_data = 
	    transport->transport_init(cm, &CMstatic_trans_svcs);
    }
    transport = add_transport_to_cm(cm, transport);
    return 1;
}
