
#ifdef MODULE
#ifndef __KERNEL__
#define __KERNEL__
#endif
#endif

#ifndef MODULE
#include "config.h"
#include <sys/types.h>
#include <sys/time.h>
#include <sys/socket.h>
#include <sys/un.h>
#include <sys/uio.h>
#include <netinet/in.h>
#include <arpa/inet.h>
#include <netdb.h>
#include <stdio.h>
#include <fcntl.h>
#include <unistd.h>
#include <assert.h>
#include <errno.h>
#include <signal.h>
#include <stdlib.h>
#include <string.h>
#include <limits.h>
#include <pthread.h>
#include <math.h>
typedef char uint_8;
typedef short uint_16;
typedef int  uint_32;

#include "gen_thread.h"
#include <atl.h>
#include "cm.h"
#include "cm_internal.h"
#include "cm_transport.h"
#else
#include <asm/uaccess.h>
#include "config.h"
#include <linux/types.h>
#include <linux/time.h>
#include <linux/socket.h>
#include <linux/in.h>
#include <linux/random.h>
#include <linux/sockios.h>
#include <net/sock.h>
#include "assert.h"
#include "../include/io.h"
#include "../include/atl.h"
#include "cm.h"
#include "cm_internal.h"
#include "cm_transport.h"
#include "kcm.h"
#include "library.h"

#define fprintf(fmt, args...)	printk(args)
#define printf		printk
#define perror		printk
#define exit		sys_exit
#define pthread_mutex_unlock(a)	up(a)
#define pthread_mutex_lock(a) 	down(a)
#define gettimeofday(a,b)	do_gettimeofday(a)
#define sendto		my_sock_sendto
#define recvfrom	my_sock_recvfrom
#define signal(a,b)	sys_signal(a,b)
#define close		sys_close
#define sched_yield	schedule
#define gethostname	sys_gethostname
#define ioctl		sys_kioctl

#define abs(x)	x < 0 ? -x : x

static inline int timersub(struct timeval *a, struct timeval *b, struct timeval *result)
{
  do { 
    (result)->tv_sec = (a)->tv_sec - (b)->tv_sec; 
    (result)->tv_usec = (a)->tv_usec - (b)->tv_usec; 
    if ((result)->tv_usec < 0) { 
      --(result)->tv_sec; 
      (result)->tv_usec += 1000000; 
    } 
  } while (0);
  return 0;
} 

static DECLARE_WAIT_QUEUE_HEAD(scd_queue);

typedef char uint_8;
typedef short uint_16;
typedef int  uint_32;

static char *
inet_ntoa(struct in_addr ina)
{
    static char buf[4*sizeof "123"];
    unsigned char *ucp = (unsigned char *)&ina;

    sprintf(buf, "%d.%d.%d.%d",
	ucp[0] & 0xff,
	ucp[1] & 0xff,
	ucp[2] & 0xff,
	ucp[3] & 0xff);
    return buf;
}

extern int timerrun_id;
extern int destroy_timerrun;
#endif

#ifndef SOCKET_ERROR 
#define SOCKET_ERROR -1
#endif

#ifndef THR_ERROR
#define THR_ERROR -1
#endif

#ifndef THR_OK
#define THR_OK 0
#endif

#ifndef TRUE
#define TRUE 1
#define FALSE 0
#endif

#define MAX_PACKET 1400
#define SEQ_MOD 32768
#define MAX_SEQ 32767

/* Threshold to enable and disable redundant transmission */
#define RED_THRESH_HIGH 30
#define RED_THRESH_LOW  10

/* Error rates that trigger RUDP adaptation when reached */
#define RUDP_CHG_HIGHMARK 30
#define RUDP_CHG_LOWMARK 10

/* consecutive number of big rtt changes needed to trigger a retrans_tm_change */

#define UPDATE_THRESH_A 10
#define UPDATE_THRESH_R 0.05

/* Application response code regarding its adaptation strategy */
#define ACTION_PASSIVE 0	/* no app adaptation */
#define ACTION_TAKEN_SLOWDOWN 1		/* adaptation has been made to slow down */
#define ACTION_TAKEN_LOWRES 2		/* adaptation has been made to send less data in one packet */
#define ACTION_INPROG 3		/* adaptation still in progress */

#define DEF_RATE -1	/* undefined, whatever available */
/* connection state */
#define CONNECTED 1
#define FIN	2	/* closed */
#define TIMED_WAIT 3

#define SYN (1 << 7)
#define ACK (1 << 6)
#define EAK (1 << 5)
#define RST (1 << 4)
#define NUL (1 << 3)
#define CHK (1 << 2)
#define TSTAMP (1 << 1)
#define L_SEG 1		/* This is the last segment of a user message */
#define FORWARD			RST

/* Composition of flags identifying packet type */
#define KEEPALIVE_FLAGS 	NUL
#define KEEPALIVE_ACK_FLAGS 	(NUL + ACK)
#define EXPSEND_FLAGS		(NUL + RST + ACK)
#define EXPSEND_ACK_FLAGS	(NUL + RST + ACK + TSTAMP)
#define FIN_FLAGS		CHK
#define FIN_ACK_FLAGS		(CHK + ACK)
#define UPDATE_FLAGS		(CHK + RST)

#define OPT_CHK 1 << 1
#define OPT_REUSE 1 << 2

/* Window Size
 * RUDP can dynamically adjust its window size to values smaller 
 *	than the initial window sizes, but not to larger values
 */
#define RWIN_DEF 91	/* Initial Receiving Window Size */
#define SWIN_DEF 45 	/* Initial Sending Window Size */
#define MAX_RWIN 128	/* Max Initial Receiving Window Size */
#define MAX_SWIN 128	/* Max Initial Sending Window Size */

#define ERR_LIMIT 0

/* Timers */
#define TIMER_ARRAY_SIZE 256

/* These two timeout values are self-adaptive and dynamically changed */
#define RETRANS   400
#define CUMACK    200

/* #define KEEPALIVE (10*RETRANS) */
#define KEEPALIVE 1000
#define FIRST_EXSEND 10		/* explicit rtt measurement packet */
#define TIMEDWAIT 600
#define LINGER 5000

/* Max number of outstanding KEEPALIVE messages before closing the conn */
#define MAXKEEPALIVE 30

/* Interval to Update Statistics(including connection record and attributes) */
#define REPORT_MOD 50	/* record data rate every 20 packets */
#define EACK_MOD 3

#define ENABLE_SND_REPORT FALSE
#define ENABLE_RCV_REPORT FALSE
#define ENABLE_ADAPT	  FALSE
#define COORD_ADAPT
#define DUMPWIN

/* Flag indicating reason of callback */
#define ERR_RATE 1
#define ERR_ERATIO 2
#define DROP_ERATIO 3 


#if !defined(BIND_DEFINED) & !defined(HAVE_WINDOWS_H)
extern int bind ARGS((int s, struct sockaddr * name, int namelen));
#endif
#if !defined(ACCEPT_DEFINED) & !defined(HAVE_WINDOWS_H)
extern int accept ARGS((int s, struct sockaddr * addr, int *addrlen));
#endif
#if !defined(CONNECT_DEFINED) & !defined(HAVE_WINDOWS_H)
extern int connect ARGS((int s, struct sockaddr * name, int namelen));
#endif
#if !defined(SETSOCKOPT_DEFINED) & !defined(HAVE_WINDOWS_H)
extern int setsockopt ARGS((int s, int level, int optname, const char *optval, int optlen));
#endif
#if !defined(GETSOCKNAME_DEFINED) & !defined(HAVE_WINDOWS_H)
extern int getsockname ARGS((int s, struct sockaddr * name, int *namelen));
#endif
#if defined(HAVE_GETDOMAINNAME) && !defined(GETDOMAINNAME_DEFINED)
extern int getdomainname ARGS((char *name, int namelen));
#endif
#if !defined(LISTEN_DEFINED) & !defined(HAVE_WINDOWS_H)
extern int listen ARGS((int s, int backlog));
#endif
#ifndef HAVE_WINDOWS_H
extern int socket ARGS((int domain, int type, int protocol));
#endif
#if !defined(GETHOSTNAME_DEFINED) & !defined(HAVE_WINDOWS_H)
extern int gethostname ARGS((char *name, int namelen));
#endif

#include "qual_hostname.c"

#ifndef timersub
# define timersub(a, b, result)				\
  do {							\
    (result)->tv_sec = (a)->tv_sec - (b)->tv_sec;	\
    (result)->tv_usec = (a)->tv_usec - (b)->tv_usec;	\
    if ((result)->tv_usec < 0) {			\
      --(result)->tv_sec;				\
      (result)->tv_usec += 1000000;			\
    }							\
  } while (0)		
#endif

int global_connid = 0;

typedef struct rudp_client_data {
	CManager	cm;
	char		*hostname;	
	int		listen_port;
	int		local_IP;
	CMtrans_services svc;
} *rudp_client_data_ptr;

typedef struct region {
	int	exp_rate;
	int	dev;
	int	eratio;
	int	(*handler)(int, int, int);
} rudp_region_spec;

typedef struct rudp_connection_data {
/* --------------general information of the connection */
	CMConnection conn;
	char	*remote_host;
	int	remote_IP;
	int	remote_contact_port;
	int	remote_conn_port;
	int	fd;
	void	*recv_buffer;
	int	recv_buffer_len;
	void	*send_buffer;
	int	send_buffer_len;
	unsigned long	connid;
	unsigned long   local_connid;

/* timeout values and measured RTT, standard deviation of RTT */
	int	retrans_to;
	int	sdev;
	int	mrtt;
	int	cumack_to;
	int	null_to;
	int	transtat_to;
/* some max(threshold) values */
	int	max_outstanding;
	int	max_segsize;
	int	max_retrans;
	int	max_cumack;
	int	max_outseq;
	int	max_auto_rst;
	int	max_keepalive;
    int min_req;	/* minimum windows required by an app packet */
	int 	max_byte;
/* receiving window size */
	int	rwin;
/* for synchronization */
#ifdef MODULE
        struct semaphore mutex;
        int die;
	struct timer_list scd_timer;
#else
	pthread_mutex_t mutex;
	pthread_cond_t die;
#endif
	
/* ----------------CM client info */
        CManager        cm;
        char            *hostname;
        int             listen_port;
        int             local_IP;
        CMtrans_services svc;

/* ----------------dynamic states */
        int     state;
	int	first_measure;
	int 	first_data;	/* send ACK when receiving the first data packet */
	int	redun_send;	/* send mode, whether redundant */
	int	fin;
	int	skip_next_L;

/* some states used for sender adaptation */
	uint_8	adapt_inprog;
	uint_8	app_adapted;
	uint_16 forward;

/* several counters */
        uint_8  retrans_cnt;
        uint_8  cumack_cnt;
        uint_8  outseq_cnt;
	int     outseq_occur;
        uint_8  rst_cnt;
        uint_8  outstanding;
	uint_8 	cont_sent;

/* some per-packet information for packets in receiving window or send window */
        int     abit[MAX_RWIN];
        int     acked[MAX_SWIN];
	int 	eacked[MAX_SWIN];
	int	sent_len[MAX_SWIN];
	int	lbit[MAX_RWIN];	/* TRUE if it's the last segment of a packet */
        struct timeval send_stamp[MAX_SWIN];
	uint_8	marked[MAX_SWIN];
        int     avail_len;
        void    *data_avail;
	int	av_seq;

/* Timers and associated variables */
        int     cumack_tm;
        int     retrans_tm;
	int	exp_send_tm;
	int	exp_send_last;
	int	last_retrans;
        int     null_tm;
	int	next_expected;
	int	next_ack;
	int	next_send;
	int	last_outseq;

/* sender statistics */
	int	total_sent;
	int	total_err;
#ifdef MODULE
	int	period_bytes;
#else
	double 	period_bytes;
#endif
	int	period_error;
	int 	period_effpkt;
	struct timeval last_s_rep;
	int 	last_sent;
	int	oneway_err;

/* receiver statistics */
#ifdef MODULE
	int	period_recvb;
#else
	double 	period_recvb;
#endif
	int	period_recvp;
	int	period_rcv_err;
	int	period_win_inc;
	int	rcv_apperr;
	struct timeval last_r_rep;
	int	total_recv;
	int 	last_recv;
	int 	senderr_cnt;	/* counting consecutive sendto err */
	int	recverr_cnt;	/* counting consecutive recvfrom err*/ 
/* SLA */
	struct region send_qos; /* receiver QoS has to be specified through conn attr list */

} *rudp_conn_data_ptr;

/*
 * Connection attribute list contains the following attributes:
 * 1. Performance related, queried by clients: WIN_INC, RECV_RATE, RECV_ERR, SEND_RATE, SEND_ERR
 * 2. Logistical information: IP_HOST, IP_PORT, ERR_RATIO, RATE, CONNECTION_FILE_DESCRIPTOR 
 * 	PEER_CONN_PORT, THIS_CONN_PORT, PEER_LISTEN_PORT, PEER_HOSTNAME 
 */

/* mark 0912 */
#define SYN_LEN 33
typedef struct {
	uint_8	flags;
	uint_8  len;			/* fixed len */
	uint_8	seq;
	uint_8	ack;
	uint_8	ver;
	uint_16	max_outstanding;
	uint_8	opt;
	uint_8	dummy;
	uint_16	max_segsize;
	uint_16	retrans_to;
	uint_16	cumack_to;
	uint_16	null_to;
	uint_16	transtat_to;
	uint_8	max_retrans;
	uint_8	max_cumack;
	uint_8	max_outseq;
	uint_8	max_auto_rst;
 	uint_16 newport;	
	unsigned long	connid;
	uint_16	cksum;
} syn_pdu;

typedef struct {
        uint_8  flags;
        uint_8  len; 
        uint_16  seq;
        uint_16  ack;
	uint_16  payload;
} gen_data_pdu;

/* mark 0912 */
typedef struct {
	uint_8	flags;
	uint_8	len;
	uint_16	seq;
	uint_16	ack;
	uint_16 payload;
	uint_8  partack;
	uint_16	forward;
} ack_pdu;

/* mark 0912 */
typedef struct {
	uint_8	flags;
	uint_8	len;
	uint_16	seq;
	uint_16	ack;
	uint_16 payload;
	uint_16	cksum;
	uint_16	dummy;
} eack_pdu;
	
typedef struct {
	uint_8	flags;
	uint_8	len;		
	uint_16	seq;
	uint_16	ack;
	uint_16	cksum;
} rst_pdu;

typedef struct {
	uint_8  flags;
	uint_8  len;
	uint_16 seq;
	uint_16 ack;
	long	newretrans;
	uint_16 cksum;
} update_pdu; 

typedef struct {
	uint_8  flags;
	uint_8  len;
	uint_16 seq;
	uint_16 ack;
	uint_16 cksum;
} fin_pdu; 

typedef struct {
	uint_8	flags;
	uint_8	len;
	uint_16	seq;
	uint_16	ack;
	long	peer_retrans;
	uint_16	cksum;
	uint_16 rnum;
} null_pdu;	

/* not used at this point */
typedef struct {
	uint_8	flags;
	uint_8	len;
	uint_16	seq;
	uint_16	ack;
	uint_8	seq_adj_fac;
	uint_8	dummy;
	unsigned long connid;
	uint_16	cksum;
} tcs_pdu;	

typedef int (*TimerHandler)(int, rudp_conn_data_ptr, int);

typedef struct {
  uint_8 	      initialized;
  int	              arraySize;
  int		      scanSize;
  uint_8	      *inUse;
  struct timeval       *expTime;
  TimerHandler	      *object;
  rudp_conn_data_ptr   *argument1;
  int	              *argument2;
#ifdef MODULE
  struct semaphore    timerMutex;
#else
  pthread_attr_t      timerAttr;
  pthread_t           timerThread;
  pthread_mutex_t     timerMutex;
#endif
} RUDPTimer;

static RUDPTimer gtimer;
static int rudp_global_init = 0;

static void TimerStart(void *);
static int TimerSet(int, TimerHandler, rudp_conn_data_ptr, int);
static int CancelTimer(int);
static void CancelObjTimers(rudp_conn_data_ptr);
static rudp_conn_data_ptr create_rudp_conn_data(CMtrans_services);
static int check_host(char *, void *);
static int TimerUpdate(int, int, TimerHandler, rudp_conn_data_ptr, int);
static void TimerLog(char *, char *);
static int my_min(int, int);
static void rudp_accept_conn(void *, void *);
static void rudp_data_available(transport_entry, CMConnection);

/* Timer handlers */
int retrans_handler(int, rudp_conn_data_ptr, int);
int null_handler(int, rudp_conn_data_ptr, int);
int cumack_handler(int, rudp_conn_data_ptr, int);
int exp_send_handler(int, rudp_conn_data_ptr, int);
int timedwait_handler(int, rudp_conn_data_ptr, int);

/* Utility functions */
int my_min(int a, int b)
{
	if (a < b)	
		return a;
	else
		return b;
}

int my_max(int a, int b)
{
	if (a > b)
		return a;
	else
		return b;
}

/* 
 * Decide whether a sequence number is in receiving window, based on the 
 * current receiving window size and the next_expected sequence number
 * For this to work correctly, receiving window size should be smaller than 
 * 1/2 sequence number space.
 */
static int in_rwin(int seq, int next_expected, int rwin)
{
	if((rwin + next_expected)%SEQ_MOD - next_expected > 0)
	{
		if(seq > next_expected && seq < rwin + next_expected)
			return TRUE;
		else
			return FALSE;
	}
	else
	{
		if(seq > next_expected || seq < (rwin + next_expected )%SEQ_MOD)
			return TRUE;
		else
			return FALSE;
	}
}		

/*
 * Decide whether a sequence number is in ACK window(i.e., the ACK is valid)
 * Have the same implication on receiving window size as in_rwin()
 */
static int in_ackwin(int this_ack, int ack_expected, int outstanding)
{
	if(outstanding == 0)
		return FALSE;

	if((outstanding + ack_expected)%SEQ_MOD - ack_expected > 0)
	{
		if( !(this_ack <ack_expected) && this_ack< outstanding + ack_expected)
			return TRUE;
		else
			return FALSE; 
	}
	else
	{
		if(!(this_ack < ack_expected) || this_ack < (outstanding + ack_expected)%SEQ_MOD)
			return TRUE;
		else
			return FALSE;
	} 
}

/* 
 * Translate a mod/SEQ_MOD sequence number to a sequence number relative to the 
 * next expected ACK number 
 */
static int unwrap(int low, int high)
{
	int i;
	i = (high < low) ? (low + (high-low)%SEQ_MOD + SEQ_MOD): high;
	return i;
}

/* 
 * Handles out of sequence packets
 * Check QoS specification, trigger any action that's necessary:
 * 1. If error ratio exceeds the maximum tolerable, send EACK, otherwise, 
 *    send ACK
 * 2. Update error ratio
 */
void outseq_handler(rudp_conn_data_ptr scd)
{
        struct sockaddr_in remote_addr;
        int onum, ocnt, ret, i, eratio=0, pktlen, latest=0, acked, tavail, window, tav_seq = -1, exp_err, spec_err;
        uint_16 *eacks, *eack_seqs;
        attr_list conn_attr_list;
        atom_t read_error_ratio;
#ifdef MODULE
        char *ack_pkt = NULL;
#else
        char *ack_pkt;
#endif
        ack_pdu *ack_hdr;
	static atom_t qos_atom=0;	
	struct attr_opaque *aop;
#ifdef MODULE
	struct socket *sock;
#endif
		
	scd->svc->trace_out(scd->cm, "outseq_handler: entering outseq_handler");
        remote_addr.sin_addr.s_addr = scd->remote_IP;
        remote_addr.sin_port = htons(scd->remote_conn_port);
        remote_addr.sin_family = AF_INET;

        conn_attr_list = scd->conn->attrs;
	/* 
	 * Try to find a max tolerable error ratio in connection attr list 
	 * If not found, fall back to the one in rudp_conn_data, which is 
	 * initialized when the connection is established.
	 */
	if(qos_atom==0)
		qos_atom = attr_atom_from_string("QOS");

	if(!query_attr(conn_attr_list, qos_atom, NULL, (attr_value *)&aop))
		exp_err = ERR_LIMIT; 
	else
	{
		spec_err = ((struct region *)(aop->buffer))->eratio;
		exp_err = spec_err == -1 ? ERR_LIMIT: spec_err;
	}

        /* count missing packets in the error window(20) */
        onum = 0;
        ocnt = 0;
	tavail = 0;
#ifdef MODULE
        eacks = (uint_16 *)DAllocMM(sizeof(uint_16) * scd->rwin);
#else
        eacks = (uint_16 *)malloc(sizeof(uint_16) * scd->rwin);
#endif

	/* count the window that we should look for out of seq packet*/
	if(scd->last_outseq > scd->next_expected)
		window = scd->last_outseq - scd->next_expected + 1;
	else
		window = scd->last_outseq - scd->next_expected + 1 + SEQ_MOD; 

        for( i = 0; i< window && i<scd->rwin; i++)
        {
            if(scd->abit[(i+ scd->next_expected)%SEQ_MOD%MAX_RWIN] == 0)
             { /* this packet is out of seq */
                 eacks[ocnt] = (i+scd->next_expected) % SEQ_MOD;
                 ocnt ++;
		 scd->lbit[(i+ scd->next_expected)%SEQ_MOD%MAX_RWIN] = 0;
              }
             if(scd->abit[(i+ scd->next_expected)%SEQ_MOD%MAX_RWIN] != 0)
	     {
		 tavail += scd->abit[(i+ scd->next_expected)%SEQ_MOD%MAX_RWIN];
		 if(scd->data_avail == NULL && tav_seq == -1)
			tav_seq = (i + scd->next_expected) % SEQ_MOD;
                 onum ++;
	     }
        }

        /* the latest packet that's seen is the one to be ACKed */
        latest = (i -1 + scd->next_expected)%SEQ_MOD;
/*
	if(ocnt + scd->rcv_apperr + scd->period_recvp == 0)
		eratio = 0;
	else
		eratio = (float)(ocnt+scd->rcv_apperr) * (float)100 /(float)(ocnt+scd->rcv_apperr+scd->period_recvp);
*/
	/* Is there a better way to compute error ratio? */
	if(window == 0)
		eratio = 0;
	else
#ifdef MODULE
		eratio = ocnt*100/window;
#else
		eratio = (float)(ocnt) * (float)100 /(float)(window);
#endif

	scd->svc->trace_out(scd->cm, "outseq_handler: latest %d, last_outseq %d, ocnt %d\n", latest, scd->last_outseq, ocnt);
	if(latest!=scd->last_outseq)
		printf("WARNING, latest!=last_outseq\n");
	
	if(ocnt > scd->rwin)
		printf("FATAL, outseq cnt > scd->rwin, will SIGSEGV\n");

	/* If the error ratio is less than the max tolerable, send ACK 
	 * otherwise, send NACK packet. In either case, list all outseq
         * packet number
         */
	ack_hdr = (ack_pdu *)ack_pkt;
        if(eratio==0 || eratio < exp_err)
        {
                ack_hdr->flags = ACK;
		ack_hdr->ack = htons(scd->last_outseq);
		ack_hdr->partack = TRUE;
        }
        else 
        {
                ack_hdr->flags = EAK + ACK;
                ack_hdr->ack = scd->next_expected==0 ?MAX_SEQ: (scd->next_expected - 1)%SEQ_MOD;
                ack_hdr->ack = htons(ack_hdr->ack);
	}

	scd->period_recvp += onum;
	scd->period_recvb += (float)tavail/(float)1000;
	scd->rcv_apperr += ocnt;

	ack_hdr->len = my_min(sizeof(eack_pdu) + ocnt*sizeof(uint_16), 127);
	ack_hdr->seq = htons(-1); /*  control packets doesn't have sequence # */
	ack_hdr->payload = htons(0);

	eack_seqs = (uint_16 *)((uint_8 *)ack_pkt + sizeof(eack_pdu));
#ifdef MODULE
	memcpy((void *)eack_seqs, (void *)eacks, my_min(127, sizeof(uint_16)*ocnt));
#else
	memcpy((void *)eack_seqs, (void *)eacks, min(127, sizeof(uint_16)*ocnt));
#endif
	pktlen = ack_hdr->len;
	acked = ntohs(ack_hdr->ack);

	if(scd->rcv_apperr + scd->period_recvp == 0)
		eratio = 0;
	else
		eratio = scd->rcv_apperr * 100/(scd->rcv_apperr + scd->period_recvp); 
#ifdef MODULE
	sock = look_up_pseudofd(scd->fd, (struct socket **)current->qfabric->fdqfabric);
	if (sock == NULL)
	    return;
        ret = sendto(sock, ack_pkt, pktlen, 0, (struct sockaddr *)&remote_addr, sizeof(struct sockaddr_in), jiffies, 0);
#else
        ret = sendto(scd->fd, ack_pkt, pktlen, 0, (struct sockaddr *)&remote_addr, sizeof(struct sockaddr_in));
#endif
        if(ret == SOCKET_ERROR)
                perror("EACK sendto error");
	else
	{
		scd->svc->trace_out(scd->cm, "outseq_handler: ACK in the packet sent: %d\n", acked);
		scd->cumack_cnt = 0;	
		if(scd->cumack_tm != -1)
			CancelTimer(scd->cumack_tm);
	}
	
        read_error_ratio = attr_atom_from_string("RECV_ERR");
        replace_attr(conn_attr_list, read_error_ratio, Attr_Int4, (attr_value)eratio);
#ifdef MODULE
	if(ack_pkt) DFreeMM((addrs_t)ack_pkt);
	if(eacks) DFreeMM((addrs_t)eacks);
#else
	if(ack_pkt) free(ack_pkt);
	if(eacks) free(eacks);
#endif

}

/* 
 * Find out whether we need to update the last_seen packet sequence 
 * called by rudp_data_available
 */
static void set_lastseq(rudp_conn_data_ptr scd, uint_16 seq)
{
	int i, next_exp, l_outseq;
	
	l_outseq = scd->last_outseq;
	next_exp = scd->next_expected;

	if(l_outseq == -1)
	{
		scd->last_outseq = seq;
		scd->outseq_cnt++;
		return;
	}

	/* if both seq and last_outseq are larger than scd->next_expected */
	if((seq - next_exp > 0) && (l_outseq - next_exp>0))
	{
		if(seq > l_outseq)
		{
			scd->last_outseq = seq;
			scd->outseq_cnt ++;
			return;
		}
		for(i = next_exp; i<l_outseq+1; i++)
			if(i==seq)
				return;
		scd->outseq_cnt++;
		return;
	}

	/* both seq and last_outseq are smaller than scd->next_expected */
	if((seq - next_exp < 0) && (l_outseq - next_exp < 0))
	{
		if(seq > l_outseq)
		{
			scd->last_outseq = seq;
			scd->outseq_cnt ++;
			return;
		}
		for(i = next_exp; i<l_outseq+1+SEQ_MOD; i++)
			if(i % SEQ_MOD==seq)
				return;
		scd->outseq_cnt ++;
		return;
	}

	if(seq < l_outseq)
	{
		scd->last_outseq = seq;
		scd->outseq_cnt ++;
		return;
	} 
	
	for(i=next_exp; i<l_outseq + 1 + SEQ_MOD; i++)
		if(i%SEQ_MOD == seq)
			return;
	scd->outseq_cnt ++;
	return;
}

/* callback by libselect, this is a very very very long routine */
static void rudp_data_available(transport_entry trans, CMConnection conn)
{
    int i=0, iget, datalen, this_ack, j, retrans_len, old_next_expected, fromlen, tolen, ocnt, temp, tavail, tav_seq = -1, missing, max_seg;
    int reset_needed = FALSE, o_retrans = 0, recv_ack=0, recv_rate, recv_err, win_inc, eratio, forward, window, stop_forward = FALSE;
    long msec, trtt;
    uint_8 flags, already_set;
    gen_data_pdu *hdr_ptr;
    char *int_buf, *retrans_buf, *ack_pkt;
    uint_16 *eack_seqs, seq;
    eack_pdu *ack_hdr;
    update_pdu update_pkt, *update_ptr;
    struct sockaddr_in remote_addr;
    rudp_conn_data_ptr scd = (rudp_conn_data_ptr)conn->transport_data;
    CMtrans_services svc = scd->svc;
    CManager cm = conn->cm;
    struct timeval tv, tdiff;
#ifndef MODULE
    struct timezone tz;
#endif
    atom_t recv_rate_atom, recv_err_atom, recv_winc_atom;
    attr_list conn_attr_list;
#ifdef MODULE
    struct socket *sock;
#endif
/*
 * Can we use recvfrom() instead of this low-level read, in case the data
 * it gives me crosses message boundary
 * I guess using read() will complicate all the handling of sequencing and
 * acknowledgement stuff. So, let's use recvfrom() and instead of read the
 * requested_len, we will read the whole packet and buffer it in case CM
 * need it later on
 */

  max_seg = scd->max_segsize;

#ifndef MODULE
/*    if(scd->null_tm != -1) */
/*        TimerUpdate(scd->null_tm, KEEPALIVE, &null_handler, scd, 0); */

   if(scd->exp_send_tm == -1)
        scd->exp_send_tm = TimerSet(FIRST_EXSEND, &exp_send_handler, scd, 0);
#endif

#ifdef MODULE
   int_buf = (char *)DAllocMM(max_seg * sizeof(char));
#else
   int_buf = (char *)malloc(max_seg * sizeof(char));
#endif

   do{
	fromlen = sizeof(struct sockaddr_in);
#ifdef MODULE
	sock = look_up_pseudofd(scd->fd, (struct socket **)current->qfabric->fdqfabric);
        iget = recvfrom(sock, (char *)int_buf,max_seg, 0, NULL, &fromlen, 0);
#else
        iget = recvfrom(scd->fd, (char *)int_buf,max_seg, 0, NULL, &fromlen);
#endif
	already_set = FALSE;

      if (iget == 0)
      {
        svc->trace_out(cm, "rudp_data_available: socket EOF");
        break;
      }
      else
      if (iget == -1) {
         int lerrno = errno;
         scd->recverr_cnt++;
         if ((lerrno != EWOULDBLOCK) && (lerrno != EAGAIN) && (lerrno != EINTR))
                svc->trace_out(cm, "rudp_data_available: err socket recvfrom");
         break;

       }

      scd->recverr_cnt = 0;

      /* get application data length, i.e., excluding RUDP header */
      hdr_ptr = (gen_data_pdu *)int_buf;
      flags = hdr_ptr->flags;

      /* if RST, NUL or CHK is in the flags, datalen must be 0 */
      if(flags & NUL || flags & CHK)
        datalen = 0;
      else
        datalen = ntohs(hdr_ptr->payload);

    /* acquire the mutex */
     pthread_mutex_lock(&(scd->mutex));

/* PART I: parste special control messages */
/*
 * I.1 FORWARD message, used for sender to instruct receiver to skip unmarked
 * messages and forward to sequence "forward"
 */
     if(flags == FORWARD || flags == FORWARD + ACK || flags == FORWARD + ACK + L_SEG|| flags == FORWARD + L_SEG)
     {
	int rwinst, last_L_SEG=-1;
	tavail = 0;
	tav_seq = -1;
	/* count the window that we should look for out of seq packet*/
	forward = ((ack_pdu *)hdr_ptr)->forward;
	forward = ntohs(forward);
/*	printf("received FORWARD to seq %d\n", forward);
	printf("avail_len %d, av_seq $d, next_expected %d\n", scd->avail_len, scd->av_seq, scd->next_expected);
	if(scd->data_avail==NULL)
		printf("scd->data_avail NULL\n");
*/
	if(forward== scd->next_expected || in_rwin(forward, scd->next_expected, scd->rwin)){
	if(!(forward < scd->next_expected))
		window = forward - scd->next_expected;
	else
		window = forward - scd->next_expected + SEQ_MOD;
	
	missing = FALSE;
	rwinst = scd->av_seq==-1? scd->next_expected: (scd->av_seq + 1)%SEQ_MOD;
	rwinst = rwinst - 1 > 0? (rwinst -1) %SEQ_MOD: (rwinst-1 + SEQ_MOD)%SEQ_MOD;
	for( i = 0; i<scd->rwin; i++)
	{
		if(scd->abit[(i+ scd->next_expected)%SEQ_MOD%MAX_RWIN] == 0)
		 {
		    if(i < window)
		    {
			    scd->lbit[(i+ scd->next_expected)%SEQ_MOD%MAX_RWIN] = 0;
			    missing = TRUE;
		    }
		    else
			    break;
		 }
		if (scd->abit[(i+ scd->next_expected)%SEQ_MOD%MAX_RWIN] != 0)
		{
		    scd->outseq_cnt--;
		    tavail += scd->abit[(i+ scd->next_expected)%SEQ_MOD%MAX_RWIN];
/*		     if(scd->av_seq == -1 && tav_seq == -1) */
		    if(tav_seq == -1)
			    tav_seq = (i + scd->next_expected) % SEQ_MOD;
		    if(!(i < window))
		    {
			    forward = (i+scd->next_expected+1)%SEQ_MOD;
			    scd->cumack_cnt++;


		    }

		if(scd->lbit[(i+ scd->next_expected)%SEQ_MOD%MAX_RWIN])
		{
/* 			printf("L_SEG found %d, tav_seq %d, av_seq %d, next_expected %d, avail_len %d\n", (i+scd->next_expected)%SEQ_MOD, tav_seq, scd->av_seq, scd->next_expected, scd->avail_len); */
			last_L_SEG = (i+scd->next_expected) %SEQ_MOD;
			if(missing)
			{
				temp = i + scd->next_expected;
				j = temp>0 ? temp % SEQ_MOD: (temp + SEQ_MOD)%SEQ_MOD;
				scd->lbit[j%MAX_RWIN] = 0;
				while(scd->lbit[j%MAX_RWIN]==0)
				{
					tavail-=scd->abit[j%MAX_RWIN];
					scd->abit[j%MAX_RWIN]=0;
					if(j==tav_seq)
					{
						tav_seq = -1;
						if(scd->av_seq==-1)
							break;
					}
					if(j==scd->av_seq)
					{
						scd->av_seq = -1;
						scd->data_avail = NULL;
						break;
					}
					if(j==rwinst) break;
					temp = temp - 1;
					j = temp>0 ? temp %SEQ_MOD:(temp+SEQ_MOD)%SEQ_MOD;
				}
					
/* 				printf("missing for dgram ending at %d, tav_seq %d, av_seq %d, tavail %d, avail_len %d\n", (i+scd->next_expected)%SEQ_MOD, tav_seq, scd->av_seq, tavail, scd->avail_len); */
			}
			missing = FALSE;
		}		
             }
        }
/* 	printf("After forward window processing, tav_seq %d, av_seq %d, tavail %d, avail_len %d, rwinst %d\n", tav_seq, scd->av_seq, tavail, scd->avail_len, rwinst); */
	temp = forward - 1;
	missing = FALSE;
	j = temp>0? temp %SEQ_MOD: (temp+SEQ_MOD)%SEQ_MOD;
	while(scd->lbit[j%MAX_RWIN]==0 && j!= last_L_SEG)
	{
		if(scd->abit[j%MAX_RWIN] == 0)
			missing = TRUE;
		if(j%SEQ_MOD == rwinst)
			break;
		temp = temp - 1;
		j = temp>0? temp %SEQ_MOD: (temp+SEQ_MOD)%SEQ_MOD;
        }
/* 	printf("stop at j %d, missing %d\n", j, missing); */
	if(missing == TRUE)
	{
		int start;
		scd->skip_next_L = TRUE;
/* 		printf("skip_next_L set %d\n", forward); */
		if(!scd->lbit[j%MAX_RWIN] && j!=last_L_SEG)
		    start = j;
		else
		    start = j+1;

		for(i = start; i < start + scd->rwin; i++)
		{
			if(i%SEQ_MOD==forward)
				break;
		        if(i%SEQ_MOD==tav_seq)
			        tav_seq = -1;
		    	if(i%SEQ_MOD==scd->av_seq)
		    	{
				scd->av_seq = -1;
				scd->data_avail = NULL;
		    	}
			scd->avail_len -= scd->abit[i%SEQ_MOD%MAX_RWIN];
			scd->abit[i%SEQ_MOD%MAX_RWIN]=0;
/* 			printf("abit[%d] reset\n", i%SEQ_MOD);			 */
		}
	}				
	else	
	{
		scd->skip_next_L = FALSE;
/* 		printf("skip_next_L unset %d\n", forward); */
	}

	old_next_expected = scd->next_expected;
        scd->next_expected = forward;
        if(old_next_expected > scd->next_expected)
            scd->period_win_inc += SEQ_MOD - old_next_expected + scd->next_expected;
        else
            scd->period_win_inc += scd->next_expected - old_next_expected;
	
	scd->avail_len += tavail;
        if(scd->data_avail == NULL && tav_seq != -1)
        {
              scd->data_avail = (char*)scd->recv_buffer + tav_seq %SEQ_MOD%MAX_RWIN * max_seg;
              scd->av_seq = tav_seq;
        }

/* 	printf("After FORWARD, av_seq %d, next_expected %d, avail_len %d\n", scd->av_seq, scd->next_expected, scd->avail_len); */

        if(forward == (scd->last_outseq+1)%SEQ_MOD)
		scd->last_outseq = -1;
	} /* in_rwin(forward, , ,) */
	else
		svc->trace_out(cm, "rudp_data_available: already moved to FORWARD sequence");
     }	

     seq = ntohs(hdr_ptr->seq); 
     remote_addr.sin_addr.s_addr = scd->remote_IP;
     remote_addr.sin_port = htons(scd->remote_conn_port);
     remote_addr.sin_family = AF_INET;

    /* I.2 Check Whether is a FIN. */
    if(flags == FIN_FLAGS)
    {
        svc->trace_out(cm, "rudp_data_available: FIN packet received");
        if(scd->outseq_cnt > 0)
                svc->trace_out(cm, "Out of sequence packets remaining");

        hdr_ptr->flags += ACK;
#ifdef MODULE
	sock = look_up_pseudofd(scd->fd, (struct socket **)current->qfabric->fdqfabric);
        if(sendto(sock, (void *)hdr_ptr, sizeof(fin_pdu), 0, 
           (struct sockaddr *)&remote_addr, sizeof(remote_addr), jiffies, 0)==0)
#else
        if(sendto(scd->fd, (void *)hdr_ptr, sizeof(fin_pdu), 0, 
           (struct sockaddr *)&remote_addr, sizeof(remote_addr))==SOCKET_ERROR)
#endif
        {
                svc->trace_out(cm, "rudp_data_available: socket error during shutdown, %d", errno);
                scd->senderr_cnt ++;
        }
        else    scd->senderr_cnt = 0;

        svc->trace_out(cm, "connection being shutdown, to TIMED WAIT state");
        scd->state = FIN;
        TimerSet(TIMEDWAIT, &timedwait_handler, scd, 0);
     }

     /* I.3 Check Whether it is the ACK to a FIN */
     if(flags == FIN_ACK_FLAGS)
     {
        svc->trace_out(cm, "rudp_data_available: ACK to FIN packet received");
        TimerSet(TIMEDWAIT, &timedwait_handler, scd, 0);
        scd->state = FIN;
#ifdef MODULE
	scd->die = 1;
#else
        pthread_cond_signal(&(scd->die));
#endif
     }

     /*
      * I.4 KeepAlive or ExplicitSend and ACK to these messages 
      * EXSEND when there is no data message and hence can't get RTT
      */
     if(flags == KEEPALIVE_FLAGS || flags == EXPSEND_FLAGS)
     {
	if(flags == KEEPALIVE_FLAGS)
	{
        	svc->trace_out(cm, "This is a KEEPALIVE message");
                if(scd->total_sent > scd->last_sent)
                {
                        eratio = 100 - ntohs(((null_pdu *)hdr_ptr)->rnum) * 100.0 / (scd->total_sent - scd->last_sent);
/*                       printf("eratio %d\n", eratio); */
			scd->oneway_err = eratio;
                        if(eratio < 50 && eratio > 0)
/*                                 scd->max_outstanding = max(scd->min_req, scd->max_outstanding * (1 - (eratio * 2.0 / 100.0))); */
                                scd->max_outstanding = scd->max_outstanding * (1 - (eratio * 2.0 / 100.0));
			if(!(eratio < 100))
/* 				scd->max_outstanding = max(scd->min_req, scd->max_outstanding/2); */
				scd->max_outstanding = scd->max_outstanding/2;
			if(eratio < 100 && !(eratio < 50)) 
/* 				scd->max_outstanding = max(scd->min_req, scd->max_outstanding *  (1 - eratio/100.0)); */
				scd->max_outstanding = scd->max_outstanding *  (1 - eratio/100.0);
                        if(!(eratio >0))
/*                                 scd->max_outstanding = max(scd->min_req, scd->max_outstanding + 2); */
                                scd->max_outstanding = min(SWIN_DEF, scd->max_outstanding + 2);

			scd->max_outstanding = max(1, scd->max_outstanding);
			scd->max_byte = scd->max_outstanding * scd->min_req;
/*                       printf("max_outstanding from %d to %d, max_byte %d\n", old_maxoutstanding, scd->max_outstanding, scd->max_byte); */
                        scd->last_sent = scd->total_sent;
                }

	        hdr_ptr->flags += ACK;
	}
	else
	{
        	svc->trace_out(cm, "This is an EXPSEND message");
		hdr_ptr->flags += TSTAMP;
	}
	recv_ack = ntohs(hdr_ptr->ack);

        hdr_ptr->ack = scd->next_expected==0? MAX_SEQ: (scd->next_expected - 1)%SEQ_MOD;
	hdr_ptr->ack = htons(hdr_ptr->ack);
        ((null_pdu *)hdr_ptr)->peer_retrans = htonl(scd->retrans_to);
        if(scd->state == CONNECTED)
        {
#ifdef MODULE
		sock = look_up_pseudofd(scd->fd, (struct socket **)current->qfabric->fdqfabric);
	        if(sendto(sock, (char*)hdr_ptr, sizeof(null_pdu), 0,(struct sockaddr *)&remote_addr, sizeof(struct sockaddr_in), jiffies, 0) == 0)
#else
	        if(sendto(scd->fd, (char*)hdr_ptr, sizeof(null_pdu), 0,(struct sockaddr *)&remote_addr, sizeof(struct sockaddr_in)) == SOCKET_ERROR)
#endif
        	{
               	    svc->trace_out(cm, "rudp_data_available: sendto null acknowledge error");
	            scd->senderr_cnt++;
        	    pthread_mutex_unlock(&(scd->mutex));
#ifdef MODULE
		    DFreeMM((addrs_t)int_buf);
#else
               	    free(int_buf);
#endif
                    return;
        	}
        	else 
			scd->senderr_cnt = 0;
        }
    }

    if(flags == KEEPALIVE_ACK_FLAGS)
    {
        svc->trace_out(cm, "ACK to KEEPALIVE message");
        scd->rst_cnt--;
    }

    if(flags == EXPSEND_ACK_FLAGS)
    {
        svc->trace_out(cm, "ACK to EXSEND message");
        scd->rst_cnt--;
        if((seq + 1) % SEQ_MOD == scd->exp_send_last)
        {
/*
                trtt = ntohl(((null_pdu *)hdr_ptr)->peer_rtt);
                o_retrans = scd->retrans_to;
                o_cumack = scd->cumack_to;
                if(!scd->first_measure)
                {
                        scd->mrtt = 0.875 * scd->mrtt + 0.125 * trtt;
                        scd->sdev = 0.875* scd->sdev + 0.125 * abs(scd->mrtt - trtt);
                }
                else
                {
                        scd->mrtt = trtt;
                        scd->first_measure = FALSE;
                }
*/
                temp = ntohl(((null_pdu *)hdr_ptr)->peer_retrans);
                scd->retrans_to = temp;
                scd->cumack_to = scd->retrans_to/2;

                svc->trace_out(cm, "ack to explicit sent packet %d new retrans_timeout %d, cumack timeout %d\n", seq, scd->retrans_to, scd->cumack_to);
        }
    }

    /*
     * PART II. Data Message Processing
     * There are several different cases, the processing logic follows: 
     * 1. there is user data and receiving window is not full
     *	  1.1. packet is next_expected, advance window and (user)avail amount 
     *	  1.2. packet is not next_expected but in receiving window
     *    1.3. packet is not next_expected or in receiving window, duplicate 
     * 2. receiving window is not full
     */
    if( datalen > 0) scd->total_recv++;

    if( datalen > 0 && (scd->av_seq == -1 || (scd->av_seq > -1 && !((scd->next_expected - scd->av_seq)%SEQ_MOD > scd->rwin))))
    {
	svc->trace_out(cm, "rudp_data_available: data packet sequence %d, len %d\n", seq, datalen);
	old_next_expected = scd->next_expected;
        if(seq == scd->next_expected )     /* we can advance receiving window*/
        {
         	svc->trace_out(cm, "rudp_data_available: seq %d is next_expected data length from packet %d\n", seq, datalen);

	     if(!scd->skip_next_L)
	     {
		/* update receiving buffer and data_avail pointer */
        	memcpy((char*)scd->recv_buffer + seq % SEQ_MOD% MAX_RWIN * max_seg, int_buf + hdr_ptr->len, datalen);
        	if(scd->data_avail == NULL)
        	{
                	scd->data_avail = (char*)scd->recv_buffer + seq % SEQ_MOD %MAX_RWIN * max_seg;
         	        scd->av_seq = seq%SEQ_MOD;
        	}
	      }
	        /* abit[i] is the length of the ith packet
       		 * 0 if this packet is not received yet
	         */
       		scd->abit[seq%SEQ_MOD%MAX_RWIN] = datalen;
		if(scd->skip_next_L && (flags & L_SEG))
		{
/* 			printf("skip_next_L and L_SEG %d\n", seq%SEQ_MOD); */
			scd->lbit[seq%SEQ_MOD%MAX_RWIN] = TRUE;
			already_set = TRUE;
		}	
	        for(i = seq; i< seq + scd->rwin; i++)
       		{
                	if(scd->abit[i%SEQ_MOD%MAX_RWIN])
                	{
                            scd->next_expected = (scd->next_expected + 1) %SEQ_MOD;
			    scd->period_win_inc++;
			    scd->period_recvb += (float)scd->abit[i%SEQ_MOD%MAX_RWIN]/(float)1000;
		            scd->period_recvp ++;
                       	    if(i != seq) scd->outseq_cnt--;
			    if(i%SEQ_MOD == scd->last_outseq)
				scd->last_outseq = -1;
                            scd->avail_len += scd->abit[i%SEQ_MOD%MAX_RWIN];
	        	    scd->cumack_cnt++;
   		         }
                	else
                        	break;
			if(scd->skip_next_L)
			{
				scd->avail_len -= scd->abit[i%SEQ_MOD%MAX_RWIN];
				scd->abit[i%SEQ_MOD%MAX_RWIN]=0;

				if(scd->lbit[i%SEQ_MOD%MAX_RWIN])
				{
					scd->skip_next_L = FALSE;
/* 					printf("skip_next_L and lbit unset %d\n", i%SEQ_MOD); */
					scd->lbit[i%SEQ_MOD%MAX_RWIN] = FALSE;
				}

			}
			if(scd->abit[i%SEQ_MOD%MAX_RWIN]&&scd->av_seq == -1)	
			{
				scd->av_seq = i%SEQ_MOD;
                		scd->data_avail = (char*)scd->recv_buffer + i % SEQ_MOD %MAX_RWIN * max_seg;
			}
        	}

		if(scd->period_recvp == -1)
		{
			scd->period_recvp = 0;
			gettimeofday(&(scd->last_r_rep), &tz);
		}

       		if(scd->first_data || !(scd->cumack_cnt <  scd->max_cumack))
        	{
			scd->first_data = FALSE;
#ifdef MODULE	
                	ack_pkt = (char *)DAllocMM(sizeof(ack_pdu));
#else
                	ack_pkt = (char *)malloc(sizeof(ack_pdu));
#endif
	                ack_hdr = (eack_pdu *)ack_pkt;

       		        ack_hdr->flags = ACK + TSTAMP;
                	ack_hdr->len = sizeof(ack_pdu);
	                ack_hdr->seq = htons(-1); 
                	ack_hdr->ack = htons(seq);
			if(scd->next_expected != (seq + 1)%SEQ_MOD)
			{
				ack_hdr->flags += FORWARD;
				((ack_pdu *)ack_hdr)->forward = htons((scd->next_expected -1)%SEQ_MOD);
			}
	                ack_hdr->payload = htons(0);
			((ack_pdu *)ack_hdr)->partack = FALSE;
       		        gettimeofday(&tv, &tz);

#ifdef MODULE
			sock = look_up_pseudofd(scd->fd, (struct socket **) current->qfabric->fdqfabric);
                	if(sendto(sock, ack_pkt, sizeof(ack_pdu), 0,(struct sockaddr *)&remote_addr, sizeof(struct sockaddr_in), jiffies, 0) == 0)
#else
                	if(sendto(scd->fd, ack_pkt, sizeof(ack_pdu), 0,(struct sockaddr *)&remote_addr, sizeof(struct sockaddr_in)) == SOCKET_ERROR)
#endif
                	{
                       	 	svc->trace_out(cm, "rudp_data_available error: sendto\n");
                        	scd->senderr_cnt ++;
	                        pthread_mutex_unlock(&(scd->mutex));
#ifdef MODULE
				DFreeMM((addrs_t)ack_pkt);
				DFreeMM((addrs_t)int_buf);
#else
       		                free(ack_pkt);
                	        free(int_buf);
				fflush(NULL);
#endif
                       		return;
                	}
	                else    
				scd->senderr_cnt = 0;

	                svc->trace_out(cm,"rudp_data_available: ACK for seq %d is sent to port %d on host %s\n", (scd->next_expected-1)%SEQ_MOD, scd->remote_conn_port, inet_ntoa(remote_addr.sin_addr));
	                scd->cumack_cnt = 0;
       		        if(scd->cumack_tm != -1)
                        	scd->cumack_tm = CancelTimer(scd->cumack_tm);
	                if(ack_pkt)
#ifdef MODULE
				DFreeMM(ack_pkt);
#else
       		                free(ack_pkt);
#endif
        	}

        /*
         ack timer if necessary
         */
	 if(scd->cumack_cnt != 0 && scd->cumack_tm == -1)
       	     scd->cumack_tm = TimerSet(scd->cumack_to, &cumack_handler, scd, 0);

      }

      if (seq != old_next_expected)     /* seq is not next_expected */
      {
        /* seq not within receiving window, assume duplicate packet*/
        if(!in_rwin(seq, scd->next_expected, scd->rwin))
        {
                svc->trace_out(cm,"rudp_data_available: duplicate packet %d, next expected %d\n", seq, scd->next_expected);
#ifdef MODULE
                ack_pkt = (char *)DAllocMM(sizeof(ack_pdu));
#else
                ack_pkt = (char *)malloc(sizeof(ack_pdu));
#endif
                ack_hdr = (eack_pdu *)ack_pkt;
                ack_hdr->flags = ACK;
                ack_hdr->len = sizeof(ack_pdu);
                ack_hdr->payload = htons(0);
                ack_hdr->seq = htons(-1);
                ack_hdr->ack = scd->next_expected==0 ? MAX_SEQ:(scd->next_expected - 1)%SEQ_MOD;
		ack_hdr->ack = htons(ack_hdr->ack);
		((ack_pdu *)ack_hdr)->partack = FALSE;

		scd->cumack_cnt++;
                if(scd->cumack_cnt % scd->max_cumack ==0 && scd->state == CONNECTED)
                {
#ifdef MODULE
			sock = look_up_pseudofd(scd->fd, (struct socket **)current->qfabric->fdqfabric);
	                if(sendto(sock, ack_pkt, sizeof(ack_pdu), 0, (struct sockaddr *)&remote_addr, sizeof(struct sockaddr_in), jiffies, 0) == 0)
#else
	                if(sendto(scd->fd, ack_pkt, sizeof(ack_pdu), 0, (struct sockaddr *)&remote_addr, sizeof(struct sockaddr_in)) == SOCKET_ERROR)
#endif
       		        {
                        	svc->trace_out(cm, "rudp_data_available error: error sendto ack");
	                        scd->senderr_cnt ++;
       		                pthread_mutex_unlock(&(scd->mutex));
#ifdef MODULE
				DFreeMM((addrs_t)int_buf);
#else
	                        free(int_buf);
#endif
       		                return;
                	}
                	else
                	{
                       		scd->senderr_cnt = 0;
	                        svc->trace_out(cm, "rudp_data_available: ACK for %d is sent to port %d on %s", ntohs(ack_hdr->ack), scd->remote_conn_port, inet_ntoa(remote_addr.sin_addr));
				scd->cumack_cnt = 0;
				if(scd->cumack_tm!=-1)
					scd->cumack_tm = CancelTimer(scd->cumack_tm);
                	}
                }

#ifdef MODULE
                if(ack_hdr) DFreeMM((addrs_t)ack_hdr);
#else
                if(ack_hdr) free(ack_hdr);
#endif

        }
        else    /* in receiving window */
        {
		if(scd->av_seq != -1 && !in_rwin(seq, scd->av_seq,scd->rwin))
		{
/* 			printf("WARNING: will overwrite unread data\n"); */
			already_set = TRUE;
		}
		else
		{
		set_lastseq(scd, seq);
		scd->outseq_occur++;
		scd->period_rcv_err++;
	        memcpy((char *)scd->recv_buffer + seq%SEQ_MOD%MAX_RWIN * max_seg, int_buf + hdr_ptr->len, datalen);
       		scd->abit[seq % SEQ_MOD %MAX_RWIN]= datalen;

		if(scd->outseq_occur % scd->max_outseq == 0)
			outseq_handler(scd);
		}
         }   
      }  /* seq is not expected */
    }   /* datalen > 0 && receiving window not full */
    else
    if(datalen > 0) /* receiving window is full */
    {
	already_set = TRUE;
#ifdef MODULE
        ack_pkt = (char *)DAllocMM(sizeof(ack_pdu));
#else
        ack_pkt = (char *)malloc(sizeof(ack_pdu));
#endif

        ack_hdr = (eack_pdu *)ack_pkt;
        ack_hdr->flags = ACK;
        ack_hdr->len = sizeof(ack_pdu);
        ack_hdr->payload = htons(0);
        ack_hdr->seq = htons(-1);
        ack_hdr->ack = scd->next_expected==0?MAX_SEQ:(scd->next_expected - 1)%SEQ_MOD;
	ack_hdr->ack = htons(ack_hdr->ack);
	((ack_pdu *)ack_hdr)->partack = FALSE;

        if(scd->state == CONNECTED)
#ifdef MODULE
	sock = look_up_pseudofd(scd->fd, (struct socket **)current->qfabric->fdqfabric);
        if(sendto(sock, ack_pkt, sizeof(ack_pdu), 0, (struct sockaddr *)&remote_addr, sizeof(struct sockaddr_in), jiffies, 0) == 0)
#else
        if(sendto(scd->fd, ack_pkt, sizeof(ack_pdu), 0, (struct sockaddr *)&remote_addr, sizeof(struct sockaddr_in)) == SOCKET_ERROR)
#endif
        {
                svc->trace_out(cm, "rudp_data_available error: ack sendto");
                scd->senderr_cnt ++;
                pthread_mutex_unlock(&(scd->mutex));
#ifdef MODULE
		DFreeMM((addrs_t)int_buf);
#else
                free(int_buf);
#endif
                return;
        };
        scd->senderr_cnt = 0;
        svc->trace_out(cm, "rudp_data_available: receiving window full, ack for %d is sent\n", ntohs(ack_hdr->ack));

#ifdef MODULE
        if(ack_hdr) DFreeMM((addrs_t)ack_hdr);
#else
        if(ack_hdr) free(ack_hdr);
#endif
    } /* datalen > 0 but receiving window is full */
	
    if(datalen > 0 && !(seq < 0 ) && !already_set)
       scd->lbit[seq % SEQ_MOD%MAX_RWIN] = flags & L_SEG ? 1: 0;

    /* PART III. Process ACK information to data messages */
    /* III.1 An ACK presence means a positive ACK */
    if((flags & ACK && flags != ACK) || (flags==ACK && !(((ack_pdu *)hdr_ptr)->partack)))
    {
	if(flags == EXPSEND_FLAGS) 
		this_ack = recv_ack;
	else
	{
		if(flags & FORWARD)
			this_ack = ntohs(((ack_pdu *)hdr_ptr)->forward);
		else
		        this_ack = ntohs(hdr_ptr->ack);
	}

        svc->trace_out(cm,"ACK for seq %d from port %d, next_ack expected %d, outstanding %d\n", this_ack, ntohs(remote_addr.sin_port), scd->next_ack, scd->outstanding);

        /*
         * from the the packet last acked, up to the sequence acked in this
         * acknowledgement, are positively acknowledged
         */
        if(in_ackwin(this_ack, scd->next_ack, scd->outstanding))
        {
                int unwrapped_ack = unwrap(scd->next_ack, this_ack);
                for(i = scd->next_ack; i< unwrapped_ack + 1; i++)
                {
			scd->eacked[i%SEQ_MOD%MAX_SWIN] = 0;
                        scd->outstanding--;
                }
                scd->next_ack = (this_ack + 1)%SEQ_MOD;
                scd->retrans_cnt = 0;
		scd->cont_sent = 0;	
        }
                reset_needed = TRUE;
     }

    /* 
     * III.2 out sequence packets received by receiver, but within
     * receiver tolerance. Whether these out seq packets need to resent
     * depends on whether they are marked on the sender
     */
    if(flags == ACK && ((ack_pdu *)hdr_ptr)->partack && in_ackwin(ntohs(hdr_ptr->ack), scd->next_ack, scd->outstanding))
    {
	int unwrapped_ack;
        svc->trace_out(cm, "rudp_data_available: PART_ACK, total_err increased by %d, partack %d, ackno %d\n", (hdr_ptr->len - sizeof(eack_pdu))/sizeof(uint_16), ((ack_pdu*)hdr_ptr)->partack, ntohs(hdr_ptr->ack));
        eack_seqs = (uint_16 *)((uint_8 *)hdr_ptr + sizeof(eack_pdu));
        scd->total_err += (hdr_ptr->len - sizeof(eack_pdu))/sizeof(uint_16);
/*
	if(scd->total_sent==0 && scd->total_err !=0)
		eratio = 0;
	else
	        eratio = (float)scd->total_err * (float)100 / (float)scd->total_sent;
        svc->trace_out(cm, "rudp_data_available: error ratio %d, total_err %d, total_sent %d\n", eratio, scd->total_err, scd->total_sent);
        conn_attr_list = conn->attrs;
        send_err_atom = attr_atom_from_string("SEND_ERR");
        replace_attr(conn_attr_list, send_err_atom, Attr_Int4, (attr_value) eratio);

        if(eratio > RED_THRESH_HIGH)
                scd->redun_send = TRUE;
        if(eratio < RED_THRESH_LOW)
                scd->redun_send = FALSE;
*/
        ocnt = 0;
	forward = scd->next_ack;
	stop_forward = FALSE;
	
	/*
	 * check if unreceived packets are marked to have to be delivered 
	 * if not, we can roll forward to some certain sequence
	 */
	unwrapped_ack = unwrap(scd->next_ack, ntohs(hdr_ptr->ack));
        for(j = scd->next_ack; j < unwrapped_ack + 1; j++)
        {
                for(i = 0; i< (hdr_ptr->len - sizeof(eack_pdu))/sizeof(uint_16); i++)
                {
                        if(j%SEQ_MOD == ntohs(eack_seqs[i]))
                        {
                                ocnt++;
                                break;
                        }
                }
                if (i< (hdr_ptr->len - sizeof(ack_pdu))/sizeof(uint_16) && (j%SEQ_MOD == ntohs(eack_seqs[i])) && scd->marked[j%SEQ_MOD%MAX_SWIN])
                {
                        retrans_buf = (char*)scd->send_buffer + j % SEQ_MOD%MAX_SWIN * max_seg;
                        retrans_len = scd->sent_len[j % SEQ_MOD%MAX_SWIN];

                        if(scd->state == CONNECTED)
                        {
#ifdef MODULE
				sock = look_up_pseudofd(scd->fd, (struct socket **)current->qfabric->fdqfabric);
                	        if( sendto(sock, retrans_buf, retrans_len, 0,(struct sockaddr *)& remote_addr, sizeof(struct sockaddr_in), jiffies, 0) == 0)
#else
                	        if( sendto(scd->fd, retrans_buf, retrans_len, 0,(struct sockaddr *)& remote_addr, sizeof(struct sockaddr_in)) == SOCKET_ERROR)
#endif
                       		{
                                	svc->trace_out(cm, "rudp_data_available: retrans for eack fail");
                                	scd->senderr_cnt++;
                        	}
                        	else    scd->senderr_cnt = 0;
                        }
                        svc->trace_out(cm, "rudp_data_available: packet %d resent for PART_ACK , outstanding %d, next_ack %d\n", j%SEQ_MOD, scd->outstanding, scd->next_ack);
			gettimeofday(&(scd->send_stamp[j % SEQ_MOD%MAX_SWIN]),&tz);

                        scd->total_sent ++;
			stop_forward = TRUE;
			reset_needed = TRUE;
                }
		else
		if(!stop_forward) 
		{
			forward = (j+1)%SEQ_MOD;
			scd->eacked[j%SEQ_MOD%MAX_SWIN] = 0;
		}

		if(j%SEQ_MOD == ntohs(hdr_ptr->ack))
                        break;
	
        }

	if(forward != scd->next_ack)
	{
		scd->outstanding -= (forward > scd->next_ack) ? forward - scd->next_ack: (forward + SEQ_MOD - scd->next_ack);
		scd->next_ack = forward;
		scd->forward = TRUE;
	}
        reset_needed = TRUE;
    }

    /* III.3. EACK, definitely need to resend */
    if(flags & EAK)
    {
/*       svc->trace_out(cm, "rudp_data_available: EACK, total_err increased by %d\n", (hdr_ptr->len - sizeof(eack_pdu))/sizeof(uint_16)); */
        svc->trace_out(cm, "rudp_data_available: EACK\n");
        eack_seqs = (uint_16 *)((uint_8 *)hdr_ptr + sizeof(eack_pdu));
/*       scd->total_err += hdr_ptr->len - sizeof(eack_pdu); */

        ocnt = 0;
        for(j = scd->next_ack; (ocnt < (hdr_ptr->len - sizeof(eack_pdu))/sizeof(uint_16)) && j < scd->next_ack + scd->outstanding; j++)
        {
                for(i = 0; i< (hdr_ptr->len - sizeof(eack_pdu))/sizeof(uint_16); i++)
                {
                        if(j%SEQ_MOD == ntohs(eack_seqs[i]))
                        {
                                ocnt++;
                                break;
                        }
                }
                if (i< (hdr_ptr->len - sizeof(eack_pdu))/sizeof(uint_16) && (j%SEQ_MOD == ntohs(eack_seqs[i])))
                {
		     if(scd->eacked[j%SEQ_MOD%MAX_SWIN] % EACK_MOD == 0)
		     {
                        retrans_buf = (char*)scd->send_buffer + j % SEQ_MOD%MAX_SWIN * max_seg;
                        retrans_len = scd->sent_len[j % SEQ_MOD%MAX_SWIN];

                        if(scd->state == CONNECTED)
                        {
#ifdef MODULE
				sock = look_up_pseudofd(scd->fd, (struct socket **)current->qfabric->fdqfabric);

                	        if( sendto(sock, retrans_buf, retrans_len, 0,(struct sockaddr *)& remote_addr, sizeof(struct sockaddr_in), jiffies, 0) == 0)
#else
                	        if( sendto(scd->fd, retrans_buf, retrans_len, 0,(struct sockaddr *)& remote_addr, sizeof(struct sockaddr_in)) == SOCKET_ERROR)
#endif
                       		{
                                	svc->trace_out(cm, "rudp_data_available: retrans for eack fail");
                                	scd->senderr_cnt++;
                        	}
                        	else    scd->senderr_cnt = 0;
                        }
                        svc->trace_out(cm, "rudp_data_available: packet %d resent for EACK ,outstanding %d, next_ack %d\n", j%SEQ_MOD, scd->outstanding, scd->next_ack);
			gettimeofday(&(scd->send_stamp[j % SEQ_MOD%MAX_SWIN]),&tz);
                        scd->total_sent ++;
			scd->total_err ++;
		     }
		     scd->eacked[j%SEQ_MOD%MAX_SWIN] ++;
                }
                reset_needed = TRUE;
        }
    }

    /* PART IV. Dealing with RTT update*/
    /*. IV.1. explicit update request by the other party */
    update_ptr = (update_pdu *)int_buf;
    if(update_ptr->flags == UPDATE_FLAGS)
    {
	scd->retrans_to = ntohl(((update_pdu *)int_buf)->newretrans);
        scd->cumack_to = scd->retrans_to/2;
	svc->trace_out(cm, "rudp_data_available: UPDATE request with new retrans_to %d\n", scd->retrans_to);
    }

    /*
     * IV.2. ACKs that are sent right after corresponding data 
     * messages are received
     */ 
    if(flags & TSTAMP && !(flags == EXPSEND_ACK_FLAGS))
    {
        /* this is a packet containing appropriate RTT information */
        gettimeofday(&tv, &tz);
        timersub(&tv, &(scd->send_stamp[ntohs(hdr_ptr->ack) % SEQ_MOD% MAX_SWIN]), &tdiff);
        trtt = tdiff.tv_sec * 1000 + tdiff.tv_usec/1000;
        o_retrans = scd->retrans_to;
	(void) o_retrans;  /* just so this isn't marked "set but unused" */
        if(!scd->first_measure)
        {
                scd->mrtt = 0.875 * scd->mrtt + 0.125 * trtt;
                scd->sdev = 0.875* scd->sdev + 0.125 * abs(scd->mrtt - trtt);
        }
        else
        {
                scd->first_measure = FALSE;
#ifdef MODULE
                scd->mrtt = my_max(trtt, 5);
#else
                scd->mrtt = max(trtt, 5);
#endif
        }

	temp = 2* scd->max_cumack * scd->mrtt + 4 * scd->sdev;	
	/* update timeout values only with some thresholds */
#ifdef MODULE
	if((temp - scd->retrans_to) /scd->retrans_to > UPDATE_THRESH_R && (temp -scd->retrans_to > UPDATE_THRESH_A))
#else
	if((float) abs(temp - scd->retrans_to) /(float)scd->retrans_to > UPDATE_THRESH_R && abs(temp -scd->retrans_to > UPDATE_THRESH_A))
#endif
	{
		memset((void *)&update_pkt, 0, sizeof(update_pdu));
		update_pkt.flags = UPDATE_FLAGS;
		update_pkt.len = sizeof(update_pdu);
		update_pkt.newretrans = htonl(temp); 
		tolen = sizeof(struct sockaddr_in );
		/* update the other party, can't change independently */
#ifdef MODULE
		sock = look_up_pseudofd(scd->fd, (struct socket **)current->qfabric->fdqfabric);
		if(sendto(sock, (char *)&update_pkt, update_pkt.len, 0, (struct sockaddr *)&remote_addr, tolen, jiffies, 0) != 0)	
#else
		if(sendto(scd->fd, (char *)&update_pkt, update_pkt.len, 0, (struct sockaddr *)&remote_addr, tolen) != SOCKET_ERROR)	
#endif
		{
			scd->retrans_to = 2* scd->max_cumack * scd->mrtt + 4 * scd->sdev;
/* 			scd->retrans_to = scd->max_cumack * scd->mrtt + 4 * scd->sdev; */
		       	scd->cumack_to = scd->retrans_to/2;
        		svc->trace_out(cm, "packet containing ack_stamp is %d trtt is %ld, new retrans_timeout value %d, new cumack timeout value %d", ntohs(hdr_ptr->ack), trtt, scd->retrans_to, scd->cumack_to);
		}
	}
    }

   /* PART V. reset timer if necessary */
   if(reset_needed)
        /*
         * update the retransmission timer whenever there is an ACK to a
         * previously unacked packet
         */
   {
        if(scd->outstanding > 0)
	{
                TimerUpdate(scd->retrans_tm, scd->retrans_to, &retrans_handler, scd, 0);
		svc->trace_out(cm, "rudp_data_available: retrans_tm reset %d\n", scd->retrans_to);
	}
        else
	{
                scd->retrans_tm = CancelTimer(scd->retrans_tm);
		svc->trace_out(cm, "rudp_data_available: retrans timer canceled\n");
	}
    }

    reset_needed = FALSE;
    pthread_mutex_unlock(&(scd->mutex));
    /* if there is data available for app, we call CMDataAvailable */
    if(scd->avail_len > 0 && scd->av_seq != -1)
    {
        int unwrapped_expected = unwrap(scd->av_seq, scd->next_expected);

        for(i = scd->av_seq; i < unwrapped_expected; i++)
	    if(scd->lbit[i % SEQ_MOD %MAX_RWIN])
		{
			svc->trace_out(cm, "scd->av_seq %d, scd->next_expected %d\n", scd->av_seq, scd->next_expected);
			svc->trace_out(cm, "DataAvail called with L_SEG %d\n", i%SEQ_MOD);
                        trans->data_available(trans, conn);
		}
	
    }

  }while(1); /*while*/

  if(scd->period_recvp > REPORT_MOD )
  {
	gettimeofday(&tv, &tz);
	timersub(&tv, &scd->last_r_rep, &tdiff);
	msec = tdiff.tv_sec *1000 + tdiff.tv_usec/1000;
	if(msec!=0)
	{
		recv_rate = scd->period_recvb * 1000 / msec;
		win_inc = scd->period_win_inc * 1000/ msec;
	}
	else
	{
		win_inc = 0;
		recv_rate = 0;
	}

	recv_err = scd->rcv_apperr * 100 /(scd->period_recvp + scd->rcv_apperr);

#ifdef ENABLE_RCV_REPORT
	svc->trace_out(cm, "rudp_data_available: periodic receiver data rate %d(KB/sec), err ratio %d, window increase %d(/sec)", recv_rate, recv_err, win_inc);
	svc->trace_out(cm, "rudp_data_available: periodic receiver err %d\n", scd->period_rcv_err *100 / scd->period_recvp);
        conn_attr_list = conn->attrs;
	recv_rate_atom = attr_atom_from_string("RECV_RATE");
	recv_err_atom = attr_atom_from_string("RECV_ERR");
	recv_winc_atom = attr_atom_from_string("WIN_INC");
	replace_attr(conn_attr_list, recv_rate_atom, Attr_Int4, (attr_value)recv_rate);
	replace_attr(conn_attr_list, recv_err_atom, Attr_Int4, (attr_value)recv_err);	
	replace_attr(conn_attr_list, recv_winc_atom, Attr_Int4, (attr_value)win_inc);
#endif
	scd->period_recvp = 0;
	scd->period_recvb = 0;
	scd->period_rcv_err = 0;
	scd->rcv_apperr = 0;
	scd->last_r_rep = tv;
	scd->period_win_inc = 0;
  }
 
/*
	TODO: sender side update
	if(scd->total_sent != 0)
	{
        eratio = (float)scd->total_err * (float)100 / (float)scd->total_sent;
        svc->trace_out(cm, "rudp_data_available: current error ratio %d, total_err %d, total_sent %d\n", eratio, scd->total_err, scd->total_sent);
        conn_attr_list = conn->attrs;
        send_err_atom = attr_atom_from_string("SEND_ERR");
        replace_attr(conn_attr_list, send_err_atom, Attr_Int4, (attr_value) eratio);

        if(eratio > RED_THRESH_HIGH)
                scd->redun_send = TRUE;
        if(eratio < RED_THRESH_LOW)
           	scd->redun_send = FALSE;
	}
*/
#ifdef MODULE
  DFreeMM((addrs_t)int_buf);
#else
  free(int_buf);
#endif
  return;
}

extern void *
libcmrudp_LTX_initialize(cm, svc)
CManager cm;
CMtrans_services svc;
{
   
    rudp_client_data_ptr rudp_data;
    svc->trace_out(cm, "Initialize RUDP transport\n");
    if (rudp_global_init == 0) {
#ifdef HAVE_WINDOWS_H
        int nErrorStatus;
        /* initialize the winsock package */
        nErrorStatus = WSAStartup(wVersionRequested, &wsaData);
        if (nErrorStatus != 0) {
            fprintf(stderr, "Could not initialize windows socket library!");
            WSACleanup();
            exit(-1);
        }
#endif
        /*
         * ignore SIGPIPE's  (these pop up when ports die.  we catch the
         * failed writes)
         */
#ifdef SIGPIPE
        signal(SIGPIPE, SIG_IGN);
#endif

	/*
	 * initialize the timer thread
	 */
        TimerStart(NULL);
    }

    rudp_data = svc->malloc(sizeof(struct rudp_client_data));

    rudp_data->cm = cm;
    rudp_data->hostname = NULL;
    rudp_data->listen_port = -1;
    rudp_data->svc = svc;
    return (void*)rudp_data;
}

/* Validate whether this is a valid SYN packet */ 
int validate_syn(syn_pdu *syn)
{
	/* just a placeholder at this moment */
	return 1;
}

void init_reg(rudp_region_spec *reg)
{
    memset((void *)reg, 0, sizeof(rudp_region_spec));
    reg->eratio = -1;
    reg->handler = NULL;	
    reg->exp_rate = -1;
}

extern CMConnection
libcmrudp_LTX_initiate_conn(cm, svc, trans, attrs)
CManager cm;
CMtrans_services svc;
transport_entry trans;
attr_list attrs;
{
    int sock;

    int rwin = RWIN_DEF, int_port_num, tolen, fromlen;
#ifndef MODULE
    int flags;
#endif
    u_short port_num;
    rudp_conn_data_ptr rudp_conn_data;
    CMConnection conn;
    syn_pdu this_syn, peer_syn;

    static atom_t host_atom = 0;
    static atom_t port_atom = 0;
    static atom_t recv_rate_atom = 0;
    static atom_t recv_err_atom = 0;
    static atom_t fd_atom = 0;
    static atom_t peer_ip_atom = 0;
    static atom_t peer_conn_port_atom = 0;
    static atom_t this_conn_port_atom = 0;
    static atom_t peer_listen_port_atom = 0;
    static atom_t peer_hostname_atom = 0;
    static atom_t send_err_atom = 0;
    static atom_t send_rate_atom = 0;
    static atom_t rt_rate_atom = 0;	
    static atom_t recv_winc_atom = 0; 
    static atom_t maxseg_atom = 0;
    char *host_name;
    int remote_IP = -1, max_seg = MAX_PACKET;
    attr_list conn_attr_list;
    int sock_len;
    struct sockaddr_in sock_addr;
    struct sockaddr_in *sock_addri = (struct sockaddr_in *)&sock_addr;
    struct timeval tv1, tv2, tdiff;
#ifdef MODULE
    struct socket *socket;
    unsigned long randbuf;
#else
    struct timezone tz;
#endif

    struct sockaddr_in my_addr;
    if (host_atom == 0) {
	/* to parse attributes from attrs */
        host_atom = attr_atom_from_string("IP_HOST");
        port_atom = attr_atom_from_string("UDP_PORT");
       	
	/* to store attributes in CM's space */ 
        fd_atom = attr_atom_from_string("CONNECTION_FILE_DESCRIPTOR");
        this_conn_port_atom = attr_atom_from_string("THIS_CONN_PORT");
        peer_conn_port_atom = attr_atom_from_string("PEER_CONN_PORT");
        peer_ip_atom = attr_atom_from_string("PEER_IP_ADDR");
        peer_listen_port_atom = attr_atom_from_string("PEER_LISTEN_PORT");
        peer_hostname_atom = attr_atom_from_string("PEER_HOSTNAME");
	send_err_atom = attr_atom_from_string("SEND_ERR");
	rt_rate_atom = attr_atom_from_string("RTSEND_RATE");
	send_rate_atom = attr_atom_from_string("SEND_RATE");	
	recv_rate_atom = attr_atom_from_string("RECV_RATE");
	recv_winc_atom = attr_atom_from_string("WIN_INC");
	recv_err_atom = attr_atom_from_string("RECV_ERR");
	maxseg_atom = attr_atom_from_string("MAX_SEG");
    }

    if (!query_attr(attrs, host_atom, /* type pointer */ NULL,
                    /* value pointer */ (attr_value *) &host_name)) {
        svc->trace_out(cm, "RUDP transport found no IP_HOST attribute");
        return NULL;
    }
    if (!query_attr(attrs, port_atom, /* type pointer */ NULL,
                    /* value pointer */ (attr_value *) &int_port_num)) {
        svc->trace_out(cm, "RUDP transport found no IP_PORT attribute");
        return NULL;
    }
    if (!query_attr(attrs, maxseg_atom, /* type pointer */ NULL,
                    /* value pointer */ (attr_value *) &max_seg)) {
        svc->trace_out(cm, "RUDP transport found no MAX_SEG attribute");
    }

    port_num = int_port_num;

    svc->trace_out(cm, "Attempting RUDP socket connection, host=\"%s\", port %d", host_name, int_port_num);
    {
        /* INET socket connection, host_name is the machine name */
        char *network_string;
#ifdef MODULE
        if (sock_create(AF_INET, SOCK_DGRAM, 0, &socket) < 0) 
#else
        if ((sock = socket(AF_INET, SOCK_DGRAM, 0)) == SOCKET_ERROR) 
#endif
	{
            svc->trace_out(cm,"CMRUDP connect FAILURE --> Couldn't create socket");
            return NULL;
        }
#ifdef MODULE
	sock = get_pseudofd(socket, (struct socket **)current->qfabric->fdqfabric);
#endif

        sock_addr.sin_family = AF_INET;
        if ((network_string = getenv("CM_NETWORK")) != NULL) 
	{
            int name_len = strlen(host_name) + 2 + strlen(network_string);
            char *new_host_name = svc->malloc(name_len);
            char *first_dot = strchr(host_name, '.');
            memset(new_host_name, 0, name_len);
            if (first_dot == NULL) 
            {
                strcpy(new_host_name, host_name);
                strcat(new_host_name, "-");
                strcat(new_host_name, network_string);
            } else {
                strncpy(new_host_name, host_name, first_dot - host_name);
                strcat(new_host_name, "-");
                strcat(new_host_name, network_string);
                strcat(new_host_name, first_dot);
            }

            if (check_host(new_host_name, (void*)&sock_addri->sin_addr) == 0)  
	    {
                /* host has no NETWORK interface */
                if (check_host(host_name, (void*)&sock_addri->sin_addr) == 0) 
		{
                    svc->trace_out(cm, "--> Host not found \"%s\"",
                                    host_name);
                    return NULL;
                }
             } else {
                svc->trace_out(cm, "--> Using non default network interface with hostname %s", new_host_name);
             }
            svc->free(new_host_name);
         } else /* network_string != NULL */
	 {
            if (check_host(host_name, (void*)&sock_addri->sin_addr) == 0) 
	    {
                svc->trace_out(cm, "CMRUDP connect FAILURE --> Host not found: %s", host_name);
                return NULL;
            }
        } /* network_string == NULL */

        sock_addri->sin_port = htons(port_num);
        remote_IP = sock_addri->sin_addr.s_addr;

	memset((void *)&this_syn, 0, sizeof(syn_pdu));
	this_syn.flags = SYN;
	this_syn.len = SYN_LEN;
#ifdef MODULE
	get_random_bytes(&randbuf, sizeof(unsigned long));
	this_syn.seq = randbuf/ULONG_MAX * ((1<<16)-1);
#else
	this_syn.seq = lrand48();
#endif
	this_syn.ver = 1;
	this_syn.max_outstanding = 4;
	this_syn.opt = 0;	 	
	this_syn.max_segsize = my_min(MAX_PACKET, max_seg);
	this_syn.retrans_to = RETRANS;
	this_syn.cumack_to = CUMACK;
	this_syn.null_to = KEEPALIVE;
	this_syn.transtat_to = 1000;
	this_syn.max_retrans = 120;
	this_syn.max_cumack = 3;
	this_syn.max_outseq = 3;
	this_syn.max_auto_rst = 120;
	this_syn.connid = global_connid ++; 
	this_syn.cksum = 0;

	tv1.tv_sec = 1;
	tv1.tv_usec = 0;
#ifndef MODULE
        setsockopt(sock, SOL_SOCKET, SO_SNDTIMEO, &tv1, sizeof(tv1));
        setsockopt(sock, SOL_SOCKET, SO_RCVTIMEO, &tv1, sizeof(tv1));
#endif

	gettimeofday(&tv1, &tz);
	tolen = sizeof(struct sockaddr_in); 	

	my_addr.sin_family = AF_INET;
	my_addr.sin_addr.s_addr = INADDR_ANY;
	my_addr.sin_port = htons(0);
#ifdef MODULE
	if (socket->ops->bind(socket, (struct sockaddr *)&my_addr, sizeof my_addr)<0) {
		printf("Error binding client socket\n");
		exit(-1);
	}
	if(sendto(socket, (void *)&this_syn, SYN_LEN, 0, (struct sockaddr *)sock_addri, tolen, jiffies, 0) == 0)
#else
	if(sendto(sock, (void *)&this_syn, SYN_LEN, 0, (struct sockaddr *)sock_addri, tolen) == SOCKET_ERROR)
#endif
	{
                svc->trace_out(cm, "CMRUDP connect FAILURE --> sendto() failed");
		close(sock);
		return NULL;
	}
	else
		svc->trace_out(cm, "Initiate connection to host %s: port %d\n", inet_ntoa(sock_addri->sin_addr), ntohs(sock_addri->sin_port));

	fromlen = sizeof(struct sockaddr_in);
	/*it's a little risky to recvfrom(.,.,.,., NULL, .) */
#ifdef MODULE
	if(recvfrom(socket, (void *)&peer_syn, SYN_LEN, 0, (struct sockaddr *)sock_addri, &fromlen, 2) < 0)
#else
	if(recvfrom(sock, (void *)&peer_syn, SYN_LEN, 0, (struct sockaddr *)sock_addri, &fromlen) == SOCKET_ERROR)
#endif
	{
                svc->trace_out(cm, "CMRUDP connect FAILURE --> recvfrom() failed");
		close(sock);
		return NULL;
	}

	gettimeofday(&tv2, &tz);
	timersub(&tv2, &tv1, &tdiff);

	if(validate_syn(&peer_syn) == -1)
	{
                svc->trace_out(cm, "CMRUDP connect FAILURE --> SYN_PDU invalid");
		close(sock);
		return NULL;
	}
	
#ifdef MODULE
	if(sendto(socket, (void *)&peer_syn, SYN_LEN, 0, (struct sockaddr *)sock_addri, fromlen, jiffies, 0) == 0)
#else
	if(sendto(sock, (void *)&peer_syn, SYN_LEN, 0, (struct sockaddr *)sock_addri, fromlen) == SOCKET_ERROR)
#endif
	{
		svc->trace_out(cm, "CMRUDP connect FAILURE --> client SYN_REPLY fails");
		close(sock);
		return NULL;
	}
   }

    svc->trace_out(cm,"peer connection port is %d", int_port_num);
    svc->trace_out(cm, "--> Connection established");
    rudp_conn_data = create_rudp_conn_data(svc);
    rudp_conn_data->remote_host = strdup(host_name);
    rudp_conn_data->remote_IP = remote_IP;
    rudp_conn_data->remote_contact_port = port_num;
    rudp_conn_data->remote_conn_port = int_port_num;
    rudp_conn_data->fd = sock;
    rudp_conn_data->cm = cm;
    rudp_conn_data->first_measure = TRUE;
    rudp_conn_data->first_data = TRUE;
    rudp_conn_data->last_outseq = -1;
    rudp_conn_data->adapt_inprog = FALSE;
    rudp_conn_data->app_adapted = FALSE;
    rudp_conn_data->max_segsize = this_syn.max_segsize;

/* the following four lines regarding buffer need more thought */
    rudp_conn_data->recv_buffer_len = MAX_RWIN;
#ifdef MODULE
    rudp_conn_data->recv_buffer = vmalloc(MAX_RWIN * rudp_conn_data->max_segsize);
#else
    rudp_conn_data->recv_buffer = svc->malloc(MAX_RWIN * rudp_conn_data->max_segsize);
#endif
    rudp_conn_data->send_buffer_len = MAX_SWIN;
#ifdef MODULE
    rudp_conn_data->send_buffer = vmalloc(MAX_SWIN * rudp_conn_data->max_segsize);
#else
    rudp_conn_data->send_buffer = svc->malloc(MAX_SWIN * rudp_conn_data->max_segsize);
#endif

    rudp_conn_data->max_outstanding = peer_syn.max_outstanding;
    rudp_conn_data->max_byte = rudp_conn_data->max_outstanding*rudp_conn_data->max_segsize;
    rudp_conn_data->retrans_to = this_syn.retrans_to;
    rudp_conn_data->mrtt = this_syn.retrans_to;
    rudp_conn_data->sdev = 0;
    rudp_conn_data->cumack_to = this_syn.cumack_to;
    rudp_conn_data->null_to = this_syn.null_to;
    rudp_conn_data->transtat_to = this_syn.transtat_to;
    rudp_conn_data->max_retrans = this_syn.max_retrans;
    rudp_conn_data->max_cumack = this_syn.max_cumack;
    rudp_conn_data->max_outseq = this_syn.max_outseq;
    rudp_conn_data->max_auto_rst = this_syn.max_auto_rst;
    rudp_conn_data->max_keepalive = MAXKEEPALIVE;
    rudp_conn_data->rwin = rwin;
    rudp_conn_data->connid = peer_syn.connid;
    rudp_conn_data->local_connid = this_syn.connid;
    rudp_conn_data->period_bytes = 0;
    rudp_conn_data->senderr_cnt = 0;
    rudp_conn_data->recverr_cnt = 0;
    rudp_conn_data->period_recvp = -1;
    rudp_conn_data->period_recvb = rudp_conn_data->period_rcv_err = 0 ;
    rudp_conn_data->rcv_apperr = 0 ;
 
/* conn state */
    rudp_conn_data->state = CONNECTED;
    rudp_conn_data->retrans_tm = -1; 
    rudp_conn_data->retrans_cnt = 0;
    rudp_conn_data->cumack_cnt = 0;
    rudp_conn_data->outseq_cnt = 0;
    rudp_conn_data->outseq_occur = 0;
    rudp_conn_data->rst_cnt = 0;
    rudp_conn_data->cumack_cnt = 0;
    rudp_conn_data->cumack_tm = -1; 
    rudp_conn_data->null_tm = TimerSet(KEEPALIVE, &null_handler, rudp_conn_data,0);
    rudp_conn_data->total_sent = 0;
    rudp_conn_data->total_err = 0;
    rudp_conn_data->exp_send_tm = -1;
    rudp_conn_data->exp_send_last = 1;
    rudp_conn_data->redun_send = FALSE;
    rudp_conn_data->svc = svc;
    rudp_conn_data->last_retrans = -1;

    init_reg(&(rudp_conn_data->send_qos));

#ifndef MODULE
    pthread_mutex_init(&(rudp_conn_data->mutex), NULL);
    pthread_cond_init(&(rudp_conn_data->die), NULL);
#endif
    rudp_conn_data->fin = FALSE;
    
    conn_attr_list = create_attr_list();

    conn = svc->connection_create(trans, rudp_conn_data, conn_attr_list);

    if (sock_addr.sin_family == AF_INET) {
#ifndef MODULE
            struct hostent *host;
            struct sockaddr_in *in_sock = (struct sockaddr_in *) &sock_addr;
            host = gethostbyaddr((char *) &in_sock->sin_addr,
                                 sizeof(struct in_addr), AF_INET);
            if (host != NULL) {
                rudp_conn_data->remote_host = strdup(host->h_name);

                add_attr(conn_attr_list, peer_hostname_atom, Attr_String,
                         (attr_value) strdup(host->h_name));
            }
#endif
        }

    add_attr(conn_attr_list, fd_atom, Attr_Int4,
             (attr_value) sock);
    add_attr(conn_attr_list, peer_conn_port_atom, Attr_Int4,
                 (attr_value) int_port_num);

    sock_len =  sizeof(sock_addr);
#ifdef MODULE
    socket->ops->getname(socket, (struct sockaddr *)&sock_addr, &sock_len, 0);
#else
    getsockname(sock, (struct sockaddr *) &sock_addr, &sock_len);
#endif
    svc->fd_add_select(cm, sock, (select_list_func)rudp_data_available, (void *)trans, (void *)conn);
    int_port_num = ntohs(((struct sockaddr_in *)&sock_addr)->sin_port);
    add_attr(conn_attr_list, this_conn_port_atom, Attr_Int4,
             (attr_value) int_port_num);
    add_attr(conn_attr_list, peer_ip_atom, Attr_Int4,
             (attr_value) rudp_conn_data->remote_IP);

    add_attr(conn_attr_list, peer_listen_port_atom, Attr_Int4,
             (attr_value) rudp_conn_data->remote_contact_port);
    add_attr(conn_attr_list, send_err_atom, Attr_Int4, (attr_value) 0);
    add_attr(conn_attr_list, send_rate_atom, Attr_Int4, (attr_value) 0);
    add_attr(conn_attr_list, recv_rate_atom, Attr_Int4, (attr_value) 0);
    add_attr(conn_attr_list, recv_winc_atom, Attr_Int4, (attr_value) 0);
    add_attr(conn_attr_list, recv_err_atom, Attr_Int4, (attr_value) 0); 
    add_attr(conn_attr_list, rt_rate_atom, Attr_Int4, (attr_value) 0);

    rudp_conn_data->conn = conn;

#ifndef MODULE
    if( (flags = fcntl(sock, F_GETFL, 0)) < 0)
	svc->trace_out(cm, "error getting socket options");

    flags |= O_NONBLOCK;
    if(fcntl(sock, F_SETFL, flags) == SOCKET_ERROR)
   	svc->trace_out(cm, "error setting nonblocking socket option");
#endif

    return conn;
}

static rudp_conn_data_ptr create_rudp_conn_data(svc)
CMtrans_services svc;
{
    int i;

    rudp_conn_data_ptr rudp_conn_data =
        svc->malloc(sizeof(struct rudp_connection_data));
    memset((void *)rudp_conn_data, 0, sizeof(struct rudp_connection_data));

    rudp_conn_data->remote_host = NULL;
    rudp_conn_data->remote_contact_port = -1;
    rudp_conn_data->fd = 0;
    rudp_conn_data->recv_buffer = NULL;
    rudp_conn_data->recv_buffer_len = MAX_RWIN;
    rudp_conn_data->send_buffer = NULL;
    rudp_conn_data->send_buffer_len = MAX_SWIN;
    rudp_conn_data->data_avail = NULL;

    for(i=0; i< MAX_RWIN; i++)
    {
	rudp_conn_data->abit[i] = 0;
	rudp_conn_data->lbit[i] = 0;
    }

    for(i=0; i< MAX_SWIN; i++)
    {
	rudp_conn_data->acked[i] = FALSE;
	rudp_conn_data->sent_len[i] = 0;
    }

    return rudp_conn_data;
}

int shutdown_conn_force(int timer_idx, rudp_conn_data_ptr scd, int arg2)
{
    scd->svc->trace_out(scd->cm, "Pending ACK but forced to shutdown_conn");
  
    pthread_mutex_lock(&(scd->mutex));
    scd->outstanding = 0;
/*  scd->state = FIN; */
    pthread_mutex_unlock(&(scd->mutex));
    return 0; 
}   

#ifdef MODULE
void scd_wake_up(unsigned long data)
{
  wake_up_interruptible(&scd_queue);
}
#endif

extern void
libcmrudp_LTX_shutdown_conn(svc, scd)
CMtrans_services svc;
rudp_conn_data_ptr scd;
{
    struct sockaddr_in remote_addr;
    fin_pdu fin_pkt;
    int tolen, tm;
#ifdef MODULE
    struct socket *sock;
    int counter;
    DECLARE_WAITQUEUE(scd_wait, current);
#else
    int ret;
#endif

    svc->trace_out(scd->cm, "shuting down conn\n");
    remote_addr.sin_addr.s_addr = scd->remote_IP;
    remote_addr.sin_port = htons(scd->remote_conn_port);
    remote_addr.sin_family = AF_INET; 
  
    memset((void *)&fin_pkt, 0, sizeof(fin_pdu));
    fin_pkt.flags = CHK;
    fin_pkt.len = sizeof(fin_pkt);
    fin_pkt.seq = htons(-1); 
    fin_pkt.ack = scd->next_expected==0?MAX_SEQ:(scd->next_expected - 1)%SEQ_MOD;  
    fin_pkt.ack = htons(fin_pkt.ack);  
    tolen = sizeof(struct sockaddr_in );
  
    pthread_mutex_lock(&(scd->mutex));
    svc->trace_out(scd->cm, "Session statistics: total err %d, total packets sent %d\n", scd->total_err, scd->total_sent);
//    printf("Session statistics: total err %d, total packets sent %d\n", scd->total_err, scd->total_sent);
    tm = TimerSet(LINGER, &shutdown_conn_force, scd, 0);
    while(scd->outstanding > 0)
    {
	pthread_mutex_unlock(&(scd->mutex));
	sched_yield();	
	pthread_mutex_lock(&(scd->mutex));
    }
    
    CancelTimer(tm);
#ifdef MODULE
    sock = look_up_pseudofd(scd->fd, (struct socket **)current->qfabric->fdqfabric);
    if(sendto(sock, (void *)&fin_pkt, sizeof(fin_pkt), 0, 
		(struct sockaddr *)&remote_addr, tolen, -1, 0) == 0)
#else
    if(sendto(scd->fd, (void *)&fin_pkt, sizeof(fin_pkt), 0, 
		(struct sockaddr *)&remote_addr, tolen) == SOCKET_ERROR)
#endif
    {
		scd->senderr_cnt++;
		svc->trace_out(scd->cm, "shutdown error: sendto");
    }
    else	
		scd->senderr_cnt = 0;

#ifdef MODULE
    counter = 0;
    if(scd->state != FIN) {
            while (scd->die != 1) {
            	up(&(scd->mutex));
		del_timer(&scd->scd_timer);
		scd->scd_timer.function = scd_wake_up;
		scd->scd_timer.data = 0;
		scd->scd_timer.expires = jiffies+PERIOD;
		add_timer(&scd->scd_timer);
		current->qfabric->sleep_queue = (unsigned long)&scd_queue;
		current->state = TASK_INTERRUPTIBLE;
		add_wait_queue(&scd_queue,&scd_wait);
		schedule();
		current->state = TASK_RUNNING;
		current->qfabric->sleep_queue = (unsigned long) 0;
		remove_wait_queue(&scd_queue,&scd_wait);
		del_timer(&scd->scd_timer);
		counter++;
		if (counter > 20) {
			printk("Wait failed!\n");
			return;
		}
	    }
	    scd->die = 0;
            down(&(scd->mutex));
    }
#else
    if(scd->state != FIN)
	    pthread_cond_wait(&(scd->die), &(scd->mutex));
#endif

    pthread_mutex_unlock(&(scd->mutex));
}   

/*
 * Create an IP socket for connection from other CMs
 */
extern attr_list
libcmrudp_LTX_non_blocking_listen(cm, svc, trans, listen_info)
CManager cm;
CMtrans_services svc;
transport_entry trans;
attr_list listen_info;
{
    rudp_client_data_ptr sd; 
    int length, max_seg=0;
    struct sockaddr_in sock_addr;
    int sock_opt_val = 1;
    int conn_sock;
    int int_port_num = 0;
    u_short port_num = 0;

    char host_name[256], int_addr[256];
    static atom_t host_atom = 0, port_atom = 0, transport_atom = 0;
    static atom_t maxseg_atom = 0;
    attr_list ret_list=NULL;
#ifdef MODULE
    struct socket *socket;
int ret;
#endif

#ifdef MODULE
    if (sock_create(AF_INET, SOCK_DGRAM, 0, &socket) < 0) {
#else
    conn_sock = socket(AF_INET, SOCK_DGRAM, 0);
    if (conn_sock == SOCKET_ERROR) {
#endif
        fprintf(stderr, "Cannot open INET socket\n");
        return NULL;
    }
#ifdef MODULE
    conn_sock = get_pseudofd(socket, (struct socket **)current->qfabric->fdqfabric);
#endif

    if (host_atom == 0) {
	host_atom = attr_atom_from_string("IP_HOST");
	transport_atom = attr_atom_from_string("CM_TRANSPORT");
	port_atom = attr_atom_from_string("UDP_PORT");
	maxseg_atom = attr_atom_from_string("MAX_SEG");
    }
    /*
     *  check to see if a bind to a specific port was requested
     */
    if (listen_info != NULL
        && !query_attr (listen_info, attr_atom_from_string ("UDP_PORT"),
                        NULL, (attr_value*) &int_port_num)) {
      if (int_port_num > USHRT_MAX || int_port_num < 0) {
        fprintf (stderr, "Requested port number %d is invalid\n", int_port_num);
	return NULL;
      }
      port_num = int_port_num;
    } else {
      port_num = 0;
    }

    int_addr[0] = 0;
#ifdef MODULE
    sock_addr.sin_addr.s_addr = INADDR_ANY;
#else
    if (listen_info != NULL
	&& query_attr (listen_info, attr_atom_from_string ("IP_HOST"),
			NULL, (attr_value*) &int_addr)) {
	struct hostent *ent = NULL;
	ent = gethostbyname(int_addr);
	sock_addr.sin_addr.s_addr = inet_addr(ent->h_addr_list[0]);
    } else {
	sock_addr.sin_addr.s_addr = INADDR_ANY;
    }
#endif

    ret_list = create_attr_list();

    if (listen_info != NULL && query_attr(listen_info, attr_atom_from_string("MAX_SEG"),
			NULL, (attr_value *)&max_seg)) 
	add_attr(ret_list, maxseg_atom, Attr_Int4,
		(attr_value) max_seg);
    else
	max_seg = MAX_PACKET;

    svc->trace_out(NULL, "CMRUDP begin listen, requested port %d", int_port_num);
    sock_addr.sin_family = AF_INET;
    sock_addr.sin_port = htons(port_num);
#ifdef MODULE
    socket->sk->reuse = 1;
#else
    if (setsockopt(conn_sock, SOL_SOCKET, SO_REUSEADDR, (char *) &sock_opt_val,
                   sizeof(sock_opt_val)) != 0) {
        fprintf(stderr, "Failed to set REUSEADDR on INET socket\n");
        return NULL;
    }
#endif
#ifdef MODULE
    if ((ret=socket->ops->bind(socket, (struct sockaddr *) &sock_addr,
             sizeof (sock_addr))) < 0) {
#else
    if (bind(conn_sock, (struct sockaddr *) &sock_addr,
             sizeof (sock_addr)) == SOCKET_ERROR) {
#endif
        fprintf(stderr, "Cannot bind INET socket (%d)\n", ret);
        return NULL;
    }
    sock_opt_val = 1;
#ifdef MODULE
    socket->sk->reuse = 1;
#else
    if (setsockopt(conn_sock, SOL_SOCKET, SO_REUSEADDR, (char *) &sock_opt_val,
                   sizeof(sock_opt_val)) != 0) {
        perror("Failed to set REUSEADDR on INET socket");
        return NULL;
    }
#endif
    length = sizeof sock_addr;
#ifdef MODULE
    socket->ops->getname(socket, (struct sockaddr *)&sock_addr, &length, 0);
#else
    if (getsockname(conn_sock, (struct sockaddr *) &sock_addr, &length) < 0) {
        fprintf(stderr, "Cannot get socket name\n");
        return NULL;
    }
#endif
    
    svc->trace_out(cm,"listening on RUDP IP %s, port %d\n", inet_ntoa(sock_addr.sin_addr), ntohs(sock_addr.sin_port));
    /*
     * by adding the socket to the select list, this socket is ready
     * to accept connections
     */
    /* set the port num as one we can be contacted at */

#ifdef MODULE
    svc->fd_add_select(cm, conn_sock+FD_SETSIZE, rudp_accept_conn,
                     (void *)trans, (void*)conn_sock);
#else
    svc->fd_add_select(cm, conn_sock, rudp_accept_conn,
                     (void *)trans, (void*)conn_sock);
#endif

    int_port_num = ntohs(sock_addr.sin_port);
    
    svc->trace_out(cm, "CMRUDP listen succeeded on port %d",
                     int_port_num);
    
    get_qual_hostname(host_name, sizeof(host_name), svc);

    sd = (rudp_client_data_ptr)trans->trans_data;
    sd->hostname = strdup(host_name);
    sd->listen_port = int_port_num;
    sd->cm = cm;

    add_attr(ret_list, host_atom, Attr_String,
                 (attr_value) strdup(host_name));
    add_attr(ret_list, port_atom, Attr_Int4,
                 (attr_value) int_port_num);
    add_attr(ret_list, transport_atom, Attr_String,
                 (attr_value) strdup("rudp"));

    return ret_list;
}


#ifdef NEED_IOVEC_DEFINE
struct  iovec {
     void *iov_base;
     int   iov_len;
};
#endif

#ifdef WSAEWOULDBLOCK
#define EWOULDBLOCK WSAEWOULDBLOCK
#endif

#if defined(HAVE_WINDOWS_H) && !defined(NEED_IOVEC_DEFINE)
#define NEED_IOVEC_DEFINE
#endif

/*
 * Accept socket connection
 */
static void
rudp_accept_conn(void_trans, void_conn_sock)
void *void_trans;
void *void_conn_sock;
{
    transport_entry trans = (transport_entry) void_trans;
    int conn_sock = (int) (long)void_conn_sock;
    rudp_client_data_ptr sd = (rudp_client_data_ptr) trans->trans_data;
    CMtrans_services svc = sd->svc;
    CManager cm = sd->cm;
    rudp_conn_data_ptr rudp_conn_data;
    struct sockaddr_in sock_addr, new_local_addr;
    int sock, ret, int_port_num, sock_len = sizeof(sock_addr), rwin;
#ifndef MODULE
    int flags;
#endif
    static int fd_atom = 0;
    static int peer_ip_atom = 0;
    static int peer_conn_port_atom = 0;
    static int this_conn_port_atom = 0;
    static int peer_hostname_atom = 0;
    static int send_err_atom = 0;
    static int send_rate_atom = 0;
    static int rt_rate_atom = 0;
    static int recv_rate_atom = 0;
    static int recv_err_atom = 0;	
    static int recv_winc_atom = 0;

    struct timeval tv1, tv2, tdiff;
    syn_pdu this_syn, peer_syn;
    attr_list conn_attr_list = NULL;
    CMConnection conn;
#ifdef MODULE
    struct socket *socket;
    struct socket *conn_socket;
    unsigned long randbuf;
#else
    struct timezone tz;
#endif

    if (fd_atom == 0) {
        fd_atom = attr_atom_from_string("CONNECTION_FILE_DESCRIPTOR");
        this_conn_port_atom = attr_atom_from_string("THIS_CONN_PORT");
        peer_conn_port_atom = attr_atom_from_string("PEER_CONN_PORT");
        peer_ip_atom = attr_atom_from_string("PEER_IP_ADDR");
        peer_hostname_atom = attr_atom_from_string("PEER_HOSTNAME");
        peer_hostname_atom = attr_atom_from_string("PEER_HOSTNAME");
	send_err_atom = attr_atom_from_string("SEND_ERR");
	rt_rate_atom = attr_atom_from_string("RTSEND_RATE");
	send_rate_atom = attr_atom_from_string("SEND_RATE");
	recv_rate_atom = attr_atom_from_string("RECV_RATE");
	recv_winc_atom = attr_atom_from_string("WIN_INC");
	recv_err_atom = attr_atom_from_string("RECV_ERR");
    }

#ifdef MODULE
    conn_socket = look_up_pseudofd(conn_sock, (struct socket **)current->qfabric->fdqfabric);
    if ((ret = recvfrom(conn_socket, (void *)&peer_syn, SYN_LEN, 0, (struct sockaddr *)&sock_addr, &sock_len, 0))  < 0)
#else
    if ((ret = recvfrom(conn_sock, (void *)&peer_syn, SYN_LEN, 0, (struct sockaddr *)&sock_addr, &sock_len)) == SOCKET_ERROR)
#endif
    {
	svc->trace_out(cm, "Failure in receiving SYN packet\n"); 
	svc->fd_remove_select(sd->cm, conn_sock);
        svc->trace_out(cm, "Failure in CMRUDP removing socket connection\n");
        return;
    }

#ifdef MODULE
    if (ret == 0) return;
#endif
 
    memcpy((void *)&this_syn, (void *)&peer_syn, sizeof(syn_pdu));
    this_syn.flags += ACK;
#ifdef MODULE
    get_random_bytes(&randbuf, sizeof(unsigned long));
    this_syn.seq = randbuf/ULONG_MAX * ((1<<16)-1);
#else
    this_syn.seq = lrand48();
#endif
    this_syn.ack = peer_syn.seq;
    this_syn.connid = global_connid ++;
    rwin = RWIN_DEF;
 
    int_port_num = ntohs(sock_addr.sin_port);
    svc->trace_out(cm, "requesting peer addr %s %d", inet_ntoa(sock_addr.sin_addr), int_port_num);

#ifdef MODULE
    sock_create(AF_INET, SOCK_DGRAM,0, &socket);
    sock = get_pseudofd(socket, (struct socket **)current->qfabric->fdqfabric);
#else
    sock = socket(AF_INET, SOCK_DGRAM,0);
#endif
    new_local_addr.sin_family = AF_INET;
    new_local_addr.sin_addr.s_addr = INADDR_ANY;
    new_local_addr.sin_port = htons(0);

   /* bind to another port */
#ifdef MODULE
    if( socket->ops->bind(socket, (struct sockaddr *)&new_local_addr, sizeof(new_local_addr)) < 0)
#else
    if( bind(sock, (struct sockaddr *)&new_local_addr, sizeof(new_local_addr)) == SOCKET_ERROR)
#endif
    {
	svc->trace_out(cm, "accept_conn socket error: bind\n");
	return ;
    }

#ifdef MODULE
    socket->ops->getname(socket, (struct sockaddr *)&new_local_addr, &sock_len, 0);
#else
    getsockname(sock, (struct sockaddr *)&new_local_addr, &sock_len); 
#endif

    this_syn.newport = new_local_addr.sin_port;
    svc->trace_out(cm, "rudp_accept_conn: local port servicing this request %d", ntohs(this_syn.newport));

    tv1.tv_sec = 1; 
    tv1.tv_usec = 0;
#ifndef MODULE
    setsockopt(conn_sock, SOL_SOCKET, SO_SNDTIMEO, &tv1, sizeof(tv1));
    setsockopt(conn_sock, SOL_SOCKET, SO_RCVTIMEO, &tv1, sizeof(tv1));
#endif

    gettimeofday(&tv1, &tz);
#ifdef MODULE
    conn_socket = look_up_pseudofd(conn_sock, (struct socket **)current->qfabric->fdqfabric);
    if(sendto(conn_socket, (void *)&this_syn, SYN_LEN, 0, (struct sockaddr *)&sock_addr, sock_len, jiffies, 0) == 0)
#else
    if(sendto(conn_sock, (void *)&this_syn, SYN_LEN, 0, (struct sockaddr *)&sock_addr, sock_len) == SOCKET_ERROR)
#endif
    {
            svc->trace_out(sd->cm, "CMRUDP connect FAILURE --> sendto() failed");
            close(conn_sock);
	    close(sock);
	    return;
    }

#ifdef MODULE
    if(recvfrom(conn_socket, (void *)&this_syn, SYN_LEN, 0, (struct sockaddr *)&sock_addr, &sock_len, 2) < 0)
#else
    if(recvfrom(conn_sock, (void *)&this_syn, SYN_LEN, 0, (struct sockaddr *)&sock_addr, &sock_len) == SOCKET_ERROR)
#endif
    {
	    svc->trace_out(sd->cm, "CMRUDP connect FAILURE --> sendto() failed");
	    close(conn_sock);
	    close(sock);
	    return;
    }
    gettimeofday(&tv2, &tz);
    timersub(&tv2, &tv1, &tdiff);
   
    tv1.tv_sec = 0; 
    tv1.tv_usec = 0;
     
    svc->trace_out(sd->cm, "--> Connection established");
    rudp_conn_data = create_rudp_conn_data(svc);
    rudp_conn_data->remote_IP = sock_addr.sin_addr.s_addr;
    rudp_conn_data->remote_contact_port = int_port_num;
    rudp_conn_data->remote_conn_port = ntohs(sock_addr.sin_port);
    rudp_conn_data->first_measure = TRUE;
    rudp_conn_data->first_data = TRUE;
    rudp_conn_data->adapt_inprog = FALSE;
    rudp_conn_data->app_adapted = FALSE;
    rudp_conn_data->svc = svc;
    rudp_conn_data->last_outseq = -1;

    conn_attr_list = create_attr_list();
    conn = svc->connection_create(trans, rudp_conn_data, conn_attr_list);
    rudp_conn_data->conn = conn;
    add_attr(conn_attr_list, peer_conn_port_atom, Attr_Int4,
                 (attr_value) int_port_num);
    add_attr(conn_attr_list, peer_ip_atom, Attr_Int4,
                 (attr_value) rudp_conn_data->remote_IP);
    
    if (sock_addr.sin_family == AF_INET) {
#ifndef MODULE
        struct hostent *host;
        struct sockaddr_in *in_sock = (struct sockaddr_in *) &sock_addr;
        host = gethostbyaddr((char *) &in_sock->sin_addr,
                                sizeof(struct in_addr), AF_INET);
        if (host != NULL) {
                rudp_conn_data->remote_host = strdup(host->h_name);
                add_attr(conn_attr_list, peer_hostname_atom, Attr_String,
                         (attr_value) strdup(host->h_name));
        }
#endif
     }
    if (rudp_conn_data->remote_host != NULL) {
        svc->trace_out(cm, "Accepted RUDP rudp connection from host \"%s\"",
                    rudp_conn_data->remote_host);
    } else {
        svc->trace_out(cm, "Accepted RUDP rudp connection from UNKNOWN host");
    }

    rudp_conn_data->fd = sock;
    rudp_conn_data->cm = sd->cm; 

/* the following four lines regarding buffer need more thought */
    rudp_conn_data->max_segsize = this_syn.max_segsize;
    rudp_conn_data->recv_buffer_len = MAX_RWIN;
#ifdef MODULE
    rudp_conn_data->recv_buffer = vmalloc(MAX_RWIN * rudp_conn_data->max_segsize);
#else
    rudp_conn_data->recv_buffer = svc->malloc(MAX_RWIN * rudp_conn_data->max_segsize);
#endif
    rudp_conn_data->send_buffer_len = MAX_SWIN;
#ifdef MODULE
    rudp_conn_data->send_buffer = vmalloc(MAX_SWIN * rudp_conn_data->max_segsize);
#else
    rudp_conn_data->send_buffer = svc->malloc(MAX_SWIN * rudp_conn_data->max_segsize);
#endif

    rudp_conn_data->max_outstanding = this_syn.max_outstanding;
    rudp_conn_data->max_byte = rudp_conn_data->max_outstanding*rudp_conn_data->max_segsize;
    rudp_conn_data->retrans_to = this_syn.retrans_to;
    rudp_conn_data->mrtt = this_syn.retrans_to;
    rudp_conn_data->sdev = 0;
    rudp_conn_data->cumack_to = this_syn.cumack_to;
    rudp_conn_data->null_to = this_syn.null_to;
    rudp_conn_data->transtat_to = this_syn.transtat_to;
    rudp_conn_data->max_retrans = this_syn.max_retrans;
    rudp_conn_data->max_cumack = this_syn.max_cumack;
    rudp_conn_data->max_outseq = this_syn.max_outseq;
    rudp_conn_data->max_auto_rst = this_syn.max_auto_rst;
    rudp_conn_data->max_keepalive = MAXKEEPALIVE;
    rudp_conn_data->rwin = rwin;
    rudp_conn_data->connid = peer_syn.connid;
    rudp_conn_data->local_connid = this_syn.connid;
    rudp_conn_data->period_bytes = 0;
    rudp_conn_data->senderr_cnt = 0;
    rudp_conn_data->recverr_cnt = 0;
   
/* conn state */
    rudp_conn_data->state = CONNECTED;
    rudp_conn_data->retrans_tm = -1; 
    rudp_conn_data->retrans_cnt = 0;
    rudp_conn_data->cumack_cnt = 0;
    rudp_conn_data->outseq_cnt = 0;
    rudp_conn_data->outseq_occur = 0;
    rudp_conn_data->rst_cnt = 0;
    rudp_conn_data->cumack_tm = -1;
    rudp_conn_data->null_tm = TimerSet(KEEPALIVE, &null_handler, rudp_conn_data, 0);
    rudp_conn_data->total_sent = 0;
    rudp_conn_data->total_err = 0;
    rudp_conn_data->exp_send_tm = -1;
    rudp_conn_data->exp_send_last = 1;
    rudp_conn_data->redun_send = FALSE;
    rudp_conn_data->last_retrans = 0;
    rudp_conn_data->period_recvp = -1;
    rudp_conn_data->period_recvb = rudp_conn_data->period_rcv_err = 0;
    rudp_conn_data->rcv_apperr = 0;
    rudp_conn_data->fin = FALSE;

    init_reg(&(rudp_conn_data->send_qos));

#ifndef MODULE
    pthread_mutex_init(&(rudp_conn_data->mutex), NULL);
    pthread_cond_init(&(rudp_conn_data->die), NULL);
#endif
    add_attr(conn_attr_list, fd_atom, Attr_Int4,
             (attr_value) sock);

    sock_len=  sizeof(sock_addr);
/* we want to have the new socket address */
#ifdef MODULE
    socket->ops->getname(socket, (struct sockaddr *)&new_local_addr, &sock_len, 0);
#else
    getsockname(sock, (struct sockaddr *) &sock_addr, &sock_len);
#endif
    int_port_num = ntohs(((struct sockaddr_in *)&sock_addr)->sin_port);
    add_attr(conn_attr_list, this_conn_port_atom, Attr_Int4,
             (attr_value) int_port_num);

    add_attr(conn_attr_list, send_err_atom, Attr_Int4, (attr_value) 0); 
    add_attr(conn_attr_list, send_rate_atom, Attr_Int4, (attr_value) 0);
    add_attr(conn_attr_list, rt_rate_atom, Attr_Int4, (attr_value) 0);
    add_attr(conn_attr_list, recv_rate_atom, Attr_Int4, (attr_value) 0);
    add_attr(conn_attr_list, recv_err_atom, Attr_Int4, (attr_value)0);
    add_attr(conn_attr_list, recv_winc_atom, Attr_Int4, (attr_value)0);

#ifndef MODULE
    if( (flags = fcntl(sock, F_GETFL, 0)) < 0)
   	svc->trace_out(cm, "error getting socket options");
    flags |= O_NONBLOCK;
    if(fcntl(sock, F_SETFL, flags) == SOCKET_ERROR)
   	svc->trace_out(cm, "error setting socket options");
#endif

    svc->fd_add_select(conn->cm, sock,
	              (select_list_func)rudp_data_available,
                     (void *)trans, (void*)conn);
}

extern int
libcmrudp_LTX_read_to_buffer_func(svc, scd, buffer, requested_len)
CMtrans_services svc;
rudp_conn_data_ptr scd;
void *buffer;
int requested_len;
{
    int i, copied, av_index;
    int cnt;
    void *tptr;	
/*
 * Can we use recvfrom() instead of this low-level read, in case the data
 * it gives me crosses message boundary 
 * I guess using read() will complicate all the handling of sequencing and 
 * acknowledgement stuff. So, let's use recvfrom() and instead of read the 
 * requested_len, we will read the whole packet and buffer it in case CM 
 * need it later on
 */

   cnt = 0;
   copied = 0;
   i = 0;

   if(scd->avail_len > 0 && requested_len > 0)
   {   
        copied = 0;
        i = 0;
        do{
               tptr = scd->data_avail;
	       av_index = scd->av_seq % SEQ_MOD % MAX_RWIN;
               cnt = my_min(requested_len - copied, scd->abit[av_index]);
               memcpy(buffer + copied, scd->data_avail, cnt);
               copied += cnt;
               scd->abit[av_index] -= cnt;
/* New Fix 04/14 */
               if( !(scd->abit[av_index]>0))
		{
		       while(!(scd->abit[av_index])> 0)
			{	
		       	scd->data_avail = scd->recv_buffer + (av_index + 1)%MAX_RWIN * scd->max_segsize;
			scd->av_seq = (scd->av_seq + 1)%SEQ_MOD;
			scd->lbit[av_index] = 0;
			if(scd->av_seq == scd->next_expected) break;
			av_index = scd->av_seq % SEQ_MOD % MAX_RWIN;
			}
		}
               else
                    memcpy(tptr, tptr + cnt, scd->abit[av_index]);

               scd->avail_len -= cnt;
               i++;
         }while(copied<requested_len && scd->avail_len > 0);
   } 

   if(scd->avail_len == 0)
   {
	scd->data_avail = NULL;
	scd->av_seq = -1;	
   }

   return copied;
}

extern int
libcmrudp_LTX_writev_attr_func(svc, scd, iov, iovcnt, attrs)
CMtrans_services svc;
rudp_conn_data_ptr scd;
struct iovec *iov;
int iovcnt;
attr_list attrs;
{
#ifdef MODULE
    int left = 0, first_seg = TRUE, iget = 0, copied = 0;
#else
    int left = 0, first_seg = TRUE, iget = 0, copied = 0, fd = scd->fd;
#endif
    int i, offset, length = 0, mark = 0, report = FALSE;
    uint_8 adapt_flag = FALSE;
    uint_16 this_seq;
    int tolen;
    char *int_buf = NULL, *send_buf, *tptr = NULL;
    ack_pdu *hdr;
    struct sockaddr_in remote_addr;
    attr_opaque *aop;
#ifndef MODULE
    struct timezone tz;
#endif
    static atom_t send_rate_atom=0, qos_atom=0, send_err_atom=0, adapt_flag_atom=0, mark_atom = 0, dumpcnt=0;
    attr_list conn_attr_list = NULL;

    /* about to start transfer */
    remote_addr.sin_addr.s_addr = scd->remote_IP;
    remote_addr.sin_port = htons(scd->remote_conn_port);
    remote_addr.sin_family = AF_INET; 

    /* sum lengths */
    for (i=0; i<iovcnt; i++) 
	left += iov[i].iov_len;

#ifdef MODULE
    send_buf = (char *)DAllocMM(left * sizeof(char) + sizeof(ack_pdu));
#else
    send_buf = (char *)malloc(left * sizeof(char) + sizeof(ack_pdu));
#endif
    hdr = (ack_pdu *)send_buf; 
    hdr->len = sizeof(ack_pdu);
    hdr->flags = 0;

    if(mark_atom == 0)
    {
	mark_atom = attr_atom_from_string("MARK");
	qos_atom = attr_atom_from_string("QOS");
	adapt_flag_atom = attr_atom_from_string("START_ADAPT");
	send_rate_atom = attr_atom_from_string("SEND_RATE");
	send_err_atom = attr_atom_from_string("SEND_ERR");	
     }
	
    if(attrs)
    {
    	if(query_attr(attrs, mark_atom, NULL, 
                     (attr_value *) &mark))
		svc->trace_out(scd->cm, "App mesg marked %d\n", mark);

    	if (query_attr(attrs, qos_atom, NULL, 
                    /* value pointer */ (attr_value *) &aop))
		scd->send_qos = *((struct region *)(aop->buffer));

	if (query_attr(attrs, adapt_flag_atom, NULL,
		    /* value pointer */ (attr_value *) &adapt_flag))
	{
		if(scd->adapt_inprog && adapt_flag)
		{
			scd->adapt_inprog = FALSE;
#ifdef COORD_ADAPT
			if(!(adapt_flag == ACTION_TAKEN_SLOWDOWN))
			{
				scd->max_outstanding = my_min(SWIN_DEF - 1, max(2 * scd->max_outstanding, 2 + scd->max_outstanding));	
				scd->max_byte = scd->max_outstanding * scd->min_req;
/* 			printf("max_outstanding recovered to %d, max_byte %d\n", scd->max_outstanding, scd->max_byte); */
			}
#endif
		}
	}
		
    }	
	
/*
  if(ceil(left/scd->max_segsize) > SWIN_DEF -1 )
	printf("Datagram exceeds maximum sending window\n");
*/
  scd->min_req = my_min(scd->max_segsize, left);

  do{
    do{
    	pthread_mutex_lock(&scd->mutex);
/*       if(!(scd->outstanding < scd->max_outstanding) || !(scd->cont_sent < scd->max_outstanding/2 * scd->max_segsize))  */
/*         if(!(scd->outstanding < SWIN_DEF-1) || !(scd->cont_sent < scd->max_outstanding * scd->max_segsize))  */
        if(!(scd->outstanding < SWIN_DEF-1) || !(scd->cont_sent < scd->max_byte)) 
    	{
		pthread_mutex_unlock(&(scd->mutex));
#ifndef MODULE
		thr_thread_yield();
#else
		sched_yield();	
#endif
	}
	else
		break;
    }while(1);

    hdr->seq = (scd->next_send)%SEQ_MOD;
    this_seq = hdr->seq;
    hdr->seq = htons(hdr->seq);
    scd->next_send ++;
    if(scd->next_send % REPORT_MOD==0)
    	report = TRUE;

    if(!(scd->cumack_cnt < scd->max_cumack - 1))
    {
    	hdr->flags += ACK;
	hdr->ack = scd->next_expected==0?MAX_SEQ:(scd->next_expected - 1)%SEQ_MOD;
	hdr->ack = htons(hdr->ack);
	scd->cumack_cnt = 0;
    }
    
    if(scd->forward)
    {
	hdr->flags += FORWARD;
	hdr->forward = htons(scd->next_ack);
	scd->forward = FALSE;
    }

/* The first seg is sent using sendto(), afterwards, it's sent by writev() */
    if(first_seg)
    {
    	int_buf = send_buf + hdr->len;
	first_seg = FALSE;
        offset = 0;
    	for (i=0; i<iovcnt; i++) 
    	{
		memcpy(int_buf + offset, iov[i].iov_base, iov[i].iov_len);
		offset += iov[i].iov_len;
    	}	
    }

    if(left > 0)
    {
		tolen = sizeof(struct sockaddr_in);
		length = min(hdr->len + left, scd->max_segsize);
		svc->trace_out(scd->cm, "write_attr: packet sequence %d, len %d", this_seq, length);
		hdr->payload = htons(length - hdr->len);
		if(length == hdr->len + left)
			hdr->flags += L_SEG;
		tptr = int_buf - hdr->len;
		if(tptr != send_buf)
			memcpy(tptr, (void *)hdr, hdr->len);	

#ifdef RUDP_WIRELESS
#ifdef MODULE
		sock = look_up_pseudofd(fd, (struct socket **)current->qfabric->fdqfabric);
		iget1 = sendto(sock, (void *)tptr, length, 0, (struct sockaddr *)&remote_addr, tolen, 0, 0);
		iget2 = sendto(sock, (void *)tptr, length, 0, (struct sockaddr *)&remote_addr, tolen, 0, 0);
		iget = iget1 > iget2 ? iget1: iget2;
		iget = sendto(sock, (void *)tptr, length, 0, (struct sockaddr *)&remote_addr, tolen, 0, 0);
#else
		iget1 = sendto(fd, (void *)tptr, length, 0, (struct sockaddr *)&remote_addr, tolen);
		iget2 = sendto(fd, (void *)tptr, length, 0, (struct sockaddr *)&remote_addr, tolen);
		iget = iget1 > iget2 ? iget1: iget2;
		iget = sendto(fd, (void *)tptr, length, 0, (struct sockaddr *)&remote_addr, tolen);
#endif
#endif		
		
		scd->total_sent++;	
		scd->period_bytes += (float)length/(float)1000; 
		scd->cont_sent += length;
		if(iget != length)
		{
			scd->senderr_cnt++;
			svc->trace_out(scd->cm, "writev_attr error: sendto, len desired %d, sent %d \n", ntohs(hdr->payload), iget);
		}
		else
			scd->senderr_cnt=0;
			
		if(hdr->flags & ACK && scd->cumack_tm!= -1)
			TimerUpdate(scd->cumack_tm, scd->cumack_to, &cumack_handler, scd, 0); 

		copied += length - hdr->len;
		left -= length - hdr->len;
    	}

	if(hdr->flags & ACK && scd->cumack_tm!= -1)
		TimerUpdate(scd->cumack_tm, scd->cumack_to, &cumack_handler, scd, 0); 
	/* save the packet to send_buffer */
	memcpy((char*)scd->send_buffer + this_seq % SEQ_MOD%MAX_SWIN * scd->max_segsize, tptr, length);
	gettimeofday(&(scd->send_stamp[this_seq % SEQ_MOD % MAX_SWIN]), &tz);
	scd->sent_len[this_seq % SEQ_MOD%MAX_SWIN] = length;

	scd->marked[this_seq % SEQ_MOD %MAX_SWIN] = mark? TRUE: FALSE;
	if(hdr->flags & L_SEG)
		scd->marked[this_seq % SEQ_MOD %MAX_SWIN] = TRUE;
	scd->outstanding ++;
	if(scd->retrans_tm == -1 && scd->outstanding != 0)
		scd->retrans_tm = TimerSet(scd->retrans_to, &retrans_handler, scd, 0);
	pthread_mutex_unlock(&(scd->mutex));
	int_buf += length - hdr->len;
    }while(left > 0);

   if(scd->exp_send_tm == -1)
        scd->exp_send_tm = TimerSet(scd->retrans_to*20, &exp_send_handler, scd, 0);
   else
	TimerUpdate(scd->exp_send_tm, scd->retrans_to*20, &exp_send_handler, scd, 0);

    if(scd->total_sent == 1) 
	gettimeofday(&(scd->last_s_rep),&tz);
	
    if(report)
    {
	int send_rate, err_ratio, perr_ratio, exp_rate;
	struct timeval tv, tdiff;
	long msec;

	gettimeofday(&tv, &tz);
	timersub(&tv, &(scd->last_s_rep), &tdiff);
	msec = tdiff.tv_sec * 1000 + tdiff.tv_usec/1000;
#ifdef MODULE
	send_rate = scd->period_bytes * 1000/ msec; 
	err_ratio = scd->total_err * 100/scd->total_sent;
#else
	send_rate = scd->period_bytes * (float)1000/ msec; 
	err_ratio = (float)scd->total_err * (float) 100/(float)scd->total_sent;
#endif
	svc->trace_out(scd->cm, "error ratio = %d\n", err_ratio);
/*
	if(scd->next_send - scd->period_effpkt == 0)
		perr_ratio = 100;
	else
		perr_ratio = (float)(scd->total_err - scd->period_error) * 100.0/(float)(scd->next_send - scd->period_effpkt);
*/
	perr_ratio = scd->oneway_err;
	/* period_error is the accumulated error at a recording point
	 * difference seen at next sampling point is the error occurred
	 * during this sampling period
	 */
/* 	printf("period_effpkt %d, next_send %d, total_err %d, period_err %d\n", scd->period_effpkt, scd->next_send, scd->total_err, scd->period_error); */
	scd->period_error = scd->total_err;
	scd->period_effpkt = scd->next_send;
#ifdef ENABLE_SND_REPORT
	conn_attr_list = scd->conn->attrs;
	replace_attr(conn_attr_list, send_rate_atom, Attr_Int4, (attr_value)send_rate); 
	replace_attr(conn_attr_list, send_err_atom, Attr_Int4, (attr_value)perr_ratio);
	svc->trace_out(scd->cm, "periodic error %d percent", perr_ratio);
#endif

#ifdef ENABLE_ADAPT
	exp_rate = (scd->send_qos).exp_rate;
	if( exp_rate != -1 && ((send_rate - exp_rate > (scd->send_qos).dev ) || (exp_rate - send_rate > (scd->send_qos).dev)) && (scd->send_qos).handler)
		(scd->send_qos).handler(ERR_RATE, send_rate, perr_ratio);

	if(perr_ratio > RUDP_CHG_HIGHMARK)
	{
	  int app_resp = ACTION_PASSIVE;
	  svc->trace_out(scd->cm, "perr_ratio %d> RUDP_CHG_HIGHMARK", perr_ratio);
#ifdef RUDP_DEBUG
	  printf("perr_ratio %d >RUDP_CHG_HIGHMARK\n", perr_ratio);
#endif
	  if((scd->send_qos).handler)
		app_resp = ((scd->send_qos).handler)(ERR_ERATIO, send_rate, perr_ratio);

	  switch(app_resp) {
		case ACTION_PASSIVE:
		case ACTION_INPROG:
		case ACTION_TAKEN_SLOWDOWN: 
/* 			scd->max_outstanding = max(max(scd->min_req, 15), scd->max_outstanding - 3); */
/* 			scd->max_outstanding = min(SWIN_DEF, max(max(scd->min_req, 20), (float) scd->max_outstanding * (1-(float)perr_ratio)/100.00)); */
/* 			printf("rudp_writev_attr: max_outstanding changed to %d\n", scd->max_outstanding); */
/* 			if(perr_ratio >100) */
/* 				scd->max_outstanding = max(10, (float)100/(float)perr_ratio *(float) scd->max_outstanding /(float) 100) ;  */
/* 			else */
/* 				scd->max_outstanding = max(10, (float)perr_ratio*(float) scd->max_outstanding /(float) 100) ;  */
				
/* 			svc->trace_out(scd->cm, "rudp_writev_attr: max_outstanding changed to %d\n", scd->max_outstanding); */
			if(app_resp == ACTION_INPROG)
				scd->adapt_inprog = TRUE;
			break;

		case ACTION_TAKEN_LOWRES:
			break;

		default:
			break;
				
	  }
	
	  if(app_resp != ACTION_PASSIVE)
			scd->app_adapted = TRUE;
	}

	if(perr_ratio < RUDP_CHG_LOWMARK)
	{
/* 		scd->max_outstanding = min( scd->max_outstanding + 3, SWIN_DEF); */
/* 		svc->trace_out(scd->cm, "rudp_writev_attr: max_outstanding increased to %d\n", scd->max_outstanding); */
/* 		printf("rudp_writev_attr: max_outstanding increased to %d\n", scd->max_outstanding); */
		if(scd->app_adapted && (scd->send_qos).handler)
		{
			(scd->send_qos).handler(DROP_ERATIO, send_rate, perr_ratio);
			scd->app_adapted = FALSE;
		}
	}	

	scd->last_s_rep = tv;
	scd->period_bytes = 0;
    }
#endif

#ifdef MODULE
    DFreeMM((addrs_t)send_buf);
#else
    free(send_buf);
#endif
#ifdef DUMPWIN
    printf("Sending window %d %d\n", dumpcnt, scd->max_outstanding);
    dumpcnt++;
#endif
    return iovcnt;
}

/*
 * There is only one Timer Thread for all the RUDP connections 
 * Every 50ms, it wakes up from a usleep() and check whether there are timers
 * for any connections that expire
 * 50ms is chosen to prevent aggregate expiration overhead and maintain 
 * reasonable timer granularity 
 */
#ifndef MODULE
void *TimerRun(void *arg)
#else 
void TimerRun(void *arg)
#endif
{
  struct timeval     tv;
#ifndef MODULE
  sigset_t           blockedSignals;
#endif
  uint_8             forever;
  uint_8             retry;
  TimerHandler       localObject;
#ifndef MODULE
  int                lastState;
  int                lastType;
#endif
  rudp_conn_data_ptr localArgument1;
  int                localArgument2;
  int                i;

  /*
   * Set deferred cancelation for this thread.
   */

#ifndef MODULE
  pthread_setcancelstate(PTHREAD_CANCEL_ENABLE, &lastState);
  pthread_setcanceltype(PTHREAD_CANCEL_DEFERRED, &lastType);
#endif

  /*
   * Block the SIGINT signal in this thread.
   */

#ifndef MODULE
  sigemptyset(&blockedSignals);
  sigaddset(&blockedSignals, SIGINT);
  pthread_sigmask(SIG_BLOCK, &blockedSignals, NULL);
#endif

  /*
   * Loop forever.
   */

  forever = TRUE;

  while (forever)
  {
      sched_yield();

    /*
     * Sleep for some amount of time using usleep().
     */

#ifdef MODULE
    if (current->qfabric == NULL)
      initialize_qfabric();
    interruptible_sleep_on_timeout(&current->qfabric->kecho_queue, 5);
    if (destroy_timerrun == 1) {
      sys_exit(0);
    }
#else
    pthread_testcancel();

#ifdef SGI
    sginap((50 * CLK_TCK) / 1000);
#else
    usleep(50000);
#endif
#endif

#ifndef MODULE
    pthread_testcancel();
#endif

    /*
     * Get the current time and perform any callbacks that are due.  Be sure
     * to aquire the mutex lock whenever accessing the static members.
     * However, the mutex lock cannot be held when actually performing any
     * callbacks.  Also, the loop that checks for expired timers must be
     * restarted after each callback since a callback might set or cancel
     * timers.
     */

#ifdef MODULE
    gettimeofday(&tv, NULL);
#else
    if (gettimeofday(&tv, NULL) == 0)
    {
#endif
      pthread_mutex_lock(&gtimer.timerMutex);

      retry = TRUE;

      while (retry)
      {
        retry = FALSE;

        for (i = 0; i < gtimer.scanSize; i++)
        {
          if (gtimer.inUse[i])
          {
            if ((tv.tv_sec > gtimer.expTime[i].tv_sec) ||
                ((tv.tv_sec == gtimer.expTime[i].tv_sec) &&
                 (tv.tv_usec >= gtimer.expTime[i].tv_usec)))
            {
              if (gtimer.object[i] != NULL)
              {
                localObject    = gtimer.object[i];
                localArgument1 = gtimer.argument1[i];

                localArgument2 = gtimer.argument2[i];

                pthread_mutex_unlock(&(gtimer.timerMutex));

                gtimer.inUse[i] = localObject(i, localArgument1, localArgument2);
                pthread_mutex_lock(&gtimer.timerMutex);

                retry = TRUE;

                break;
              }
            }
          }
        }
        pthread_mutex_unlock(&(gtimer.timerMutex));
        sched_yield();
        if (destroy_timerrun == 1)
	  sys_exit(0);
        pthread_mutex_lock(&(gtimer.timerMutex));
	
      }

      pthread_mutex_unlock(&gtimer.timerMutex);
#ifndef MODULE
    }
#endif
  }

#ifdef MODULE
  return;
#else
  return ((void *)NULL);
#endif
}

static void TimerStart(void *arg)
{
  /*
   * Allocate the arrays.
   */ 
  char *mn = "TimerStart";
  int i;

  if(gtimer.inUse == NULL)
  {
#ifdef MODULE
    gtimer.inUse = (uint_8 *)DAllocMM(sizeof(uint_8) * TIMER_ARRAY_SIZE);
#else
    gtimer.inUse = (uint_8 *)malloc(sizeof(uint_8) * TIMER_ARRAY_SIZE);
#endif

    if (gtimer.inUse == NULL)
    {
      TimerLog(mn, "Unable to allocate inUse array.");
      return;

    }
  }

  if (gtimer.expTime == NULL)
  {
#ifdef MODULE
    gtimer.expTime = (struct timeval *)DAllocMM(sizeof(struct timeval) * TIMER_ARRAY_SIZE);
#else
    gtimer.expTime = (struct timeval *)malloc(sizeof(struct timeval) * TIMER_ARRAY_SIZE);
#endif

    if (gtimer.expTime == NULL)
    {
 /*    log(mn, "Unable to allocate expiration time array."); */
      return;

    }
  }

  if (gtimer.object == NULL)
  {
#ifdef MODULE
    gtimer.object = (TimerHandler *)DAllocMM(sizeof(TimerHandler)* TIMER_ARRAY_SIZE);
#else
    gtimer.object = (TimerHandler *)malloc(sizeof(TimerHandler)* TIMER_ARRAY_SIZE);
#endif

    if (gtimer.object == NULL)
    {
/*      log(mn, "Unable to allocate object array."); */
      return;

    }
  }

  if (gtimer.argument1 == NULL)
  {
#ifdef MODULE
    gtimer.argument1 = (rudp_conn_data_ptr *)DAllocMM(sizeof(rudp_conn_data_ptr) * TIMER_ARRAY_SIZE);
#else
    gtimer.argument1 = (rudp_conn_data_ptr *)malloc(sizeof(rudp_conn_data_ptr) * TIMER_ARRAY_SIZE);
#endif

    if (gtimer.argument1 == NULL)
      return;

  }

  if (gtimer.argument2 == NULL)
  {
#ifdef MODULE
    gtimer.argument2 = (int *)DAllocMM(sizeof(int) * TIMER_ARRAY_SIZE);
#else
    gtimer.argument2 = (int *)malloc(sizeof(int) * TIMER_ARRAY_SIZE);
#endif

    if (gtimer.argument2 == NULL)
      return;
  }

  /*
   * Initialize the sizes and array elements.
   */

  gtimer.arraySize = TIMER_ARRAY_SIZE;
  gtimer.scanSize  = 0;

  for (i = 0; i < TIMER_ARRAY_SIZE; i++)
  {
    gtimer.inUse[i]           = FALSE;
    gtimer.expTime[i].tv_sec  = 0;
    gtimer.expTime[i].tv_usec = 0;
    gtimer.object[i]          = NULL;
    gtimer.argument1[i]       = NULL;
    gtimer.argument2[i]       = 0;
  }

  /*
   * Create the timer thread.  Make it a detached thread.
   */ 

  /*  initialize the mutex */
#ifdef MODULE
  sema_init(&gtimer.timerMutex, 1);
#else
  pthread_mutex_init(&gtimer.timerMutex, NULL);
#endif

#ifndef MODULE
  if (pthread_attr_init(&gtimer.timerAttr) != 0)
    return;
#endif

#ifndef MODULE
  if (pthread_attr_setdetachstate(&gtimer.timerAttr,
                                  PTHREAD_CREATE_DETACHED) != 0)
    return;
#endif

#ifdef MODULE
  timerrun_id = kernel_thread((int (*)(void *)) TimerRun, NULL, 0);
#else
  if (pthread_create(&gtimer.timerThread, &gtimer.timerAttr,
                     &TimerRun, NULL) != 0)
  {
    TimerLog(mn, "Pthread creation error.");
    return;
  }
#endif
  /*
   * Mark the initialization as complete.
   */ 

  gtimer.initialized = TRUE;

}

static int TimerSet(int msec, TimerHandler obj, rudp_conn_data_ptr arg1,
                               int arg2)
{
  char            *mn = "SetTimer";
  struct timeval  tv;
  int             i;

  /*
   * Synchronize the threads.
   */

  pthread_mutex_lock(&gtimer.timerMutex);

  /*
   * If the timer class has not been initialized yet, call Start().
   */

  if (!gtimer.initialized)
  {
    TimerStart(NULL);

    if (!gtimer.initialized)
    {
      pthread_mutex_unlock(&gtimer.timerMutex);
      return(THR_ERROR);
    }
  }

  /*
   * Add the timer information to the first open array element.
   */

  for (i = 0; i < gtimer.arraySize; i++)
  {
    if (!gtimer.inUse[i])
    {
#ifdef MODULE
      gettimeofday(&tv, NULL);
#else
      if (gettimeofday(&tv, NULL) != 0)
      {
        pthread_mutex_unlock(&gtimer.timerMutex);
        return(THR_ERROR);
      }
#endif

      tv.tv_sec  += (msec / 1000);
      tv.tv_usec += ((msec % 1000) * 1000);

      tv.tv_sec  += (tv.tv_usec / 1000000);
      tv.tv_usec  = (tv.tv_usec % 1000000);

      gtimer.inUse[i]     = TRUE;
      gtimer.expTime[i]   = tv;
      gtimer.object[i]    = obj;
      gtimer.argument1[i] = arg1;
      gtimer.argument2[i] = arg2;

      /*
       * Update the scan size.
       */

      if (i >= gtimer.scanSize)
      {
        gtimer.scanSize = (i + 1);
      }

      pthread_mutex_unlock(&gtimer.timerMutex);
     
      return(i);
    }
  }

  TimerLog(mn, "No room in timer array.");
  pthread_mutex_unlock(&gtimer.timerMutex);
  return(THR_ERROR);
}

/* Cancel a particular timer */
static int CancelTimer(int TimerIdx)
{
  int    i;


  pthread_mutex_lock(&gtimer.timerMutex);

  /*
   * If the timer class has not been initialized yet, call Start().
   */

  if (!gtimer.initialized)
  {
    TimerStart(NULL);

    if (!gtimer.initialized)
    {
      pthread_mutex_unlock(&gtimer.timerMutex);
      return -1;
    }
  }

  /*
   * Find all of the timer entries for the callback object and delete them.
   */

  if (gtimer.inUse[TimerIdx])
       gtimer.inUse[TimerIdx] = FALSE;

  /*
   * Update the scan size.
   */

  for (i = (gtimer.arraySize - 1); i >= 0; i--)
  {
    if (gtimer.inUse[i])
    {
      gtimer.scanSize = (i + 1);
      pthread_mutex_unlock(&gtimer.timerMutex);
      return -1;
    }
  }

  gtimer.scanSize = 0;
  pthread_mutex_unlock(&gtimer.timerMutex);
  return -1;
}

/* ============================================================================ */
/* Cancel all timers related to an connection */
static void CancelObjTimers(rudp_conn_data_ptr obj)
{
  char*  mn = "CancelObjTimers";
  int    i;

  /* Synchronize the threads. */

  pthread_mutex_lock(&gtimer.timerMutex);

  /* If the timer class has not been initialized yet, call Start(). */

  if (!gtimer.initialized)
  {
    TimerStart(NULL);

    if (!gtimer.initialized)
    {
      TimerLog(mn, "Timer thread start failure.");
      pthread_mutex_unlock(&gtimer.timerMutex);
      return;
    }
  }

  /* Find all of the timer entries for the callback object and delete them. */

  for (i = 0; i < gtimer.arraySize; i++)
  {
    if (gtimer.inUse[i])
    {
      if (gtimer.argument1[i] == obj)
      {
        gtimer.inUse[i] = FALSE;
      }
    }
  }

  /* Update the scan size. */

  for (i = (gtimer.arraySize - 1); i >= 0; i--)
  {
    if (gtimer.inUse[i])
    {
      gtimer.scanSize = (i + 1);
      pthread_mutex_unlock(&gtimer.timerMutex);
      return;
    }
  }

  gtimer.scanSize = 0;
  pthread_mutex_unlock(&gtimer.timerMutex);

}

/*
 * all timer handlers should return != 0 the timer is still in use after the 
 * handling 
 */
int retrans_handler(int timer_idx, rudp_conn_data_ptr arg1, int arg2)
{
	struct sockaddr_in remote_addr;
	int ret, i, seq, num_resend;
#ifdef MODULE
	struct socket *sock;
#else
        struct timezone tz;
#endif

	arg1->svc->trace_out(arg1->cm, "entering retrans_handler, next_ack %d\n", arg1->next_ack);
	if(arg1->state == TIMED_WAIT)
		return 0;

	pthread_mutex_lock(&(arg1->mutex));

	/*
	 * when retransmission timer expires, 
	 * assume error of all unacked packets
	 */
/*	arg1->total_err += arg1->outstanding;
#ifdef MODULE
	eratio = arg1->total_err*100/arg1->total_sent;
#else
	eratio = (float)arg1->total_err * (float)100/ (float)arg1->total_sent;
#endif
	arg1->svc->trace_out(arg1->cm, "current error rate %d, total_err %d, total_sent %d, retrans_to %d\n", eratio, arg1->total_err, arg1->total_sent, arg1->retrans_to);

	send_error_ratio = attr_atom_from_string("SEND_ERR");
	conn_attr_list = CMConnection_get_attrs(arg1->conn);
	replace_attr(conn_attr_list, send_error_ratio, Attr_Int4, (attr_value)eratio); 
	if(eratio > RED_THRESH_HIGH)
		arg1->redun_send = TRUE;	
	if(eratio < RED_THRESH_LOW)
		arg1->redun_send = FALSE;
*/
	num_resend = my_min(arg1->outstanding, arg1->max_cumack);
	for(i = arg1->next_ack; i< arg1->next_ack + num_resend; i++)
	{
		remote_addr.sin_addr.s_addr = arg1->remote_IP;
		remote_addr.sin_port = htons(arg1->remote_conn_port);
		remote_addr.sin_family = AF_INET;

#ifdef MODULE
		pthread_mutex_unlock(&(arg1->mutex));
    		sock = look_up_pseudofd(arg1->fd, (struct socket **)current->qfabric->fdqfabric);
		if (sock == NULL) {
		    return 0;
		}
		ret = sendto(sock, arg1->send_buffer + i % SEQ_MOD%MAX_SWIN * arg1->max_segsize, arg1->sent_len[i % SEQ_MOD% MAX_SWIN], 0, (struct sockaddr *)&remote_addr, sizeof(struct sockaddr_in), jiffies, 0);
		pthread_mutex_lock(&(arg1->mutex));
		if(ret == 0)
#else
		ret = sendto(arg1->fd, arg1->send_buffer + i % SEQ_MOD%MAX_SWIN * arg1->max_segsize, arg1->sent_len[i % SEQ_MOD% MAX_SWIN], 0, (struct sockaddr *)&remote_addr, sizeof(struct sockaddr_in));
		if(ret == SOCKET_ERROR)
#endif
		{
			arg1->svc->trace_out(arg1->cm, "retrans_handler: sendto error, %d\n", errno);
			arg1->senderr_cnt++;
/*
			if(arg1->senderr_cnt > arg1->max_retrans)
			{
				pthread_mutex_unlock(&(arg1->mutex));
				libcmrudp_LTX_shutdown_conn(arg1->svc, arg1);
				timedwait_handler(-1, arg1, 0);
				return 0;
			}
*/
			if (errno == 11)
				TimerUpdate(timer_idx, ((arg1->senderr_cnt/3)+2)*arg1->retrans_to, &retrans_handler, arg1, 0);
			else
				TimerUpdate(timer_idx, arg1->retrans_to, &retrans_handler, arg1, 0);
			pthread_mutex_unlock(&(arg1->mutex));
			return 1;
			break;
		}
		seq = ((gen_data_pdu *)((char *)arg1->send_buffer + i% SEQ_MOD%MAX_SWIN * arg1->max_segsize))->seq;
		seq = ntohs(seq);
		arg1->svc->trace_out(arg1->cm,"packet %d resent to port %d, len %d", ((gen_data_pdu *)(arg1->send_buffer + i% SEQ_MOD%MAX_SWIN * arg1->max_segsize))->seq, ntohs(remote_addr.sin_port), arg1->sent_len[i % SEQ_MOD % MAX_SWIN]);	 

		gettimeofday(&(arg1->send_stamp[i %SEQ_MOD % MAX_SWIN]), &tz);
		if(ret != arg1->sent_len[i % SEQ_MOD%MAX_SWIN])
		{
			arg1->svc->trace_out(arg1->cm, "retrans_handler: sendto error, %d\n", errno);
			pthread_mutex_unlock(&(arg1->mutex));
#ifndef MODULE
			fflush(NULL);
#endif
			return 1;
		}	
		arg1->total_sent ++;
	}
	if(arg1->outstanding > 0)
	{
		if(arg1->last_retrans == -1)
			arg1->retrans_cnt = 0;
		else
		{
			if(arg1->next_ack == arg1->last_retrans) 
				arg1->retrans_cnt ++;
			else
				arg1->retrans_cnt = 0;
		}

		arg1->last_retrans = arg1->next_ack;

/*
		if(arg1->retrans_cnt == arg1->max_retrans)
		{
			pthread_mutex_unlock(&(arg1->mutex));
			libcmrudp_LTX_shutdown_conn(arg1->svc, arg1);
			timedwait_handler(-1, arg1, 0);
			return 0;
		}
*/
		TimerUpdate(timer_idx, arg1->retrans_to, &retrans_handler, arg1, 0); 
		if(arg1->exp_send_tm)
			TimerUpdate(arg1->exp_send_tm, 20*arg1->retrans_to, &exp_send_handler, arg1, 0);
		pthread_mutex_unlock(&(arg1->mutex));
		return 1;
	} 

	pthread_mutex_unlock(&(arg1->mutex));
	return 0;
}

int later(struct timeval *a, struct timeval *b)
{
   return (((a)->tv_sec == (b)->tv_sec) ? 
   ((a)->tv_usec > (b)->tv_usec) : 
   ((a)->tv_sec > (b)->tv_sec));
}

static int TimerUpdate(int timer_idx, int msec, TimerHandler obj, rudp_conn_data_ptr arg1, int arg2)
{
#ifndef MODULE
  char*           mn = "UpdateTimer";
#endif
  struct timeval  tv;
  int             i;

#ifdef MODULE
      gettimeofday(&tv, NULL);
#else
      if (gettimeofday(&tv, NULL) != 0)
      {
        TimerLog(mn, "Unable to get current time.");
        return(THR_ERROR);
      }
#endif

      tv.tv_sec  += (msec / 1000);
      tv.tv_usec += ((msec % 1000) * 1000);

      tv.tv_sec  += (tv.tv_usec / 1000000);
      tv.tv_usec  = (tv.tv_usec % 1000000);

      i = timer_idx;
      gtimer.expTime[i]   = tv;
      gtimer.object[i]    = obj;
      gtimer.argument1[i] = arg1;
      gtimer.argument2[i] = arg2;

      return(THR_OK);

}

static int
check_host(hostname, sin_addr)
char *hostname;
void *sin_addr;
{
#ifdef MODULE
    struct hostent *host_addr;

    host_addr = lookup_name(hostname);
    if (host_addr == NULL) return 0;
    memcpy(sin_addr, host_addr->h_addr, host_addr->h_length);
    DFreeMM((addrs_t)host_addr->h_name);
    DFreeMM((addrs_t)host_addr->h_addr_list);
    DFreeMM((addrs_t)host_addr);

    return 1;
#else
    struct hostent *host_addr;
    host_addr = gethostbyname(hostname);
    if (host_addr == NULL) {
        unsigned long addr = inet_addr(hostname);
        if (addr == -1) {
            /*
             *  not translatable as a hostname or
             * as a dot-style string IP address
             */
            return 0;
        }
        assert(sizeof(int) == sizeof(struct in_addr));
        *((int*)sin_addr) = (int)addr;
    } else {
        memcpy(sin_addr, host_addr->h_addr, host_addr->h_length);
    }
    return 1;
#endif
}


static void TimerLog(char *func, char *diag)
{
	return;
} 

int null_handler(int timer_idx, rudp_conn_data_ptr scd, int arg2)
{
	null_pdu null_pkt;
	struct sockaddr_in remote_addr;
	int ret, tolen;
#ifdef MODULE
	struct socket *sock;
#endif

	if(scd->state == TIMED_WAIT)
		return 0;

/*	if(scd->rst_cnt == scd->max_keepalive) */
/*		libcmrudp_LTX_shutdown_conn(scd->svc, scd); */
/*		timedwait_handler(-1, scd, 0); */
		
	remote_addr.sin_addr.s_addr = scd->remote_IP;
	remote_addr.sin_port = htons(scd->remote_conn_port);
	remote_addr.sin_family = AF_INET;	

	memset((void *)&null_pkt, 0, sizeof(null_pdu));
	null_pkt.flags = NUL;
	null_pkt.len = sizeof(null_pkt);
	null_pkt.seq = htons(-1);
	null_pkt.ack = scd->next_expected==0?MAX_SEQ:(scd->next_expected - 1)%SEQ_MOD;  
	null_pkt.ack = htons(null_pkt.ack);
	null_pkt.rnum = htons(scd->total_recv - scd->last_recv);
	scd->last_recv = scd->total_recv;

	tolen = sizeof(struct sockaddr_in );
#ifdef MODULE
    	sock = look_up_pseudofd(scd->fd, (struct socket **)current->qfabric->fdqfabric);
	if (sock == NULL) return 0;
	ret = sendto(sock, (char *)&null_pkt, null_pkt.len, 0, (struct sockaddr *)&remote_addr, tolen, jiffies, 0);	
#else
	ret = sendto(scd->fd, (char *)&null_pkt, null_pkt.len, 0, (struct sockaddr *)&remote_addr, tolen);	
#endif

	pthread_mutex_lock(&(scd->mutex));
	if(ret == SOCKET_ERROR)
	{
		scd->senderr_cnt++;
		if(scd->senderr_cnt > scd->max_retrans)
		{
/*			pthread_mutex_unlock(&(scd->mutex)); */
/*			libcmrudp_LTX_shutdown_conn(scd->svc, scd); */
/*			timedwait_handler(-1, scd, 0); */
/*			return 0; */
		}
		perror("null_handler: sendto socket error\n");
/*		return 1; */
	}
	if(ret != null_pkt.len)
	{
		perror("null_handler: sendto error\n");
/*		pthread_mutex_unlock(&(scd->mutex)); */
/*		return 1; */
	}		
	scd->senderr_cnt = 0;
	scd->rst_cnt++;
	pthread_mutex_unlock(&(scd->mutex));
	TimerUpdate(timer_idx, KEEPALIVE, &null_handler, scd, arg2);
	return 1;
}

int cumack_handler(int timer_idx, rudp_conn_data_ptr scd, int arg2)
{
	ack_pdu ack_pkt;
	int ret;
	int tolen;
	struct sockaddr_in remote_addr;
#ifdef MODULE
	struct socket *sock;
#endif
	
	remote_addr.sin_addr.s_addr = scd->remote_IP;
	remote_addr.sin_port = htons(scd->remote_conn_port);
	remote_addr.sin_family = AF_INET;	

	ack_pkt.flags = ACK;
	ack_pkt.len = sizeof(ack_pdu);
	ack_pkt.ack = scd->next_expected==0?MAX_SEQ:(scd->next_expected - 1)%SEQ_MOD;
	ack_pkt.ack = htons(ack_pkt.ack);
	ack_pkt.seq = htons(-1);
	ack_pkt.payload = htons(0);
	ack_pkt.partack = FALSE;
	/* anyway, if length of upper layer data is 0,seq is ignored	 */

	tolen = sizeof(struct sockaddr_in);
#ifdef MODULE
    	sock = look_up_pseudofd(scd->fd, (struct socket **)current->qfabric->fdqfabric);
	if (sock == NULL) return 0;
	ret = sendto(sock, (char *)&ack_pkt, ack_pkt.len, 0, (struct sockaddr *)&remote_addr, tolen, jiffies, 0);
#else
	ret = sendto(scd->fd, (char *)&ack_pkt, ack_pkt.len, 0, (struct sockaddr *)&remote_addr, tolen);
#endif

	pthread_mutex_lock(&(scd->mutex));
	if(ret == SOCKET_ERROR && errno!= EINVAL)
	{
		scd->senderr_cnt++;
		perror("cumack_handler: sendto socket error\n");
		TimerUpdate(timer_idx, scd->cumack_to, &cumack_handler, scd, 0);
		pthread_mutex_unlock(&(scd->mutex));
		return 0;
	}
	scd->senderr_cnt = 0;	
	scd->svc->trace_out(scd->cm, "cumack_handler: ACK to seq %d is sent to port %d", ack_pkt.ack, ntohs(remote_addr.sin_port));

	scd->cumack_cnt = 0;
	scd->cumack_tm = CancelTimer(scd->cumack_tm);
	pthread_mutex_unlock(&(scd->mutex));
	return 0;
}	 

/*
 * this is essentially the same as null_handler, except that it updates the 
 * timer with EXSEND instead of KEEPALIVE
 */
int exp_send_handler(int timer_idx, rudp_conn_data_ptr arg1, int arg2)
{
	
	null_pdu null_pkt;
	struct sockaddr_in remote_addr;
	int ret, tolen;
#ifdef MODULE
	struct socket *sock;
#endif

	if(arg1->state == TIMED_WAIT)
		return 0;

	remote_addr.sin_addr.s_addr = arg1->remote_IP;
	remote_addr.sin_port = htons(arg1->remote_conn_port);
	remote_addr.sin_family = AF_INET;	

	memset((void *)&null_pkt, 0, sizeof(null_pdu));
	null_pkt.flags = NUL + RST + ACK;
	null_pkt.len = sizeof(null_pkt);
	null_pkt.seq = htons(arg1->exp_send_last); /*we do look at a NUL packet's seq*/
	arg1->exp_send_last = (arg1->exp_send_last + 1) % SEQ_MOD;
	null_pkt.ack = arg1->next_expected==0?MAX_SEQ:(arg1->next_expected - 1)%SEQ_MOD;
	null_pkt.ack = htons(null_pkt.ack);

	tolen = sizeof(struct sockaddr_in );
#ifdef MODULE
    	sock = look_up_pseudofd(arg1->fd, (struct socket **)current->qfabric->fdqfabric);
	if (sock == NULL) {
          return 0;
	}
	ret = sendto(sock, (char *)&null_pkt, null_pkt.len, 0, (struct sockaddr *)&remote_addr, tolen, jiffies, 0);	
#else
	ret = sendto(arg1->fd, (char *)&null_pkt, null_pkt.len, 0, (struct sockaddr *)&remote_addr, tolen);	
#endif
	pthread_mutex_lock(&(arg1->mutex));
	if(ret == SOCKET_ERROR)
	{
		arg1->senderr_cnt++;
		perror("exp_send_handler: sendto socket error\n");
		pthread_mutex_unlock(&(arg1->mutex));
		return 1;
	}
	if(ret != null_pkt.len)
	{
		perror("exp_send_handler: sendto error\n");
		pthread_mutex_unlock(&(arg1->mutex));
		return 1;
	}		
	arg1->senderr_cnt=0;
	TimerUpdate(timer_idx, arg1->retrans_to * 20, &exp_send_handler, arg1, arg2);
	arg1->rst_cnt++;
	pthread_mutex_unlock(&(arg1->mutex));	
	return 1;
}

/*
 * Check to see that if we were to attempt to initiate a connection as
 * indicated by the attribute list, would we be connecting to ourselves?
 * For sockets, this involves checking to see if the host name is the
 * same as ours and if the IP_PORT matches the one we are listening on.
 */
extern int
libcmrudp_LTX_self_check(cm, svc, trans, attrs)
CManager cm;
CMtrans_services svc;
transport_entry trans;
attr_list attrs;
{

    rudp_client_data_ptr sd = trans->trans_data;
    int int_port_num;
    atom_t host_atom = 0;
    atom_t port_atom = 0;
    char *host_name;
    char my_host_name[256];

    if (host_atom == 0) {
        host_atom = attr_atom_from_string("IP_HOST");
        port_atom = attr_atom_from_string("UDP_PORT");
    }
    if (!query_attr(attrs, host_atom, /* type pointer */ NULL,
                    /* value pointer */ (attr_value *) &host_name)) {
        svc->trace_out(cm, "CMself check TCP/IP transport found no IP_HOST attribute");
        return 0;

    }
    if (!query_attr(attrs, port_atom, /* type pointer */ NULL,
                    /* value pointer */ (attr_value *) &int_port_num)) {
        svc->trace_out(cm, "CMself check TCP/IP transport found no IP_PORT attribute");
        return 0;
    }

    gethostname(my_host_name, sizeof(my_host_name));

    if (strcmp(host_name, my_host_name) != 0) {
        svc->trace_out(cm, "CMself check - Hostnames don't match");
        return 0;
    }
    if (int_port_num != sd->listen_port) {
        svc->trace_out(cm, "CMself check - Ports don't match");
        return 0;
    }
    svc->trace_out(cm, "CMself check returning TRUE");
    return 1;
}

extern int
libcmrudp_LTX_connection_eq(cm, svc, trans, attrs, scd)
CManager cm;
CMtrans_services svc;
transport_entry trans;
attr_list attrs;
rudp_conn_data_ptr scd;
{

    int int_port_num;
    int requested_IP;
    atom_t host_atom = 0;
    atom_t port_atom = 0;
    char *host_name;

    if (host_atom == 0) {
        host_atom = attr_atom_from_string("IP_HOST");
        port_atom = attr_atom_from_string("UDP_PORT");
    }
    if (!query_attr(attrs, host_atom, /* type pointer */ NULL,
                    /* value pointer */ (attr_value *) &host_name)) {
        svc->trace_out(cm, "RUDP transport found no IP_HOST attribute");
        return 0;
    }
    if (!query_attr(attrs, port_atom, /* type pointer */ NULL,
                    /* value pointer */ (attr_value *) &int_port_num)) {
        svc->trace_out(cm, "RUDP transport found no UDP_PORT attribute");
        return 0;
    }

    check_host(host_name, (void*)&requested_IP);

    if ((scd->remote_IP == requested_IP) &&

        (scd->remote_contact_port == int_port_num)) {
        svc->trace_out(cm, "rudp Conn_eq returning TRUE");
        return 1;
    }
    svc->trace_out(cm, "rudp Conn_eq returning FALSE");
    return 0;
}

int timedwait_handler(int timer_idx, rudp_conn_data_ptr scd, int arg2)
{
	scd->svc->trace_out(scd->cm, "Timed Wait finishes");
	close(scd->fd);
        scd->svc->fd_remove_select(scd->cm, scd->fd);
#ifndef MODULE
        pthread_mutex_destroy(&(scd->mutex));
#endif
        CancelObjTimers(scd);
#ifdef MODULE
        DFreeMM((addrs_t)scd->remote_host);
	vfree(scd->send_buffer);
	vfree(scd->recv_buffer);
        DFreeMM((addrs_t)scd);
#else
        free(scd->remote_host);
	free(scd->send_buffer);
	free(scd->recv_buffer);
        free(scd);
#endif
	return 0;
}

extern int
libcmrudp_LTX_writev_func(svc, scd, iov, iovcnt, flag)
CMtrans_services svc;
rudp_conn_data_ptr scd;
struct iovec *iov;
int iovcnt;
int flag;
{
    return libcmrudp_LTX_writev_attr_func(svc, scd, iov, iovcnt, NULL);
}

