
#ifdef MODULE
#ifndef __KERNEL__
#define __KERNEL__
#endif
#endif

#ifndef MODULE
/***** Includes *****/
#include "config.h"
#include <sys/types.h>

#ifdef HAVE_WINDOWS_H
#include <windows.h>
#include <winsock.h>
#define getpid()	_getpid()
#else
#ifdef HAVE_SYS_TIME_H
#include <sys/time.h>
#endif
#ifdef HAVE_SYS_TIMES_H
#include <sys/times.h>
#endif
#include <sys/socket.h>
#ifdef HAVE_SYS_SELECT_H
#include <sys/select.h>
#endif
#ifdef HAVE_SYS_UN_H
#include <sys/un.h>
#endif
#ifdef HAVE_SYS_UIO_H
#include <sys/uio.h>
#endif
#ifdef HAVE_HOSTLIB_H
#include "hostLib.h"
#endif
#ifdef HAVE_STREAMS_UN_H
#include <streams/un.h>
#endif
#include <netinet/in.h>
#include <arpa/inet.h>
#ifdef HAVE_NETDB_H
#include <netdb.h>
#endif
#endif
#include <stdio.h>
#include <fcntl.h>
#ifdef HAVE_UNISTD_H
#include <unistd.h>
#endif
#include <assert.h>
#include <errno.h>
#include <signal.h>
#include <stdlib.h>
#include <string.h>
#include <limits.h>
#ifdef HAVE_MEMORY_H
#include <memory.h>
#endif

#include "gen_thread.h"
#include <atl.h>
#include "cm.h"
#include "cm_internal.h"
#include "cm_transport.h"

//#define DO_TCP_NODELAY 1

#if !defined(BIND_DEFINED) & !defined(HAVE_WINDOWS_H)
extern int bind ARGS((int s, struct sockaddr * name, int namelen));
#endif
#if !defined(ACCEPT_DEFINED) & !defined(HAVE_WINDOWS_H)
extern int accept ARGS((int s, struct sockaddr * addr, int *addrlen));
#endif
#if !defined(CONNECT_DEFINED) & !defined(HAVE_WINDOWS_H)
extern int connect ARGS((int s, struct sockaddr * name, int namelen));
#endif
#if !defined(SETSOCKOPT_DEFINED) & !defined(HAVE_WINDOWS_H)
extern int setsockopt ARGS((int s, int level, int optname, const char *optval, int optlen));
#endif
#if !defined(GETSOCKNAME_DEFINED) & !defined(HAVE_WINDOWS_H)
extern int getsockname ARGS((int s, struct sockaddr * name, int *namelen));
#endif
#if defined(HAVE_GETDOMAINNAME) && !defined(GETDOMAINNAME_DEFINED)
extern int getdomainname ARGS((char *name, int namelen));
#endif
#if !defined(LISTEN_DEFINED) & !defined(HAVE_WINDOWS_H)
extern int listen ARGS((int s, int backlog));
#endif
#ifndef HAVE_WINDOWS_H
extern int socket ARGS((int domain, int type, int protocol));
#endif
#if !defined(GETHOSTNAME_DEFINED) & !defined(HAVE_WINDOWS_H)
extern int gethostname ARGS((char *name, int namelen));
#endif
#else
#include "config.h"
#include <linux/types.h>
#include <linux/time.h>
#include <linux/fcntl.h>
#include <linux/unistd.h>
#include <linux/errno.h>
#include <linux/signal.h>
#include <linux/net.h>
#include <linux/inet.h>
#include <linux/socket.h>
#include <linux/skbuff.h>
#include <linux/un.h>
#include <linux/sockios.h>
#include <linux/in.h>
#include <linux/poll.h>
#include <linux/uio.h>
#include <linux/qfabric.h>
#include <net/sock.h>
#include <asm/uaccess.h>
#include <linux/limits.h>
#include "../include/atl.h"
#include "cm.h"
#include "cm_internal.h"
#include "cm_transport.h"
#include "kcm.h"
#include "library.h"
#include "assert.h"

#define printf         printk
#define perror         printk
#define fprintf(fmt, args...) printk (args) 
#define exit           sys_exit
#define qsort          _quicksort
#define getpid()       current->pid
#define signal(a,b)    sys_signal(a,b)    
#define gethostname    sys_gethostname
#define gethostbyname  lookup_name
#define getdomainname  sys_getdomainname
#define select         sys_select
#endif

#ifndef SOCKET_ERROR
#define SOCKET_ERROR -1
#endif

typedef struct func_list_item {
    select_list_func func;
    void *arg1;
    void *arg2;
} FunctionListElement;

typedef struct socket_client_data {
    CManager	cm;
    char 	*hostname;
    int		listen_port;
    CMtrans_services svc;
} *socket_client_data_ptr;

typedef struct socket_connection_data {
    char *remote_host;
    int   remote_IP;
    int   remote_contact_port;
    int   fd;
    void *read_buffer;
    int   read_buffer_len;
    socket_client_data_ptr sd;
#ifdef MODULE
    struct task_struct *parent;
#endif
} *socket_conn_data_ptr;

#ifndef MODULE
#ifdef WSAEWOULDBLOCK
#define EWOULDBLOCK WSAEWOULDBLOCK
/*#define EAGAIN WSAEINPROGRESS*/
/*#define EINTR WSAEINTR*/
#define errno GetLastError()
#define read(fd, buf, len) recv(fd, buf, len, 0)
#define write(fd, buf, len) send(fd, buf, len, 0)
#endif
#endif

static int
check_host(hostname, sin_addr)
char *hostname;
void *sin_addr;
{
#ifdef MODULE
    struct hostent *host_addr;

    host_addr = lookup_name(hostname);
    if (host_addr == NULL) return 0;
    memcpy(sin_addr, host_addr->h_addr, host_addr->h_length);
    DFreeMM((addrs_t)host_addr->h_name);
    DFreeMM((addrs_t)host_addr->h_addr_list);
    DFreeMM((addrs_t)host_addr);
    return 1;
#else
#ifdef HAS_STRUCT_HOSTENT
    struct hostent *host_addr;
    host_addr = gethostbyname(hostname);
    if (host_addr == NULL) {
	unsigned long addr = inet_addr(hostname);
	if (addr == -1) {
	    /* 
	     *  not translatable as a hostname or 
	     * as a dot-style string IP address
	     */
	    return 0;
	}
	assert(sizeof(int) == sizeof(struct in_addr));
	*((int*)sin_addr) = (int)addr;
    } else {
	memcpy(sin_addr, host_addr->h_addr, host_addr->h_length);
    }
    return 1;
#else
    /* VxWorks ? */
    *((int *)sin_addr) = hostGetByName(hostname);
    return (*(int*)sin_addr) != -1;
#endif
#endif
}

static socket_conn_data_ptr create_socket_conn_data(svc)
CMtrans_services svc;
{
    socket_conn_data_ptr socket_conn_data = 
	svc->malloc(sizeof(struct socket_connection_data));
    socket_conn_data->remote_host = NULL;
    socket_conn_data->remote_contact_port = -1;
    socket_conn_data->fd = 0;
    socket_conn_data->read_buffer = svc->malloc(1);
    socket_conn_data->read_buffer_len = 1;
    return socket_conn_data;
}

#ifdef NOTDEF
static 
void dump_sockaddr(who,sa)
char *who;
struct sockaddr_in *sa;
{
    unsigned char *addr;

    addr = (unsigned char *) &(sa->sin_addr.s_addr);

    printf("%s: family=%d port=%d addr=%d.%d.%d.%d\n",
	   who,
	   ntohs(sa->sin_family),
	   ntohs(sa->sin_port),
	   addr[0], addr[1], addr[2], addr[3]);
}

static 
void dump_sockinfo(msg, fd)
char *msg;
int fd;
{
#ifndef MODULE
    int nl;
    struct sockaddr_in peer, me;

    printf("Dumping sockinfo for fd=%d: %s\n", fd, msg);

    nl = sizeof(me);
    getsockname(fd, (struct sockaddr *) &me, &nl);
    dump_sockaddr("Me", &me);
	   
    nl = sizeof(peer);
    getpeername(fd, (struct sockaddr *) &peer, &nl);
    dump_sockaddr("Peer",&peer);
#endif
}
#endif

/* 
 * Accept socket connection
 */
static void
socket_accept_conn(void_trans, void_conn_sock)
void *void_trans;
void *void_conn_sock;
{
    transport_entry trans = (transport_entry) void_trans;
    int conn_sock = (int) (long)void_conn_sock;
    socket_client_data_ptr sd = (socket_client_data_ptr) trans->trans_data;
    CMtrans_services svc = sd->svc;
    socket_conn_data_ptr socket_conn_data;
#ifndef MODULE
    int sock;
#else
    //struct tcp_opt *tp;
    struct socket *sock = NULL;
    struct socket *newsock = NULL;
    int newsockfd;
#endif
    struct sockaddr sock_addr;
    int sock_len =  sizeof(sock_addr);
    int int_port_num;
#ifndef MODULE
    struct linger linger_val;
#endif
    static int fd_atom = 0;
    static int peer_ip_atom = 0;
    static int peer_conn_port_atom = 0;
    static int this_conn_port_atom = 0;
    static int peer_listen_port_atom = 0;
    static int peer_hostname_atom = 0;

#ifdef DO_TCP_NODELAY
    int delay_value = 1;
#endif
    CMConnection conn;
    attr_list conn_attr_list = NULL;;

    if (fd_atom == 0) {
	fd_atom = attr_atom_from_string("CONNECTION_FILE_DESCRIPTOR");
	this_conn_port_atom = attr_atom_from_string("THIS_CONN_PORT");
	peer_conn_port_atom = attr_atom_from_string("PEER_CONN_PORT");
	peer_ip_atom = attr_atom_from_string("PEER_IP_ADDR");
	peer_listen_port_atom = attr_atom_from_string("PEER_LISTEN_PORT");
	peer_hostname_atom = attr_atom_from_string("PEER_HOSTNAME");
    }

#ifndef MODULE
    linger_val.l_onoff = 1;
    linger_val.l_linger = 60;
    if ((sock = accept(conn_sock, (struct sockaddr *) 0, (int *) 0)) == SOCKET_ERROR) {
	perror("Cannot accept socket connection");
	svc->fd_remove_select(sd->cm, conn_sock);
	fprintf(stderr, "failure in CMsockets  removing socket connection\n");
	return;
    }
    if (setsockopt(sock, SOL_SOCKET, SO_LINGER, (char *) &linger_val,
		   sizeof(struct linger)) != 0) {
	perror("set SO_LINGER");
	return;
    }
#else
    sock = look_up_pseudofd(conn_sock, 
	(struct socket **)current->qfabric->fdqfabric);

    if ((newsock = my_sock_accept(sock, O_NONBLOCK)) == NULL) {
	return;
    }

    newsockfd = get_pseudofd(newsock, 
	(struct socket **)current->qfabric->fdqfabric);
#endif
#ifdef DO_TCP_NODELAY
#ifdef MODULE
    tp = &(newsock->sk->tp_pinfo.af_tcp);
    tp->nonagle = (delay_value == 0) ? 0 : 1;
#else
    setsockopt(sock, IPPROTO_TCP, TCP_NODELAY, (char *) &delay_value,
	       sizeof(delay_value));
#endif
#endif
    socket_conn_data = create_socket_conn_data(svc);
    socket_conn_data->sd = sd;
#ifdef MODULE
    socket_conn_data->fd = newsockfd;
#else
    socket_conn_data->fd = sock;
#endif
    conn_attr_list = create_attr_list();
    conn = svc->connection_create(trans, socket_conn_data, conn_attr_list);

    add_attr(conn_attr_list, fd_atom, Attr_Int4,
	     (attr_value) sock);

    sock_len=  sizeof(sock_addr);
#ifdef MODULE
    newsock->ops->getname(newsock, &sock_addr,
		    &sock_len, 0);
#else
    getsockname(sock, (struct sockaddr *) &sock_addr, &sock_len);
#endif
    int_port_num = ntohs(((struct sockaddr_in *)&sock_addr)->sin_port);
    add_attr(conn_attr_list, this_conn_port_atom, Attr_Int4,
	     (attr_value) int_port_num);

    memset(&sock_addr, 0, sizeof(sock_addr));
    sock_len=  sizeof(sock_addr);
#ifdef MODULE
    if (newsock->ops->getname(newsock, &sock_addr,
		    &sock_len, 1) == 0) {
#else
    if (getpeername(sock, &sock_addr,
		    &sock_len) == 0) {
#endif
	int_port_num = ntohs(((struct sockaddr_in *)&sock_addr)->sin_port);
	add_attr(conn_attr_list, peer_conn_port_atom, Attr_Int4,
		 (attr_value) int_port_num);
	socket_conn_data->remote_IP = 
	    ((struct sockaddr_in *) &sock_addr)->sin_addr.s_addr;
	socket_conn_data->parent = current;
	add_attr(conn_attr_list, peer_ip_atom, Attr_Int4,
		 (attr_value) socket_conn_data->remote_IP);
	if (sock_addr.sa_family == AF_INET) {
#ifndef MODULE
#ifdef HAS_STRUCT_HOSTENT
	    struct hostent *host;
	    struct sockaddr_in *in_sock = (struct sockaddr_in *) &sock_addr;
	    host = gethostbyaddr((char *) &in_sock->sin_addr,
				 sizeof(struct in_addr), AF_INET);
	    if (host != NULL) {
		socket_conn_data->remote_host = strdup(host->h_name);
		add_attr(conn_attr_list, peer_hostname_atom, Attr_String,
			 (attr_value) strdup(host->h_name));
	    }
#endif
#endif
	}
    }
    
    if (socket_conn_data->remote_host != NULL) {
	svc->trace_out(NULL, "Accepted TCP/IP socket connection from host \"%s\"", 
		    socket_conn_data->remote_host);
    } else {
	svc->trace_out(NULL, "Accepted TCP/IP socket connection from UNKNOWN host");
    }

#ifdef MODULE
    my_sock_recvmsg(newsock, (char *)(&socket_conn_data->remote_contact_port),
       4);
#else	
    read(sock, &socket_conn_data->remote_contact_port, 4);
#endif
    socket_conn_data->remote_contact_port = 
	ntohs(socket_conn_data->remote_contact_port);
    add_attr(conn_attr_list, peer_listen_port_atom, Attr_Int4,
	     (attr_value) socket_conn_data->remote_contact_port);
    svc->trace_out(NULL, "Remote host (IP %x) is listening at port %d\n", 
		   socket_conn_data->remote_IP,
		   socket_conn_data->remote_contact_port);

/*    dump_sockinfo("accept ", sock);*/
#ifdef MODULE
    svc->fd_add_select(conn->cm, newsockfd, 
		     (void(*)(void*,void*)) trans->data_available,
		     (void *)trans, (void*)conn);
#else
    svc->fd_add_select(conn->cm, sock, 
		     (void(*)(void*,void*)) trans->data_available,
		     (void *)trans, (void*)conn);
#endif
}

extern void
libcmsockets_LTX_shutdown_conn(svc, scd)
CMtrans_services svc;
socket_conn_data_ptr scd;
{
#ifdef MODULE
    mm_segment_t fs;

    fs = get_fs();
    set_fs (get_ds());
    sys_close((int)(long)scd->fd);
    set_fs (fs);
#else
    close(scd->fd);
#endif
    svc->fd_remove_select(scd->sd->cm, scd->fd);
#ifdef MODULE
    DFreeMM((addrs_t)scd->remote_host);
    DFreeMM((addrs_t)scd);
#else
    free(scd->remote_host);
    free(scd);
#endif
}


/* 
 * Initiate a socket connection with another data exchange.  If port_num is -1,
 * establish a unix socket connection (name_str stores the file name of
 * the waiting socket).  Otherwise, establish an INET socket connection
 * (name_str stores the machine name).
 */
extern CMConnection
libcmsockets_LTX_initiate_conn(cm, svc, trans, attrs)
CManager cm;
CMtrans_services svc;
transport_entry trans;
attr_list attrs;
{
    int sock;

#ifdef DO_TCP_NODELAY
    int delay_value = 1;
#endif
    struct linger linger_val;
#ifndef MODULE
    int sock_opt_val = 1;
#endif
    int int_port_num;
    u_short port_num;
    socket_conn_data_ptr socket_conn_data;
    socket_client_data_ptr sd = (socket_client_data_ptr) trans->trans_data;
    CMConnection conn;
    static atom_t host_atom = 0;
    static atom_t port_atom = 0;
    static int fd_atom = 0;
    static int peer_ip_atom = 0;
    static int peer_conn_port_atom = 0;
    static int this_conn_port_atom = 0;
    static int peer_listen_port_atom = 0;
    static int peer_hostname_atom = 0;
    char *host_name;
    int remote_IP = -1;
    attr_list conn_attr_list;
    int sock_len;
    struct sockaddr sock_addr;
    struct sockaddr_in *sock_addri = (struct sockaddr_in *)&sock_addr;
#ifdef MODULE
    struct socket *socket;
    mm_segment_t fs;
#endif

    if (host_atom == 0) {
	host_atom = attr_atom_from_string("IP_HOST");
	port_atom = attr_atom_from_string("IP_PORT");
	fd_atom = attr_atom_from_string("CONNECTION_FILE_DESCRIPTOR");
	this_conn_port_atom = attr_atom_from_string("THIS_CONN_PORT");
	peer_conn_port_atom = attr_atom_from_string("PEER_CONN_PORT");
	peer_ip_atom = attr_atom_from_string("PEER_IP_ADDR");
	peer_listen_port_atom = attr_atom_from_string("PEER_LISTEN_PORT");
	peer_hostname_atom = attr_atom_from_string("PEER_HOSTNAME");
    }
    if (!query_attr(attrs, host_atom, /* type pointer */ NULL,
		    /* value pointer */ (attr_value *) &host_name)) {
	svc->trace_out(cm, "TCP/IP transport found no IP_HOST attribute");
	return NULL;
    }
    if (!query_attr(attrs, port_atom, /* type pointer */ NULL,
		    /* value pointer */ (attr_value *) &int_port_num)) {
	svc->trace_out(cm, "TCP/IP transport found no IP_PORT attribute");
	return NULL;
    }
    
    port_num = int_port_num;
    linger_val.l_onoff = 1;
    linger_val.l_linger = 60;

    svc->trace_out(cm, "Attempting TCP/IP socket connection, host=\"%s\", port %d",
		    host_name, int_port_num);
    if (int_port_num == -1) {
#if defined(AF_UNIX) && !defined(HAVE_WINDOWS_H)
	/* unix socket connection, host_name is the file name */
	struct sockaddr_un sock_addru;
#ifdef MODULE
	if ((sock_create(AF_UNIX, SOCK_STREAM, 0, &socket)) < 0) {
	    return NULL;
	}
        sock = get_pseudofd(socket, 
	    (struct socket **)current->qfabric->fdqfabric);
	sock_addru.sun_family = AF_UNIX;
	strcpy(sock_addru.sun_path, host_name);
	if (socket->ops->connect(socket, (struct sockaddr *) &sock_addru,
		    sizeof sock_addru, O_NONBLOCK) < 0) {
	    return NULL;
	}
#else
	if ((sock = socket(AF_UNIX, SOCK_STREAM, 0)) < 0) {
	    return NULL;
	}
	sock_addru.sun_family = AF_UNIX;
	strcpy(sock_addru.sun_path, host_name);
	if (connect(sock, (struct sockaddr *) &sock_addru,
		    sizeof sock_addru) < 0) {
	    return NULL;
	}
#endif
#else
	fprintf(stderr, "socket initiate_conn port_num parameter == -1 and unix sockets not available.\n");
	return NULL;
#endif
    } else {
	/* INET socket connection, host_name is the machine name */
	char *network_string;
#ifdef MODULE
	if ((sock_create(AF_INET, SOCK_STREAM, 0, &socket)) < 0) {
#else
	if ((sock = socket(AF_INET, SOCK_STREAM, 0)) == SOCKET_ERROR) {
#endif
	    svc->trace_out(cm, " CMSocket connect FAILURE --> Couldn't create socket");
	    return NULL;
	}
#ifdef MODULE
        sock = get_pseudofd(socket, 
	    (struct socket **)current->qfabric->fdqfabric);
#endif
	((struct sockaddr_in *)&sock_addr)->sin_family = AF_INET;
	if ((network_string = getenv("CM_NETWORK")) != NULL) {
	    int name_len = strlen(host_name) + 2 + strlen(network_string);
	    char *new_host_name = svc->malloc(name_len);
	    char *first_dot = strchr(host_name, '.');
	    memset(new_host_name, 0, name_len);
	    if (first_dot == NULL) {
		strcpy(new_host_name, host_name);
		strcat(new_host_name, "-");
		strcat(new_host_name, network_string);
	    } else {
		strncpy(new_host_name, host_name, first_dot - host_name);
		strcat(new_host_name, "-");
		strcat(new_host_name, network_string);
		strcat(new_host_name, first_dot);
	    }
	    if (check_host(new_host_name, (void*)&sock_addri->sin_addr) == 0)  {
		/* host has no NETWORK interface */
		if (check_host(host_name, (void*)&sock_addri->sin_addr) == 0) {
		    svc->trace_out(cm, "--> Host not found \"%s\"",
				    host_name);
		    return NULL;
		}
	    } else {
		svc->trace_out(cm, "--> Using non default network interface with hostname %s",
				new_host_name);
	    }
	    svc->free(new_host_name);
	} else {
	    if (check_host(host_name, (void*)&sock_addri->sin_addr) == 0) {
		svc->trace_out(cm, "CMSocket connect FAILURE --> Host not found \"%s\"", host_name);
		return NULL;
	    }
	}
	sock_addri->sin_port = htons(port_num);
	remote_IP = sock_addri->sin_addr.s_addr;
#ifdef MODULE
	if ((socket->ops->connect(socket, (struct sockaddr *) &sock_addr,
		    sizeof sock_addr, 0)) < 0) {
#ifdef WSAEWOULDBLOCK
	    int err = WSAGetLastError();
	    if (err != WSAEWOULDBLOCK || err != WSAEINPROGRESS) {
#endif
		svc->trace_out(cm, "--> Connection failed");
		fs = get_fs();
		set_fs (get_ds());
		sys_close((int)(long)sock);
		set_fs (fs);
		return NULL;
#ifdef WSAEWOULDBLOCK
	    }
#endif
	}
#else
	if (connect(sock, (struct sockaddr *) &sock_addr,
		    sizeof sock_addr) == SOCKET_ERROR) {
#ifdef WSAEWOULDBLOCK
	    int err = WSAGetLastError();
	    if (err != WSAEWOULDBLOCK || err != WSAEINPROGRESS) {
#endif
		svc->trace_out(cm, "CMSocket connect FAILURE --> Connect() failed");
		close(sock);
		return NULL;
#ifdef WSAEWOULDBLOCK
	    }
#endif
	}
#endif
    }
#ifndef MODULE
    sock_opt_val = 1;
    setsockopt(sock, SOL_SOCKET, SO_KEEPALIVE, (char *) &sock_opt_val,
	       sizeof(sock_opt_val));
    setsockopt(sock, SOL_SOCKET, SO_LINGER, (char *) &linger_val,
	       sizeof(struct linger));
#endif

#ifdef DO_TCP_NODELAY
#ifdef MODULE
    //socket->ops->setsockopt(socket, IPPROTO_TCP, TCP_NODELAY, 
//		(char *) &delay_value,
//	        sizeof(delay_value));
    tp = &(socket->sk->tp_pinfo.af_tcp);
    tp->nonagle = (delay_value == 0) ? 0 : 1;
#else
    setsockopt(sock, IPPROTO_TCP, TCP_NODELAY, (char *) &delay_value,
	       sizeof(delay_value));
#endif
#endif

    {
	int local_listen_port = htons(sd->listen_port);
#ifdef MODULE
	my_sock_sendmsg(socket, (char *)&local_listen_port, 4, 0, 0);
#else
	write(sock, &local_listen_port, 4);
#endif
   }

    svc->trace_out(cm, "--> Connection established");
    socket_conn_data = create_socket_conn_data(svc);
    socket_conn_data->remote_host = strdup(host_name);
    socket_conn_data->remote_IP = remote_IP;
    socket_conn_data->remote_contact_port = int_port_num;
    socket_conn_data->parent = current;
    socket_conn_data->fd = sock;
    socket_conn_data->sd = sd;
    
    conn_attr_list = create_attr_list();
    add_attr(conn_attr_list, fd_atom, Attr_Int4,
	     (attr_value) sock);
    sock_len=  sizeof(sock_addr);
#ifdef MODULE
    socket->ops->getname(socket, (struct sockaddr *)&sock_addr,
		    &sock_len, 0);
#else
    getsockname(sock, (struct sockaddr *) &sock_addr, &sock_len);
#endif
    int_port_num = ntohs(((struct sockaddr_in *)&sock_addr)->sin_port);
    add_attr(conn_attr_list, this_conn_port_atom, Attr_Int4,
	     (attr_value) int_port_num);
    add_attr(conn_attr_list, peer_ip_atom, Attr_Int4,
	     (attr_value) socket_conn_data->remote_IP);
#ifdef MODULE
    if (socket->ops->getname(socket, &sock_addr,
		    &sock_len, 1) == 0) {
#else
    if (getpeername(sock, &sock_addr,
		    &sock_len) == 0) {
#endif
	int_port_num = ntohs(((struct sockaddr_in *)&sock_addr)->sin_port);
	add_attr(conn_attr_list, peer_conn_port_atom, Attr_Int4,
		 (attr_value) int_port_num);
	if (sock_addr.sa_family == AF_INET) {
#ifndef MODULE
#ifdef HAS_STRUCT_HOSTENT
	    struct hostent *host;
	    struct sockaddr_in *in_sock = (struct sockaddr_in *) &sock_addr;
	    host = gethostbyaddr((char *) &in_sock->sin_addr,
				 sizeof(struct in_addr), AF_INET);
	    if (host != NULL) {
		socket_conn_data->remote_host = strdup(host->h_name);
		add_attr(conn_attr_list, peer_hostname_atom, Attr_String,
			 (attr_value) strdup(host->h_name));
	    }
#endif
#endif
	}
    }
    add_attr(conn_attr_list, peer_listen_port_atom, Attr_Int4,
	     (attr_value) socket_conn_data->remote_contact_port);
    conn = svc->connection_create(trans, socket_conn_data, conn_attr_list);

    svc->fd_add_select(cm, sock, (select_list_func)trans->data_available,
		     (void *)trans, (void*)conn);

/*    dump_sockinfo("initiate ", sock);*/
    return conn;
}

/* 
 * Check to see that if we were to attempt to initiate a connection as
 * indicated by the attribute list, would we be connecting to ourselves?
 * For sockets, this involves checking to see if the host name is the 
 * same as ours and if the IP_PORT matches the one we are listening on.
 */
extern int
libcmsockets_LTX_self_check(cm, svc, trans, attrs)
CManager cm;
CMtrans_services svc;
transport_entry trans;
attr_list attrs;
{

    socket_client_data_ptr sd = trans->trans_data;
    int int_port_num;
    atom_t host_atom = 0;
    atom_t port_atom = 0;
    char *host_name;
    char my_host_name[256];

    if (host_atom == 0) {
	host_atom = attr_atom_from_string("IP_HOST");
	port_atom = attr_atom_from_string("IP_PORT");
    }
    if (!query_attr(attrs, host_atom, /* type pointer */ NULL,
		    /* value pointer */ (attr_value *) &host_name)) {
	svc->trace_out(cm, "CMself check TCP/IP transport found no IP_HOST attribute");
	return 0;
    }
    if (!query_attr(attrs, port_atom, /* type pointer */ NULL,
		    /* value pointer */ (attr_value *) &int_port_num)) {
	svc->trace_out(cm, "CMself check TCP/IP transport found no IP_PORT attribute");
	return 0;
    }
    
    gethostname(my_host_name, sizeof(my_host_name));

    if (strcmp(host_name, my_host_name) != 0) {
	svc->trace_out(cm, "CMself check - Hostnames don't match");
	return 0;
    }
    if (int_port_num != sd->listen_port) {
	svc->trace_out(cm, "CMself check - Ports don't match");
	return 0;
    }
    svc->trace_out(cm, "CMself check returning TRUE");
    return 1;
}

extern int
libcmsockets_LTX_connection_eq(cm, svc, trans, attrs, scd)
CManager cm;
CMtrans_services svc;
transport_entry trans;
attr_list attrs;
socket_conn_data_ptr scd;
{

    int int_port_num;
    int requested_IP;
    atom_t host_atom = 0;
    atom_t port_atom = 0;
    char *host_name;

    if (host_atom == 0) {
	host_atom = attr_atom_from_string("IP_HOST");
	port_atom = attr_atom_from_string("IP_PORT");
    }
    if (!query_attr(attrs, host_atom, /* type pointer */ NULL,
		    /* value pointer */ (attr_value *) &host_name)) {
	svc->trace_out(cm, "TCP/IP transport found no IP_HOST attribute");
	return 0;
    }
    if (!query_attr(attrs, port_atom, /* type pointer */ NULL,
		    /* value pointer */ (attr_value *) &int_port_num)) {
	svc->trace_out(cm, "TCP/IP transport found no IP_PORT attribute");
	return 0;
    }
    
    check_host(host_name, (void*)&requested_IP);

    if ((scd->remote_IP == requested_IP) && 
	(scd->remote_contact_port == int_port_num)) {
	svc->trace_out(cm, "Socket Conn_eq returning TRUE");
	return 1;
    }
    svc->trace_out(cm, "Socket Conn_eq returning FALSE");
    return 0;
}

/* 
 * Create an IP socket for connection from other CMs
 */
extern attr_list
libcmsockets_LTX_non_blocking_listen(cm, svc, trans, listen_info)
CManager cm;
CMtrans_services svc;
transport_entry trans;
attr_list listen_info;
{
    socket_client_data_ptr sd = trans->trans_data;
    int length;
    struct sockaddr_in sock_addr;
    int sock_opt_val = 1;
    int conn_sock;
    int int_port_num = 0;
    u_short port_num = 0;
#ifdef MODULE
    struct socket *sock;

    conn_sock = sock_create(AF_INET, SOCK_STREAM, 0, &sock);
#else
    conn_sock = socket(AF_INET, SOCK_STREAM, 0);
#endif
    if (conn_sock == SOCKET_ERROR) {
	fprintf(stderr, "Cannot open INET socket\n");
	return NULL;
    }

    /*
     *  Check to see if a bind to a specific port was requested
     */
    if (listen_info != NULL 
        && !query_attr (listen_info, attr_atom_from_string ("IP_PORT"),
                        NULL, (attr_value*) &int_port_num)) {
      port_num = 0;
    } else {
      if (int_port_num > USHRT_MAX || int_port_num < 0) {
        fprintf (stderr, "Requested port number %d is invalid\n", int_port_num);
        return NULL;
      }
      port_num = int_port_num;
    }

    svc->trace_out(cm, "CMSocket begin listen, requested port %d", int_port_num);
    sock_addr.sin_family = AF_INET;
    sock_addr.sin_addr.s_addr = INADDR_ANY;
    sock_addr.sin_port = htons(port_num);
#ifdef MODULE
    sock->sk->reuse = 2;
    sock->sk->linger = 0;
#else
    if (setsockopt(conn_sock, SOL_SOCKET, SO_REUSEADDR, (char *) &sock_opt_val,
		   sizeof(sock_opt_val)) != 0) {
	fprintf(stderr, "Failed to set 1REUSEADDR on INET socket\n");
	return NULL;
    }
#endif
#ifdef MODULE
    if (sock->ops->bind(sock, (struct sockaddr *) &sock_addr,
	     sizeof sock_addr) < 0) {
#else
    if (bind(conn_sock, (struct sockaddr *) &sock_addr,
	     sizeof sock_addr) == SOCKET_ERROR) {
#endif
	fprintf(stderr, "Cannot bind INET socket\n");
	return NULL;
    }
    sock_opt_val = 1;
#ifndef MODULE
    if (setsockopt(conn_sock, SOL_SOCKET, SO_REUSEADDR, (char *) &sock_opt_val,
		   sizeof(sock_opt_val)) != 0) {
	perror("Failed to set 2REUSEADDR on INET socket");
	return NULL;
    }
#endif
    length = sizeof sock_addr;
#ifdef MODULE
    if (sock->ops->getname(sock,(struct sockaddr *)&sock_addr,&length,0)<0) {
	printk("Cannot get socket name\n");
	return NULL;
    }
#else
    if (getsockname(conn_sock, (struct sockaddr *) &sock_addr, &length) < 0) {
	fprintf(stderr, "Cannot get socket name\n");
	return NULL;
    }
#endif
    /* begin listening for conns and set the backlog */
#ifdef MODULE
    if (sock->ops->listen(sock, FD_SETSIZE) < 0) {
#else
    if (listen(conn_sock, FD_SETSIZE)) {
#endif
	fprintf(stderr, "listen failed\n");
	return NULL;
    }

#ifdef MODULE
    conn_sock = get_pseudofd(sock, 
	(struct socket **)current->qfabric->fdqfabric);
#endif

    /* set the port num as one we can be contacted at */

#ifdef MODULE
    svc->fd_add_select(cm, conn_sock+FD_SETSIZE, socket_accept_conn,
		     (void *)trans, (void*)conn_sock);
#else
    svc->fd_add_select(cm, conn_sock, socket_accept_conn,
		     (void *)trans, (void*)conn_sock);
#endif

    /* in the event the DE is shut down, close the socket */
    /* 
     *  -- Don't do this...  Close() seems to hang on sockets after 
     *  listen() for some reason.  I haven't found anywhere that defines 
     *  this behavior, but it seems relatively uniform. 
     */
    /* DExchange_add_close(de, close_socket_fd, (void*)conn_sock, NULL); */

    {
	char host_name[256];
#ifndef MODULE
	char domain_name[256];
#endif
	int int_port_num = ntohs(sock_addr.sin_port);
	atom_t host_atom = 0;
	atom_t port_atom = 0;
	attr_list ret_list;

	if (host_atom == 0) {
	    host_atom = attr_atom_from_string("IP_HOST");
	    port_atom = attr_atom_from_string("IP_PORT");
	}
	svc->trace_out(cm, "CMSocket listen succeeded on port %d", 
		       int_port_num);
	ret_list = create_attr_list();
	gethostname(host_name, sizeof(host_name));
	if (memchr((void *)host_name, '.', sizeof(host_name)) == NULL) {
	    strcat(host_name, ".");
#ifdef MODULE
	    strcat(host_name, system_utsname.domainname);
#else
	    getdomainname(domain_name, sizeof(domain_name));
	    strcat(host_name, domain_name);
#endif 
	}
	sd->hostname = strdup(host_name);
	sd->listen_port = int_port_num;
	add_attr(ret_list, host_atom, Attr_String,
		 (attr_value) strdup(host_name));
	add_attr(ret_list, port_atom, Attr_Int4,
		 (attr_value) int_port_num);
	return ret_list;
    }
}

#if defined(HAVE_WINDOWS_H) && !defined(NEED_IOVEC_DEFINE)
#define NEED_IOVEC_DEFINE
#endif

#ifdef NEED_IOVEC_DEFINE
struct	iovec {
     void *iov_base;
     int   iov_len;
};
#endif

/*
 *  This function will not be used unless there is no read_to_buffer function
 *  in the transport.  It is an example, meant to be copied in transports 
 *  that are more efficient if they allocate their own buffer space.
 */
extern void *
libcmsockets_LTX_read_func(svc, scd, requested_len, actual_len)
CMtrans_services svc;
socket_conn_data_ptr scd;
int requested_len;
int *actual_len;
{
    int left, iget;
#ifdef MODULE
    mm_segment_t fs;
    struct socket *sock;

    sock = look_up_pseudofd(scd->fd, 
	(struct socket **)current->qfabric->fdqfabric);
#endif

    if (scd->read_buffer_len < requested_len) {
	scd->read_buffer = svc->realloc(scd->read_buffer, requested_len);
    }
#ifdef MODULE
    fs = get_fs();
    set_fs(get_ds());
    iget = my_sock_recvmsg(sock, (char *) scd->read_buffer, requested_len);
    set_fs(fs);
#else
    iget = read(scd->fd, (char *) scd->read_buffer, requested_len);
#endif
    if (iget == 0) {
	*actual_len = 0;
	return 0;		/* end of file */
    } else if (iget == -1) {
	int lerrno = errno;
	if ((lerrno != EWOULDBLOCK) &&
	    (lerrno != EAGAIN) &&
	    (lerrno != EINTR)) {
	    /* serious error */
	    *actual_len = 0;
	    return NULL;
	} else {
	    iget = 0;
	}
    }
    left = requested_len - iget;
    while (left > 0) {
#ifdef MODULE
        fs = get_fs();
        set_fs(get_ds());
	iget = my_sock_recvmsg(sock, (char *)scd->read_buffer + 
                    requested_len - left,
		    left);
        set_fs(fs);
        if ((iget == 0) || (iget == -EAGAIN) || (iget == -EWOULDBLOCK)
		|| (iget == -EINTR)) {
	  *actual_len = requested_len - left;	
	  return scd->read_buffer;
        } else if (iget < 0) {
          iget = 0;
	  *actual_len = (requested_len - left);
	  return scd->read_buffer;
        }
#else
	iget = read(scd->fd, (char *)scd->read_buffer + requested_len - left,
		    left);
	if (iget == 0) {
	    *actual_len = requested_len - left;	/* end of file */
	    return scd->read_buffer;
	} else if (iget == -1) {
	    int lerrno = errno;
	    if ((lerrno != EWOULDBLOCK) &&
		(lerrno != EAGAIN) &&
		(lerrno != EINTR)) {
		/* serious error */
		*actual_len = (requested_len - left);
		return scd->read_buffer;
	    } else {
		iget = 0;
	    }
	}
#endif
	left -= iget;
    }
    *actual_len = iget;
    return scd->read_buffer;
}


#ifdef MODULE
extern int
libcmsockets_LTX_read_to_buffer_func(svc, scd, buffer, requested_len, tsk, flag)
CMtrans_services svc;
socket_conn_data_ptr scd;
void *buffer;
int requested_len;
struct task_struct *tsk;
int flag;
#else
extern int
libcmsockets_LTX_read_to_buffer_func(svc, scd, buffer, requested_len)
CMtrans_services svc;
socket_conn_data_ptr scd;
void *buffer;
int requested_len;
#endif
{
    int left, iget;
#ifdef MODULE
    struct socket *sock;
    int counter = 0;

    sock = look_up_pseudofd(scd->fd, 
	(struct socket **)current->qfabric->fdqfabric);
#endif

#ifdef MODULE
    iget = my_sock_recvmsg(sock, (char *) buffer, requested_len);
    if ((iget == -EAGAIN) || (iget == -EWOULDBLOCK) || (iget == -EINTR)) {
	if (flag == 0) return 0;
    }
#else
    iget = read(scd->fd, (char *) buffer, requested_len);
#endif
    if (iget == 0) {
	return 0;		/* end of file */
    } else if (iget == -1) {
	int lerrno = errno;
	if ((lerrno != EWOULDBLOCK) &&
	    (lerrno != EAGAIN) &&
	    (lerrno != EINTR)) {
	    /* serious error */
	    return 0;
	} else {
	    iget = 0;
	}
    }
#ifdef MODULE
    if (iget < 0) iget = 0;
#endif
    left = requested_len - iget;
    while (left > 0) {
#ifdef MODULE
	iget = my_sock_recvmsg(sock, (char *)buffer + requested_len - left,
	    left);
	if (iget == left)
	{
	  return requested_len;
	}
	if ((counter < 19) && (iget <= 0)) {
	  interruptible_sleep_on_timeout(&current->qfabric->kecho_queue, HZ/2);
	}
	if (iget <= 0)
	  counter++;
	else counter = 0;
        if (((iget == 0) || (iget == -EAGAIN) || (iget == -EWOULDBLOCK)
		|| (iget == -EINTR)) && (counter > 19)) {
	  return requested_len - left;	/* end of file */
        } else if ((iget < 0) && (counter > 19)) {
          iget = 0;
	  return (requested_len - left);
	}
	if (iget < 0) iget = 0;
#else
	iget = read(scd->fd, (char *)buffer + requested_len - left,
		    left);
	if (iget == 0) {
	    return requested_len - left;	/* end of file */
	} else if (iget == -1) {
	    int lerrno = errno;
	    if ((lerrno != EWOULDBLOCK) &&
		(lerrno != EAGAIN) &&
		(lerrno != EINTR)) {
		/* serious error */
		return (requested_len - left);
	    } else {
		iget = 0;
	    }
	}
#endif
	left -= iget;
    }
    return requested_len;
}


extern int
libcmsockets_LTX_write_func(svc, scd, buffer, length)
CMtrans_services svc;
socket_conn_data_ptr scd;
void *buffer;
int length;
{
    int left = length;
    int iget = 0;
    int fd = scd->fd;
#ifdef MODULE
    struct socket *sock;

    sock = look_up_pseudofd(fd, (struct socket **)current->qfabric->fdqfabric);
#endif

    while (left > 0) {
#ifdef MODULE
	iget = my_sock_sendmsg(sock, (char *) buffer + length - left, left, -1, 0);
#else
	iget = write(fd, (char *) buffer + length - left, left);
#endif
	if (iget == -1) {
	    int lerrno = errno;
	    if ((lerrno != EWOULDBLOCK) &&
		(lerrno != EAGAIN) &&
		(lerrno != EINTR)) {
		/* serious error */
		return (length - left);
	    } else {
		iget = 0;
	    }
	}
	left -= iget;
    }
    return length;
}

#ifndef IOV_MAX
/* this is not defined in some places where it should be.  Conservative. */
#define IOV_MAX 16
#endif 

#ifndef HAVE_WRITEV
static 
int writev(fd, iov, iovcnt)
int fd;
struct iovec *iov;
int iovcnt;
{
    int wrote = 0;
    int i;
#ifdef MODULE
    mm_segment_t fs;
    struct socket *sock;

    sock = look_up_pseudofd(fd, (struct socket **)current->qfabric->fdqfabric);
#endif

    for(i=0; i<iovcnt; i++) {
	int left = iov[i].iov_len;
	int iget = 0;

	while (left > 0) {
#ifdef MODULE
	    iget = my_sock_sendmsg(sock, (char *) iov[i].iov_base + 
                   iov[i].iov_len - left, left, -1, 0); 
#else
	    iget = write(fd, (char *) iov[i].iov_base + iov[i].iov_len - left, left);
#endif
	    if (iget == -1) {
		int lerrno = errno;
		if ((lerrno != EWOULDBLOCK) &&
		    (lerrno != EAGAIN) &&
		    (lerrno != EINTR)) {
		    /* serious error */
		    return -1;
		} else {
		    iget = 0;
		}
	    }
	    left -= iget;
	}
	wrote += iov[i].iov_len;
    }
    return wrote;
}
#endif

extern int
libcmsockets_LTX_writev_attr_func(svc, scd, iov, iovcnt, attrs, flag)
CMtrans_services svc;
socket_conn_data_ptr scd;
struct iovec *iov;
int iovcnt;
attr_list attrs;
int flag;
{
    struct task_struct *parent = scd->parent;
    int fd = scd->fd;
    int left = 0;
    int iget = 0;
    int iovleft, i;
#ifdef MODULE
    struct socket *sock;
    int deadline = -1;
    int flags = flag;
    int retries = 5;

    if (flag == 1919)
      sock=look_up_pseudofd(fd, (struct socket **)parent->qfabric->fdqfabric);
    else
      sock=look_up_pseudofd(fd, (struct socket **)current->qfabric->fdqfabric);

    if (sock == NULL)
      return iovcnt;
#endif

    iovleft = iovcnt;

    /* sum lengths */
    for (i=0; i<iovcnt; i++) left += iov[i].iov_len;

#ifdef MODULE
    while (left > 0) {
	int write_count = iovleft;
	if (write_count > IOV_MAX) write_count = IOV_MAX;
	iget = my_sock_sendmsgv(sock, (struct iovec *) &iov[iovcnt - iovleft],
		      write_count, deadline, flags);
	if (iget < 0) {
	    if ((iget != -EWOULDBLOCK) && (iget != -EAGAIN)) {
		/* serious error */
		return (iovcnt - iovleft);
	    } else {
		retries--;
		if (retries == 0) {
			return (iovcnt - iovleft);
		}
		iget = 0;
	    }
	}
#else
	iget = writev(fd, (struct iovec *) &iov[iovcnt - iovleft],
		      write_count);
	if (iget == -1) {
	    if ((errno != EWOULDBLOCK) && (errno != EAGAIN)) {
		/* serious error */
		return (iovcnt - iovleft);
	    } else {
		iget = 0;
	    }
	}
#endif
	if (iget == left) {
	    return iovcnt;
	}
	left -= iget;
	while (iget > 0) {
	    iget -= iov[iovcnt - iovleft].iov_len;
	    iovleft--;
	}

	if (iget < 0) {
	    /*
	     * Only part of the last block was written.  Modify IO 
	     * vector to indicate the remaining block to be written.
	     */
	    /* restore iovleft and iget to cover remaining block*/
	    iovleft++;
	    iget += iov[iovcnt - iovleft].iov_len;

	    /* adjust count down and base up by number of bytes written */
	    iov[iovcnt-iovleft].iov_len -= iget;
	    iov[iovcnt-iovleft].iov_base = 
		(char*)(iov[iovcnt-iovleft].iov_base) + iget;
	}
    }
    return iovcnt;
}

extern int
libcmsockets_LTX_writev_func(svc, scd, iov, iovcnt, flag)
CMtrans_services svc;
socket_conn_data_ptr scd;
struct iovec *iov;
int iovcnt;
int flag;
{
    return libcmsockets_LTX_writev_attr_func(svc, scd, iov, iovcnt, NULL, flag);
}

int socket_global_init = 0;

#ifdef HAVE_WINDOWS_H
/* Winsock init stuff, ask for ver 1.1 */
static WORD wVersionRequested = MAKEWORD(1, 1);
static WSADATA wsaData;
#endif

extern void *
libcmsockets_LTX_initialize(cm, svc)
CManager cm;
CMtrans_services svc;
{
    
    socket_client_data_ptr socket_data;
    svc->trace_out(cm, "Initialize TCP/IP Socket transport");
    if (socket_global_init == 0) {
#ifdef HAVE_WINDOWS_H
	int nErrorStatus;
	/* initialize the winsock package */
	nErrorStatus = WSAStartup(wVersionRequested, &wsaData);
	if (nErrorStatus != 0) {
	    fprintf(stderr, "Could not initialize windows socket library!");
	    WSACleanup();
	    exit(-1);
	}
#endif
	/* 
	 * ignore SIGPIPE's  (these pop up when ports die.  we catch the 
	 * failed writes) 
	 */
#ifdef SIGPIPE
	signal(SIGPIPE, SIG_IGN);
#endif
    }
    socket_data = svc->malloc(sizeof(struct socket_client_data));
    socket_data->cm = cm;
    socket_data->hostname = NULL;
    socket_data->listen_port = -1;
    socket_data->svc = svc;
    return (void*) socket_data;
}

