/* -*-C-*-
*******************************************************************************
* 
* File:        comm_group2.h
* RCS:         $Id: comm_group2.h,v 1.1.1.1 2000/04/30 21:09:48 fabianb Exp $
* Description: New cm-based comm_group I/F, derived from greg's code.
* Author:      Fabian E. Bustamante (fabianb@cc.gatech.edu)
*              Systems Research Group
*              College of Computing
*              Georgia Institute of Technology
* Created:
* Modified:
*  
* (C) Copyright 2000, all rights reserved.
*
*******************************************************************************
*/
#ifndef __COMM_GROUP__H__
#define __COMM_GROUP__H__

#ifdef MODULE
#ifndef __KERNEL__
#define __KERNEL__
#endif
#endif

#if defined(FUNCPROTO) || defined(__STDC__) || defined(__cplusplus) || defined(c_plusplus)
#ifndef ARGS
#define ARGS(args) args
#endif
#else
#ifndef ARGS
#define ARGS(args) (/*args*/)
#endif
#endif

typedef struct _comm_group_struct {
  char *user_name;
  char *application_name;
  char *group_type;
  char *group_id;
  char *host;
  int port;
  int info_len;
  void *info;
} comm_group_struct, *comm_group_list;

typedef struct _comm_group_return {
  int cond;			/* no ideal but could be worst */
  int count;
  comm_group_list list;
} ret_group_struct, *comm_group_return;

//#define MAX_STR 256

typedef struct _syscall_attrs_cg {
  char name[MAX_STR];
  int name_len;
  char username[MAX_STR];
  char appname[MAX_STR];
  char grouptype[MAX_STR];
  int exp_time;
  char host[MAX_STR];
  int port;
  char groupid[MAX_STR];
  char *group_id;
  void *data;
  int data_len;
  int retval;
  char applist[MAX_STR];
  char typelist[MAX_STR];
  comm_group_return ret;
  char *info;
  char groupinfo[MAX_STR];
}syscall_attrs_cg, *syscall_attrs_cg_ptr;


#ifndef KECHO
extern char * 
setup_comm_group ARGS((char *user_name, char *application_name, char *group_type, 
		       int expiration_time, char *host, int port));

extern char * 
setup_specific_comm_group ARGS((char *application_name, char *group_type, 
				int expiration_time, char *host, int port));

extern int 
init_comm_group_contact ARGS((char *user_name, char *application_name,
			      char *group_type, char **host, int *port));

extern int 
init_specific_comm_group_contact ARGS((char *group_id, char **host, int *port));

extern int 
group_server_present();

extern int 
group_server_test();

extern void 
assoc_group_info ARGS((char *group_name, void *data, int data_len));

extern comm_group_return 
matching_comm_groups ARGS((char *app_list, char *type_list));

extern CMConnection 
CMConnection_initiate_first ARGS((CManager cm, comm_group_return groups));

typedef void (*group_error_routine_type) ();

extern void 
set_comm_group_error_handler ARGS((group_error_routine_type error_routine));

extern void 
comm_group_set_app_name ARGS((char *app_name));

extern void 
comm_group_set_app_list ARGS((char *list));

extern 
void comm_group_set_comm_type ARGS((char *comm_type));

extern 
void comm_group_set_type_list ARGS((char *list));

#else
#include <linux/qfabric.h>

syscall_attrs_cg gattrs1;
syscall_attrs_cg gattrs2;
syscall_attrs_cg gattrs3;
syscall_attrs_cg gattrs4;
syscall_attrs_cg gattrs5;
syscall_attrs_cg gattrs6;
syscall_attrs_cg gattrs7;
syscall_attrs_cg gattrs8;

char *setup_comm_group(char *user_name, char *application_name, 
	char *group_type, int expiration_time, char *host, int port)
{
  char *name = "setup_comm_group";

  memcpy((void *)&gattrs1.name, (void *)name, strlen(name)+1);
  gattrs1.name_len = strlen(gattrs1.name);
  if (user_name != NULL) 
    memcpy((void *)&gattrs1.username, (void *)user_name, strlen(user_name)+1);
  else gattrs1.username[0] = '0';
  if (application_name != NULL) 
    memcpy((void *)&gattrs1.appname, (void *)application_name, 
	strlen(application_name)+1);
  else gattrs1.appname[0] = '0';
  if (group_type != NULL) 
    memcpy((void *)&gattrs1.grouptype, (void *)group_type, strlen(group_type)+1);
  else gattrs1.grouptype[0] = '0';
  gattrs1.exp_time = expiration_time;
  if (host != NULL)
    memcpy((void *)&gattrs1.host, (void *)host, strlen(host)+1);
  else gattrs1.host[0] = '0';
  gattrs1.port = port;

  KECho_syscall_cg((void *)&gattrs1);

  return gattrs1.group_id;
}

char *setup_specific_comm_group(char *application_name, char *group_type,
	int expiration_time, char *host, int port) 
{
  char *name = "setup_specific_comm_group";

  memcpy((void *)&gattrs2.name, (void *)name, strlen(name)+1);
  gattrs2.name_len = strlen(gattrs2.name);
  memcpy((void *)&gattrs2.appname, (void *)application_name, 
	strlen(application_name)+1);
  memcpy((void *)&gattrs2.grouptype, (void *)group_type, strlen(group_type)+1);
  gattrs2.exp_time = expiration_time;
  memcpy((void *)&gattrs2.host, (void *)host, strlen(host)+1);
  gattrs2.port = port;

  KECho_syscall_cg((void *)&gattrs2);

  return gattrs2.group_id;
}

int init_comm_group_contact(char *user_name, char *application_name,
     char *group_type, char **host, int *port)
{
  char *name = "init_comm_group_contact";

  memcpy((void *)&gattrs3.name, (void *)name, strlen(name)+1);
  gattrs3.name_len = strlen(gattrs3.name);
  memcpy((void *)&gattrs3.username, (void *)user_name, strlen(user_name)+1);
  memcpy((void *)&gattrs3.appname, (void *)application_name, 
	strlen(application_name)+1);
  memcpy((void *)&gattrs3.grouptype, (void *)group_type, strlen(group_type)+1);
  memcpy((void *)&gattrs3.host, (void *)*host, strlen(*host)+1);
  gattrs3.port = *port;

  KECho_syscall_cg((void *)&gattrs3);

  return gattrs3.retval;
}

int init_specific_comm_group_contact(char *group_id, char **host, int *port)
{
  char *name = "init_specific_comm_group_contact";

  memcpy((void *)&gattrs4.name, (void *)name, strlen(name)+1);
  gattrs4.name_len = strlen(attrs4.name);
  memcpy((void *)&gattrs4.groupid, (void *)group_id, strlen(group_id)+1);
  memcpy((void *)&gattrs4.host, (void *)*host, strlen(*host)+1);
  gattrs4.port = *port;

  KECho_syscall_cg((void *)&gattrs4);

  return gattrs4.retval;
}

int group_server_present()
{
  char *name = "group_server_present";

  memcpy((void *)&gattrs5.name, (void *)name, strlen(name)+1);
  gattrs5.name_len = strlen(gattrs5.name);

  KECho_syscall_cg((void *)&gattrs5);

  return gattrs5.retval;
}

int group_server_test()
{
  char *name = "group_server_test";

  memcpy((void *)&gattrs6.name, (void *)name, strlen(name)+1);
  gattrs6.name_len = strlen(gattrs6.name);

  KECho_syscall_cg((void *)&gattrs6);

  return gattrs6.retval;
}

void assoc_group_info(char *group_name, void *data, int data_len)
{
  char *name = "assoc_group_info";

  memcpy((void *)&gattrs7.name, (void *)name, strlen(name)+1);
  gattrs7.name_len = strlen(gattrs7.name);
  memcpy((void *)&gattrs7.grouptype, (void *)group_name, strlen(group_name)+1);
  gattrs7.data = data;
  gattrs7.data_len = data_len;

  KECho_syscall_cg((void *)&gattrs7);
}

comm_group_return matching_comm_groups(char *app_list, char *type_list)
{
  char *name = "matching_comm_groups";

  memcpy((void *)&gattrs8.name, (void *)name, strlen(name)+1);
  gattrs8.name_len = strlen(gattrs8.name);
  memcpy((void *)&gattrs8.applist, (void *)app_list, strlen(app_list)+1);
  memcpy((void *)&gattrs8.typelist, (void *)type_list, strlen(type_list)+1);

  gattrs8.ret = malloc(sizeof(ret_group_struct));
  gattrs8.ret->list = malloc(sizeof(comm_group_struct));

  KECho_syscall_cg((void *)&gattrs8);

  if (gattrs8.groupinfo != NULL) {
    name = (char *)malloc(strlen(gattrs8.groupinfo)+1);
    memcpy(name, gattrs8.groupinfo, strlen(gattrs8.groupinfo)+1);
    gattrs8.ret->list[0].info = name;
    gattrs8.ret->count = 1;
  } else {
    gattrs8.ret->list[0].info = NULL;
    gattrs8.ret->count = 0;
  }

  return gattrs8.ret;
}

#endif
#endif
