/*
 *   ECalls.c
 *
 *   Copyright (C) 1999/2000 Christian Poellabauer
 *
 */

/* 
 *   Kernel-loadable module for ECalls
 */

/* Required definitions. */
#ifndef MODULE
#define MODULE
#endif
#ifndef __KERNEL__
#define __KERNEL__
#endif

/* Various required header files. */
#include <linux/config.h> 
#include <linux/module.h> 
#include <linux/version.h> 
#include <linux/modversions.h> 
#include <linux/kernel.h> 
#include <linux/types.h> 
#include <linux/tty.h> 
#include <linux/kernel_stat.h> 
#include <linux/unistd.h>
//#include <linux/skbuff.h>
#include <linux/socket.h>
#include <linux/sched.h>
#include <linux/utsname.h>
#include <linux/types.h>
#include <linux/uio.h>
#include <linux/proc_fs.h>
#include <linux/mm.h>
#include <linux/pagemap.h>
#include <linux/iobuf.h>
//#include <linux/pagemap.h>
#include <asm/errno.h>
#include <asm/current.h>
#include <asm/uaccess.h>
#include <asm/pgtable.h>
#include <asm/io.h>
#ifndef ecalls_module_h
#include "ecalls_module.h"
#endif

MODULE_AUTHOR("CERCS - GATECH");
MODULE_DESCRIPTION("ECalls Module (KECho)");
MODULE_LICENSE("GPL");

#define EXIT_SUCCESS 0
#define EXIT_FAILURE -1
#ifdef CONFIG_DWCS
#define DWCS_SERVICED_THIS_PERIOD 0x10000000
#endif

#define myrdtsc(x)	__asm__ __volatile__("rdtsc \n\t" : "=A" (*(x)))

/* Global variables. */
int thr_pool[2][255];
int pool_size;
int applications_registered = 0;
char *services[MAX_SERVICES];
struct service_data service_list[MAX_SERVICES]; 
struct pid_list pidlist[MAX_PROCESSES];
struct hook_node hook_list[MAX_HOOKS];
struct proc_dir_entry *hooks_dir;

int pidlist_size = 0;
struct timer_list service_timer[MAX_SERVICES];
struct upcall_task_struct *head = NULL;
struct place_holder_struct {
  unsigned long period;
  unsigned long deadline;
  unsigned long service_time;
  unsigned long ready_time;
};
struct place_holder_struct place_holder;
struct kernel_functions mod_func_list[MAX_FUNCS];
void *handler_list[255];
int i, j;
int mypid;
int queuelength = 0;

DECLARE_MUTEX(ecalls_sem);
DECLARE_MUTEX(heap_sem);

/*
static DECLARE_WAIT_QUEUE_HEAD(wait_head0);
static DECLARE_WAIT_QUEUE_HEAD(wait_head1);
static DECLARE_WAIT_QUEUE_HEAD(wait_head2);
*/

/* Prototypes. */
int init_module(void);		       
void cleanup_module(void);
int (*original_start_ecalls)(int);
int (*original_end_ecalls)();
int (*original_register_ecall)(char* service, int len,
	struct ecall_down *downcall,
	struct ecall_up *upcall);
int (*original_do_downcall)(struct ecall_down *downcall , void *data, int len);
void (*original_run_event_handler)(void);
//int (*original_fast_do_downcall)(char *service, int len);
int (*original_upcall_scheduler)(void);
int (*original_download_code)(void);
int (*sys_mlock) (unsigned long start, size_t len);
int internal_end_ecalls(void); 

/* arch/i386/kernel/entry.S */
extern void *sys_call_table[];
//extern void *fast_sys_call_table[];

extern int upcall_sched_active;
extern int DWCS_module_loaded;
extern int ECalls_current_flag;
extern int ECalls_active;
extern struct task_struct *ECalls_current;

/* Strdup function. */
char *strdup(char *s)
{
  char *p;

  p = (char *)kmalloc(strlen(s)+1, GFP_KERNEL);
  if (p != NULL)
    strcpy(p, s);
  return p;
}

/* External functions. */
extern int HCompare(struct pid_list entry1, struct pid_list entry2);
extern void HHeapify(struct pid_list pidlist[], int i, int size);
extern int HExist(struct pid_list pidlist[], int pid, int start, int size);
extern void 
    HInsert(struct pid_list pidlist[], struct pid_list tmp_heap, int size);
extern void HExtractMin(struct pid_list pidlist[], int size);
extern void 
    HDelete(struct pid_list pidlist[], int pid, int size);



/* read /proc/kernel-hooks */
static int hooks_read(char *page, char **start, off_t off, int count, int *eof,
	void *data) 
{
  int len = 0;
  int i = 0;
  static char my_buffer[4096];
 
  if (off > 0) return 0; 
  MOD_INC_USE_COUNT;
  len = sprintf(my_buffer, "Available Hooks:\n");
  while (i < MAX_HOOKS) {
    if (hook_list[i].callback != NULL) 
      len += sprintf(my_buffer+len, "%s: %s (%s)\n", hook_list[i].hookname, 
		hook_list[i].description, hook_list[i].parm_list); 
    i++;
  }
  *start = my_buffer;
  MOD_DEC_USE_COUNT;
  return len;
}

/* read /proc/kservices */
static int kservices_read(char *page, char **start, off_t off, int count, int *eof,
	void *data) 
{
  int len = 0;
  int i = 0;
  static char my_buffer[4096];
 
  if (off > 0) return 0; 
  MOD_INC_USE_COUNT;
  len = sprintf(my_buffer, "Available Services:\n");
  while ((i < MAX_SERVICES) && (services[i] != NULL)) {
    len += sprintf(my_buffer+len, "%s\n", services[i]);
    i++;
  }
  *start = my_buffer;
  MOD_DEC_USE_COUNT;
  return len;
}

/* Register hook for extension */
int register_hook(char *hookname, char *description, char *parm_list, void *callback)
{
  int i = 0;

  while ((i < MAX_SERVICES) && (hook_list[i].callback != NULL))
    i++;

  if (i >= MAX_SERVICES) 
    return EXIT_FAILURE;

  strcpy(hook_list[i].hookname, hookname);
  strcpy(hook_list[i].description, description);
  strcpy(hook_list[i].parm_list, parm_list);
  hook_list[i].callback = callback;

  return EXIT_SUCCESS;
}

/* Unregister hook for extension */
int unregister_hook(char *hookname)
{
  int i = 0;

  while ((i < MAX_SERVICES) && (strcmp(hook_list[i].hookname, hookname) != 0)) 
    i++;

  if (i < MAX_SERVICES) {
    hook_list[i].callback = NULL;
  }

  return EXIT_SUCCESS;
}

/* Register kernel service. */
int register_ecalls_service(char *service_name, void *fast_handler_func,
	void *slow_handler_func, int type)
{
  int i = 0;
  int j;

  down(&ecalls_sem);
  while ((i < MAX_SERVICES) && (services[i] != NULL))
    i++;

  if (i < MAX_SERVICES) {
    services[i] = strdup(service_name);	
    service_list[i].service_id = i;
    service_list[i].type = type;
    service_list[i].service_name = strdup(service_name);
    service_list[i].clients = 0;
    service_list[i].fast_handler_func = fast_handler_func;
    service_list[i].slow_handler_func = slow_handler_func;
    for (j = 0; j < MAX_CLIENTS; j++)
      service_list[i].client_pid[j] = -1;
      service_list[i].context[j] = NULL;
  }
  else {
    up(&ecalls_sem);
    return EXIT_FAILURE;
  }

  if ((fast_handler_func != NULL) && (type == DEFERRED))
    ECalls_active++;

  up(&ecalls_sem);
  return i;
}

/* Unregister service from ecalls. */
int unregister_ecalls_service(char *service_name)
{
  int i = 0;
  int j;

  down(&ecalls_sem);
  while (i < MAX_SERVICES) {
    if ((services[i] != NULL) && 
	(strcmp(services[i], service_name) == 0))
      break;
    i++;
  }

  if (i < MAX_SERVICES) {
    kfree(services[i]); 
    if ((service_list[i].fast_handler_func != NULL) && 
	(service_list[i].type == DEFERRED))
      ECalls_active--;

    service_list[i].service_id = -1;
    service_list[i].type = -1;
    service_list[i].service_name = NULL;
    service_list[i].clients = 0;
    service_list[i].fast_handler_func = NULL;
    service_list[i].slow_handler_func = NULL;
    for (j = 0; j < MAX_CLIENTS; j++)
      service_list[i].client_pid[j] = -1;
  }
  else {
    up(&ecalls_sem);
    return EXIT_FAILURE;
  }

  up(&ecalls_sem);
  return EXIT_SUCCESS;
}

/* Register kernel function. */
int register_function(char *function_name, void *function_addr)
{
  int i = 0;

  down(&ecalls_sem);
  while ((mod_func_list[i].flag != -1) && (i < MAX_FUNCS))
    i++;

  if (i == MAX_FUNCS) {
    up(&ecalls_sem);
    return EXIT_FAILURE;
  }

  mod_func_list[i].flag = 1;
  strcpy((char *)mod_func_list[i].identifier, function_name);
  mod_func_list[i].function_ptr = function_addr;

  up(&ecalls_sem);
  return i;
}

/* Unregister kernel function. */
int unregister_function(char *function_name)
{
  int i = 0;

  down(&ecalls_sem);
  while ((strcmp(mod_func_list[i].identifier, function_name) != 0) 
	&& (i < MAX_FUNCS))
    i++;

  if (i == MAX_FUNCS) {
    up(&ecalls_sem);
    return EXIT_FAILURE;
  }

  mod_func_list[i].flag = -1;

  up(&ecalls_sem);
  return EXIT_SUCCESS;
}

/* Event handler. */
void run_event_handler(void)
{
  int a = -1;
  int flag;
  asmlinkage int (*service_function)(unsigned long data);

  if (current->state == 4) {
    internal_end_ecalls(); 
  } else {
    mypid = current->pid;
    while ((a = HExist(pidlist, mypid, a + 1, pidlist_size)) != -1) {
      i = pidlist[a].i;
      j = pidlist[a].j;
      if ((flag = *(int*)service_list[i].flag_down[j])) {
        service_function = service_list[i].fast_handler_func;
        service_function(flag);
        *(int *)service_list[i].flag_down[j] = 0;
      }
    }
  }
}

/* Fast system call. */
int handle_event(struct ecall_down *downcall, void *data, int len)
{
  asmlinkage int (*service_function)(unsigned long data);

  service_function = service_list[downcall->service_id].fast_handler_func;
  if (service_function != NULL) {
    return service_function(downcall->service_id);
  } else {
    return EXIT_FAILURE; 
  }
}

/* Request number of clients for service. */
int request_num_clients(int id)
{
  return service_list[id].clients;
}

/* Poll clients in behalf of service. */
void poll_clients(unsigned long data) 
{
  int i = 0;
  int flag;
  int clients;
  asmlinkage int (*service_function)(unsigned long data);

  clients = service_list[data].clients;
  if (clients > 0) {
    while ((i < MAX_CLIENTS) && (clients > 0)) {
      if (service_list[data].client_pid[i] > 0) {
        memcpy(&flag, 
          (void *)service_list[data].flag_down_phys[i], sizeof (int));
        if (flag > 0) {
          flag = 0;
          memcpy((void *)service_list[data].flag_down_phys[i],
	    &flag,sizeof (int));
          service_function = service_list[data].fast_handler_func;
          service_function(i);
        }
        clients--;
      }
      i++;
    }
  }

  service_timer[data].function = poll_clients;
  service_timer[data].data = data;
  service_timer[data].expires = jiffies+service_list[data].poll_period;
  add_timer(&service_timer[data]);
}

/* Poll request from service. */
int start_polling(int id, int period)
{
  service_list[id].poll_period = period;
  init_timer(&service_timer[id]);
  service_timer[id].function = poll_clients;
  service_timer[id].data = id;
  service_timer[id].expires = jiffies+period;
  add_timer(&service_timer[id]);

  return EXIT_SUCCESS;
}

/* Stop polling. */
int stop_polling(int id)
{
  del_timer(&service_timer[id]);

  return EXIT_SUCCESS;
}

/* Poll if service requests available. */
int poll_service_requests(int id, int index) 
{
  int i = 0;
  int clients;
  int flag;

  down(&ecalls_sem); 
  clients = service_list[id].clients;
  if (clients < 1) {
    up(&ecalls_sem); 
    return -1;
  }

  while ((i < MAX_CLIENTS) && (clients > 0)) {
    if (service_list[id].client_pid[i] > 0) {
      if (index == 0) {
        memcpy(&flag, (void *)service_list[id].flag_down_phys[i], sizeof (int));
        if (flag > 0) {
          flag = 0;
          memcpy((void *)service_list[id].flag_down_phys[i],&flag,sizeof (int));
          up(&ecalls_sem); 
          return service_list[id].client_pid[i];
 	}
      }
      else 
        index--;
      clients--;
    }
    i++;
  }

  up(&ecalls_sem); 
  return -1;
}

/* Start ECalls mechanism. */
int internal_start_ecalls(int flag)
{
  return 0;
}  

/* End ECalls mechanism. */
int internal_end_ecalls(void) 
{
  int i, j;
  int count = 1;

  current->ecalls_flag = 0;
  /* cleanup data structures */
  down(&ecalls_sem); 
  while (HExist(pidlist, current->pid, 0, pidlist_size) != -1) {
    pidlist_size--;
    HDelete(pidlist, current->pid, pidlist_size);
    if (count == 1) {
      applications_registered--;
      count = 0;
    }
  }

  for (i = 0; i < MAX_SERVICES; i++) 
    for (j = 0; j < MAX_CLIENTS; j++) 
      if (service_list[i].client_pid[j] == current->pid) {
        service_list[i].client_pid[j] = -1; 
        service_list[i].tptr[j] = NULL; 
        service_list[i].clients--;
      }

  up(&ecalls_sem);
  return EXIT_SUCCESS;
}

/* Register downcall with application. */
int internal_register_ecall(char *service, int len,
	struct ecall_down *downcall, struct ecall_up *upcall) 
{
  int i = 0;
  int j = 0;
  char kservice[MAX_LEN];
  char kfunction[MAX_LEN];
  struct pid_list tmp_pid;
  int flag = 0;

  /* Copy service name into kernel space. */
  copy_from_user(kservice, service, len);
  kservice[len] = '\0';

  /* Check if service present. */
  down(&ecalls_sem);
  while (i < MAX_SERVICES) {
    if (services[i] != NULL)  
      if (strcmp(kservice, services[i]) == 0)
        break;
    i++;
  }

  if (i >= MAX_SERVICES) {
    printk("Service not available!\n");
    up(&ecalls_sem);
    return EXIT_FAILURE;
  }
 
  if ((upcall == NULL) && (downcall == NULL)) {
    j = 0;
    while (service_list[i].client_pid[j] != current->pid)
      j++;
    if (j < MAX_CLIENTS) {
      service_list[i].client_pid[j] = -1;
      service_list[i].tptr[j] = NULL;
      service_list[i].clients--;
      up(&ecalls_sem);
      return EXIT_SUCCESS;
    }
  }
 
  j = 0;
  while (service_list[i].client_pid[j] != -1 )
    j++;

  if (j >= MAX_CLIENTS) {
    up(&ecalls_sem);
    return EXIT_FAILURE;
  }

  applications_registered++;

  if ((downcall != NULL) && (downcall->flag != 0)) {
    current->suser_flag = 1;
    sys_mlock(downcall->flag, downcall->total_size);
    current->suser_flag = 0;
    service_list[i].memory_ptr_down[j] = downcall->data;
    service_list[i].memory_ptr_down_phys[j] = (unsigned long)
      generic_virt_to_phys((void *)service_list[i].memory_ptr_down[j], current);
    service_list[i].flag_down[j] = downcall->flag;
    service_list[i].flag_down_phys[j] = (unsigned long)
      generic_virt_to_phys((void *)service_list[i].flag_down[j], current);
    memcpy((void *)service_list[i].flag_down[j], &flag, sizeof(int));
    if (downcall->bitpattern != 0) {
      service_list[i].bitpattern_down[j] =  
	downcall->bitpattern;
      service_list[i].bitpattern_down_phys[j] = (unsigned long)
        generic_virt_to_phys((void*)service_list[i].bitpattern_down[j],current);
    }
    else {
      service_list[i].bitpattern_down[j] = -1;
      service_list[i].bitpattern_down_phys[j] = -1;
    }
    if (downcall->front != 0) {
      service_list[i].front_down[j] = downcall->front;
      service_list[i].front_down_phys[j] = (unsigned long)
        generic_virt_to_phys((void*)service_list[i].front_down[j],current);
    }
    else {
      service_list[i].front_down[j] = -1;
      service_list[i].front_down_phys[j] = -1;
    }
    if (downcall->back != 0) {
      service_list[i].back_down[j] = downcall->back;
      service_list[i].back_down_phys[j] = (unsigned long)
        generic_virt_to_phys((void*)service_list[i].back_down[j],current);
    }
    else {
      service_list[i].back_down[j] = -1;
      service_list[i].back_down_phys[j] = -1;
    }
  }

  if ((upcall != NULL) && (upcall->flag != 0)) {
    current->suser_flag = 1;
    sys_mlock(upcall->flag, upcall->total_size);
    current->suser_flag = 0;
    service_list[i].memory_ptr_up[j]= upcall->data;
    service_list[i].memory_ptr_up_phys[j] = (unsigned long)
      generic_virt_to_phys((void *)service_list[i].memory_ptr_up[j], current);
    service_list[i].flag_up[j] = upcall->flag;
    service_list[i].flag_up_phys[j] = (unsigned long)
      generic_virt_to_phys((void *)service_list[i].flag_up[j], current);
    memcpy((void *)service_list[i].flag_down[j], &flag, sizeof(int));
    if (upcall->bitpattern != 0) {
      service_list[i].bitpattern_up[j] = upcall->bitpattern;
      service_list[i].bitpattern_up_phys[j] = (unsigned long)
        generic_virt_to_phys((void*)service_list[i].bitpattern_up[j],current);
    }
    else {
      service_list[i].bitpattern_up[j] = -1;
      service_list[i].bitpattern_up_phys[j] = -1;
    }
    if (upcall->front != 0) {
      service_list[i].front_up[j] = upcall->front;
      service_list[i].front_up_phys[j] = (unsigned long)
        generic_virt_to_phys((void*)service_list[i].front_up[j],current);
    }
    else {
      service_list[i].front_up[j] = -1;
      service_list[i].front_up_phys[j] = -1;
    }
    if (upcall->back != 0) {
      service_list[i].back_up[j] = upcall->back;
      service_list[i].back_up_phys[j] = (unsigned long)
        generic_virt_to_phys((void*)service_list[i].back_up[j],current);
    }
    else {
      service_list[i].back_up[j] = -1;
      service_list[i].back_up_phys[j] = -1;
    }
  }
  if (upcall != NULL) {
    service_list[i].handler_id[j] = upcall->process_id;
    service_list[i].signal[j] = upcall->signal;
    service_list[i].kernel_function[j] = NULL;
    service_list[i].kfunc_id[j] = -1;
    if (upcall->kfct_string_len > 0) {
      int k = 0;
      copy_from_user(kfunction,upcall->kernel_function,upcall->kfct_string_len);
      kfunction[upcall->kfct_string_len] = '\0';
      while ((strcmp((char *)mod_func_list[k].identifier, kfunction) != 0)
	&& (k < MAX_FUNCS))
	k++;
      if (k == MAX_FUNCS) {
 	/* Must be ECL code. */
	service_list[i].context[j] = new_ecl_parse_context();
	ecl_add_param("in", "unsigned long", 0, service_list[i].context[j]);
	service_list[i].gen_code[j] = ecl_code_gen(kfunction, 
			service_list[i].context[j]);
	service_list[i].kernel_function[j] = 
		(unsigned long(*)(unsigned long))(unsigned long)
		service_list[i].gen_code[j]->func;
        service_list[i].kfunc_id[j] = MAX_FUNCS;
      }
      else {
        service_list[i].kernel_function[j] = mod_func_list[k].function_ptr;
        service_list[i].kfunc_id[j] = k;
      }
    }
    if (upcall->handler_function != NULL) {
      service_list[i].handler_function[j] = 
	      upcall->handler_function;
//        generic_virt_to_phys(upcall->handler_function, current);
    }
    else
      service_list[i].handler_function[j] = NULL;
  }

  tmp_pid.pid = current->pid;
  tmp_pid.i = i;
  tmp_pid.j = j;
  pidlist_size++;

  HInsert(pidlist, tmp_pid, pidlist_size);

  service_list[i].client_pid[j] = current->pid;
  service_list[i].tptr[j] = current;
  service_list[i].clients++;

  up(&ecalls_sem);
  return i;
}

/* Request upcall pointer outside process context. */
unsigned long request_upcall_pointer(int id, int pid)
{
  int index, i, j;
  unsigned long ret;

  if ((index = HExist(pidlist, pid, 0, pidlist_size)) == -1) 
    return -1;
  i = id;
  j = pidlist[index].j;

  ret = service_list[i].memory_ptr_up_phys[j];
  return ret;
}

/* Request upcall pointer in process context. */
unsigned long request_upcall_pointer2(int id, int pid)
{
  int index, i, j;
  unsigned long ret;

  if ((index = HExist(pidlist, pid, 0, pidlist_size)) == -1) 
    return -1;
  i = id;
  j = pidlist[index].j;

  ret = service_list[i].memory_ptr_up[j];
  return ret;
}

/* Request downcall pointer outside process context. */
unsigned long request_downcall_pointer(int id, int pid)
{
  int index, i, j;
  unsigned long ret;

  if ((index = HExist(pidlist, pid, 0, pidlist_size)) == -1) 
    return -1;
  i = id;
  j = pidlist[index].j;

  ret = service_list[i].memory_ptr_down_phys[j];
  return ret;
}

/* Request upcall pointer flag (in process context). */
unsigned long request_upcall_flag(int id, int pid)
{
  int index, i, j;
  unsigned long ret;

  if ((index = HExist(pidlist, pid, 0, pidlist_size)) == -1) 
    return -1;
  i = id;
  j = pidlist[index].j;

  ret = service_list[i].flag_up[j];
  return ret;
}

/* Request upcall pointer flag. */
unsigned long request_upcall_flag2(int id, int pid)
{
  int index, i, j;
  unsigned long ret;

  if ((index = HExist(pidlist, pid, 0, pidlist_size)) == -1) 
    return -1;
  i = id;
  j = pidlist[index].j;
  ret = service_list[i].flag_up_phys[j];
  return ret;
}

/* Request downcall pointer in process context. */
unsigned long request_downcall_pointer2(int id, int pid)
{
  int index, i, j;
  unsigned long ret;

  if ((index = HExist(pidlist, pid, 0, pidlist_size)) == -1) 
    return -1;
  i = id;
  j = pidlist[index].j;

  ret = service_list[i].memory_ptr_down[j];
  return ret;
}

/* Request downcall flag (in process context). */
unsigned long request_downcall_flag(int id, int pid)
{
  int index, i, j;
  unsigned long ret;

  if ((index = HExist(pidlist, pid, 0, pidlist_size)) == -1) 
    return -1;
  i = id;
  j = pidlist[index].j;

  ret = service_list[i].flag_down[j];
  return ret;
}

unsigned long request_downcall_flag2(int id, int pid)
{
  int index, i, j;
  unsigned long ret;

  if ((index = HExist(pidlist, pid, 0, pidlist_size)) == -1) 
    return -1;
  i = id;
  j = pidlist[index].j;

  ret = service_list[i].flag_down_phys[j];
  return ret;
}

/* Code download function. */
int download_code(char *hookname, char *parameter1, int parameter2, int parameter3,
	char *code) 
{
  int i = 0;
  char khookname[MAX_LEN];
  char kparameter[MAX_LEN];
  char *kcode;
  void (*callback)(char *, int, int, char *);
 
  copy_from_user(khookname, hookname, strlen(hookname));
  copy_from_user(kparameter, parameter1, strlen(parameter1));
  kcode = kmalloc(strlen(code), GFP_KERNEL);
  copy_from_user(kcode, code, strlen(code));

  while ((i < MAX_HOOKS) && (strcmp(hook_list[i].hookname, khookname) != 0)) 
    i++;

  if (i == MAX_HOOKS) return -1;

  callback = hook_list[i].callback;
  callback(kparameter, parameter2, parameter3, kcode); 

  kfree(kcode);
 
  return 0;
}

/* Do Downcall. */
int internal_do_downcall(struct ecall_down *downcall, void *data, int len) 
{
  void *kdata = NULL;
  //asmlinkage int (*service_function)(struct ecall_down *downcall,
//		void *data, int len);
  asmlinkage int (*service_function)(int id, unsigned long data, int len);

  if (service_list[downcall->service_id].slow_handler_func == NULL)
    return EXIT_FAILURE;

  service_function = service_list[downcall->service_id].slow_handler_func;

  if ((len > 0) && (data != NULL)) {
    kdata = kmalloc(len, GFP_KERNEL);
    copy_from_user(kdata, data, len);
  }

  service_function(downcall->service_id, (unsigned long)kdata, len);

  if (kdata != NULL)  
    kfree(kdata);

  return EXIT_SUCCESS;
}

/* Do Upcall. */
int do_upcall(int id, int pid, unsigned long pointer, long priority1, long priority2, int cpu) 
{
  int index;
  int i, j;
  unsigned long return_value = 0;
  int sink_pid = pid; 
  struct task_struct *task_ptr = current;
  struct upcall_task_struct *upcall_ptr = head;
  struct upcall_task_struct *hptr = NULL;
  struct upcall_task_struct *hptr2 = NULL;

  i = id;
  j = 0;

  if (pid != -1) {
    if ((index = HExist(pidlist, pid, 0, pidlist_size)) == -1) 
      return -1;

    j = pidlist[index].j;
    if (service_list[i].tptr[j] == NULL) return 0;
  } 

  while ((j >= 0) && (j < MAX_CLIENTS)) {
    if (pid == -1)
      sink_pid = service_list[i].client_pid[j];

    if (find_task_by_pid(sink_pid) == NULL) {
	    if (pid == -1) {
	      j++;
	      break;
	    } else return 0;
    }

    if ((service_list[i].tptr[j] == NULL) || (service_list[i].tptr[j]->state ==
			    TASK_ZOMBIE) || (service_list[i].tptr[j]->state ==
				    TASK_STOPPED)) {
	    if (pid == -1) {
	      j++;
	      break;
	    } else return 0;
    }

    /* Handler function in user space. */
    if (service_list[i].handler_function[j] != NULL) {
      int (*handler_func)(unsigned long);
      struct task_struct *hptr;
      mm_segment_t fs;

      hptr = service_list[i].tptr[j];

      handler_func = service_list[i].handler_function[j];
//      ECalls_current = hptr;

      fs = get_fs();
      set_fs(hptr->addr_limit);
//      ECalls_current_flag = 1;
      if (handler_func != NULL)
//printk("address=%ld\n", (unsigned long) handler_func);
        return_value = handler_func(pointer);
//      ECalls_current_flag = 0;
      set_fs(fs);
    }

  /* Handler function in kernel space. */
  if (service_list[i].kernel_function[j] != NULL) {
    asmlinkage int (*kernel_fct)(unsigned long);
    unsigned long (*func) (unsigned long);
    if (service_list[i].kfunc_id[j] == MAX_FUNCS) {
	func = (unsigned long(*)()) (unsigned long)
		service_list[i].kernel_function[j];
	if (func != NULL)
          return_value = func(pointer);
    }
    else if ((mod_func_list[service_list[i].kfunc_id[j]].flag != -1) &&
      	(mod_func_list[service_list[i].kfunc_id[j]].identifier != NULL)) { 
      kernel_fct = service_list[i].kernel_function[j];
      return_value = kernel_fct(pointer);
    }
  }

  /* Dynamic handler function in kernel space. */
  if (service_list[i].upcall_func[j] != NULL) {
    long (*func)();
    func = (long(*)()) (long) service_list[i].upcall_func[j];
    return_value = func();
  }

  /* Raise signal if signal attribute is set. */
  if ((return_value == 0) && (service_list[i].signal[j] >= SIGRTMIN) && 
	(service_list[i].signal[j] <= SIGRTMAX)) {
//printk("upcall signal:%d\n", service_list[i].signal[j]);
    while ((task_ptr != NULL) && (task_ptr->pid != sink_pid))
      task_ptr = task_ptr->next_task;
    if (task_ptr == NULL) {
      task_ptr = current;
      while ((task_ptr != NULL) && (task_ptr->pid != sink_pid))
        task_ptr = task_ptr->prev_task;
    }
    if (task_ptr != NULL) {
      send_sig(service_list[i].signal[j], task_ptr, 1);
    }
    return_value = 1;
  } 

  if (task_ptr == NULL)
    return EXIT_FAILURE;

  /* Put upcall-receiving process into upcall scheduler queue. */
  if ((cpu == 1) && (service_list[i].handler_id[j] > 0)) {
    if (head == NULL) {
      queuelength++;
      head = kmalloc(GFP_KERNEL, sizeof(struct upcall_task_struct));
      head->tptr = task_ptr;
      head->seconds = priority1;
      head->microseconds = priority2;
      head->next = NULL; 
    } else {
      while (upcall_ptr->next != NULL) {
        if (upcall_ptr->tptr == task_ptr) 
	  break;
        upcall_ptr = upcall_ptr->next;
      }
      if (upcall_ptr->tptr != task_ptr) {
        upcall_ptr = head;
	hptr = head;
        while (upcall_ptr != NULL) {
          if (priority1 < upcall_ptr->seconds) {
	    if (upcall_ptr == head) {
              queuelength++;
              hptr2 = kmalloc(GFP_KERNEL,sizeof(struct upcall_task_struct));
              hptr2->next = head;
              head = hptr2;
              hptr2->tptr = task_ptr;
              hptr2->seconds = priority1;
              hptr2->microseconds = priority2;
	    } else {
              queuelength++;
              hptr2 = kmalloc(GFP_KERNEL,sizeof(struct upcall_task_struct));
              hptr2->next = upcall_ptr;
              hptr->next = hptr2;
              hptr2->tptr = task_ptr;
              hptr2->seconds = priority1;
              hptr2->microseconds = priority2;
	    }
            return EXIT_SUCCESS;
          } else if ((priority1 == upcall_ptr->seconds) &&
            (priority2 < upcall_ptr->microseconds)) {
	    if (upcall_ptr == head) {
              queuelength++;
              hptr2 = kmalloc(GFP_KERNEL,sizeof(struct upcall_task_struct));
              hptr2->next = head;
              head = hptr2;
              hptr2->tptr = task_ptr;
              hptr2->seconds = priority1;
              hptr2->microseconds = priority2;
	    } else {
              queuelength++;
              hptr2 = kmalloc(GFP_KERNEL,sizeof(struct upcall_task_struct));
              hptr2->next = upcall_ptr;
              hptr->next = hptr2;
              hptr2->tptr = task_ptr;
              hptr2->seconds = priority1;
              hptr2->microseconds = priority2;
	    }
            return EXIT_SUCCESS;
          }
	  hptr = upcall_ptr;
          upcall_ptr = upcall_ptr->next;
        }
        queuelength++;
        hptr2 = kmalloc(GFP_KERNEL,sizeof(struct upcall_task_struct));
        hptr->next = hptr2;
        hptr2->next = NULL;
        hptr2->tptr = task_ptr;
        hptr2->seconds = priority1;
        hptr2->microseconds = priority2;
      }
    }
  }
    if (pid != -1) return EXIT_SUCCESS;
    j++;
    if (service_list[i].tptr[j] == NULL) {
	    return 0;
    } else {
    }
  }
  return EXIT_SUCCESS;
}

/* Placeholder task for thread pool. */
void placeholder_task(void *data) 
{
  int identifier;
/*
  int i;
*/

  asmlinkage int (*kernel_fct)(void);

/*
  sprintf(current->comm, "ECalls_server_thread");
  lock_kernel();
  exit_mm(current);
  for (i = 255; i >= 0; i--)
    if (current->files->fd[i])
      close(i);
*/

  memcpy(&identifier, data, sizeof(int));
  thr_pool[1][identifier] = 0;

/*
  DECLARE_WAITQUEUE(w_queue0, current);
  DECLARE_WAITQUEUE(w_queue1, current);
  DECLARE_WAITQUEUE(w_queue2, current);
*/

  while (1) {
/*
    sleep_on(&w_queue[identifier]);
*/
/*
    if (identifier == 0)
      add_wait_queue(&wait_head0, &w_queue[identifier]);
    else if (identifier == 1)
      add_wait_queue(&wait_head1, &w_queue[identifier]);
    else 
      add_wait_queue(&wait_head2, &w_queue[identifier]);
    current->state = TASK_INTERRUPTIBLE;
    schedule();
    current->state = TASK_RUNNING;
    if (identifier == 0)
      remove_wait_queue(&wait_head0, &w_queue[identifier]);
    else if (identifier == 1)
      remove_wait_queue(&wait_head1, &w_queue[identifier]);
    else
      remove_wait_queue(&wait_head2, &w_queue[identifier]);
*/
    kernel_fct = handler_list[identifier];
    kernel_fct();
    thr_pool[1][identifier] = 0;
  }
}

/* Compute goodness of upcall-receiving processes. */
int upcall_goodness(struct task_struct *tptr)
{
  int weight = 0;

  if (tptr->policy != SCHED_OTHER) {
    weight = 1000 + tptr->rt_priority;
    weight = weight * 2 + 1;
    return weight;
  } 
  weight += tptr->nice + 1000; 

  return weight;
}

/* Update placeholder structure. */
#ifdef CONFIG_DWCS
int update_placeholder(int flag, int utilization) 
{
  struct upcall_task_struct *upcall_ptr = head;
  int task_util = MAX_UTILIZATION - utilization;
  unsigned long period = 0;

  if (flag == 0) { 
    if (place_holder.service_time == 0) {
      if (upcall_ptr != NULL) {
        struct task_struct *tptr = upcall_ptr->tptr;
        place_holder.service_time = tptr->dwcs_params->service_time; 
      } else
        place_holder.service_time = 1; 
      place_holder.period = MAX_UTILIZATION * place_holder.service_time; 
      if (place_holder.period/task_util*task_util == place_holder.period)
	place_holder.period = place_holder.period / task_util;
      else
	place_holder.period = place_holder.period / task_util + 1;
      place_holder.deadline = jiffies + place_holder.period; 
      place_holder.ready_time = jiffies;
    } else {
      if (upcall_ptr != NULL) {
        struct task_struct *tptr = upcall_ptr->tptr;
        if (place_holder.service_time != tptr->dwcs_params->service_time) 
          place_holder.service_time = tptr->dwcs_params->service_time; 
        place_holder.period = MAX_UTILIZATION * place_holder.service_time; 
        if (place_holder.period/task_util*task_util == place_holder.period)
	  place_holder.period = place_holder.period / task_util;
        else
	  place_holder.period = place_holder.period / task_util + 1;
        if (place_holder.period != period) {
          if (period > place_holder.period) {
            place_holder.deadline += (period - place_holder.period); 
            place_holder.ready_time += (period - place_holder.period); 
  	  } else {
            place_holder.deadline -= (place_holder.period - period); 
            place_holder.ready_time -= (period - place_holder.period); 
	  }
          place_holder.period = period; 
        }
      }
    }
  } else if (flag == 1) {
    place_holder.deadline += place_holder.period; 
    place_holder.ready_time += place_holder.period;
  } else if (flag == 2) {
    if (place_holder.deadline < jiffies) 
      place_holder.deadline += place_holder.period; 
      place_holder.ready_time += place_holder.period;
  }

  return EXIT_SUCCESS;
}
#endif

/* Upcall scheduler. */
int upcall_scheduler(struct task_struct **old_next,
		     struct task_struct **new_next,
		     int *weight,
		     int utilization)
{
  struct upcall_task_struct *upcall_ptr = head;
  struct upcall_task_struct *hptr1 = head;
  struct task_struct *tptr;
#ifdef CONFIG_DWCS
  struct task_struct *task_ptr;
#endif
  int c, w;
  unsigned long long t1, t2;

printk("in upcall scheduler\n");
myrdtsc(&t1);
  memcpy(&w, weight, sizeof(int));
  new_next = old_next;
  tptr = *old_next;

  /* DWCS Real-Time Scheduler. */
#ifdef CONFIG_DWCS
  if (DWCS_module_loaded) {
    update_placeholder(0, utilization);
    if (upcall_ptr == NULL)
      return EXIT_SUCCESS;
    if ((tptr->dwcs_params->deadline == (unsigned long)-1) ||
        (tptr->dwcs_params->flags & DWCS_SERVICED_THIS_PERIOD)) {
      new_next = &(upcall_ptr->tptr);
      head = upcall_ptr->next;
      kfree(upcall_ptr);
      c = (upcall_goodness(head->tptr)-1)/2;
      memcpy(weight, &c, sizeof(int));
    } else if ((place_holder.deadline <= tptr->dwcs_params->deadline) &&
	(place_holder.ready_time <= jiffies)) {
      new_next = &(upcall_ptr->tptr);
      head = upcall_ptr->next;
      kfree(upcall_ptr);
      c = (upcall_goodness(head->tptr)-1)/2;
      memcpy(weight, &c, sizeof(int));
      update_placeholder(1, utilization);
    } else if ((tptr->dwcs_params->deadline > jiffies) &&
	(tptr->dwcs_params->cwn > 0)) {
      upcall_ptr = head;
      hptr1 = head;
      while (upcall_ptr != NULL) {
        task_ptr = upcall_ptr->tptr;
	if ((task_ptr->dwcs_params->deadline == tptr->dwcs_params->deadline)
	  && (task_ptr->dwcs_params->ready_time <= jiffies)) {
          new_next = &(upcall_ptr->tptr);
          if (upcall_ptr == head) {
	    head = head->next;
            kfree(upcall_ptr);
	  } else {
            hptr1->next = upcall_ptr->next;
            kfree(upcall_ptr);
	  }
        }
        hptr1 = upcall_ptr;
	upcall_ptr = upcall_ptr->next;
      }
    } else {
      upcall_ptr = head;
      hptr1 = head;
      while (upcall_ptr != NULL) {
	if (upcall_ptr->tptr == tptr) {
          if (upcall_ptr == head) {
	    head = head->next;
            kfree(upcall_ptr);
	  } else {
            hptr1->next = upcall_ptr->next;
            kfree(upcall_ptr);
	  }
        }
        hptr1 = upcall_ptr;
	upcall_ptr = upcall_ptr->next;
      }
      update_placeholder(2, utilization);
    }
    return EXIT_SUCCESS;
  } 
#endif

  /* Standard Unix Scheduler. */
  if (upcall_ptr != NULL) {
    c = upcall_goodness(upcall_ptr->tptr);
    if (c > (2 * w)) {
      new_next = &(upcall_ptr->tptr);
      upcall_ptr = upcall_ptr->next;
      w = (c-1)/2;
      memcpy(weight, &w, sizeof(int));
      hptr1 = head->next;
      kfree(head);
      head = hptr1;
      queuelength--;
    } 
  }
myrdtsc(&t2);

  return EXIT_SUCCESS;
}

void upcall_callback(char *parameter1, int parameter2, int parameter3, char *code)
{
  int i = 0;
  int j = 0;

  while (i < MAX_SERVICES) {
    if (services[i] != NULL)  
      if (strcmp(parameter1, services[i]) == 0)
        break;
    i++;
  }

  if (i < MAX_SERVICES) {
    while (j < MAX_CLIENTS) {
      if (service_list[i].client_pid[j] == parameter2) 
	break;
    }
    if (j < MAX_CLIENTS) {
      service_list[i].context[j] = new_ecl_parse_context();
      service_list[i].gen_code[j] = ecl_code_gen(code, service_list[i].context[j]);
      service_list[i].upcall_func[j] = service_list[i].gen_code[j]->func;
    }
  }
}

/* Initialize module. */
int init_module(void) 
{
  int i;
  struct proc_dir_entry *t;
  struct proc_dir_entry *t2;

  t = create_proc_read_entry("kernel-hooks", S_IFREG | S_IRUGO, NULL, 
	hooks_read, NULL);
  t->owner = THIS_MODULE;
  t2 = create_proc_read_entry("kservices", S_IFREG | S_IRUGO, NULL, 
	kservices_read, NULL);
  t2->owner = THIS_MODULE;

  ECalls_active = 0;
  //upcall_sched_active = 1;
  place_holder.service_time = 0;

  /* Save original system call entries. */
  original_start_ecalls       =    sys_call_table[__NR_start_ecalls];
  original_end_ecalls         =    sys_call_table[__NR_end_ecalls];
  original_register_ecall     =    sys_call_table[__NR_register_ecall];
  original_do_downcall        =    sys_call_table[__NR_do_downcall];
  original_run_event_handler  =    sys_call_table[__NR_run_event_handler];
  original_upcall_scheduler   =    sys_call_table[__NR_upcall_scheduler];
  original_download_code      =    sys_call_table[__NR_download_code];
//  original_fast_do_downcall   =    fast_sys_call_table[__NR_fast_do_downcall];

  /* Redirect system calls. */
  sys_call_table[__NR_start_ecalls]      =   internal_start_ecalls;
  sys_call_table[__NR_end_ecalls]        =   internal_end_ecalls;
  sys_call_table[__NR_register_ecall]    =   internal_register_ecall;
  sys_call_table[__NR_do_downcall]       =   internal_do_downcall;
  sys_call_table[__NR_run_event_handler] =   run_event_handler;
  sys_call_table[__NR_upcall_scheduler]  =   upcall_scheduler;
  sys_call_table[__NR_download_code]     =   download_code;
//  fast_sys_call_table[__NR_fast_do_downcall]  =  handle_event;

  sys_mlock   =    sys_call_table[__NR_mlock];

  /* Create pool of kernel-threads. */
/* FIXME! */
/*
  pool_size = THR_POOL;
  for (i = 0; i < pool_size; i++) {
    thr_pool[1][i] = i;
    thr_pool[0][i] = kernel_thread((int (*)(void *)) placeholder_task, 
	&thr_pool[1][i], 0);
  }
*/
  
  for (i = 0; i < MAX_FUNCS; i++) 
    mod_func_list[i].flag = -1;

  internal_start_ecalls(0);
  ECalls_current = current;

  register_hook("upcall", "Upcall function", "service name, pid",(void *)upcall_callback);

  return EXIT_SUCCESS;
}

/* Remove module. */
void cleanup_module(void) 
{
  int i = 0;

  ECalls_active = 0;
  
  upcall_sched_active = 0;

  unregister_hook("upcall");

  for (i = 0; i < MAX_SERVICES; i++) 
    hook_list[i].callback = NULL;

  /* Restore original system call entries. */
  sys_call_table[__NR_start_ecalls]      =   original_start_ecalls;
  sys_call_table[__NR_end_ecalls]        =   original_end_ecalls;
  sys_call_table[__NR_register_ecall]    =   original_register_ecall;
  sys_call_table[__NR_do_downcall]       =   original_do_downcall;
  sys_call_table[__NR_run_event_handler] =   original_run_event_handler;
  sys_call_table[__NR_upcall_scheduler]  =   original_upcall_scheduler;
  sys_call_table[__NR_download_code]     =   original_download_code;
 // fast_sys_call_table[__NR_fast_do_downcall]  =  original_fast_do_downcall;

  /* Terminate threads. */
/*
  for (i = 0; i < pool_size; i++) {
    kill_proc(thr_pool[0][i], SIGKILL, 1);
  }
*/

  remove_proc_entry("kernel-hooks", NULL);
  remove_proc_entry("kservices", NULL);
}

