/*********************************************/
/* Project:      Ecall mechanism             */
/* File:         heap.c                      */
/* Date:         9/16/1999                   */ 
/* Modified:                                 */
/* Author:       cp (chris@cc)               */
/* Contents:     heap functions              */

/* Required definitions. */
#ifndef MODULE
#define MODULE
#endif
#ifndef __KERNEL__
#define __KERNEL__
#endif

#include <linux/unistd.h>
#include <linux/errno.h>
#include <linux/types.h>
#include "ecalls_module.h"

extern struct semaphore heap_sem;

/* Floor function. */
double floor(double x) {

  int a = (int) x;

  return (double) a;
}

/* Return parent element of i in heap array. */
int HParent(int i) {

  return (floor(i/2));
}

/* Return left child element. */
int HLeft(int i) {
 
  return (2*i);
}

/* Return right child element. */
int HRight(int i) {
  
  return (2*i+1);
}

/* Compare two elements in the heap structure. */
int HCompare(struct pid_list entry1, struct pid_list entry2) {

  if (entry2.pid < entry1.pid) 
    return TRUE;
  else
    return FALSE;
}

/* Reorder heap. */
void HHeapify(struct pid_list pidlist[], int i, int size) {

  int l, r;
  int smallest;
  struct pid_list tmp_heap;

  l = HLeft(i);
  r = HRight(i);
 
  if ((l <= size) && (HCompare(pidlist[i-1], pidlist[l-1]) == TRUE))
    smallest = l;
  else
    smallest = i;
  
  if ((r <= size) && (HCompare(pidlist[smallest-1], pidlist[r-1]) == TRUE))
    smallest = r;

  if (smallest != i) {
    tmp_heap = pidlist[i-1];
    pidlist[i-1] = pidlist[smallest-1];
    pidlist[smallest-1] = tmp_heap;     
    HHeapify(pidlist, smallest, size);
  }
}

/* Look for element in heap. */
int HExist(struct pid_list pidlist[], int pid, int start, int size) {

  int i = start;

  while ((i<(size))&&
	(pidlist[i].pid != pid))
    i++;

  if (i == size) return -1;
  else return i;
}

/* Insert element into heap. */
void HInsert(struct pid_list pidlist[],struct pid_list tmp_heap, int size) {
  
  int i = size;

  down(&heap_sem);
  while (i > 1) {

    if (HCompare(pidlist[HParent(i)-1], tmp_heap)) {
          pidlist[i-1] = pidlist[HParent(i)-1];
          i = HParent(i);
    }
    else
      break;
  }
  
  pidlist[i-1] = tmp_heap;
  up(&heap_sem);
}

/* Remove first element and reorder heap. */
void HExtractMin(struct pid_list pidlist[], int size) {

  down(&heap_sem);
  pidlist[0] = pidlist[size];
  pidlist[size].pid = -1;   
  pidlist[size].i = -1; 
  pidlist[size].j = -1; 
  HHeapify(pidlist, 1, size);  
  up(&heap_sem);
}

/* Remove element from heap. */
void HDelete(struct pid_list pidlist[], int pid, int size) {

  int i=1;

  down(&heap_sem);
  while ((i < size+1) && (pidlist[i-1].pid != pid))
    i++;
  if (pidlist[i-1].pid == pid) { 
    pidlist[i-1] = pidlist[size];   
    pidlist[size].pid = -1;   
    pidlist[size].i = -1; 
    pidlist[size].j = -1; 
    HHeapify(pidlist, i, size); 
  } 
  up(&heap_sem);
}

