%{
#include "config.h"
#ifdef LINUX_KERNEL_MODULE
#ifndef MODULE
#define MODULE
#endif
#ifndef __KERNEL__
#define __KERNEL__
#endif
#include <linux/kernel.h>
#include <linux/module.h>
#endif
#include "ecl.h"
#include "ecl_internal.h"
#include "assert.h"
#ifndef LINUX_KERNEL_MODULE
#include <ctype.h>
#include <malloc.h>
#include <string.h>
#else
#include <linux/ctype.h>
#include <linux/string.h>
#include <linux/mm.h>
#endif
#include "structs.h"
#ifdef HAVE_DRISC_H
#include "drisc.h"
#else
#ifdef HAVE_ICODE_H
#include "drisc_icode.h"
#include "icode.h"
#else
enum {
    DR_C,    /* char */
    DR_UC,   /* unsigned char */
    DR_S,    /* short */
    DR_US,   /* unsigned short */
    DR_I,    /* int */
    DR_U,    /* unsigned */
    DR_L,    /* long */
    DR_UL,   /* unsigned long */
    DR_P,    /* pointer */
    DR_F,    /* floating */
    DR_D,    /* double */
    DR_V,    /* void */
    DR_B,    /* block structure */
    DR_ERR   /* no type */
};
#endif
#endif
#ifndef LINUX_KERNEL_MODULE
#ifdef STDC_HEADERS
#include <stdarg.h>
#else
#include <varargs.h>
#endif
#else
#include "kecl.h"
#define malloc (void *)DAllocMM
#define free(a) DFreeMM((addrs_t)a)
#define realloc(a,b) DReallocMM((addrs_t)a,b)
#define fprintf(fmt, args...) printk(args);
#define printf printk
char *strdup(const char *s)
{
	char *p;

	p = (char *)kmalloc(strlen(s)+1, GFP_KERNEL);
	if (p != NULL)
		strcpy(p,s);
	return p;
}
#endif
static char *create_string_from_yytext();
extern int yyparse();
static sm_ref yyparse_value;
static int yyerror_count = 1;
extern void yyerror(char *str);
static int parsing_type = 0;
static ecl_parse_context yycontext;
%}

%union {
    lx_info info;
    int number;
    sm_ref reference;
    sm_list list;
    char *string;
};

%token <info> ARROW
%token <info> LPAREN
%token <info> RPAREN
%token <info> LCURLY
%token <info> RCURLY
%token <info> LBRACKET
%token <info> RBRACKET
%token <info> DOT
%token <info> STAR
%token <info> SLASH
%token <info> MODULUS
%token <info> PLUS
%token <info> MINUS
%token <info> LEQ
%token <info> LT
%token <info> GEQ
%token <info> GT
%token <info> EQ
%token <info> NEQ
%token <info> LEFT_SHIFT
%token <info> RIGHT_SHIFT
%token <info> ASSIGN
%token <info> LOG_OR
%token <info> LOG_AND
%token <info> ARITH_OR
%token <info> ARITH_AND
%token <info> ARITH_XOR
%token <info> BANG
%token <info> SEMI
%token <info> IF
%token <info> ELSE
%token <info> FOR
%token <info> CHAR
%token <info> SHORT
%token <info> INT
%token <info> LONG
%token <info> UNSIGNED
%token <info> SIGNED
%token <info> FLOAT
%token <info> DOUBLE
%token <info> VOID
%token <info> STRING
%token <info> STATIC
%token <info> RETURN_TOKEN
%token <info> PRINT
%token <info> COMMA
%token <info> DOTDOTDOT
%token <info> integer_constant
%token <info> string_constant
%token <info> floating_constant
%token <info> identifier_ref
%token <info> type_id

%type <reference> compound_statement;
%type <list> declaration_list statement_list;
%type <reference> declaration;
%type <reference> statement;
%type <reference> init_declarator;
%type <number> static_opt;
%type <reference> declarator;
%type <list> type_specifier_list;
%type <list> parameter_list;
%type <list> parameter_type_list;
%type <list> argument_expression_list;
%type <reference> parameter_declaration;
%type <reference> type_specifier;
%type <reference> initializer;
%type <reference> assignment_expression;
%type <reference> expression_statement;
%type <reference> selection_statement;
%type <reference> iteration_statement;
%type <reference> jump_statement;
%type <reference> conditional_expression;
%type <reference> unary_expression;
%type <reference> cast_expression;
%type <reference> postfix_expression;
%type <reference> primary_expression;
%type <reference> expression;
%type <reference> expression_opt;
%type <reference> constant;
%type <reference> multiplicative_expression;
%type <reference> additive_expression;
%type <reference> shift_expression;
%type <reference> relational_expression;
%type <reference> equality_expression;
%type <reference> logical_or_expression;
%type <reference> logical_and_expression;
%type <reference> inclusive_or_expression;
%type <reference> exclusive_or_expression;
%type <reference> and_expression;
%type <reference> start
%%

start :
	declaration_list {
	    yyparse_value = (sm_ref)$1;
	}
	|
	compound_statement {
	    yyparse_value = $1;
	}
	;

compound_statement:
	LCURLY declaration_list statement_list RCURLY {
	    $$ = ecl_new_compound_statement();
	    $$->node.compound_statement.decls = $2;
	    $$->node.compound_statement.statements = $3;
	};

declaration_list:
	/* empty */ { $$ = NULL; }
	|
	declaration_list declaration {
	    if ($2 == NULL) {
		$$ = $1;
	    } else {
		if ($1 == NULL) {
		    $$ = malloc(sizeof(struct list_struct));
		    $$->node = $2;
		    $$->next = NULL;
		} else {
		    sm_list tmp = $1;
		    while (tmp->next != NULL) {
			tmp = tmp->next;
		    }
		    tmp->next = malloc(sizeof(struct list_struct));
		    tmp->next->node = $2;
		    tmp->next->next = NULL;
		    $$ = $1;
		}
	    }
	};

declaration:
	static_opt type_specifier_list 
	     { 
		 if (parsing_type) {
		     yyparse_value = (sm_ref) $2;
		     YYACCEPT;
		 }
	     }
	init_declarator SEMI
	    {
		$$ = $4;
		$$->node.declaration.static_var = $1;
		$$->node.declaration.type_spec = $2;
	    }
        | error SEMI {
	    $$ = NULL;
	}
	;

static_opt:
	/* null */ {$$ = 0;}
	| STATIC { $$ = 1;}
	;

init_declarator:
	declarator
	|
	declarator ASSIGN initializer
	    {
		assert($1->node_type == ecl_declaration);
		$1->node.declaration.init_value = $3;
	    }
	;

declarator:
	identifier_ref	    {
		$$ = ecl_new_declaration();
		$$->node.declaration.param_num = -1;
		$$->node.declaration.id = $1.string;
		$$->node.declaration.init_value = NULL;
		$$->node.declaration.lx_srcpos = $1.lx_srcpos;
		$$->node.declaration.is_subroutine = 0;
		$$->node.declaration.params = NULL;
	    }
	| identifier_ref LPAREN parameter_type_list RPAREN {
		$$ = ecl_new_declaration();
		$$->node.declaration.param_num = -1;
		$$->node.declaration.id = $1.string;
		$$->node.declaration.init_value = NULL;
		$$->node.declaration.lx_srcpos = $1.lx_srcpos;
		$$->node.declaration.is_subroutine = 1;
		$$->node.declaration.params = $3;
	}
	| identifier_ref LPAREN RPAREN {
		$$ = ecl_new_declaration();
		$$->node.declaration.param_num = -1;
		$$->node.declaration.id = $1.string;
		$$->node.declaration.init_value = NULL;
		$$->node.declaration.lx_srcpos = $1.lx_srcpos;
		$$->node.declaration.is_subroutine = 1;
		$$->node.declaration.params = NULL;
	}
	;

parameter_type_list:
	parameter_list |
	parameter_list COMMA DOTDOTDOT {
	    sm_list tmp = $1;
	    sm_ref id = ecl_new_declaration();
	    while (tmp->next != NULL) {
		tmp = tmp->next;
	    }
	    tmp->next = malloc(sizeof(struct list_struct));
	    tmp->next->node = id;
	    tmp->next->next = NULL;
	    id->node.declaration.id = strdup("...");
	    $$ = $1;
	}
	;

parameter_list:
	parameter_declaration {
		$$ = malloc(sizeof(struct list_struct));
		$$->node = $1;
		$$->next = NULL;
	}
	|
	parameter_list COMMA parameter_declaration {
	    sm_list tmp = $1;
	    while (tmp->next != NULL) {
		tmp = tmp->next;
	    }
	    tmp->next = malloc(sizeof(struct list_struct));
	    tmp->next->node = $3;
	    tmp->next->next = NULL;
	    $$ = $1;
	};
	
parameter_declaration:
	type_specifier_list declarator {
		$$ = $2;
		$$->node.declaration.static_var = 0;
		$$->node.declaration.type_spec = $1;
	};

initializer:
	assignment_expression { $$ = $1;}
	;

type_specifier_list:
	type_specifier {
	    $$ = malloc(sizeof(struct list_struct));
	    $$->node = $1;
	    $$->next = NULL;
	}
	| type_specifier_list type_specifier {
	    sm_list tmp = $1;
	    while (tmp->next != NULL) {
		tmp = tmp->next;
	    }
	    tmp->next = malloc(sizeof(struct list_struct));
	    tmp->next->node = $2;
	    tmp->next->next = NULL;
	    $$ = $1;
	};

type_specifier:
	CHAR {
	    $$ = ecl_new_type_specifier();
	    $$->node.type_specifier.lx_srcpos = $1.lx_srcpos;
	    $$->node.type_specifier.token = CHAR;
	}
	| SHORT {
	    $$ = ecl_new_type_specifier();
	    $$->node.type_specifier.lx_srcpos = $1.lx_srcpos;
	    $$->node.type_specifier.token = SHORT;
	}
	| INT {
	    $$ = ecl_new_type_specifier();
	    $$->node.type_specifier.lx_srcpos = $1.lx_srcpos;
	    $$->node.type_specifier.token = INT;
	}
	| LONG {
	    $$ = ecl_new_type_specifier();
	    $$->node.type_specifier.lx_srcpos = $1.lx_srcpos;
	    $$->node.type_specifier.token = LONG;
	}
	| FLOAT {
	    $$ = ecl_new_type_specifier();
	    $$->node.type_specifier.lx_srcpos = $1.lx_srcpos;
	    $$->node.type_specifier.token = FLOAT;
	}
	| DOUBLE {
	    $$ = ecl_new_type_specifier();
	    $$->node.type_specifier.lx_srcpos = $1.lx_srcpos;
	    $$->node.type_specifier.token = DOUBLE;
	}
	| VOID {
	    $$ = ecl_new_type_specifier();
	    $$->node.type_specifier.lx_srcpos = $1.lx_srcpos;
	    $$->node.type_specifier.token = VOID;
	}
	| SIGNED {
	    $$ = ecl_new_type_specifier();
	    $$->node.type_specifier.lx_srcpos = $1.lx_srcpos;
	    $$->node.type_specifier.token = SIGNED;
	}
	| UNSIGNED {
	    $$ = ecl_new_type_specifier();
	    $$->node.type_specifier.lx_srcpos = $1.lx_srcpos;
	    $$->node.type_specifier.token = UNSIGNED;
	}
	| STRING {
	    $$ = ecl_new_type_specifier();
	    $$->node.type_specifier.lx_srcpos = $1.lx_srcpos;
	    $$->node.type_specifier.token = STRING;
	}
	| type_id {
	    $$ = ecl_new_identifier();
	    $$->node.identifier.lx_srcpos = $1.lx_srcpos;
	    $$->node.identifier.id = $1.string;
	}
	;

statement_list:
	/* NULL */ {$$ = NULL;}
	| statement_list statement {
	    if ($2 == NULL) {
		$$ = $1;
	    } else {
		if ($1 == NULL) {
		    $$ = malloc(sizeof(struct list_struct));
		    $$->node = $2;
		    $$->next = NULL;
		} else {
		    sm_list tmp = $1;
		    while (tmp->next != NULL) {
			tmp = tmp->next;
		    }
		    tmp->next = malloc(sizeof(struct list_struct));
		    tmp->next->node = $2;
		    tmp->next->next = NULL;
		    $$ = $1;
		}
	    }
	};

/* missing labeled statement */
statement:
	  compound_statement
	| expression_statement
	| selection_statement
	| iteration_statement
	| jump_statement
	| error SEMI {
	      $$ = NULL;
	  }
	;

/* Missing goto, continue, break */
jump_statement:
	RETURN_TOKEN expression_opt SEMI {
	    $$ = ecl_new_return_statement();
	    $$->node.return_statement.expression = $2;
	    $$->node.return_statement.lx_srcpos = $1.lx_srcpos;
	}
	;

expression_statement:
	expression SEMI	{ 
	    $$ = ecl_new_expression_statement();
	    $$->node.expression_statement.expression = $1;
	}
	;

/* missing switch */
selection_statement:
	IF LPAREN expression RPAREN statement
	{ 
	    $$ = ecl_new_selection_statement();
	    $$->node.selection_statement.lx_srcpos = $1.lx_srcpos;
	    $$->node.selection_statement.conditional = $3;
	    $$->node.selection_statement.then_part = $5;
	    $$->node.selection_statement.else_part = NULL;
	}
	| 
	IF LPAREN expression RPAREN statement ELSE statement
	{ 
	    $$ = ecl_new_selection_statement();
	    $$->node.selection_statement.lx_srcpos = $1.lx_srcpos;
	    $$->node.selection_statement.conditional = $3;
	    $$->node.selection_statement.then_part = $5;
	    $$->node.selection_statement.else_part = $7;
	}
	;

/* missing while, do while */
iteration_statement:
	FOR LPAREN expression_opt SEMI expression_opt SEMI expression_opt RPAREN statement
	{ 
	    $$ = ecl_new_iteration_statement();
	    $$->node.selection_statement.lx_srcpos = $1.lx_srcpos;
	    $$->node.iteration_statement.init_expr = $3;
	    $$->node.iteration_statement.test_expr = $5;
	    $$->node.iteration_statement.iter_expr = $7;
	    $$->node.iteration_statement.statement = $9;
	} 
	;

expression_opt:
	{ $$ = NULL; }
	/* null */
	| expression;

constant :
	integer_constant {
	    $$ = ecl_new_constant();
	    $$->node.constant.token = integer_constant;
	    $$->node.constant.const_val = $1.string;
	    $$->node.constant.lx_srcpos = $1.lx_srcpos;
	}
	|
	floating_constant {
	    $$ = ecl_new_constant();
	    $$->node.constant.token = floating_constant;
	    $$->node.constant.const_val = $1.string;
	    $$->node.constant.lx_srcpos = $1.lx_srcpos;
	}
	|
	string_constant {
	    $$ = ecl_new_constant();
	    $$->node.constant.token = string_constant;
	    $$->node.constant.const_val = $1.string;
	    $$->node.constant.lx_srcpos = $1.lx_srcpos;
	}
	;

primary_expression:
	identifier_ref {
	    $$ = ecl_new_identifier();
	    $$->node.identifier.id = $1.string;
	    $$->node.identifier.lx_srcpos = $1.lx_srcpos;
	}
	|
	constant
	|
	LPAREN expression RPAREN
	{ $$ = $2; }
	;

/* missing ->, ++, -- */
postfix_expression:
	primary_expression
	|
	postfix_expression LBRACKET expression RBRACKET {
	    $$ = ecl_new_element_ref();
	    $$->node.element_ref.lx_srcpos = $2.lx_srcpos;
	    $$->node.element_ref.expression = $3;
	    $$->node.element_ref.array_ref = $1;
	}
	|
	postfix_expression DOT identifier_ref {
	    $$ = ecl_new_field_ref();
	    $$->node.field_ref.lx_srcpos = $2.lx_srcpos;
	    $$->node.field_ref.lx_field = $3.string;
	    $$->node.field_ref.struct_ref = $1;
	}
	|
	postfix_expression LPAREN argument_expression_list RPAREN {
	    $$ = ecl_new_subroutine_call();
	    $$->node.subroutine_call.lx_srcpos = $2.lx_srcpos;
	    $$->node.subroutine_call.arguments = $3;
	    $$->node.subroutine_call.sm_func_ref = $1;
	}
	|
	postfix_expression LPAREN  RPAREN {
	    $$ = ecl_new_subroutine_call();
	    $$->node.subroutine_call.lx_srcpos = $2.lx_srcpos;
	    $$->node.subroutine_call.arguments = NULL;
	    $$->node.subroutine_call.sm_func_ref = $1;
	}
	;

argument_expression_list:
	assignment_expression {
		$$ = malloc(sizeof(struct list_struct));
		$$->node = $1;
		$$->next = NULL;
	}
	|
	argument_expression_list COMMA assignment_expression {
	    sm_list tmp = $1;
	    while (tmp->next != NULL) {
		tmp = tmp->next;
	    }
	    tmp->next = malloc(sizeof(struct list_struct));
	    tmp->next->node = $3;
	    tmp->next->next = NULL;
	    $$ = $1;
	};


/* missing ++, --, sizeof, ~, & */
unary_expression:
	postfix_expression
	| STAR cast_expression {
	    $$ = ecl_new_operator();
	    $$->node.operator.lx_srcpos = $1.lx_srcpos;
	    $$->node.operator.op = op_deref;
	    $$->node.operator.right = $2;
	    $$->node.operator.left = NULL;
	}
	| BANG cast_expression {
	    $$ = ecl_new_operator();
	    $$->node.operator.lx_srcpos = $1.lx_srcpos;
	    $$->node.operator.op = op_log_neg;
	    $$->node.operator.right = $2;
	    $$->node.operator.left = NULL;
	}
	| PLUS cast_expression {
	    $$ = ecl_new_operator();
	    $$->node.operator.lx_srcpos = $1.lx_srcpos;
	    $$->node.operator.op = op_plus;
	    $$->node.operator.right = $2;
	    $$->node.operator.left = NULL;
	}
	| MINUS cast_expression {
	    $$ = ecl_new_operator();
	    $$->node.operator.lx_srcpos = $1.lx_srcpos;
	    $$->node.operator.op = op_minus;
	    $$->node.operator.right = $2;
	    $$->node.operator.left = NULL;
	}
	;

cast_expression:
	unary_expression
	| LPAREN type_specifier_list RPAREN cast_expression {
	    $$ = ecl_new_cast();
	    $$->node.cast.lx_srcpos = $1.lx_srcpos;
	    $$->node.cast.type_spec = $2;
	    $$->node.cast.expression = $4;
	}
	;

multiplicative_expression:
	cast_expression
	|
	multiplicative_expression STAR cast_expression
	{
	    $$ = ecl_new_operator();
	    $$->node.operator.lx_srcpos = $2.lx_srcpos;
	    $$->node.operator.op = op_mult;
	    $$->node.operator.right = $3;
	    $$->node.operator.left = $1;
	}
	|
	multiplicative_expression SLASH cast_expression
	{
	    $$ = ecl_new_operator();
	    $$->node.operator.lx_srcpos = $2.lx_srcpos;
	    $$->node.operator.op = op_div;
	    $$->node.operator.right = $3;
	    $$->node.operator.left = $1;
	}
	|
	multiplicative_expression MODULUS cast_expression
	{
	    $$ = ecl_new_operator();
	    $$->node.operator.lx_srcpos = $2.lx_srcpos;
	    $$->node.operator.op = op_modulus;
	    $$->node.operator.right = $3;
	    $$->node.operator.left = $1;
	}
	;

additive_expression:
	multiplicative_expression
	|
	additive_expression PLUS multiplicative_expression
	{
	    $$ = ecl_new_operator();
	    $$->node.operator.op = op_plus;
	    $$->node.operator.lx_srcpos = $2.lx_srcpos;
	    $$->node.operator.right = $3;
	    $$->node.operator.left = $1;
	}
	|
	additive_expression MINUS multiplicative_expression
	{
	    $$ = ecl_new_operator();
	    $$->node.operator.lx_srcpos = $2.lx_srcpos;
	    $$->node.operator.op = op_minus;
	    $$->node.operator.right = $3;
	    $$->node.operator.left = $1;
	}
	;

shift_expression:
	additive_expression
	|
	shift_expression LEFT_SHIFT additive_expression
	{
	    $$ = ecl_new_operator();
	    $$->node.operator.lx_srcpos = $2.lx_srcpos;
	    $$->node.operator.op = op_left_shift;
	    $$->node.operator.right = $3;
	    $$->node.operator.left = $1;
	}
	|
	shift_expression RIGHT_SHIFT additive_expression
	{
	    $$ = ecl_new_operator();
	    $$->node.operator.lx_srcpos = $2.lx_srcpos;
	    $$->node.operator.op = op_right_shift;
	    $$->node.operator.right = $3;
	    $$->node.operator.left = $1;
	}
	;

relational_expression:
	shift_expression
	|
	relational_expression LT shift_expression
	{
	    $$ = ecl_new_operator();
	    $$->node.operator.lx_srcpos = $2.lx_srcpos;
	    $$->node.operator.op = op_lt;
	    $$->node.operator.right = $3;
	    $$->node.operator.left = $1;
	}
	|
	relational_expression GT shift_expression
	{
	    $$ = ecl_new_operator();
	    $$->node.operator.lx_srcpos = $2.lx_srcpos;
	    $$->node.operator.op = op_gt;
	    $$->node.operator.right = $3;
	    $$->node.operator.left = $1;
	}
	|
	relational_expression LEQ shift_expression
	{
	    $$ = ecl_new_operator();
	    $$->node.operator.lx_srcpos = $2.lx_srcpos;
	    $$->node.operator.op = op_leq;
	    $$->node.operator.right = $3;
	    $$->node.operator.left = $1;
	}
	|
	relational_expression GEQ shift_expression
	{
	    $$ = ecl_new_operator();
	    $$->node.operator.lx_srcpos = $2.lx_srcpos;
	    $$->node.operator.op = op_geq;
	    $$->node.operator.right = $3;
	    $$->node.operator.left = $1;
	}
	;

equality_expression:
	relational_expression
	|
	equality_expression EQ relational_expression
	{
	    $$ = ecl_new_operator();
	    $$->node.operator.lx_srcpos = $2.lx_srcpos;
	    $$->node.operator.op = op_eq;
	    $$->node.operator.right = $3;
	    $$->node.operator.left = $1;
	}
	|
	equality_expression NEQ relational_expression
	{
	    $$ = ecl_new_operator();
	    $$->node.operator.lx_srcpos = $2.lx_srcpos;
	    $$->node.operator.op = op_neq;
	    $$->node.operator.right = $3;
	    $$->node.operator.left = $1;
	}
	;

/* missing expression, assignment_expression production */
expression:
	assignment_expression
	    {$$ = $1;}
	;

/* missing other assignment operators */
assignment_expression:
	conditional_expression
	{ $$ = $1;} 
	|
	unary_expression ASSIGN assignment_expression
	{
	    $$ = ecl_new_assignment_expression();
	    $$->node.assignment_expression.lx_srcpos = $2.lx_srcpos;
	    $$->node.assignment_expression.left = $1;
	    $$->node.assignment_expression.right = $3;
	}
	;

/* missing ?: construction */
conditional_expression:
	logical_or_expression
	;

logical_or_expression:
	logical_and_expression
	|
	logical_or_expression LOG_OR logical_and_expression
	{
	    $$ = ecl_new_operator();
	    $$->node.operator.lx_srcpos = $2.lx_srcpos;
	    $$->node.operator.op = op_log_or;
	    $$->node.operator.right = $3;
	    $$->node.operator.left = $1;
	}
	;

logical_and_expression:
	inclusive_or_expression
	|
	logical_and_expression LOG_AND inclusive_or_expression
	{
	    $$ = ecl_new_operator();
	    $$->node.operator.lx_srcpos = $2.lx_srcpos;
	    $$->node.operator.op = op_log_and;
	    $$->node.operator.right = $3;
	    $$->node.operator.left = $1;
	}
	;

inclusive_or_expression:
	exclusive_or_expression
	|
	inclusive_or_expression ARITH_OR exclusive_or_expression
	{
	    $$ = ecl_new_operator();
	    $$->node.operator.lx_srcpos = $2.lx_srcpos;
	    $$->node.operator.op = op_arith_or;
	    $$->node.operator.right = $3;
	    $$->node.operator.left = $1;
	}
	;

exclusive_or_expression:
	and_expression
	|
	exclusive_or_expression ARITH_XOR and_expression
	{
	    $$ = ecl_new_operator();
	    $$->node.operator.lx_srcpos = $2.lx_srcpos;
	    $$->node.operator.op = op_arith_xor;
	    $$->node.operator.right = $3;
	    $$->node.operator.left = $1;
	}
	;

and_expression:
	equality_expression
	|
	and_expression ARITH_AND equality_expression
	{
	    $$ = ecl_new_operator();
	    $$->node.operator.lx_srcpos = $2.lx_srcpos;
	    $$->node.operator.op = op_arith_and;
	    $$->node.operator.right = $3;
	    $$->node.operator.left = $1;
	}
	;

%%
#include "lex.yy.c"

typedef struct scope *scope_ptr;

struct parse_struct {
    sm_list decls;
    scope_ptr scope;
    int defined_type_count;
    char **defined_types;
    err_out_func_t error_func;
    void *client_data;
    sm_list return_type_list;
    int return_cg_type;
};


static int
semanticize_compound_statement(ecl_parse_context context, sm_ref compound, 
			       scope_ptr containing_scope);
static int semanticize_decls_list(ecl_parse_context context, sm_list decls, 
				  scope_ptr scope);
static void add_decl(char *id, sm_ref node, scope_ptr scope);
static sm_ref find_complex_type(sm_ref node, scope_ptr scope);
static char *ecl_code_string;
static int is_string(sm_ref expr);


int
ecl_parse_for_context(code, context)
char *code;
ecl_parse_context context;
{
    sm_list decls;
    int ret;
#if defined(YYDEBUG) && (YYDEBUG != 0)
    extern int yydebug;
    yydebug = 1;
#endif
    if (code != NULL) {
	setup_for_string_parse(code, context->defined_type_count, context->defined_types);
	ecl_code_string = code;
    }
    yyerror_count = 0;
    yycontext = context;
    yyparse();
    terminate_string_parse();

    if ((yyparse_value == NULL) || (yyerror_count != 0)) {
	return 0;
    }

    decls = (sm_list) yyparse_value;
    context->decls = decls;
    ret = semanticize_decls_list(context, decls, context->scope);
    if (ret == 0) {
	ecl_rfree_list(decls, NULL);
	context->decls = NULL;
    }
    return ret;
}

ecl_code
ecl_code_gen(code, context)
char *code;
ecl_parse_context context;
{
    sm_ref tmp;
    ecl_code ret_code;
    unsigned int offset;
    void *func;

    if (code != NULL) {
	setup_for_string_parse(code, context->defined_type_count, context->defined_types);
	ecl_code_string = code;
    }

    yyerror_count = 0;
    yycontext = context;
    yyparse();
    terminate_string_parse();

    if ((yyparse_value == NULL) || (yyerror_count != 0)) {
	return 0;
    }
    tmp = ecl_new_compound_statement();
    tmp->node.compound_statement.decls = context->decls;
    tmp->node.compound_statement.statements =
	malloc(sizeof(struct list_struct));
    tmp->node.compound_statement.statements->next = NULL;
    tmp->node.compound_statement.statements->node = yyparse_value;
    if (!semanticize_compound_statement(context, tmp, context->scope)) {
	tmp->node.compound_statement.decls = NULL;
	ecl_rfree(tmp);
	return NULL;
    }
    ret_code = malloc(sizeof(struct _ecl_code_struct));
    ret_code->code_memory_block = NULL;
    func = ecl_cg_net(tmp, &offset, ret_code);
    tmp->node.compound_statement.decls = NULL;
    ecl_rfree(tmp);
    ret_code->func = (void(*)())(long)func;
    ret_code->data = NULL;
    return ret_code;
}

void 
ecl_dump(ecl_code code)
{
    printf("ECL CODE structure %lx - \n", (long) code);
    printf("  function pointer %lx, code memory block %lx, data %lx, static size %d\n",
	   (long) code->func, (long) code->code_memory_block,
	   (long) code->data, code->static_size_required);
#ifndef HAVE_ICODE_H
#ifdef HAVE_DRISC_H
    dr_dump((drisc_ctx) code->drisc_context);
#endif
#else
    v_dump((void*)(long)code->func);
#endif
}
    

int
ecl_code_verify(code, context)
char *code;
ecl_parse_context context;
{
    sm_ref tmp;

    if (code != NULL) {
	setup_for_string_parse(code, context->defined_type_count, context->defined_types);
	ecl_code_string = code;
    }

    yyerror_count = 0;
    yycontext = context;
    yyparse();
    terminate_string_parse();

    if ((yyparse_value == NULL) || (yyerror_count != 0)) {
	return 0;
    }

    tmp = ecl_new_compound_statement();
    tmp->node.compound_statement.decls = context->decls;
    tmp->node.compound_statement.statements =
	malloc(sizeof(struct list_struct));
    tmp->node.compound_statement.statements->next = NULL;
    tmp->node.compound_statement.statements->node = yyparse_value;
    if (semanticize_compound_statement(context, tmp, context->scope) == 0) {
	tmp->node.compound_statement.decls = NULL;
	ecl_rfree(tmp);
	return 0;
    }
    tmp->node.compound_statement.decls = NULL;
    ecl_rfree(tmp);
    return 1;
}

extern void 
ecl_code_free(code)
ecl_code code;
{
    if (code->code_memory_block) free(code->code_memory_block);
    if (code->data) free(code->data);
#if !defined(HAVE_ICODE_H) && defined(HAVE_DRISC_H)
    if (code->drisc_context) {
	dr_free_context((drisc_ctx) code->drisc_context);
    }
#endif
    free(code);
}

static char *
copy_line(char *line_begin)
{
    char *line_end;
    if ((line_end = strchr(line_begin, 10)) == NULL) {
	/* no CR */
	return strdup(line_begin);
    } else {
	char *tmp = malloc(line_end - line_begin + 1);
	strncpy(tmp, line_begin, line_end - line_begin);
	tmp[line_end - line_begin] = 0;
	return tmp;
    }
}

static void
default_error_out(void *client_data, char *string)
{
    fprintf(stderr, "%s", string);
}

static void
print_context(ecl_parse_context context, int line, int character)
{
    char *tmp = ecl_code_string;
    char *line_begin = ecl_code_string;
    char *line_copy = NULL;
    int i, line_len, offset = 0;

    while (line > 1) {
	switch(*tmp) {
	case 10:
	    line_begin = tmp + 1;
	    line--;
	    break;
	case 0:
	    line = 1;   /* end of src */
	    break;
	}
	tmp++;
    }
    if (character > 40) {
	offset = character - 40;
    }
    line_copy = copy_line(line_begin + offset);
    line_len = strlen(line_copy);
    if (line_len > 60) {
	line_copy[60] = 0;
    }
    context->error_func(context->client_data, line_copy);
    context->error_func(context->client_data, "\n");
    free(line_copy);
    for(i=offset + 1; i< character; i++) {
	if (line_begin[i-1] == '\t') {
	    context->error_func(context->client_data, "\t");
	} else {
	    context->error_func(context->client_data, " ");
	}
    }
    context->error_func(context->client_data, "^\n");
}

void yyerror(str)
char *str;
{
    char tmp_str[100];
    yycontext->error_func(yycontext->client_data, "## Error ");
    yycontext->error_func(yycontext->client_data, str);
    yycontext->error_func(yycontext->client_data, "\n");
    yycontext->error_func(yycontext->client_data, "## While parsing near ");
    yycontext->error_func(yycontext->client_data, yytext);
    sprintf(tmp_str, ", offset = %d, line = %d ####\n",lex_offset,line_count);
    yycontext->error_func(yycontext->client_data, tmp_str);
    print_context(yycontext, line_count, lex_offset);
    yyerror_count++;
}

#ifdef STDC_HEADERS
static void
ecl_src_error(ecl_parse_context context, sm_ref expr, char *format, ...)
#else
static void
ecl_src_error(context, expr, format, va_alist)
ecl_parse_context context;
sm_ref expr;
char *format;
va_dcl
#endif
{

    va_list ap;
    char *tmp = malloc(10240); /* arbitrarily large */
    srcpos lx_srcpos = ecl_get_srcpos(expr);
#ifdef STDC_HEADERS
    va_start(ap, format);
#else
    va_start(ap);
#endif
    context->error_func(context->client_data, "## Ecode Error:  ");
    vsprintf(tmp, format, ap);
    context->error_func(context->client_data, tmp);
    sprintf(tmp, " at line %d, char %d\n", lx_srcpos.line, lx_srcpos.character);
    context->error_func(context->client_data, tmp);
    free(tmp);
    print_context(context, lx_srcpos.line, lx_srcpos.character);
}

extern void
ecl_print_operator_t(operator_t o)
{
    switch (o) {
    case  op_modulus:
	printf("MODULUS");
	break;
    case  op_plus:
	printf("PLUS");
	break;
    case  op_minus:
	printf("MINUS");
	break;
    case  op_leq:
	printf("LEQ");
	break;
    case  op_lt:
	printf("LESS THAN");
	break;
    case  op_geq:
	printf("GEQ");
	break;
    case  op_gt:
	printf("GREATER THAN");
	break;
    case  op_eq:
	printf("EQUAL");
	break;
    case  op_neq:
	printf("NOT EQUAL");
	break;
    case  op_log_or:
	printf("LOGICAL OR");
	break;
    case  op_log_and:
	printf("LOGICAL AND");
	break;
    case op_log_neg:
	printf("LOGICAL NEGATION");
	break;
    case op_arith_and:
	printf("ARITH AND");
	break;
    case op_arith_or:
	printf("ARITH OR");
	break;
    case op_arith_xor:
	printf("ARITH XOR");
	break;
    case op_left_shift:
	printf("LEFT SHIFT");
	break;
    case op_right_shift:
	printf("RIGHT SHIFT");
	break;
    case  op_mult:
	printf("MULTIPLY");
	break;
    case  op_div:
	printf("DIVISION");
	break;
    case  op_deref:
	printf("DEREFERENCE");
	break;
    }
}

extern void
ecl_print_srcpos(srcpos pos)
{
    printf("line %d, char %d", pos.line, pos.character);
}

typedef struct st_entry {
    char *id;
    sm_ref node;
    struct st_entry *next;
} *st_entry;

struct scope {
    ecl_extern_list externs;
    struct st_entry *entry_list;
    struct scope *containing_scope;
};


extern ecl_parse_context
ecl_copy_context(context)
ecl_parse_context context;
{
    int i;
    ecl_parse_context new_context = new_ecl_parse_context();
    new_context->decls = ecl_copy_list(context->decls);
    new_context->scope->externs = context->scope->externs;
    new_context->error_func = context->error_func;
    new_context->client_data = context->client_data;
    semanticize_decls_list(new_context, new_context->decls, 
			   new_context->scope);
    new_context->defined_types = malloc(sizeof(char*) *
					context->defined_type_count);
    for (i=0; i< context->defined_type_count; i++) {
	new_context->defined_types[i] = context->defined_types[i];
    }
    return new_context;
}

static void *
resolve_extern(char *id, scope_ptr scope)
{
    if (scope == NULL) return NULL;
    if (scope->externs != NULL) {
	ecl_extern_list externs = scope->externs;
	while(externs->extern_name != NULL) {
	    if (strcmp(id, externs->extern_name) == 0) {
		return externs->extern_value;
	    }
	    externs++;
	}
    }
    return resolve_extern(id, scope->containing_scope);
}

static scope_ptr
push_scope(scope_ptr containing_scope)
{
    scope_ptr new_scope = malloc(sizeof(*new_scope));
    new_scope->externs = NULL;
    new_scope->entry_list = NULL;
    new_scope->containing_scope = containing_scope;
    return new_scope;
}

static void
pop_scope(scope_ptr scope)
{
    st_entry list = scope->entry_list;
    while (list != NULL) {
	st_entry tmp = list->next;
	free(list);
	list = tmp;
    }
    free(scope);
}

static void
add_decl(char *id, sm_ref node, scope_ptr scope)
{
    st_entry entry = malloc(sizeof(*entry));
    entry->node = node;
    entry->id = id;
    entry->next = scope->entry_list;
    scope->entry_list = entry;
}

static sm_ref
resolve_local(char *id, scope_ptr scope)
{
    st_entry list = scope->entry_list;
    while(list != NULL) {
	if (strcmp(list->id, id) == 0) {
	    return list->node;
	}
	list = list->next;
    }
    return NULL;
}

static sm_ref
resolve(char *id, scope_ptr scope)
{
    sm_ref tmp;
    if (scope == NULL) return NULL;
    if ((tmp = resolve_local(id, scope)) != NULL) {
	return tmp;
    }
    return resolve(id, scope->containing_scope);
}

static int 
determine_op_type(ecl_parse_context context, sm_ref expr, 
		  int left_type, int right_type)
{
    int unsigned_used = 0;
    operator_t op = expr->node.operator.op;
    if (left_type == DR_P) {
	if (is_string(expr->node.operator.left) &&
	    is_string(expr->node.operator.right) &&
	    (op == op_eq)) {
	    return DR_I;
	} else {
	    ecl_src_error(context, expr, 
			  "Illegal arithmetic.  Left side is a pointer value");
	    return DR_ERR;
	}
    }
    if (right_type == DR_P) {
	ecl_src_error(context, expr, 
		      "Illegal arithmetic.  Right side is a pointer value");
	return DR_ERR;
    }
    if (left_type == DR_B) {
	ecl_src_error(context, expr, 
		      "Illegal arithmetic.  Left side is a structured type");
	return DR_ERR;
    }
    if (right_type == DR_B) {
	ecl_src_error(context, expr, 
		      "Illegal arithmetic.  Right side is a structured type");
	return DR_ERR;
    }
    if ((left_type == DR_D) || (right_type == DR_D)) {
	if ((op == op_modulus) || (op == op_log_or) || (op == op_log_and)) {
	    ecl_src_error(context, expr, "Operands must be integral.");
	    return DR_ERR;
	} else {
	    return DR_D;
	}
    }
    if ((left_type == DR_F) || (right_type == DR_F)) {
	if ((op == op_modulus) || (op == op_log_or) || (op == op_log_and)) {
	    ecl_src_error(context, expr, "Operands must be integral.");
	    return DR_ERR;

	} else {
	    return DR_F;
	}
    }
    switch(left_type) {
    case DR_C: case DR_UC: case DR_S: case DR_US:
	left_type = DR_I;   /* integer promotion */
    }
    switch(right_type) {
    case DR_C: case DR_UC: case DR_S: case DR_US:
	right_type = DR_I;   /* integer promotion */
    }
    switch(left_type) {
    case DR_UC: case DR_US: case DR_U: case DR_UL:
	unsigned_used++;
    }
    switch(right_type) {
    case DR_UC: case DR_US: case DR_U: case DR_UL:
	unsigned_used++;
    }
    if ((op == op_left_shift) || (op == op_right_shift)) return left_type;
    if ((left_type == DR_UL) || (right_type == DR_UL)) return DR_UL;
    if ((left_type == DR_L) || (right_type == DR_L)) {
	/* GSE -bug  This test should be for *generated* target, not host */
	if (sizeof(long) > sizeof(unsigned int)) {
	    /* Long can represent all values of unsigned int */
	    return DR_L;
	} else {
	    return unsigned_used? DR_UL : DR_L;
	}
    }
    if ((left_type == DR_U) || (right_type == DR_U)) return DR_U;
    return unsigned_used? DR_U: DR_I;
}

static int reduce_type_list(ecl_parse_context context, sm_list type_list);

static int
is_n_dimen_array(int dimen, sm_ref expr)
{
    if (expr->node_type == ecl_field_ref) {
	return is_n_dimen_array(dimen, expr->node.field_ref.sm_field_ref);
    }
    if (expr->node_type == ecl_element_ref) {
	return is_n_dimen_array(dimen + 1, expr);
    }
    if (expr->node_type != ecl_field) return 0;
    if ((expr->node.field.sm_var_control_field != NULL) && (dimen != 1)) return 0;

    switch(dimen) {
    case 1:
	if ((expr->node.field.dimen1 != 1) && (expr->node.field.dimen2 == 1))
	    return 1;
	return 0;
    case 2:
	if ((expr->node.field.dimen1 != 1) && (expr->node.field.dimen2 != 1))
	    return 1;
	return 0;
    }
    return 0;
}

static int 
is_string(sm_ref expr)
{
    if (expr->node_type == ecl_field) {
	return (strcmp(expr->node.field.string_type, "string") == 0);
    } else if (expr->node_type == ecl_field_ref) {
	return is_string(expr->node.field_ref.sm_field_ref);
    } else if (expr->node_type == ecl_identifier) {
	return is_string(expr->node.identifier.sm_declaration);
    } else if (expr->node_type == ecl_declaration) {
	if (expr->node.declaration.cg_type != DR_P) return 0;
	if (expr->node.declaration.sm_complex_type != NULL) return 0;
	/* only strings have pointers without complex types */
	return 1;
    } else if (expr->node_type == ecl_constant) {
	return (expr->node.constant.token == string_constant);
    }
    return 0;
}

extern int 
ecl_expr_is_string(sm_ref expr)
{
    return is_string(expr);
}

extern int 
is_control_value(sm_ref expr, sm_ref strct)
{
    sm_list fields;
    if (expr->node_type == ecl_field_ref) {
	return is_control_value(expr->node.field_ref.sm_field_ref,
				expr->node.field_ref.struct_ref);
    }
    if (expr->node_type != ecl_field) return 0;
    assert(strct != NULL);
    strct = get_complex_type(0, strct);	
    assert(strct->node_type == ecl_struct_type_decl);
    fields =  strct->node.struct_type_decl.fields;
    while(fields != NULL) {
	if (fields->node->node.field.sm_var_control_field == expr) return 1;
	fields = fields->next;
    }
    return 0;
}


static int 
is_array(sm_ref expr)
{
    if (expr->node_type == ecl_field_ref) {
	return is_array(expr->node.field_ref.sm_field_ref);
    }
    if (expr->node_type == ecl_element_ref) {
	return is_n_dimen_array(2, expr->node.element_ref.array_ref);
    }
    if (expr->node_type != ecl_field) return 0;
    if (expr->node.field.sm_var_control_field != NULL) return 1;

    if ((expr->node.field.dimen1 == 1) &&
	(expr->node.field.dimen2 == 1)) return 0;
    return 1;
}
    
static int semanticize_expr(ecl_parse_context context, sm_ref expr, 
			    scope_ptr scope) 
{
    switch(expr->node_type) {
    case ecl_identifier: {
	sm_ref tmp = resolve(expr->node.identifier.id, scope);
	if (tmp != NULL) {
	    expr->node.identifier.sm_declaration = tmp;
	    return 1;
	} else {
	    ecl_src_error(context, expr,
			  "Undefined Symbol \"%s\"", 
			  expr->node.identifier.id);
	    return 0;
	}
    }
    case ecl_cast: {
	if (!semanticize_expr(context, expr->node.cast.expression, scope)) {
	    return 0;
	}
	expr->node.cast.cg_type = 
	    reduce_type_list(context, expr->node.cast.type_spec);
	if (expr->node.cast.cg_type == DR_ERR) {
	    return 0;
	}
	return 1;
    }
    case ecl_operator: {
	int ret = 1;
	if (expr->node.operator.left != NULL) {
	    if (!semanticize_expr(context, expr->node.operator.left, scope)) {
		ret = 0;
	    }
	}
	if (!semanticize_expr(context, expr->node.operator.right, scope)) {
	    ret = 0;
	}
	if (ret == 0) return 0;
	if (expr->node.operator.left != NULL) {
	    expr->node.operator.operation_type = 
		determine_op_type(context, expr, 
				  ecl_sm_get_type(expr->node.operator.left),
				  ecl_sm_get_type(expr->node.operator.right));
	    if (expr->node.operator.operation_type == DR_ERR) {
		return 0;
	    }
	} else {
	    expr->node.operator.operation_type = ecl_sm_get_type(expr->node.operator.right);
	}
	switch (expr->node.operator.op) {
	case op_leq: case op_lt: case op_geq: case op_gt: case op_neq: 
	case op_eq:  case op_log_neg: case op_log_or: case op_log_and:
	    expr->node.operator.result_type = DR_I;
	    break;
	default:
	    expr->node.operator.result_type=expr->node.operator.operation_type;
	}
	return ret;
    }
    case ecl_constant:
	return 1;
    case ecl_assignment_expression: {
	int ret = 1;
	if (!semanticize_expr(context, expr->node.assignment_expression.left, scope)) {
	    ret = 0;
	} else {
	    expr->node.assignment_expression.cg_type = 
		ecl_sm_get_type(expr->node.assignment_expression.left);
	}
	if ((expr->node.assignment_expression.cg_type == DR_P) ||
	    (expr->node.assignment_expression.cg_type == DR_ERR)) {
	    if (!is_string(expr->node.assignment_expression.left)) {
		ecl_src_error(context, expr->node.assignment_expression.left, "Invalid assignment, left side must be simple, non-pointer type");
		ret = 0;
	    }
	}
	if (!semanticize_expr(context, expr->node.assignment_expression.right, scope)){
	    ret = 0;
	} else {
	    int right_type = 
		ecl_sm_get_type(expr->node.assignment_expression.right);
	    if ((right_type == DR_P) && 
		(is_string(expr->node.assignment_expression.right))) {
		if (expr->node.assignment_expression.cg_type != DR_P) {
		    ecl_src_error(context, expr, "assignment mixes string and non-string types");
		    ret = 0;
		}
	    } else if ((right_type == DR_B) || (right_type == DR_ERR)) {
		ecl_src_error(context, expr->node.assignment_expression.right, "Invalid assignment, right side must be simple type");
		ret = 0;
	    }
	}
	return ret;
    }
    case ecl_field_ref: {
	sm_ref typ;
	sm_list fields;
	if (!semanticize_expr(context, expr->node.field_ref.struct_ref, scope)) {
	    return 0;
	}
	typ = get_complex_type(context, expr->node.field_ref.struct_ref);
	if (typ == NULL) {
	    ecl_src_error(context, expr->node.field_ref.struct_ref, 
			  "Reference must be structured type", 
			  expr->node.field_ref.lx_field);
	    return 0;
	}
	fields =  typ->node.struct_type_decl.fields;
	while(fields != NULL) {
	    if (strcmp(expr->node.field_ref.lx_field,
		       fields->node->node.field.name) == 0) {
		break;
	    }
	    fields = fields->next;
	}
	if (fields == NULL) {
	    ecl_src_error(context, expr, 
			  "Unknown field reference, \"%s\".",
			  expr->node.field_ref.lx_field);
	    return 0;
	}
	expr->node.field_ref.sm_field_ref = fields->node;
	return 1;
    }
    case ecl_element_ref: {
	if (semanticize_expr(context, expr->node.element_ref.array_ref, scope)) {
	    int cg_type;
	    if (!is_array(expr->node.element_ref.array_ref)) {
		ecl_src_error(context, expr, "Indexed element must be array");
		return 0;
	    }
	    if (!semanticize_expr(context, expr->node.element_ref.expression, scope)) {
		return 0;
	    }
	    
	    cg_type = ecl_sm_get_type(expr->node.element_ref.expression);
	    switch(cg_type) {
	    case DR_C:
	    case DR_UC:
	    case DR_S:
	    case DR_US:
	    case DR_I: 
	    case DR_U: 
	    case DR_L:
	    case DR_UL:
		return 1;
		break;
	    }
	    ecl_src_error(context, expr, 
			  "Index for element reference must be integer type");
	    return 0;
	}
	return 0;
    }
    case ecl_subroutine_call: {
	sm_ref func_ref = expr->node.subroutine_call.sm_func_ref;
	sm_ref tmp = resolve(func_ref->node.identifier.id, scope);
	sm_list args = expr->node.subroutine_call.arguments;
	sm_list formals;
	int ret = 1;
	if (tmp != NULL) {
	    if ((tmp->node_type != ecl_declaration) ||
		!tmp->node.declaration.is_subroutine) {
		ecl_src_error(context, expr, 
			      "Identifier is not subroutine \"%s\".", 
			func_ref->node.identifier.id);
		return 0;
	    }
	    free(func_ref->node.identifier.id);
	    free(func_ref);
	    expr->node.subroutine_call.sm_func_ref = func_ref = tmp;
	    formals = func_ref->node.declaration.params;
	} else {
	    ecl_src_error(context, func_ref, "Undefined Subroutine \"%s\".", 
			  func_ref->node.identifier.id);

	    return 0;
	}
	while (args != NULL) {
	    sm_ref arg = args->node;
	    sm_ref formal = NULL;
	    if (formals != NULL) {
		formal = formals->node;
	    }
	    if (!semanticize_expr(context, arg, scope) ) {
		args = args->next;
		continue;
	    }
	    if (formal == NULL) {
		ecl_src_error(context, arg, "To many arguments to subroutine");
		ret = 0;
		return ret;
	    }
	    if (strcmp(formal->node.declaration.id, "...") != 0) {
		/* we've got a real formal to check against */
		/* do some checking... */
		int mismatch = 0;
		switch (ecl_sm_get_type(arg)) {
		case DR_D: case DR_F:
		    if ((formal->node.declaration.cg_type != DR_D) &&
			(formal->node.declaration.cg_type != DR_F)) {
			mismatch++;
		    }
		    break;
		case DR_I: case DR_U:
		case DR_L: case DR_UL:
		    if ((formal->node.declaration.cg_type == DR_D) ||
			(formal->node.declaration.cg_type == DR_F) ||
			(formal->node.declaration.cg_type == DR_P)) {
			mismatch++;
		    }
		    break;
		case DR_P:
		    if (formal->node.declaration.cg_type != DR_P) mismatch++;
		    break;
		}

		if (mismatch) {
		    ecl_src_error(context, arg, 
				  "Type mismatch, parameter \"%s\".",
			    formal->node.declaration.id);
		    ret = 0;
		}
	    }
	    args = args->next;
	    if ((formals != NULL) &&
		(strcmp(formal->node.declaration.id, "...") != 0)) {
		formals = formals->next;
	    }
	}
	return ret;
    }
    default:
	fprintf(stderr, "Unknown case in semanticize_expression\n");
	ecl_print(expr);
    }
    return 0;
}

extern int
ecl_sm_get_type(sm_ref node)
{
    switch(node->node_type) {
    case ecl_identifier:
	if (node->node.identifier.sm_declaration != NULL) {
	    return ecl_sm_get_type(node->node.identifier.sm_declaration);
	}
	return node->node.identifier.cg_type;
    case ecl_operator:
	return node->node.operator.result_type;
    case ecl_cast:
	return node->node.cast.cg_type;
    case ecl_assignment_expression:
	return node->node.assignment_expression.cg_type;
    case ecl_declaration:
	return node->node.declaration.cg_type;
    case ecl_constant:
	/* need to handle bigger constants */
	if (node->node.constant.token == string_constant) {
	    return DR_P;
	} else if (node->node.constant.token == floating_constant) {
	    return DR_F;
	} else {
	    return DR_I;
	}
    case ecl_field_ref:
	return ecl_sm_get_type(node->node.field_ref.sm_field_ref);
    case ecl_element_ref:
    {
	sm_ref array = node->node.element_ref.array_ref;
	if (!is_array(array)) return DR_ERR;
	if (array->node_type == ecl_field_ref) {
	    array = array->node.field_ref.sm_field_ref;
	}
	if (array->node_type == ecl_field) {
	    /* single dimen reference */
	    if ((array->node.field.sm_var_control_field != 0) ||
		(array->node.field.dimen2 == 1)) {
		return array->node.field.cg_type;
	    } else {
		/* must be multi dimen array */
		return DR_P;
	    }
	} else if (array->node_type == ecl_element_ref) {
	    /* multi dimen reference */
	    sm_ref array2 = array->node.element_ref.array_ref;
	    if (array2->node_type == ecl_field_ref) {
		array2 = array2->node.field_ref.sm_field_ref;
	    }
	    if (array2->node_type != ecl_field) return DR_ERR;
	    if ((array2->node.field.sm_var_control_field != 0) ||
		(array2->node.field.dimen2 == 1)) {
		return DR_ERR;  /* too many dimen reference */
	    }
	    return array2->node.field.cg_type;
	} else {
	    return DR_ERR;
	}
	break;
    }
/*	return ecl_sm_get_type(node->node.element_ref.array_ref);*/
    case ecl_field:
	if (is_array(node)) {
	    return DR_P;
	} else {
	    return node->node.field.cg_type;
	}
    case ecl_subroutine_call:
	return ecl_sm_get_type(node->node.subroutine_call.sm_func_ref);
    default:
	fprintf(stderr, "Unknown case in ecl_sm_get_type()\n");
	ecl_print(node);
    }
    return DR_ERR;
}

extern sm_ref
get_complex_type(ecl_parse_context context, sm_ref node)
{
    switch(node->node_type) {
    case ecl_identifier:
	return get_complex_type(context, 
				node->node.identifier.sm_declaration);
    case ecl_element_ref:
    {
	sm_ref array = node->node.element_ref.array_ref;
	return get_complex_type(context, array->node.field_ref.sm_field_ref);
    }
    case ecl_field:
	return node->node.field.sm_complex_type;
    case ecl_declaration:
	return node->node.declaration.sm_complex_type;
    case ecl_field_ref:{
	sm_ref typ;
	sm_list fields;
	typ = get_complex_type(context, node->node.field_ref.struct_ref);
	fields =  typ->node.struct_type_decl.fields;
	while ((fields != NULL) && 
	       (strcmp(node->node.field_ref.lx_field,
		       fields->node->node.field.name) != 0)) {
	    fields = fields->next;
	}
	if (fields == NULL) {
	    ecl_src_error(context, node, "Unknown field reference \"%s\".",
		    node->node.field_ref.lx_field);
	    return NULL;
	}
	return fields->node->node.field.sm_complex_type;
    }
    default:
	fprintf(stderr, "Unknown case in get_complex_type()\n");
	ecl_print(node);
    }
    return NULL;
}

static int
reduce_type_list(ecl_parse_context context, sm_list type_list)
{
    sm_list orig_list = type_list;
    int short_appeared = 0;
    int long_appeared = 0;
    int int_appeared = 0;
    int double_appeared = 0;
    int float_appeared = 0;
    int char_appeared = 0;
    int signed_appeared = 0;
    int unsigned_appeared = 0;
    int void_appeared = 0;
    int string_appeared = 0;
    int spec_count = 0;

    if (type_list && type_list->node->node_type != ecl_type_specifier) {
	return DR_ERR;
    }
    while (type_list != NULL) {
	int typ = type_list->node->node.type_specifier.token;
	spec_count++;
	switch (typ) {
	case INT:
	    int_appeared++;
	    break;
	case LONG:
	    long_appeared++;
	    break;
	case SHORT:
	    short_appeared++;
	    break;
	case DOUBLE:
	    double_appeared++;
	    break;
	case STRING:
	    string_appeared++;
	    break;
	case VOID:
	    void_appeared++;
	    break;
	case FLOAT:
	    float_appeared++;
	    break;
	case CHAR:
	    char_appeared++;
	    break;
	case SIGNED:
	    signed_appeared++;
	    break;
	case UNSIGNED:
	    unsigned_appeared++;
	    break;
	}
	type_list = type_list->next;
    }
    if (spec_count == 0) return DR_I;   /* default to int */
    if (void_appeared && (spec_count > 1)) {
	ecl_src_error(context, orig_list->node, 
		      "Void type may not appear with other specifiers");
	return DR_ERR;
    }
    if (string_appeared && (spec_count > 1)) {
	ecl_src_error(context, orig_list->node, 
		      "String type may not appear with other specifiers");
	return DR_ERR;
    }
    if (void_appeared) return DR_V;
    if (string_appeared) return DR_P;
    if (short_appeared + long_appeared > 1) {
	ecl_src_error(context, orig_list->node, 
		      "Only one of long or short permitted");
	return DR_ERR;
    }
    if (short_appeared && (double_appeared + float_appeared)) {
	ecl_src_error(context, orig_list->node, 
		      "Short may not be specified with double or float");
	return DR_ERR;
    }
    if (double_appeared + float_appeared) {
	if (double_appeared + float_appeared + short_appeared + signed_appeared + unsigned_appeared + char_appeared + int_appeared > 1) {
	    ecl_src_error(context, orig_list->node, "Bad type spec");
	    return DR_ERR;
	} else {
	    /* not handling LONG plus one of these */
	    if (double_appeared) {
		return DR_D;
	    } else {
		return DR_F;
	    }
	}
    }

    /* neither float or double appeared */
    if (short_appeared + char_appeared + long_appeared >= 2) {
	ecl_src_error(context, orig_list->node, 
		      "Only one integer size spec may be specified");
	return DR_ERR;
    }
    if (unsigned_appeared + signed_appeared > 1) {
	ecl_src_error(context, orig_list->node, "Bad type spec");
	return DR_ERR;
    }
    if (unsigned_appeared) {
	if (char_appeared) {
	    return DR_UC;
	} else if (short_appeared) {
	    return DR_US;
	} else if (long_appeared) {
	    return DR_UL;
	} else {
	    return DR_U;
	}
    } else {
	if (char_appeared) {
	    return DR_C;
	} else if (short_appeared) {
	    return DR_S;
	} else if (long_appeared) {
	    return DR_L;
	} else {
	    return DR_I;
	}
    }
}

static int semanticize_type_node(ecl_parse_context context, sm_ref decl, 
				 scope_ptr scope);

static int semanticize_decl(ecl_parse_context context, sm_ref decl, 
			    scope_ptr scope)
{
    switch(decl->node_type) {
    case ecl_declaration: 
	if (resolve_local(decl->node.declaration.id, scope) != NULL) {
	    ecl_src_error(context, decl, "Duplicate Symbol \"%s\"", 
		    decl->node.declaration.id);
	    return 0;
	} else {
	    add_decl(decl->node.declaration.id, decl, scope);
	}
	if (scope->containing_scope == NULL) {
	    /* must be external variable */
	    void *extern_value = 
		resolve_extern(decl->node.declaration.id, scope);
	    if (extern_value == NULL) {
		ecl_src_error(context, decl, 
			      "External symbol lacking address \"%s\"", 
			decl->node.declaration.id);
		return  0;
	    }
	    decl->node.declaration.cg_address = extern_value;
	}
	if (decl->node.declaration.static_var) {
	    if (decl->node.declaration.init_value != NULL) {
		sm_ref const_val = decl->node.declaration.init_value;
		if (const_val->node_type != ecl_constant) {
		    ecl_src_error(context, const_val, 
				  "Only simple constants allowed as static initializers.  Variable \"%s\"",
			decl->node.declaration.id);
		    return 0;
		}
	    }
	}
	decl->node.declaration.cg_type = 
	    reduce_type_list(context, decl->node.declaration.type_spec);
	if (decl->node.declaration.cg_type == DR_ERR) {
	    if (decl->node.declaration.type_spec->node->node_type == ecl_identifier) {
		decl->node.declaration.sm_complex_type = find_complex_type(decl->node.declaration.type_spec->node, scope);
		decl->node.declaration.cg_type = DR_P;
	    } else {
		return 0;
	    }
	} else {
	    decl->node.declaration.sm_complex_type = NULL;
	}
	if (decl->node.declaration.init_value != NULL) {
	    return semanticize_expr(context, decl->node.declaration.init_value, 
				    scope);
	}
	if (decl->node.declaration.is_subroutine) {
	    int ret;
	    scope_ptr sub_scope = push_scope(scope);
	    ret = semanticize_decls_list(context,
					 decl->node.declaration.params, 
					 sub_scope);
	    pop_scope(sub_scope);
	    return ret;
	}
	return 1;
	break;
    case ecl_struct_type_decl:
	return semanticize_type_node(context, decl, scope);
	break;
    default:
	printf("Unhandled case in semantize decls_list\n");
	ecl_print(decl);
    }
    return 0;
}

static int
semanticize_decls_list(ecl_parse_context context, sm_list decls, 
		       scope_ptr scope)
{
    int ret = 1;
    while (decls != NULL) {
	if (!semanticize_decl(context, decls->node, scope)) {
	    ret = 0;
	}
	decls = decls->next;
    }
    return ret;
}

static int semanticize_statement(ecl_parse_context context, sm_ref stmt, 
				 scope_ptr scope);

static int
semanticize_selection_statement(ecl_parse_context context, sm_ref selection,
				scope_ptr scope)
{
    int ret = 1;
    if (!semanticize_expr(context, 
			  selection->node.selection_statement.conditional,
			  scope)) {
	ret = 0;
    }
    if (!semanticize_statement(context,
			       selection->node.selection_statement.then_part,
			       scope)) {
	ret = 0;
    }
    if (selection->node.selection_statement.else_part) {
	if (!semanticize_statement(context,
				   selection->node.selection_statement.else_part,
				   scope)) {
	    ret = 0;
	}
    }
    return ret;
}

static int
semanticize_iteration_statement(ecl_parse_context context, sm_ref iteration,
				scope_ptr scope)
{
    int ret = 1;
    if (iteration->node.iteration_statement.init_expr != NULL) {
	if (!semanticize_expr(context, 
			      iteration->node.iteration_statement.init_expr,
			      scope)) {
	    ret = 0;
	}
    }

    if (iteration->node.iteration_statement.test_expr != NULL) {
	if (!semanticize_expr(context, 
			      iteration->node.iteration_statement.test_expr,
			      scope)) {
	    ret = 0;
	}
    }

    if (iteration->node.iteration_statement.iter_expr != NULL) {
	if (!semanticize_expr(context, 
			      iteration->node.iteration_statement.iter_expr,
			      scope)) {
	    ret = 0;
	}
    }

    if (iteration->node.iteration_statement.statement != NULL) {
	if (!semanticize_statement(context,
				   iteration->node.iteration_statement.statement,
				   scope)) {
	    ret = 0;
	}
    }
    return ret;
}

static int 
semanticize_statement(ecl_parse_context context, sm_ref stmt, 
		      scope_ptr scope)
{
    switch (stmt->node_type) {
    case ecl_selection_statement:
	return semanticize_selection_statement(context, stmt, scope);
    case ecl_iteration_statement:
	return semanticize_iteration_statement(context, stmt, scope);
    case ecl_expression_statement: {
	return semanticize_expr(context, 
				stmt->node.expression_statement.expression,
				scope);
    }	
    case ecl_compound_statement:
	return semanticize_compound_statement(context, stmt, scope);
    case ecl_return_statement:{
	stmt->node.return_statement.cg_func_type = context->return_cg_type;
	return semanticize_expr(context, 
				stmt->node.return_statement.expression,
				scope);
    }	
    default:
	printf("unhandled case in semanticize statement\n");
	return 1;
    }
}

static int
semanticize_statements(ecl_parse_context context, sm_list stmts, 
		       scope_ptr scope)
{
    int ret = 1;
    while (stmts != NULL) {
	if (!semanticize_statement(context, stmts->node, scope)) {
	    ret = 0;
	}
	stmts = stmts->next;
    }
    return ret;
}

extern int
semanticize_compound_statement(ecl_parse_context context, sm_ref compound, 
			       scope_ptr containing_scope)
{
    int ret = 1;
    scope_ptr current_scope = push_scope(containing_scope);
    ret &= semanticize_decls_list(context,
				  compound->node.compound_statement.decls,
				  current_scope);
    ret &= semanticize_statements(context,
				  compound->node.compound_statement.statements,
				  current_scope);
    pop_scope(current_scope);
    return ret;
}

extern sm_ref
ecl_build_type_node(const char *name, IOFieldList field_list)
{
    sm_ref decl = ecl_new_struct_type_decl();
    sm_list *end_ptr = &decl->node.struct_type_decl.fields;

    decl->node.struct_type_decl.id = strdup(name);
    while ((field_list != NULL) && (field_list->field_name != NULL)) {
	sm_list new_elem;
	new_elem = malloc(sizeof(*new_elem));
	new_elem->next = NULL;
	new_elem->node = ecl_new_field();
	new_elem->node->node.field.name = strdup(field_list->field_name);
	new_elem->node->node.field.string_type = strdup(field_list->field_type);
	new_elem->node->node.field.cg_size = field_list->field_size;
	new_elem->node->node.field.cg_offset = field_list->field_offset;
	new_elem->node->node.field.cg_type = DR_ERR;
	*end_ptr = new_elem;
	end_ptr = &new_elem->next;
	field_list++;
    }
    return decl;
}


static
void
add_defined_type(id, context)
char *id;
ecl_parse_context context;
{
    int count = context->defined_type_count;
    if (count == 0) {
	context->defined_types = malloc(sizeof(char*));
    } else {
	context->defined_types = realloc(context->defined_types,
					 (count+1)*sizeof(char*));
    }
    context->defined_types[count] = id;
    context->defined_type_count++;
}

extern void
ecl_add_struct_type(const char *name, IOFieldList field_list, 
		    ecl_parse_context context)
{
    sm_ref node = ecl_build_type_node(name, field_list);
    add_defined_type(name, context);
    ecl_add_decl_to_parse_context(name, node, context);
}

static int
str_to_data_type(str, size)
char *str;
int size;
{
    char *tmp = malloc(strlen(str) + 1);
    strcpy(tmp, str);
    str = tmp;			/* make a copy of str parameter */

    while (isspace((int)*str)) {	/* skip preceeding space */
	str++;
    }
    tmp = str + strlen(str) - 1;
    while (isspace((int)*tmp)) {	/* kill trailing space */
	*tmp = 0;
	tmp--;
    }
    tmp = str;
    while (*tmp) {		/* map to lower case */
	*tmp = tolower(*tmp);
	tmp++;
    }
    if ((strcmp(str, "integer") == 0) || (strcmp(str, "enumeration") == 0)) {
	free(str);
	if (size == sizeof(long)) {
	    return DR_L;
	} else if (size == sizeof(int)) {
	    return DR_I;
	} else if (size == sizeof(short)) {
	    return DR_S;
	} else if (size == sizeof(char)) {
	    return DR_C;
	} else {
	    fprintf(stderr, "unsupported integer size %d\n", size);
	    return DR_I;
	}
    } else if (strcmp(str, "unsigned integer") == 0) {
	free(str);
	if (size == sizeof(long)) {
	    return DR_UL;
	} else if (size == sizeof(int)) {
	    return DR_U;
	} else if (size == sizeof(short)) {
	    return DR_US;
	} else if (size == sizeof(char)) {
	    return DR_UC;
	} else {
	    fprintf(stderr, "unsupported unsigned integer size %d\n", size);
	    return DR_I;
	}
    } else if ((strcmp(str, "float") == 0) || (strcmp(str, "double") == 0)) {
	free(str);
	if (size == sizeof(double)) {
	    return DR_D;
	} else if (size == sizeof(float)) {
	    return DR_F;
	} else {
	    fprintf(stderr, "unsupported float size %d\n", size);
	    return DR_D;
	}
    } else if (strcmp(str, "char") == 0) {
	free(str);
	assert(size == 1);
	return DR_C;
    } else if (strcmp(str, "string") == 0) {
	free(str);
	assert(size == sizeof(char*));
	return DR_P;
    } else {
	free(str);
	return DR_ERR;
    }
}

static int
array_str_to_data_type(str, size, dimen1_ptr, dimen2_ptr)
char *str;
int size;
int *dimen1_ptr;
int *dimen2_ptr;
{
    int ret_type;
    char field_type[1024];
    char *left_paren;
    if ((left_paren = strchr(str, '[')) == NULL) {
	*dimen1_ptr = *dimen2_ptr = 1;
	ret_type = str_to_data_type(str, size);
    } else if (strchr(left_paren + 1, '[') == NULL) {
	/* single dimen array */
	int dimension = 0;
	int i = 0;
	char *tmp = str;
	for( ; tmp < left_paren; tmp++) {
	    field_type[i++] = *tmp;
	}
	field_type[i] = 0;
	if (sscanf(left_paren + 1, "%d]", &dimension) != 1) {
	    ret_type = str_to_data_type(field_type, size);
	    *dimen1_ptr = -1;
	    *dimen2_ptr = 1;
	} else {
	    ret_type = str_to_data_type(field_type, size);
	    *dimen1_ptr = dimension;
	    *dimen2_ptr = 1;
	}
    } else {
	/* double dimen array */
	int dimension1 = 0;
	int dimension2 = 0;
	int i = 0;
	char *tmp = str;
	for( ; tmp < left_paren; tmp++) {
	    field_type[i++] = *tmp;
	}
	field_type[i] = 0;
	if (sscanf(left_paren + 1, "%d][%d]", &dimension1, &dimension2) != 2) {
	    fprintf(stderr, "Couldn't parse array size in \"%s\"\n", str);
	    *dimen1_ptr = 0;
	    *dimen2_ptr = 0;
	    ret_type = str_to_data_type(field_type, size);
	} else {
	    ret_type = str_to_data_type(field_type, size);
	    *dimen1_ptr = dimension1;
	    *dimen2_ptr = dimension2;
	}
    }
    return ret_type;
}

static int
is_var_array(context, decl, f, fields, err)
ecl_parse_context context;
sm_ref decl;
sm_ref f;
sm_list fields;
int *err;
{
    char field_name[1024];
    char *left_paren;
    int junk;
    char *str = f->node.field.string_type;

    *err = 0;
    if ((left_paren = strchr(str, '[')) == NULL) {
	return 0;
    } else if (strchr(left_paren + 1, '[') == NULL) {
	/* single dimen array */

	if (sscanf(left_paren + 1, "%d]", &junk) != 1) {
	    int count = 0;
	    int i = 0;
	    while (((left_paren+1)[count] != ']') &&
		   ((left_paren+1)[count] != 0)) {
		field_name[count] = (left_paren+1)[count];
		count++;
	    }
	    field_name[count] = 0;
	    while(fields != NULL) {
		if (strcmp(field_name, fields->node->node.field.name) == 0) {
		    switch (str_to_data_type(fields->node->node.field.string_type, (int)sizeof(int))) {
		    case DR_C: case DR_UC: case DR_S: case DR_US: 
		    case DR_I: case DR_U: case DR_L: case DR_UL:
			f->node.field.sm_var_control_field = fields->node;
			return 1;
			break;
		    default:
			ecl_src_error(context, decl, 
				      "Variable length control field \"%s\"not of integer type.", field_name);
			*err = 1;
			return 0;
			break;
		    }
		}
		i++;
		fields = fields->next;
	    }
	    ecl_src_error(context, decl, 
			  "Variable length control field \"%s\" not found.", field_name);
	    *err = 1;
	    return 0;
	} else {
	    return 0;
	}
    }
    return 0;
}

#define Max(i,j) ((i<j) ? j : i)

static int
semanticize_type_node(ecl_parse_context context, sm_ref decl, 
		      scope_ptr scope)
{
    int ret = 1;
    int struct_size = 0;
    sm_list fields = decl->node.struct_type_decl.fields;
    add_decl(decl->node.struct_type_decl.id, decl, scope);
    while(fields != NULL) {
	field *f = &fields->node->node.field;
	int field_size = 0;
	int err;

	if (is_var_array(context, decl, fields->node,
			 decl->node.struct_type_decl.fields, &err)) {
	    field_size = sizeof(char*);
	}
	if (err == 1) ret = 0;
	f->cg_type = array_str_to_data_type(f->string_type, f->cg_size, 
					    &f->dimen1, &f->dimen2);
	f->element_size = f->cg_size;
	if (f->sm_var_control_field == NULL) {
	    field_size = f->dimen1 * f->dimen2 * f->element_size;
	}
	if (f->cg_type == DR_ERR) {
	    char *tmp_str = strdup(f->string_type);
	    char *rbracket = strchr(tmp_str, '[');
	    if (rbracket != NULL) *rbracket = 0;
	    f->sm_complex_type = resolve(tmp_str, scope);
	    if (f->sm_complex_type == NULL) {
		ecl_src_error(context, decl, 
			      "Field \"%s\" has unknown type \"%s\".",
			      f->name, tmp_str);
		ret = 0;
	    }
	    if (rbracket != NULL) *rbracket = '[';
	    free(tmp_str);	    
	}
	assert(field_size > 0);
	struct_size = Max(struct_size,
			  (f->cg_offset + field_size));
	fields = fields->next;
    }
    decl->node.struct_type_decl.cg_size = struct_size;
    return ret;
}

extern sm_ref
ecl_build_param_node(const char *id, sm_ref typ, int param_num)
{
    sm_ref node = ecl_new_declaration();
    sm_ref ident;
    node->node.declaration.param_num = param_num;
    node->node.declaration.id = strdup(id);
    node->node.declaration.sm_complex_type = typ;
    if (typ != NULL) {
	ident = ecl_new_identifier();
	node->node.declaration.type_spec = malloc(sizeof(struct list_struct));
	node->node.declaration.type_spec->next = NULL;
	node->node.declaration.type_spec->node = ident;
	ident->node.identifier.id = strdup(typ->node.struct_type_decl.id);
    }
    return node;
}

extern void
ecl_add_param(const char *id, const char *typ, int param_num, 
	      ecl_parse_context context)
{
    sm_list type_list;
    sm_ref node;
    setup_for_string_parse(typ, context->defined_type_count, context->defined_types);
    parsing_type = 1;
    yyerror_count = 0;
    yycontext = context;
    yyparse();
    parsing_type = 0;
    terminate_string_parse();

    if ((yyparse_value == NULL) || (yyerror_count != 0)) {
	return;
    }
    type_list = (sm_list) yyparse_value;

    node = ecl_build_param_node(id, NULL, param_num);
    node->node.declaration.type_spec = type_list;
    ecl_add_decl_to_parse_context(id, node, context);
}

extern void
ecl_set_return_type(char *typ, ecl_parse_context context)
{
    sm_list type_list;
    setup_for_string_parse(typ, context->defined_type_count, context->defined_types);
    parsing_type = 1;
    yyerror_count = 0;
    yycontext = context;
    yyparse();
    parsing_type = 0;
    terminate_string_parse();

    if ((yyparse_value == NULL) || (yyerror_count != 0)) {
	return;
    }
    type_list = (sm_list) yyparse_value;

    context->return_type_list = type_list;
    context->return_cg_type = reduce_type_list(context, type_list);
}

static sm_ref
find_complex_type(sm_ref node, scope_ptr scope)
{
    assert(node->node_type == ecl_identifier);
    return resolve(node->node.identifier.id, scope);
}

extern ecl_parse_context
new_ecl_parse_context()
{
    ecl_parse_context context = malloc(sizeof(struct parse_struct));
    context->decls = NULL;
    context->scope = push_scope(NULL);
    context->defined_type_count = 0;
    context->defined_types = NULL;
    context->error_func = default_error_out;
    context->client_data = NULL;
    context->return_type_list = NULL;
    context->return_cg_type = DR_I;
    return context;
}

extern void
ecl_free_parse_context(ecl_parse_context parse_context)
{
    pop_scope(parse_context->scope);
    if (parse_context->defined_types) {
	free(parse_context->defined_types);
    }
    if (parse_context->decls) {
	ecl_rfree_list(parse_context->decls, NULL);
    }
    free(parse_context);
}

extern void
ecl_assoc_externs(ecl_parse_context context, ecl_extern_list externs)
{
    context->scope->externs = externs;
}

extern void
ecl_add_decl_to_parse_context(const char *name, sm_ref item, ecl_parse_context context)
{
    sm_list *last_ptr = &context->decls;
    sm_list list = context->decls;
    while (list != NULL) {
	last_ptr = &list->next;
	list = list->next;
    }
    *last_ptr = malloc(sizeof(*list));
    (*last_ptr)->next = NULL;
    (*last_ptr)->node = item;
}
