/* Required definitions. */
#ifndef MODULE
#define MODULE
#endif
#ifndef __KERNEL__
#define __KERNEL__
#endif

#define ecalls_module_h 1

/* Various required header files. */
#include <linux/kernel.h>
#include <linux/types.h> 
#include <linux/unistd.h>
#include <linux/types.h>
#include <linux/slab.h>
#include <linux/highmem.h>
#include <asm/segment.h> 
#include <asm/signal.h> 
#include "../include/ecl.h"

#ifndef FALSE
#define FALSE 0
#endif
#ifndef TRUE
#define TRUE 1
#endif

#define MAX_LEN 256 
#define MAX_ECALLS 512
#define ECALL_FILTER 1
#define ECALL_CONT   2
#define THR_POOL   3
#define MAX_THR  255
#define ECALL_ID  1111 
#define MAX_SERVICES 64
#define MAX_HOOKS 128
#define MAX_HOOKS_LEN 16
#define MAX_DESC_STR 80
#define MAX_CLIENTS 64
#define POLL 0
#define EPOLL 1
#define DEFERRED 3
#define IMMEDIATE 2
#define MAX_PROCESSES 512
#define RET_BH 1
#define RET_SCHED 2
#define RET_SIG 4
#define RET_SLOW 8
#define MAX_FUNCS 64

#define KECHO_API 1

static inline unsigned long generic_virt_to_phys(void *address, 
	struct task_struct *process) {
  unsigned long ret;
  pmd_t *pmd;
  pte_t *ptep, pte;
  pgd_t *pgd;

  pgd = pgd_offset(process->mm, (unsigned long) address);
  pmd = pmd_offset(pgd, (unsigned long) address);
  ptep = pte_offset(pmd, (unsigned long)address);
  if (!pte_present(*ptep)) { printk("Not locked!\n"); return 0; }
  pte = *ptep;
#ifdef CONFIG_HIGHMEM
  kmap_high(pte_page(pte));
#endif
  ret = (unsigned long) (page_address(pte_page(pte)));
  ret |= ((unsigned long)address & (PAGE_SIZE - 1));
  return ret;
}

extern int register_ecalls_service(char *service_name, 
	void *fast_handler_func,
	void *slow_handler_func,
	int type);
extern int unregister_ecalls_service(char *service_name);
extern int register_function(char *function_name, void *function_addr);
extern int unregister_function(char *function_name);
extern int internal_do_ecall(int id, int pid);
extern int poll_service_requests(int id, int index);
extern int request_num_clients(int id);
extern int start_polling(int id, int period);
extern int stop_polling(int id);
extern int do_upcall(int id, int pid, unsigned long pointer, long priority1, 
	long priority2, int cpu);
extern unsigned long request_upcall_pointer1(int id, int pid);
extern unsigned long request_upcall_pointer2(int id, int pid);
extern unsigned long request_downcall_pointer1(int id, int pid);
extern unsigned long request_downcall_pointer2(int id, int pid);
extern unsigned long request_downcall_flag(int id, int pid);
extern unsigned long request_downcall_flag2(int id, int pid);
extern unsigned long request_upcall_flag(int id, int pid);
extern unsigned long request_upcall_flag2(int id, int pid);
extern int set_sock_ecalls(struct socket *, int, int, int, void *ptr,int,int);
extern int get_sock_info(void *, int);
extern int set_sock_info(void *, int, int);
extern int internal_start_ecalls(int);
extern int register_hook(char *, char *, char *, void *);
extern int unregister_hook(char *);

struct service_data {
	int service_id;
	char *service_name;
	void *fast_handler_func;
	void *slow_handler_func;
        int type;
        int clients;
	int client_pid[MAX_CLIENTS];
	int poll_period;
        int handler_id[MAX_CLIENTS];
	int signal[MAX_CLIENTS];
	unsigned long memory_ptr_down[MAX_CLIENTS];
	unsigned long memory_ptr_up[MAX_CLIENTS];
	unsigned long memory_ptr_down_phys[MAX_CLIENTS];
	unsigned long memory_ptr_up_phys[MAX_CLIENTS];
	unsigned long flag_down[MAX_CLIENTS];
	unsigned long flag_down_phys[MAX_CLIENTS];
	unsigned long flag_up[MAX_CLIENTS];
	unsigned long flag_up_phys[MAX_CLIENTS];
	unsigned long bitpattern_down[MAX_CLIENTS];
	unsigned long bitpattern_down_phys[MAX_CLIENTS];
	unsigned long bitpattern_up[MAX_CLIENTS];
	unsigned long bitpattern_up_phys[MAX_CLIENTS];
	unsigned long front_down[MAX_CLIENTS];
	unsigned long front_down_phys[MAX_CLIENTS];
	unsigned long front_up[MAX_CLIENTS];
	unsigned long front_up_phys[MAX_CLIENTS];
	unsigned long back_down[MAX_CLIENTS];
	unsigned long back_down_phys[MAX_CLIENTS];
	unsigned long back_up[MAX_CLIENTS];
	unsigned long back_up_phys[MAX_CLIENTS];
	void *handler_function[MAX_CLIENTS];
	void *kernel_function[MAX_CLIENTS];
	int kfunc_id[MAX_CLIENTS];
	struct task_struct *tptr[MAX_CLIENTS];
	ecl_parse_context context[MAX_CLIENTS];
	ecl_code gen_code[MAX_CLIENTS];
	void *upcall_func[MAX_CLIENTS];
};
typedef struct service_data *t_service_data;

struct ecall_down {
	int service_id;
	unsigned long flag;
	unsigned long bitpattern;
	unsigned long front;
	unsigned long back;
	unsigned long data;
	char *service_name;
	int string_len;
	int total_size;
	int size1;
	int size2;
};
typedef struct ecall_down *t_downcall_data;

struct ecall_up {
	int service_id;
	unsigned long flag;
	unsigned long bitpattern;
	unsigned long front;
	unsigned long back;
	unsigned long data;
	char *service_name;
	int string_len;
	int kfct_string_len;
	int total_size;
        int process_id;
	int signal;
	void *handler_function;
	char *kernel_function;
	int size1;
	int size2;
};
typedef struct ecall_up *t_upcall_data;

struct ecall_data {
	int    	proc_id;
	int	ecall_id;
	int	handler_id;
	void   *handler_func;
        int     signal;
	int	flags;
};
typedef struct ecall_data *t_ecall_data;

struct pid_list {
	int pid;
	int i;
	int j;
};

struct upcall_task_struct {
	struct task_struct *tptr;
	struct upcall_task_struct *next;
  	long seconds;
	long microseconds;
};

struct kernel_functions {
 	char identifier[MAX_LEN];
	void *function_ptr;
	int  flag;
};

struct hook_node {
	char hookname[MAX_HOOKS_LEN];
	char description[MAX_DESC_STR];
	char parm_list[MAX_DESC_STR];
	void *callback;
};
