#include "config.h"
#include "stdio.h"
#include "io.h"
#ifdef HAVE_SYS_UIO_H
#include <sys/types.h>
#include <sys/uio.h>
#endif
#ifdef HAVE_MEMORY_H
#include <memory.h>
#endif
#include "assert.h"
#include "io_interface.h"
#include "io_internal.h"
#ifdef HAVE_WCODE_H
#include "wcode.h"
#define static_ctx c 
#define _vrr(x) x
#else
#include "vcode.h"
typedef void *wcode_ctx;
#endif
#include "io_gen.h"

#define REG_DEBUG(x) printf x ;
/* #define REG_DEBUG(x) */

iogen_oprnd
gen_operand(src_reg, offset, size, data_type, aligned, byte_swap)
v_reg_t src_reg;
int offset;
int size;
IOdata_type data_type;
int aligned;
int byte_swap;
{
    iogen_oprnd ret_val;
    ret_val.address = 1;
    ret_val.size = size;
    ret_val.data_type = data_type;
    ret_val.offset = offset;
    ret_val.aligned = aligned;
    ret_val.byte_swap = byte_swap;
    ret_val.vc_reg = src_reg;
    ret_val.vc_reg2 = src_reg;
    return ret_val;
}

void
gen_load(c, src_oprnd)
wcode_ctx c;
iogen_oprnd_ptr src_oprnd;
{
    iogen_oprnd tmp_val;
    tmp_val = gen_fetch(c, 
			src_oprnd->vc_reg, src_oprnd->offset, src_oprnd->size,
			src_oprnd->data_type, src_oprnd->aligned,
			src_oprnd->byte_swap);
    *src_oprnd = tmp_val;
}

void
gen_bswap_load(c, src_oprnd)
wcode_ctx c;
iogen_oprnd_ptr src_oprnd;
{
    iogen_oprnd tmp_val;
    tmp_val = gen_fetch(c, 
			src_oprnd->vc_reg, src_oprnd->offset, src_oprnd->size,
			src_oprnd->data_type, src_oprnd->aligned,
			src_oprnd->byte_swap);
    *src_oprnd = tmp_val;
}

#ifdef v_ldbsi
    /* have byte swap load extension */
iogen_oprnd
gen_bswap_fetch(src_reg, offset, size, data_type, aligned)
v_reg_t src_reg;
int offset;
int size;
IOdata_type data_type;
int aligned;
{
    iogen_oprnd ret_val;
    ret_val.address = 0;
    ret_val.size = size;
    ret_val.data_type = data_type;
    ret_val.offset = 0;
    ret_val.aligned = 0;
    ret_val.byte_swap = 0;
    switch (data_type) {
    case unknown_type:
	assert(FALSE);
	break;
    case integer_type:
    case enumeration_type:
	switch (size) {
	case 1:		/* sizeof char */
	    if (!v_getreg(&ret_val.vc_reg, V_C, V_TEMP))
		v_fatal("gen fetch out of registers \n");
	    REG_DEBUG(("get %d in gen_Fetch\n", _vrr(ret_val.vc_reg)));
	    v_ldbsc(ret_val.vc_reg, src_reg, offset);
	    break;
	case 2:		/* sizeof short */
	    if (!v_getreg(&ret_val.vc_reg, V_S, V_TEMP))
		v_fatal("gen fetch out of registers \n");
	    REG_DEBUG(("get %d in gen_Fetch\n", _vrr(ret_val.vc_reg)));
	    v_ldbssi(ret_val.vc_reg, src_reg, offset);
	    break;
	case 4:		/* sizeof int */
	    if (!v_getreg(&ret_val.vc_reg, V_I, V_TEMP))
		v_fatal("gen fetch out of registers A\n");
	    REG_DEBUG(("get %d in gen_Fetch\n", _vrr(ret_val.vc_reg)));
	    v_ldbsii(ret_val.vc_reg, src_reg, offset);
	    break;
#if SIZEOF_LONG != 4
	case SIZEOF_LONG:
	    if (!v_getreg(&ret_val.vc_reg, V_L, V_TEMP))
		v_fatal("gen fetch out of registers \n");
	    REG_DEBUG(("get %d in gen_Fetch\n", _vrr(ret_val.vc_reg)));
	    if (!aligned || ((offset & 0x7) != 0)) {
		/* misaligned */
		if ((offset & 0x3) == 0) {
		    /* 4 byte aligned */
		    v_reg_t high_reg;
		    if (!v_getreg(&high_reg, V_L, V_TEMP))
			v_fatal("gen fetch out of registers \n");
#ifdef WORDS_BIGENDIAN
		    /* vc_reg2 holds high value */
		    v_ldbsui(ret_val.vc_reg, src_reg, offset);
		    v_ldbsii(high_reg, src_reg, offset + 4);
#else
		    v_ldbsii(high_reg, src_reg, offset);
		    v_ldbsui(ret_val.vc_reg, src_reg, offset + 4);
#endif
		    v_lshli(high_reg, high_reg, 32);
		    v_orl(ret_val.vc_reg, high_reg, ret_val.vc_reg);
		    v_putreg(high_reg, V_L);
		} else {
		    assert(FALSE);
		}
	    } else {
		v_ldbsli(ret_val.vc_reg, src_reg, offset);
	    }
	    break;
#else
	case 8:
	    /* simulate with double reg */
	    if (!v_getreg(&ret_val.vc_reg, V_I, V_TEMP) ||
		!v_getreg(&ret_val.vc_reg2, V_I, V_TEMP))
		v_fatal("gen fetch out of registers B \n");
	    REG_DEBUG(("get %d in gen_Fetch\n", _vrr(ret_val.vc_reg)));
	    REG_DEBUG(("get %d in gen_Fetch\n", _vrr(ret_val.vc_reg2)));

#ifdef WORDS_BIGENDIAN
	    /* vc_reg2 holds high value */
	    v_ldbsii(ret_val.vc_reg, src_reg, offset);
	    v_ldbsii(ret_val.vc_reg2, src_reg, offset + 4);
#else
	    v_ldbsii(ret_val.vc_reg2, src_reg, offset);
	    v_ldbsii(ret_val.vc_reg, src_reg, offset + 4);
#endif
	    break;
#endif
	}
	break;
    case unsigned_type:
    case char_type:
	switch (size) {
	case 1:		/* sizeof char */
	    if (!v_getreg(&ret_val.vc_reg, V_UC, V_TEMP))
		v_fatal("gen fetch out of registers \n");
	    REG_DEBUG(("get %d in gen_Fetch\n", _vrr(ret_val.vc_reg)));
	    v_ldbsuci(ret_val.vc_reg, src_reg, offset);
	    break;
	case 2:		/* sizeof short */
	    if (!v_getreg(&ret_val.vc_reg, V_US, V_TEMP))
		v_fatal("gen fetch out of registers \n");
	    REG_DEBUG(("get %d in gen_Fetch\n", _vrr(ret_val.vc_reg)));
	    v_ldbsusi(ret_val.vc_reg, src_reg, offset);
	    break;
	case 4:		/* sizeof int */
	    if (!v_getreg(&ret_val.vc_reg, V_U, V_TEMP))
		v_fatal("gen fetch out of registers \n");
	    REG_DEBUG(("get %d in gen_Fetch\n", _vrr(ret_val.vc_reg)));
	    v_ldbsui(ret_val.vc_reg, src_reg, offset);
	    break;
#if SIZEOF_LONG != 4
	case SIZEOF_LONG:
	    if (!v_getreg(&ret_val.vc_reg, V_UL, V_TEMP))
		v_fatal("gen fetch out of registers \n");
	    REG_DEBUG(("get %d in gen_Fetch\n", _vrr(ret_val.vc_reg)));
	    v_ldbsuli(ret_val.vc_reg, src_reg, offset);
	    break;
#else
	case 8:
	    /* simulate with double reg */
	    if (!v_getreg(&ret_val.vc_reg2, V_U, V_TEMP) ||
		!v_getreg(&ret_val.vc_reg, V_U, V_TEMP))
		v_fatal("gen fetch out of registers \n");
	    REG_DEBUG(("get %d in gen_Fetch\n", _vrr(ret_val.vc_reg)));
	    REG_DEBUG(("get %d in gen_Fetch\n", _vrr(ret_val.vc_reg2)));
	    /* vc_reg2 holds high value */
#ifdef WORDS_BIGENDIAN
	    v_ldbsui(ret_val.vc_reg, src_reg, offset);
	    v_ldbsui(ret_val.vc_reg2, src_reg, offset + 4);
#else
	    v_ldbsui(ret_val.vc_reg2, src_reg, offset);
	    v_ldbsui(ret_val.vc_reg, src_reg, offset + 4);
#endif
	    break;
#endif
	}
	break;
    case float_type:
    case string_type:
	assert(FALSE);
    }
    return ret_val;
}
#endif

iogen_oprnd
gen_fetch(c, src_reg, offset, size, data_type, aligned, byte_swap)
wcode_ctx c;
v_reg_t src_reg;
int offset;
int size;
IOdata_type data_type;
int aligned;
int byte_swap;
{
    iogen_oprnd ret_val;
#ifdef v_ldbsi
    /* have byte swap load extension */
    if (byte_swap && (data_type != float_type)) {
	return gen_bswap_fetch(src_reg, offset, size, data_type, aligned);
    }
#endif
    ret_val.address = 0;
    ret_val.size = size;
    ret_val.data_type = data_type;
    ret_val.offset = 0;
    ret_val.aligned = 0;
    ret_val.byte_swap = 0;
    switch (data_type) {
    case unknown_type:
	assert(FALSE);
	break;
    case integer_type:
    case enumeration_type:
	switch (size) {
	case 1:		/* sizeof char */
	    if (!v_getreg(&ret_val.vc_reg, V_C, V_TEMP))
		v_fatal("gen fetch out of registers \n");
	    REG_DEBUG(("get %d in gen_Fetch\n", _vrr(ret_val.vc_reg)));
	    v_ldci(ret_val.vc_reg, src_reg, offset);
	    break;
	case 2:		/* sizeof short */
	    if (!v_getreg(&ret_val.vc_reg, V_S, V_TEMP))
		v_fatal("gen fetch out of registers \n");
	    REG_DEBUG(("get %d in gen_Fetch\n", _vrr(ret_val.vc_reg)));
	    v_ldsi(ret_val.vc_reg, src_reg, offset);
	    break;
	case 4:		/* sizeof int */
	    if (!v_getreg(&ret_val.vc_reg, V_I, V_TEMP))
		v_fatal("gen fetch out of registers C\n");
	    REG_DEBUG(("get %d in gen_Fetch\n", _vrr(ret_val.vc_reg)));
	    v_ldii(ret_val.vc_reg, src_reg, offset);
	    break;
#if SIZEOF_LONG != 4
	case SIZEOF_LONG:
	    if (!v_getreg(&ret_val.vc_reg, V_L, V_TEMP))
		v_fatal("gen fetch out of registers \n");
	    REG_DEBUG(("get %d in gen_Fetch\n", _vrr(ret_val.vc_reg)));
	    v_ldli(ret_val.vc_reg, src_reg, offset);
	    break;
#else
	case 8:
	    /* simulate with double reg */
	    if (!v_getreg(&ret_val.vc_reg, V_I, V_TEMP) ||
		!v_getreg(&ret_val.vc_reg2, V_I, V_TEMP))
		v_fatal("gen fetch out of registers D \n");
	    REG_DEBUG(("get %d in gen_Fetch\n", _vrr(ret_val.vc_reg)));
	    REG_DEBUG(("get %d in gen_Fetch\n", _vrr(ret_val.vc_reg2)));
	    /* vc_reg2 holds high value */
#ifdef WORDS_BIGENDIAN
	    v_ldii(ret_val.vc_reg2, src_reg, offset);
	    v_ldii(ret_val.vc_reg, src_reg, offset + 4);
#else
	    v_ldii(ret_val.vc_reg, src_reg, offset);
	    v_ldii(ret_val.vc_reg2, src_reg, offset + 4);
#endif
	    break;
#endif
	}
	break;
    case unsigned_type:
    case char_type:
	switch (size) {
	case 1:		/* sizeof char */
	    if (!v_getreg(&ret_val.vc_reg, V_UC, V_TEMP))
		v_fatal("gen fetch out of registers \n");
	    REG_DEBUG(("get %d in gen_Fetch\n", _vrr(ret_val.vc_reg)));
	    v_lduci(ret_val.vc_reg, src_reg, offset);
	    break;
	case 2:		/* sizeof short */
	    if (!v_getreg(&ret_val.vc_reg, V_US, V_TEMP))
		v_fatal("gen fetch out of registers \n");
	    REG_DEBUG(("get %d in gen_Fetch\n", _vrr(ret_val.vc_reg)));
	    v_ldusi(ret_val.vc_reg, src_reg, offset);
	    break;
	case 4:		/* sizeof int */
	    if (!v_getreg(&ret_val.vc_reg, V_U, V_TEMP))
		v_fatal("gen fetch out of registers \n");
	    REG_DEBUG(("get %d in gen_Fetch\n", _vrr(ret_val.vc_reg)));
	    v_ldui(ret_val.vc_reg, src_reg, offset);
	    break;
#if SIZEOF_LONG != 4
	case SIZEOF_LONG:
	    if (!v_getreg(&ret_val.vc_reg, V_UL, V_TEMP))
		v_fatal("gen fetch out of registers \n");
	    REG_DEBUG(("get %d in gen_Fetch\n", _vrr(ret_val.vc_reg)));
	    v_lduli(ret_val.vc_reg, src_reg, offset);
	    break;
#else
	case 8:
	    /* simulate with double reg */
	    if (!v_getreg(&ret_val.vc_reg2, V_U, V_TEMP) ||
		!v_getreg(&ret_val.vc_reg, V_U, V_TEMP))
		v_fatal("gen fetch out of registers \n");
	    REG_DEBUG(("get %d in gen_Fetch\n", _vrr(ret_val.vc_reg)));
	    REG_DEBUG(("get %d in gen_Fetch\n", _vrr(ret_val.vc_reg2)));
	    /* vc_reg2 holds high value */
#ifdef WORDS_BIGENDIAN
	    v_ldui(ret_val.vc_reg2, src_reg, offset);
	    v_ldui(ret_val.vc_reg, src_reg, offset + 4);
#else
	    v_ldui(ret_val.vc_reg, src_reg, offset);
	    v_ldui(ret_val.vc_reg2, src_reg, offset + 4);
#endif
	    break;
#endif
	}
	break;
    case float_type:
	if (byte_swap) {
	    /* best to byte_swap floating point things as we're loading */
	    ret_val.offset = offset;
	    ret_val.aligned = aligned;
	    ret_val.address = TRUE;
	    ret_val.vc_reg = src_reg;
	    ret_val.byte_swap = TRUE;
	    gen_byte_swap(c, &ret_val);
	    byte_swap = FALSE;	/* taken care of */
	} else {
	    switch (size) {
	    case SIZEOF_FLOAT:	/* sizeof char */
		if (!v_getreg(&ret_val.vc_reg, V_F, V_TEMP))
		    v_fatal("gen fetch out of registers \n");
		REG_DEBUG(("get %d in gen_Fetch\n", _vrr(ret_val.vc_reg)));
		v_ldfi(ret_val.vc_reg, src_reg, offset);
		break;
	    case SIZEOF_DOUBLE:	/* sizeof short */
		if (!v_getreg(&ret_val.vc_reg, V_D, V_TEMP))
		    v_fatal("gen fetch out of registers \n");
		REG_DEBUG(("get %d in gen_Fetch\n", _vrr(ret_val.vc_reg)));
		v_lddi(ret_val.vc_reg, src_reg, offset);
		break;
	    }
	}
	break;
    case string_type:
	assert(FALSE);
    }
    if (byte_swap) {
	gen_byte_swap(c, &ret_val);
    }
    return ret_val;
}

/* v_bswap8 exchanges adjacent bytes in a register */
#define v_bswap8(rd, rs, tmp) do {		\
    v_rshli(tmp, rs, 8);  /* odd bytes to even in tmp */\
    v_andli(tmp, tmp, 0x00ff00ff);  /* clear odd bytes in tmp */\
    v_lshli(rd, rs, 8);  /* even bytes to odd dest */\
    v_anduli(rd, rd, (long)(unsigned)0xff00ff00);  /* clear even bytes in  dest */\
    v_orl(rd, rd, tmp);  /* combine in dest */\
} while(0)

/* v_bswap16 exchanges adjacent 16-bit values in a register */
#define v_bswap16(rd, rs, tmp) do {		\
    v_rshii(tmp, rs, 16);  /* top 16 to bottom 16 in tmp */\
    v_andli(tmp, tmp, 0x0000ffff);  /* clear top 16 in tmp */\
    v_lshli(rd, rs, 16);  /* bottom 16 to top in dest */\
    v_anduli(rd, rd, (long)(unsigned)0xffff0000);  /* clear bottom bytes in  dest */\
    v_orl(rd, rd, tmp);  /* combine in dest */\
    v_bswap8(rd, rd, tmp);\
} while(0)

#define v_bswaps(rd, rs) do {			\
    v_reg_t tmp;				\
    if(v_getreg(&tmp, V_L, V_TEMP) == 0)	\
	v_fatal("bswaps: out of registers\n");	\
    v_bswap8(rd, rs, tmp);			\
    v_putreg(tmp, V_L);				\
} while(0)

#define v_bswapi(rd,rs) do {			\
    v_reg_t tmp;				\
    if(v_getreg(&tmp, V_L, V_TEMP) == 0)	\
	v_fatal("bswaps: out of registers\n");	\
    v_bswap16(rd, rs, tmp);			\
    v_putreg(tmp, V_L);				\
} while(0)

#if sizeof_long == 8
/* v_bswap32 exchanges adjacent 32-bit values in a register */
#define v_bswap32(rd,rs,tmp) do {		\
    v_rshii(tmp, rs, 32);  /* top 32 to bottom 32 in tmp */\
    v_andli(tmp, tmp, 0x00000000ffffffff);  /* clear top 32 in tmp */\
    v_lshii(rd, rs, 32);  /* bottom 32 to top in dest */\
    v_anduli(rd, rd, 0xffffffff00000000);  /* clear bottom 32 in dest */\
    v_ori(rd, rd, tmp);  /* combine in dest */\
    v_bswap16(rd, rd, tmp);\
} while(0)

#define v_bswapi(rd,rs) do {			\
    v_reg_t tmp;				\
    if(v_getreg(&tmp, V_L, V_TEMP) == 0)	\
	v_fatal("bswaps: out of registers\n");	\
    v_bswap32(rd, rs, tmp);			\
    v_putreg(tmp, V_L);				\
} while(0)

#else
#define v_bswapl(rd, rs) v_bswapi(rd, rs)
#endif

void
gen_byte_swap(c, src_oprnd)
wcode_ctx c;
iogen_oprnd_ptr src_oprnd;
{
    iogen_oprnd swap_oprnd;
    if (src_oprnd->address) {
	if (src_oprnd->data_type == float_type) {
	    swap_oprnd = *src_oprnd;
	    swap_oprnd.data_type = integer_type;
	    gen_load(c, &swap_oprnd);
#ifdef v_ldbsi
	    /* gen_load did everything for us */
	    goto do_float_load_store;
#endif
	} else {
	    swap_oprnd = *src_oprnd;
	    gen_load(c, &swap_oprnd);
#ifdef v_ldbsi
	    /* gen_load did everything for us */
	    return;
#endif
	}
    } else {
	assert(src_oprnd->data_type != float_type);
	swap_oprnd = *src_oprnd;
    }
    /* no byte swap load extension, do it manually */
    switch (swap_oprnd.size) {
    case 1:
	break;
    case 2:
	v_bswaps(swap_oprnd.vc_reg, swap_oprnd.vc_reg);
	break;
    case 4:
	v_bswapi(swap_oprnd.vc_reg, swap_oprnd.vc_reg);
	break;
    case 8:
	if (sizeof(long) == 4) {
	    /* swap top and bottom */
	    v_reg_t tmp_reg = swap_oprnd.vc_reg;
	    swap_oprnd.vc_reg = swap_oprnd.vc_reg2;
	    swap_oprnd.vc_reg2 = tmp_reg;
	    /* then byte swap each */
	    v_bswapi(swap_oprnd.vc_reg, swap_oprnd.vc_reg);
	    v_bswapi(swap_oprnd.vc_reg2, swap_oprnd.vc_reg2);
	} else {
	    v_bswapl(swap_oprnd.vc_reg, swap_oprnd.vc_reg);
	}
	break;
    default:
	assert(FALSE);
    }
#ifdef v_ldbsi
 do_float_load_store:
#endif
    if (src_oprnd->address) {
	if (src_oprnd->data_type == float_type) {
	    /* lose type info by storing to memory and retrieving */
	    int tmp_base = v_localb(src_oprnd->size);
	    gen_store(c, swap_oprnd, v_lp, tmp_base, src_oprnd->size,
		      integer_type, TRUE);
	    free_oprnd(c, swap_oprnd);
	    swap_oprnd = gen_fetch(c, v_lp, tmp_base, src_oprnd->size,
				   src_oprnd->data_type, TRUE, FALSE);
	}
    }
    *src_oprnd = swap_oprnd;
}

void
gen_store(c, src, dest_reg, offset, size, data_type, aligned)
wcode_ctx c;
iogen_oprnd src;
v_reg_t dest_reg;
int offset;
int size;
IOdata_type data_type;
int aligned;
{
    assert(src.size == size);

    switch (data_type) {
    case unknown_type:
	assert(FALSE);
	break;
    case integer_type:
    case enumeration_type:
	switch (size) {
	case 1:		/* sizeof char */
	    v_stci(src.vc_reg, dest_reg, offset);
	    break;
	case 2:		/* sizeof short */
	    v_stsi(src.vc_reg, dest_reg, offset);
	    break;
	case 4:		/* sizeof int */
	    v_stii(src.vc_reg, dest_reg, offset);
	    break;
#if SIZEOF_LONG != 4
	case SIZEOF_LONG:
	    v_stli(src.vc_reg, dest_reg, offset);
	    break;
#else
	case 8:
	    /* simulate with double reg */
	    /* vc_reg2 holds high value */
#ifdef WORDS_BIGENDIAN
	    v_stii(src.vc_reg2, dest_reg, offset);
	    v_stii(src.vc_reg, dest_reg, offset + 4);
#else
	    v_stii(src.vc_reg, dest_reg, offset);
	    v_stii(src.vc_reg2, dest_reg, offset + 4);
#endif
	    break;
#endif
	}
	break;
    case unsigned_type:
    case char_type:
	switch (size) {
	case 1:		/* sizeof char */
	    v_stuci(src.vc_reg, dest_reg, offset);
	    break;
	case 2:		/* sizeof short */
	    v_stusi(src.vc_reg, dest_reg, offset);
	    break;
	case 4:		/* sizeof int */
	    v_stui(src.vc_reg, dest_reg, offset);
	    break;
#if SIZEOF_LONG != 4
	case SIZEOF_LONG:
	    v_stuli(src.vc_reg, dest_reg, offset);
	    break;
#else
	case 8:
	    /* simulate with double reg */
	    /* vc_reg2 holds high value */
#ifdef WORDS_BIGENDIAN
	    v_stui(src.vc_reg2, dest_reg, offset);
	    v_stui(src.vc_reg, dest_reg, offset + 4);
#else
	    v_stui(src.vc_reg, dest_reg, offset);
	    v_stui(src.vc_reg2, dest_reg, offset + 4);
#endif
	    break;
#endif
	}
	break;
    case float_type:
	switch (size) {
	case SIZEOF_FLOAT:	/* sizeof char */
	    v_stfi(src.vc_reg, dest_reg, offset);
	    break;
	case SIZEOF_DOUBLE:	/* sizeof short */
	    v_stdi(src.vc_reg, dest_reg, offset);
	    break;
	}
	break;
    case string_type:

	break;
    }
}

void
gen_memcpy(c, src, src_offset, dest, dest_offset, size)
wcode_ctx c;
v_reg_t src;
int src_offset;
v_reg_t dest;
int dest_offset;
v_reg_t size;
{
    v_reg_t final_src, final_dest;
    if (src_offset != 0) {
	if (v_getreg(&final_src, V_P, V_TEMP) == 0)
	    v_fatal("gen memcpy convert out of registers \n");
	v_addpi(final_src, src, src_offset);
    } else {
	final_src = src;
    }
    if (dest_offset != 0) {
	if (v_getreg(&final_dest, V_P, V_TEMP) == 0)
	    v_fatal("gen memcpy convert out of registers \n");
	v_addpi(final_dest, dest, dest_offset);
    } else {
	final_dest = dest;
    }
#ifdef HAVE_WCODE_H
    w_scalli(c, (v_iptr) memcpy, "%p%p%i", final_dest, final_src, size);
#else
    v_scalli((v_iptr) memcpy, "%p%p%i", final_dest, final_src, size);
#endif
    if (src_offset != 0) {
	v_putreg(final_src, V_P);
    }
    if (dest_offset != 0) {
	v_putreg(final_dest, V_P);
    }
}

void
free_oprnd(c, oprnd)
wcode_ctx c;
iogen_oprnd oprnd;
{
    REG_DEBUG(("put %d in free\n", _vrr(oprnd.vc_reg)));
    switch (oprnd.data_type) {
    case unknown_type:
	assert(FALSE);
	break;
    case integer_type:
    case enumeration_type:
	switch (oprnd.size) {
	case 1:		/* sizeof char */
	    v_putreg(oprnd.vc_reg, V_C);
	    break;
	case 2:		/* sizeof short */
	    v_putreg(oprnd.vc_reg, V_S);
	    break;
	case 4:		/* sizeof int */
	    v_putreg(oprnd.vc_reg, V_I);
	    break;
#if SIZEOF_LONG != 4
	case SIZEOF_LONG:
	    v_putreg(oprnd.vc_reg, V_L);
	    break;
#else
	case 8:
	    /* simulate with double reg */
	    v_putreg(oprnd.vc_reg, V_I);
	    v_putreg(oprnd.vc_reg2, V_I);
	    REG_DEBUG(("put %d in free\n", _vrr(oprnd.vc_reg2)));
	    break;
#endif
	}
	break;
    case unsigned_type:
    case char_type:
	switch (oprnd.size) {
	case 1:		/* sizeof char */
	    v_putreg(oprnd.vc_reg, V_UC);
	    break;
	case 2:		/* sizeof short */
	    v_putreg(oprnd.vc_reg, V_US);
	    break;
	case 4:		/* sizeof int */
	    v_putreg(oprnd.vc_reg, V_U);
	    break;
#if SIZEOF_LONG != 4
	case SIZEOF_LONG:
	    v_putreg(oprnd.vc_reg, V_UL);
	    break;
#else
	case 8:
	    /* simulate with double reg */
	    v_putreg(oprnd.vc_reg, V_U);
	    v_putreg(oprnd.vc_reg2, V_U);
	    /* vc_reg2 holds high value */
	    REG_DEBUG(("put %d in free\n", _vrr(oprnd.vc_reg2)));
	    break;
#endif
	}
	break;
    case float_type:
	switch (oprnd.size) {
	case SIZEOF_FLOAT:	/* sizeof char */
	    v_putreg(oprnd.vc_reg, V_F);
	    break;
	case SIZEOF_DOUBLE:	/* sizeof short */
	    v_putreg(oprnd.vc_reg, V_D);
	    break;
	}
	break;
    case string_type:
	break;
    }
}

iogen_oprnd
gen_type_conversion(c, src_oprnd, data_type)
wcode_ctx c;
iogen_oprnd src_oprnd;
IOdata_type data_type;
{
    iogen_oprnd result_oprnd = src_oprnd;
    v_reg_t at;  /* temporary */
    result_oprnd.data_type = data_type;
    switch (data_type) {
    case unknown_type:
    case char_type:
	assert(FALSE);
	break;
    case integer_type:
	result_oprnd.size = sizeof(long);
	if (!v_getreg(&result_oprnd.vc_reg, V_L, V_TEMP))
	    v_fatal("gen type convert out of registers \n");
	REG_DEBUG(("get %d in type_convert\n", _vrr(result_oprnd.vc_reg)));
	switch (src_oprnd.data_type) {
	case integer_type:
	case unknown_type:
	    assert(FALSE);
	    break;
	case enumeration_type:
	case unsigned_type:
	case char_type:
	    switch (src_oprnd.size) {
	    case 1:
		v_cvc2l(result_oprnd.vc_reg, src_oprnd.vc_reg);
		break;
	    case 2:
		v_cvus2l(result_oprnd.vc_reg, src_oprnd.vc_reg);
		break;
	    case 4:
		v_cvu2l(result_oprnd.vc_reg, src_oprnd.vc_reg);
		break;
	    case 8:
#if SIZEOF_LONG == 8
		v_cvul2l(result_oprnd.vc_reg, src_oprnd.vc_reg);
#else
		result_oprnd.size = 8;
		if (!v_getreg(&result_oprnd.vc_reg2, V_L, V_TEMP))
		    v_fatal("gen type convert out of registers \n");
		REG_DEBUG(("get %d in type_convert\n", _vrr(result_oprnd.vc_reg2)));
		v_cvul2l(result_oprnd.vc_reg2, src_oprnd.vc_reg2);
		v_movul(result_oprnd.vc_reg, src_oprnd.vc_reg);
#endif
		break;
	    }
	    break;
	case float_type:
	    switch (src_oprnd.size) {
	    case SIZEOF_FLOAT:
		v_cvf2l(result_oprnd.vc_reg, src_oprnd.vc_reg);
		break;
	    case SIZEOF_DOUBLE:
		v_cvd2l(result_oprnd.vc_reg, src_oprnd.vc_reg);
		break;
	    default:
		assert(FALSE);
	    }
	    break;
	default:
	    assert(FALSE);
	    break;
	}
	break;
    case enumeration_type:
    case unsigned_type:
	result_oprnd.size = sizeof(unsigned long);
	if (!v_getreg(&result_oprnd.vc_reg, V_UL, V_TEMP))
	    v_fatal("gen type convert out of registers \n");
	REG_DEBUG(("get %d in type_convert\n", _vrr(result_oprnd.vc_reg)));
	switch (src_oprnd.data_type) {
	case enumeration_type:
	case unsigned_type:
	case unknown_type:
	    assert(FALSE);
	    break;
	case char_type:
	case integer_type:
	    switch (src_oprnd.size) {
	    case 1:
		v_cvc2ul(result_oprnd.vc_reg, src_oprnd.vc_reg);
		break;
	    case 2:
		v_cvs2ul(result_oprnd.vc_reg, src_oprnd.vc_reg);
		break;
	    case 4:
		v_cvi2ul(result_oprnd.vc_reg, src_oprnd.vc_reg);
		break;
	    case 8:
#if SIZEOF_LONG == 8
		v_cvl2ul(result_oprnd.vc_reg, src_oprnd.vc_reg);
#else
		result_oprnd.size = 8;
		if (!v_getreg(&result_oprnd.vc_reg2, V_UL, V_TEMP))
		    v_fatal("gen type convert out of registers \n");
		REG_DEBUG(("get %d in type_convert\n", _vrr(result_oprnd.vc_reg2)));
		v_cvl2ul(result_oprnd.vc_reg2, src_oprnd.vc_reg2);
		v_movul(result_oprnd.vc_reg, src_oprnd.vc_reg);
#endif
		v_cvul2l(result_oprnd.vc_reg, src_oprnd.vc_reg);
		break;
	    }
	    break;
	case float_type:
	    switch (src_oprnd.size) {
	    case SIZEOF_FLOAT:
		v_cvf2ul(result_oprnd.vc_reg, src_oprnd.vc_reg);
		break;
	    case SIZEOF_DOUBLE:
		v_cvd2ul(result_oprnd.vc_reg, src_oprnd.vc_reg);
		break;
	    default:
		assert(FALSE);
	    }
	    break;
	default:
	    assert(FALSE);
	}
	break;
    case float_type:
	result_oprnd.size = sizeof(double);
	if (!v_getreg(&result_oprnd.vc_reg, V_D, V_TEMP))
	    v_fatal("gen type convert out of registers \n");
	REG_DEBUG(("get %d in type_convert\n", _vrr(result_oprnd.vc_reg)));
	switch (src_oprnd.data_type) {
	case enumeration_type:
	case unsigned_type:
	    switch (src_oprnd.size) {
	    case 1:
		if (!v_getreg(&at, V_L, V_TEMP))
		    v_fatal("gen type convert2 out of registers \n");
		v_cvc2l(at, src_oprnd.vc_reg);
		v_cvl2d(result_oprnd.vc_reg, at);
		v_putreg(at, V_L);
		break;
	    case 2:
		if (!v_getreg(&at, V_L, V_TEMP))
		    v_fatal("gen type convert2 out of registers \n");
		v_cvs2l(at, src_oprnd.vc_reg);
		v_cvl2d(result_oprnd.vc_reg, at);
		v_putreg(at, V_L);
		break;
	    case 4:
		if (!v_getreg(&at, V_L, V_TEMP))
		    v_fatal("gen type convert2 out of registers \n");
		v_cvi2l(at, src_oprnd.vc_reg);
		v_cvl2d(result_oprnd.vc_reg, at);
		v_putreg(at, V_L);
		break;
	    case 8:
#if SIZEOF_LONG == 8
		v_cvl2d(result_oprnd.vc_reg, src_oprnd.vc_reg);
#else
		{
		    v_reg_t dat; /* temporary */
		    if (!v_getreg(&dat, V_D, V_TEMP))
			v_fatal("gen type convert2 out of registers \n");
		    v_cvu2d(dat, src_oprnd.vc_reg);
		    v_setd(result_oprnd.vc_reg, 4294967296.0);
		    v_muld(result_oprnd.vc_reg, result_oprnd.vc_reg, dat);
		    v_cvi2d(dat, src_oprnd.vc_reg);
		    v_addd(result_oprnd.vc_reg, result_oprnd.vc_reg, dat);
		    v_putreg(dat, V_D);
		}
#endif
		break;
	    }
	    break;
	case char_type:
	case integer_type:
	    switch (src_oprnd.size) {
	    case 1:
		if (!v_getreg(&at, V_L, V_TEMP))
		    v_fatal("gen type convert2 out of registers \n");
		v_cvc2l(at, src_oprnd.vc_reg);
		v_cvl2d(result_oprnd.vc_reg, at);
		v_putreg(at, V_L);
		break;
	    case 2:
		if (!v_getreg(&at, V_L, V_TEMP))
		    v_fatal("gen type convert2 out of registers \n");
		v_cvc2l(at, src_oprnd.vc_reg);
		v_cvl2d(result_oprnd.vc_reg, at);
		v_putreg(at, V_L);
		break;
	    case 4:
		v_cvi2d(result_oprnd.vc_reg, src_oprnd.vc_reg);
		break;
	    case 8:
#if SIZEOF_LONG == 8
		v_cvl2d(result_oprnd.vc_reg, src_oprnd.vc_reg);
#else
		{
		    v_reg_t dat; /* temporary */
		    if (!v_getreg(&dat, V_D, V_TEMP))
			v_fatal("gen type convert2 out of registers \n");
		    v_cvu2d(dat, src_oprnd.vc_reg);
		    v_setd(result_oprnd.vc_reg, 4294967296.0);
		    v_muld(result_oprnd.vc_reg, result_oprnd.vc_reg, dat);
		    v_cvi2d(dat, src_oprnd.vc_reg);
		    v_addd(result_oprnd.vc_reg, result_oprnd.vc_reg, dat);
		    v_putreg(dat, V_D);
		}
#endif
		break;
	    }
	    break;
	case float_type:
	    assert(FALSE);
	    break;
	case unknown_type:
	default:
	    assert(FALSE);
	    break;
	}
	break;

    case string_type:
	assert(FALSE);
	break;
    }
    return result_oprnd;
}

iogen_oprnd
gen_size_conversion(c, src_oprnd, size)
wcode_ctx c;
iogen_oprnd src_oprnd;
int size;
{
    iogen_oprnd result_oprnd = src_oprnd;
    v_reg_t at;  /* temporary */
    result_oprnd.size = size;
    switch (src_oprnd.data_type) {
    case integer_type:
	switch (size) {
	case 1:
	    if (!v_getreg(&result_oprnd.vc_reg, V_C, V_TEMP))
		v_fatal("gen size convert out of registers \n");
	    REG_DEBUG(("get %d in size convert\n", _vrr(result_oprnd.vc_reg)));
	    switch (src_oprnd.size) {
	    case 2:
		v_cvs2l(result_oprnd.vc_reg, src_oprnd.vc_reg);
		v_cvl2c(result_oprnd.vc_reg, src_oprnd.vc_reg);
		break;
	    case 4:
		v_cvi2c(result_oprnd.vc_reg, src_oprnd.vc_reg);
		break;
	    case 8:
		v_cvl2c(result_oprnd.vc_reg, src_oprnd.vc_reg);
		break;
	    }
	    break;
	case 2:
	    if (!v_getreg(&result_oprnd.vc_reg, V_S, V_TEMP))
		v_fatal("gen size convert out of registers \n");
	    REG_DEBUG(("get %d in size convert\n", _vrr(result_oprnd.vc_reg)));
	    switch (src_oprnd.size) {
	    case 1:
		if (!v_getreg(&at, V_L, V_TEMP))
		    v_fatal("gen type convert2 out of registers \n");
		v_cvc2l(at, src_oprnd.vc_reg);
		v_cvl2s(result_oprnd.vc_reg, at);
		v_putreg(at, V_L);
		break;
	    case 4:
		v_cvi2s(result_oprnd.vc_reg, src_oprnd.vc_reg);
		break;
	    case 8:
		v_cvl2s(result_oprnd.vc_reg, src_oprnd.vc_reg);
		break;
	    }
	    break;
	case 4:
	    if (!v_getreg(&result_oprnd.vc_reg, V_I, V_TEMP))
		v_fatal("gen size convert out of registers E\n");
	    REG_DEBUG(("get %d in size convert\n", _vrr(result_oprnd.vc_reg)));
	    switch (src_oprnd.size) {
	    case 1:
		v_cvc2i(result_oprnd.vc_reg, src_oprnd.vc_reg);
		break;
	    case 2:
		v_cvs2i(result_oprnd.vc_reg, src_oprnd.vc_reg);
		break;
	    case 8:
		v_cvl2i(result_oprnd.vc_reg, src_oprnd.vc_reg);
		break;
	    }
	    break;
	case 8:
#if SIZEOF_LONG == 8
	    if (!v_getreg(&result_oprnd.vc_reg, V_L, V_TEMP))
		v_fatal("gen size convert out of registers \n");
	    REG_DEBUG(("get %d in size convert\n", _vrr(result_oprnd.vc_reg)));
	    switch (src_oprnd.size) {
	    case 1:
		v_cvc2l(result_oprnd.vc_reg, src_oprnd.vc_reg);
		break;
	    case 2:
		v_cvs2l(result_oprnd.vc_reg, src_oprnd.vc_reg);
		break;
	    case 4:
		v_cvi2l(result_oprnd.vc_reg, src_oprnd.vc_reg);
		break;
	    }
	    break;
#else
	    if (!v_getreg(&result_oprnd.vc_reg, V_L, V_TEMP) ||
		(!v_getreg(&result_oprnd.vc_reg2, V_L, V_TEMP)))
		v_fatal("gen size convert out of registers \n");
	    REG_DEBUG(("get %d in size convert\n", _vrr(result_oprnd.vc_reg)));
	    REG_DEBUG(("get %d in size convert\n", _vrr(result_oprnd.vc_reg2)));
	    v_setl(result_oprnd.vc_reg2, 0);
	    switch (src_oprnd.size) {
	    case 1:
		v_cvc2l(result_oprnd.vc_reg, src_oprnd.vc_reg);
		break;
	    case 2:
		v_cvs2l(result_oprnd.vc_reg, src_oprnd.vc_reg);
		break;
	    case 4:
		v_cvi2l(result_oprnd.vc_reg, src_oprnd.vc_reg);
		break;
	    }
	    break;
#endif
	}
	break;
    case enumeration_type:
    case unsigned_type:
    case char_type:
	switch (size) {
	case 1:
	    if (!v_getreg(&result_oprnd.vc_reg, V_UC, V_TEMP))
		v_fatal("gen size convert out of registers \n");
	    REG_DEBUG(("get %d in size convert\n", _vrr(result_oprnd.vc_reg)));
	    switch (src_oprnd.size) {
	    case 2:
		v_cvs2l(result_oprnd.vc_reg, src_oprnd.vc_reg);
		v_cvl2c(result_oprnd.vc_reg, src_oprnd.vc_reg);
		break;
	    case 4:
		v_cvi2l(result_oprnd.vc_reg, src_oprnd.vc_reg);
		v_cvl2c(result_oprnd.vc_reg, src_oprnd.vc_reg);
		break;
	    case 8:
		v_cvl2c(result_oprnd.vc_reg, src_oprnd.vc_reg);
		break;
	    }
	    break;
	case 2:
	    if (!v_getreg(&result_oprnd.vc_reg, V_US, V_TEMP))
		v_fatal("gen size convert out of registers \n");
	    REG_DEBUG(("get %d in size convert\n", _vrr(result_oprnd.vc_reg)));
	    switch (src_oprnd.size) {
	    case 1:
		if (!v_getreg(&at, V_L, V_TEMP))
		    v_fatal("gen type convert2 out of registers \n");
		v_cvc2l(at, src_oprnd.vc_reg);
		v_cvl2s(result_oprnd.vc_reg, at);
		v_putreg(at, V_L);
		break;
	    case 4:
		v_cvi2s(result_oprnd.vc_reg, src_oprnd.vc_reg);
		break;
	    case 8:
		v_cvl2s(result_oprnd.vc_reg, src_oprnd.vc_reg);
		break;
	    }
	    break;
	case 4:
	    if (!v_getreg(&result_oprnd.vc_reg, V_U, V_TEMP))
		v_fatal("gen size convert out of registers \n");
	    REG_DEBUG(("get %d in size convert\n", _vrr(result_oprnd.vc_reg)));
	    switch (src_oprnd.size) {
	    case 1:
		v_cvc2u(result_oprnd.vc_reg, src_oprnd.vc_reg);
		break;
	    case 2:
		v_cvs2u(result_oprnd.vc_reg, src_oprnd.vc_reg);
		break;
	    case 8:
		v_cvl2u(result_oprnd.vc_reg, src_oprnd.vc_reg);
		break;
	    }
	    break;
	case 8:
#if SIZEOF_LONG == 8
	    if (!v_getreg(&result_oprnd.vc_reg, V_UL, V_TEMP))
		v_fatal("gen size convert out of registers \n");
	    REG_DEBUG(("get %d in size convert\n", _vrr(result_oprnd.vc_reg)));
	    switch (src_oprnd.size) {
	    case 1:
		v_cvc2ul(result_oprnd.vc_reg, src_oprnd.vc_reg);
		break;
	    case 2:
		v_cvs2ul(result_oprnd.vc_reg, src_oprnd.vc_reg);
		break;
	    case 4:
		v_cvi2ul(result_oprnd.vc_reg, src_oprnd.vc_reg);
		break;
	    }
	    break;
#else
	    if (!v_getreg(&result_oprnd.vc_reg, V_UL, V_TEMP) ||
		!v_getreg(&result_oprnd.vc_reg2, V_UL, V_TEMP))
		v_fatal("gen size convert out of registers \n");
	    REG_DEBUG(("get %d in size convert\n", _vrr(result_oprnd.vc_reg)));
	    REG_DEBUG(("get %d in size convert\n", _vrr(result_oprnd.vc_reg2)));
	    v_setl(result_oprnd.vc_reg2, 0);
	    switch (src_oprnd.size) {
	    case 1:
		v_cvc2ul(result_oprnd.vc_reg, src_oprnd.vc_reg);
		break;
	    case 2:
		v_cvs2ul(result_oprnd.vc_reg, src_oprnd.vc_reg);
		break;
	    case 4:
		v_cvi2ul(result_oprnd.vc_reg, src_oprnd.vc_reg);
		break;
	    }
	    break;
#endif
	}
	break;
    case float_type:
	switch (size) {
	case SIZEOF_FLOAT:
	    if (!v_getreg(&result_oprnd.vc_reg, V_F, V_TEMP))
		v_fatal("gen size convert out of registers \n");
	    REG_DEBUG(("get %d in size convert\n", _vrr(result_oprnd.vc_reg)));
	    switch (src_oprnd.size) {
	    case SIZEOF_FLOAT:
		assert(FALSE);
		break;
	    case SIZEOF_DOUBLE:
		v_cvd2f(result_oprnd.vc_reg, src_oprnd.vc_reg);
		break;
	    default:
		assert(FALSE);
	    }
	    break;
	case SIZEOF_DOUBLE:
	    if (!v_getreg(&result_oprnd.vc_reg, V_D, V_TEMP))
		v_fatal("gen size convert out of registers \n");
	    REG_DEBUG(("get %d in size convert\n", _vrr(result_oprnd.vc_reg)));
	    REG_DEBUG(("get %d in size convert\n", _vrr(result_oprnd.vc_reg)));
	    switch (src_oprnd.size) {
	    case SIZEOF_DOUBLE:
		assert(FALSE);
		break;
	    case SIZEOF_FLOAT:
		v_cvf2d(result_oprnd.vc_reg, src_oprnd.vc_reg);
		break;
	    default:
		assert(FALSE);
	    }
	    break;
	default:
	    /* really should fail */
	    if (!v_getreg(&result_oprnd.vc_reg, V_D, V_TEMP))
		v_fatal("gen size convert out of registers \n");
	    REG_DEBUG(("get %d in size convert\n", _vrr(result_oprnd.vc_reg)));
	    REG_DEBUG(("get %d in size convert\n", _vrr(result_oprnd.vc_reg)));
	    switch (src_oprnd.size) {
	    case SIZEOF_DOUBLE:
/*		assert(FALSE);*/
		break;
	    case SIZEOF_FLOAT:
		v_cvf2d(result_oprnd.vc_reg, src_oprnd.vc_reg);
		break;
	    default:
		assert(FALSE);
	    }
	    break;
	    
	}
	break;
    case unknown_type:
    default:
	assert(FALSE);
	break;
    }
    return result_oprnd;
}
