#ifdef MODULE
#ifndef __KERNEL__
#define __KERNEL__
#endif
#endif

#include <linux/types.h>
#include <linux/fcntl.h>
#include <linux/string.h>
#include <linux/errno.h>
#include <linux/unistd.h>
#include <linux/ctype.h>
#include <linux/fs.h>
#include <asm/uaccess.h>
#include "kpbio.h"
#include "io.h"
#include "config.h"
#include "library.h"

#define STRMAX 256
#define INADDRSZ 4

typedef struct __node {
  struct __node *next;
  char hostname[STRMAX];
  char ipaddr[STRMAX];
} node;

struct __node *dns_list_head = NULL;

char buffer2[STRMAX];
char wbuffer[STRMAX];
char str0[STRMAX];
char str1[STRMAX];
char str2[STRMAX];
char str3[STRMAX];

struct hostent *lookup_name(const char *name)
{
  char *h_addr_ptrs[8];
  char *h_addr_ptrs2[8];
  int a,b;
  int len;
  int count;
  int i1, i2, i3, i4;
  unsigned int itotal;
  unsigned int itotal2;
  int retval;
  char *membuf;
  node *new_node;
  char *format_server_host;
  node *p;
  struct hostent *hp;

  hp = (struct hostent *)DAllocMM(sizeof(struct hostent));
  membuf = (char *)DAllocMM(8);

  if (dns_list_head == NULL) {
    new_node = (struct __node *)DAllocMM(sizeof(struct __node));
    new_node->next = NULL;
    format_server_host = getenv("FORMAT_SERVER_HOST");
    if (format_server_host == NULL) 
      format_server_host = FORMAT_SERVER_HOST;
    strncpy(new_node->hostname, format_server_host, 
	strlen(format_server_host)+1);
    strncpy(new_node->ipaddr, main_ipaddr, 
	strlen(main_ipaddr)+1);
    dns_list_head = new_node;
  }

  p = dns_list_head;
  while ((p != NULL) && (strcmp(p->hostname, name) != 0))
    p = p->next;
  if (p != NULL) {
    a = 0;
    count = 0;
    strncpy(buffer2, p->ipaddr, strlen(p->ipaddr));
    buffer2[strlen(p->ipaddr)] = '\0';
    len = strlen(buffer2);
    for (b = 0; b <= len; b++) {
      if (isdigit(buffer2[b]) == 0) {
        if (count == 0) {
	  strncpy(str0, &buffer2[a], b-a);
	  str0[b-a] = '\0';
	}
	else if (count == 1) {
	  strncpy(str1, &buffer2[a], b-a);
	  str1[b-a] = '\0';
	}
	else if (count == 2) {
	  strncpy(str2, &buffer2[a], b-a);
	  str2[b-a] = '\0';
	}
	else {
	  strncpy(str3, &buffer2[a], b-a);
	  str3[b-a] = '\0';
	}
	a = b+1;
	count++;
      }
    }
    i1 = atoi(str0);
    i2 = atoi(str1);
    i3 = atoi(str2);
    i4 = atoi(str3);
    itotal = i1+(i2<<8)+(i3<<16)+(i4<<24);
    h_addr_ptrs[0] = (char *)&itotal;
    hp->h_name = DAllocMM(STRMAX);
    strcpy(hp->h_name, name);
    hp->h_addrtype = AF_INET;
    hp->h_length = INADDRSZ;
    hp->h_addr_list = (char **) DAllocMM(INADDRSZ);
    memcpy(hp->h_addr_list, h_addr_ptrs, INADDRSZ);
  } else {
    retval = server_get_ipaddr(membuf, (void *)name); 
    if (retval != -1) {
      memcpy(&itotal2, membuf, sizeof(unsigned int));
      h_addr_ptrs2[0] = (char *)&itotal2;
      hp->h_name = DAllocMM(STRMAX);
      strcpy(hp->h_name, name);
      hp->h_addrtype = AF_INET;
      hp->h_length = INADDRSZ;
      hp->h_addr_list = (char **) DAllocMM(INADDRSZ);
      memcpy(hp->h_addr_list, h_addr_ptrs2, INADDRSZ);
      sprintf(wbuffer, "%d.%d.%d.%d%c", itotal2-((itotal2>>8)<<8),
		(itotal2>>8)-((itotal2>>16)<<8),
		(itotal2>>16)-((itotal2>>24)<<8), itotal2>>24, '\0');
      new_node = (struct __node *)DAllocMM(sizeof(struct __node));
      new_node->next = dns_list_head;
      strncpy(new_node->hostname, name, strlen(name)+1);
      strncpy(new_node->ipaddr, wbuffer, strlen(wbuffer)+1); 
      dns_list_head = new_node;
      return hp;
    } else {
      printk("Hostname not found!\n");
    }
  } 
  return hp;
}
