/*
 *   test.c
 *
 *   Copyright (C) 2000 Christian Poellabauer
 *
 */

/* Required definitions. */
#ifndef MODULE
#define MODULE
#endif
#ifndef __KERNEL__
#define __KERNEL__
#endif

/* Various required header files. */
#include <linux/module.h> 
#include <linux/kernel.h> 
#include <linux/tty.h>
#include <linux/unistd.h>
#include <linux/skbuff.h>
#include <linux/socket.h>
#include <linux/sched.h>
#include <linux/mm.h>
#include <linux/utsname.h>
#include <linux/types.h>
#include <linux/shm.h>
#include <linux/ipc.h>
#include <linux/uio.h>
#include <asm/errno.h>
#include <asm/current.h>
#include <asm/uaccess.h>
#include <asm/io.h>
#include "../include/ecalls_module.h"
#ifndef sw_h
#include "sw.h"
#endif
#include "sockwatch.h"

MODULE_AUTHOR("CERCS - GATECH");
MODULE_DESCRIPTION("Sockwatch Module (KECho)");
MODULE_LICENSE("GPL");

#define EXIT_SUCCESS 0
#define EXIT_FAILURE -1

#define my_rdtsc(x)	__asm__ __volatile__ (		\
				"rdtsc \n\t"		\
				"movl %%eax,%0 \n\t"	\
				: "=r" (x)		\
				:			\
				: "%eax","%edx");

extern int SOCKWATCH_module_loaded;
extern void *sys_call_table[];
extern struct socket *look_up_pseudofd(int fd, struct socket *fd_array[]);
int (*callback_fct)(int, int, void *);

int id;

/* Prototypes. */
int init_module(void);		       
void cleanup_module(void);
void poll_clients(unsigned long data);
void (*original_wake_up_ecalls)(void);

struct file * my_fcheck(unsigned int fd)
{
	struct file * file = NULL;
	struct files_struct *files = current->files;

	if (fd < files->max_fds)
		file = files->fd[fd];
	return file;
}

struct file * my_fget(unsigned int fd)
{
	struct file * file = my_fcheck(fd);

	if (file)
		atomic_inc(&(file)->f_count);
	return file;
}

struct socket *my_socki_lookup(struct inode *inode)
{
	return &inode->u.socket_i;
}

struct socket *my_sockfd_lookup(int fd, int *err)
{
	struct file *file;
	struct inode *inode;
	struct socket *sock;

	if (!(file = my_fget(fd)))
	{
		*err = -EBADF;
		return NULL;
	}

	inode = file->f_dentry->d_inode;
	if (!inode || !inode->i_sock || !(sock = my_socki_lookup(inode)))
	{
		*err = -ENOTSOCK;
		return NULL;
	}

	if (sock->file != file) {
		printk(KERN_ERR "socki_lookup: socket file changed!\n");
		sock->file = file;
	}
	return sock;
}

/* Upcall raising function. */
void raise_upcall(unsigned long data)
{
  do_upcall(id, data, 0, 0, 0, 0);
}

/* Internal wake_up_ecalls function. */
void internal_wake_up_ecalls(int fd, int pid, int type, int kecho, void *ptr)
{
  unsigned long data_ptr;
  struct skwatch_uc_data *uc_data;

//printk("SW:%d:%d\n", fd, pid);
  if (kecho == 1) {
    if (callback_fct == NULL)
      callback_fct = sys_call_table[__NR_KECho_sw_callback];
    callback_fct(fd, pid, ptr);
  } else { 
    data_ptr = request_upcall_flag2(id, pid);
    uc_data = (struct skwatch_uc_data *)data_ptr;
    if (type == 0) 
      uc_data->readable[fd/32] |= (1 << (fd%32));
    else 
      uc_data->writable[fd/32] |= (1 << (fd%32));
    //if (uc_data->flag <= 0) 
    uc_data->flag++;
    raise_upcall((unsigned long) pid);
  }
}

/* Register interest in sockets through kecho. */
int register_sock_interest(int fd, void *ptr)
{
  struct socket *socket;

  socket = look_up_pseudofd(fd, (struct socket **)current->qfabric->fdqfabric);

  if (socket != NULL)
    set_sock_ecalls(socket, 1, 0, 1, ptr, fd, current->pid); 
  else
    return 0;

  return 1;
}

/* Unregister interest in sockets through kecho. */
int unregister_sock_interest(int fd)
{
  struct socket *socket;

  socket = look_up_pseudofd(fd, (struct socket **)current->qfabric->fdqfabric);

  if (socket != NULL)
    set_sock_ecalls(socket, 0, 0, 1, NULL, fd, current->pid); 
  else
    return 0;

  return 1;
}

/* Fast handler function. */
int fast_handler_func(int id, unsigned long data, int len) 
{
  unsigned long data_ptr;
  struct skwatch_dc_data *dc_data;
  int i = 0;
  int res = 0;
  int flag;
  int next = 0;
  int fd;
  int error;
  struct socket *socket;

  /* Look for changes in fd interest and apply changes. */
  data_ptr = request_downcall_flag(id, current->pid);
  dc_data = (struct skwatch_dc_data *)data_ptr;
  flag = dc_data->flag;
  while((i < SKWATCH_DC_ENTRIES) && (flag > 0)) {
    res = find_next_zero_bit((void *)&dc_data->changes[i],
	sizeof(unsigned long),next);  
    if (res < 32) {
      flag--;
      next = res;
      fd = i*32+res;
      socket = my_sockfd_lookup(fd, &error);
      if ((~(dc_data->readable[i]) & (1 << res)) > 0) {
        if ((~(dc_data->writable[i]) & (1 << res)) > 0) { 
//printk("set r/w for %d:%d\n", fd, current->pid);
          set_sock_ecalls(socket, 1, 1, 0, NULL, fd, current->pid); 
        } else {
//printk("set r for %d:%d\n", fd, current->pid);
          set_sock_ecalls(socket, 1, 0, 0, NULL, fd, current->pid); 
	}
      } else {
        if ((~(dc_data->writable[i]) & (1 << res)) > 0) {
//printk("set w for %d:%d\n", fd, current->pid);
          set_sock_ecalls(socket, 0, 1, 0, NULL, fd, current->pid); 
        } else {
//printk("set x for %d:%d\n", fd, current->pid);
          set_sock_ecalls(socket, 0, 0, 0, NULL, fd, current->pid); 
	}
      }
    }
    else {
      i++;
      next = 0;
    }
  }
/*
  for (i = 0; i < SKWATCH_DC_ENTRIES; i++) 
    dc_data->changes[i] = ~((unsigned long)0);
*/
  dc_data->flag = 0;
  return 0;
}

/* Slow handler function. */
void slow_handler_func(int id, unsigned long data, int len) 
{
  fast_handler_func(id, data, len);
}

/* Initialize module. */
int init_module(void) 
{
  SOCKWATCH_module_loaded = 1;

  original_wake_up_ecalls = sys_call_table[__NR_wake_up_ecalls];
  sys_call_table[__NR_wake_up_ecalls] = internal_wake_up_ecalls;

  id = register_ecalls_service("sockwatch", fast_handler_func, slow_handler_func,
	IMMEDIATE);

  return EXIT_SUCCESS;
}

/* Remove module. */
void cleanup_module(void) 
{
  SOCKWATCH_module_loaded = 0;

  sys_call_table[__NR_wake_up_ecalls] = original_wake_up_ecalls;

  unregister_ecalls_service("sockwatch");
}

