/*
 * Copyright 1999, Silicon Graphics, Inc.
 *
 * Written October 1999 by Rajagopal Ananthanarayanan (ananth)
 *  			at Silicon Graphics, Inc.
 *
 * Made into a kernel module by CP 2003
 */

#ifndef MODULE
#define MODULE
#endif
#ifndef __KERNEL__
#define __KERNEL__
#endif

#if defined(MODVERSIONS)
#include <linux/modversions.h>
#endif
#include <linux/kernel.h>
#include <linux/module.h>
#include <linux/unistd.h>
#include <linux/types.h>
#include <linux/time.h>
#include <linux/slab.h>
#include <linux/mm.h>
#include <linux/sched.h>
#include <linux/wait.h>
#include <linux/signal.h>
#include <linux/string.h>

#define AIO_LARGEFILE64

#include <linux/aio.h>
#include "utils.h"

#define exit	sys_exit
#define fprintf(fmt, args...)	printk(args)
#define malloc(a)	kmalloc(a, GFP_KERNEL)
#define perror	printk

static int aio_cpid[AIO_MAX_THREADS];
static kthread_t aio_threads[AIO_MAX_THREADS];

int
child(void *arg)
{
	/* child */
//	init_kthread(&aio_threads[child_i], "aio_thread");
	sys_kaio(AIOCMD_SLAVE, 0, 0, 0, 0);
//	exit_kthread(&aio_threads[child_i]);
	return 0;
}


static void
aio_init(int nslaves)
{
	int i;
	char *cstack;

	if (nslaves == 0) 
		nslaves = AIO_DEFAULT_THREADS;

	if (nslaves <= 0 || nslaves > AIO_MAX_THREADS) {
		fprintf(stderr, "aio_init: Warning! nthreads %d not in "
				"valid range 1 .. %d (see AIO_MAX_THREADS)\n",
				nslaves, AIO_MAX_THREADS); 
		fprintf(stderr, "aio_init: defaulting to %d threads "
				"(see AIO_DEFAULT_THREADS)\n",
				AIO_DEFAULT_THREADS);
		nslaves = AIO_DEFAULT_THREADS;
	}

	for (i = 0; i < nslaves; i++) {
		cstack = (char *) malloc(64 * 1024);
		cstack += (32 * 1024);
		cstack = (char *)((unsigned long)cstack & 0xfffff000);

		/*
		aio_cpid[i] =
			aio(AIOCMD_SLAVE, 0, 0, 0, 0);
		*/
		//aio_cpid[i] = clone(child, cstack,
	//		SIGCHLD|CLONE_VM|CLONE_FS|CLONE_FILES, 0);
		aio_cpid[i] = kernel_thread(child, NULL, SIGCHLD|CLONE_VM|
			CLONE_FS|CLONE_FILES);
printk("go away\n");
		set_task_state(current, TASK_UNINTERRUPTIBLE);
		schedule_timeout(100);
printk("back\n");
//		aio_cpid[i] = start_kthread(child, 
//			&aio_threads[i]);
		if (aio_cpid[i] < 0) {
			perror("clone");
			exit(-1);
		}
	}
}

int
aio_read64(aiocb64_t *aiocb)
{
	int ret;

	ret = sys_kaio(AIOCMD_READ, (unsigned long)aiocb, 0, 0, 0);
	if (ret < 0 && errno == ENOTCONN) {
		aio_init(0);
		ret = sys_kaio(AIOCMD_READ, (unsigned long)aiocb, 0, 0, 0);
	}
	return ret;
}

int
aio_write64(aiocb64_t *aiocb)
{
	int ret;

printk("first call!\n");
	ret = sys_kaio(AIOCMD_WRITE, (unsigned long)aiocb, 0, 0, 0);
printk("first call done:%d\n", ret);
	if (ret == -ENOTCONN) {
printk("init call!\n");
		aio_init(0);
printk("init call done!\n");
		ret = sys_kaio(AIOCMD_WRITE, (unsigned long)aiocb, 0, 0, 0);
printk("second call done:%d\n", ret);
	}
	return ret;
}


int
aio_cancel64(int fildes, aiocb64_t *aiocb)
{
	int ret;

	ret = sys_kaio(AIOCMD_CANCEL, (unsigned long)fildes,
			(unsigned long)aiocb, 0, 0);
	if (ret < 0 && errno == ENOTCONN) {
		aio_init(0);
		ret = sys_kaio(AIOCMD_CANCEL, (unsigned long)fildes,
				(unsigned long)aiocb, 0, 0);
	}
	return ret;
}

int
aio_suspend64(const aiocb64_t *const list[], int nent, const struct timespec *timeout)
{
	int ret;

	ret = sys_kaio(AIOCMD_SUSPEND, (unsigned long)list, (unsigned long)nent,
			(unsigned long)timeout, 0);
	if (ret < 0 && errno == ENOTCONN) {
		aio_init(0);
		ret = sys_kaio(AIOCMD_SUSPEND, (unsigned long)list,
				(unsigned long)nent,
				(unsigned long)timeout, 0);
	}
	return ret;
}

ssize_t
kaio_return64(aiocb64_t *aiocb)
{
	return(aiocb->aio_reserved[3]);
}

int
lio_listio64(int mode, aiocb64_t *const list[], int nent, sigevent_t *sig)
{
	int ret;

	ret = sys_kaio(AIOCMD_LIST_IO,
			(unsigned long)mode, (unsigned long)list,
			(unsigned long)nent, (unsigned long)sig);
	if (ret < 0 && errno == ENOTCONN) {
		aio_init(0);
		ret = sys_kaio(AIOCMD_LIST_IO,
				(unsigned long)mode, (unsigned long)list,
				(unsigned long)nent, (unsigned long)sig);
	}
	return ret;
}

//#ifndef __USE_FILE_OFFSET64
//#define PAD_AIOCB(aiocb)	*(unsigned int *)&((aiocb)->__aio_pad) = 0;
//#else
#define PAD_AIOCB(aiocb)
//#endif

int
aio_read(aiocb_t *aiocb)
{
	PAD_AIOCB(aiocb);
	return(aio_read64((aiocb64_t *)aiocb));
}

int
aio_write(aiocb_t *aiocb)
{
	PAD_AIOCB(aiocb);
	return(aio_write64((aiocb64_t *)aiocb));
}


int
aio_cancel(int fildes, aiocb_t *aiocb)
{
	if (aiocb)
		PAD_AIOCB(aiocb);
	return(aio_cancel64(fildes, (aiocb64_t *)aiocb));
}


typedef const aiocb64_t * const *CLISTC;
typedef       aiocb64_t * const	*LISTC;

int
aio_suspend(const aiocb_t *const list[], int nent, const struct timespec *timeout)
{
#ifndef __USE_FILE_OFFSET64
	int i;

	for (i = 0; i < nent; i++) 
		if (list[i])
			PAD_AIOCB(list[i]);
#endif
	return(aio_suspend64((CLISTC)list, nent, timeout));
}

ssize_t
kaio_return(aiocb_t *aiocb)
{
	return kaio_return64((aiocb64_t *)aiocb);
}

int
lio_listio(int mode, aiocb_t *const list[], int nent, sigevent_t *sig)
{
#ifndef __USE_FILE_OFFSET64
	int i;

	for (i = 0; i < nent; i++) 
		if (list[i])
			PAD_AIOCB(list[i]);
#endif
	return(lio_listio64(mode, (LISTC)list, nent, sig));
}

void
aio_pcinvalidate(int fd)
{
	if (sys_kaio(AIOCMD_PCINVALIDATE, fd, 0, 0, 0)) {
		if (errno == ENOTCONN) {
			aio_init(0);
			if (sys_kaio(AIOCMD_PCINVALIDATE, fd, 0, 0, 0) == 0)
				return;
		}
		perror("AIOCMD_PCINVALIDATE");
		exit(-1);
	}
}
