#ifndef _UTILS_H
#define _UTILS_H
#include <linux/config.h>
#include <linux/version.h>

#include <linux/kernel.h>
#include <linux/sched.h>
#include <linux/tqueue.h>
#include <linux/wait.h>
#include <asm/unistd.h>
#include <asm/semaphore.h>

typedef struct kthread {
	char *name;
	struct task_struct *task;
	struct completion done;
#define KTH_RUNNING	1
#define KTH_SHUTDOWN	2
	long state;
	int (*init)(struct kthread *, void *);
	void (*cleanup)(struct kthread *, void *);
	int (*main)(struct kthread *, void *);
	void *data;
	int used;
} kthread_t;

#define KTHREAD_FLAGS (CLONE_FS|CLONE_FILES|CLONE_SIGHAND)

typedef struct kthread_args {
	struct kthread *kth;
	void *data;
} kthread_args_t;

/* Prototypes */

int start_kthread(void *func, kthread_t *kth);
int start_kthread_args(void *func, kthread_t *kth, void *data);

void stop_kthread(kthread_t *kthread);

void init_kthread(kthread_t *kthread, char *name);

void exit_kthread(kthread_t *kthread);

int getthishostname(char *name, int len);

int(*sys_kaio)(int cmd, unsigned long arg1, unsigned long arg2, unsigned long arg3,
	unsigned long arg4);
int (*sys_exit)(int error_code);

extern struct semaphore utils_sem;

#ifdef CONFIG_ARCH_SA1100
int freq_levels[12] = {59000, 73700, 88500, 103200, 118000, 132700, 147500,
	162200, 176900, 191700, 206400, 221200};

int get_freq_level(void);

int set_freq_level(int level);

int inc_freq_level(void);

int dec_freq_level(void);

int rst_freq_level(void);
#endif

#endif
