
/*
 *
 *   Copyright (C) 2002 Christian Poellabauer
 *
 */

/* Required definitions. */
#ifndef MODULE
#define MODULE
#endif
#ifndef __KERNEL__
#define __KERNEL__
#endif

/* Various required header files. */
#include <linux/module.h> 
#include <linux/kernel.h> 
#include <linux/tty.h>
#include <linux/unistd.h>
#include <linux/sysrq.h>
#include <linux/socket.h>
#include <linux/kmod.h>
#include <linux/sched.h>
#include <linux/timer.h>
#include <linux/wait.h>
#include <linux/init.h>
#include <linux/types.h>
#include <linux/stat.h>
#include <linux/utsname.h>
#include <linux/mm.h>
#include <linux/fcntl.h>
#include <linux/slab.h>
#include <linux/vmalloc.h>
#include <asm/current.h>
#include <asm/uaccess.h>
#include <linux/kernel_stat.h>
#include <linux/skbuff.h>
#include <net/sock.h>
#include <net/pkt_sched.h>
#include <linux/proc_fs.h>
#include <linux/unistd.h>
#include <linux/sch_dwcs.h>
#include "../kecho/include/io.h"
#include "../kecho/include/atl.h"
#include "../kecho/include/cm.h"
#include "../kecho/include/echo.h"
#include "../kecho/include/comm_group2.h"
#include "../kecho/include/utils.h"
#include "../kecho/include/ecalls_module.h"
#include "../dproc/monitor.h"
#include "../adaptors/adaptor.h"
#include "qos_cm.h"

MODULE_AUTHOR("Christian Poellabauer");
MODULE_DESCRIPTION("Qfabric Module");
MODULE_LICENSE("GPL");

#ifndef FALSE
#define FALSE 0
#endif
#ifndef TRUE
#define TRUE 1
#endif

#define PORT_START 10000
#define PORT_RANGE 100

#define MAX_HOSTS 100

extern void *sys_call_table[];
int (*original_qos_conf)();

struct down_info {
	int type;
	int pid;
};

/* Prototypes. */
int init_module(void);		       
void cleanup_module(void);
void q_mon_callback(char *, int, int, char *); 
void q_con_callback(char *, int, int, char *); 
void q_com_callback(char *, int, int, char *); 

/* Global variables. */
static int loop_flag = 1;
DECLARE_WAIT_QUEUE_HEAD(sleep_queue);
DECLARE_WAIT_QUEUE_HEAD(my_queue);
static qos_apps head_qos_ptr = NULL;
static int qos_apps_registered = 0;
char hostname[MAXHOSTNAMELEN];
static int port_num = PORT_START;
kthread_t qfabric_thread;
int active_qchannels = 0;
char *this_node_name;
struct _hosts qos_hosts[MAX_HOSTS];
list_of_res reslist;
comm_group_return groups;
char *channel_id;
char *group_id;
int handler_called = 0;
int send_type2_event = 0;
struct qfabric_event2 *type2_event = NULL;
struct qfabric_event qf_event;
int qf_id;

IOField struct_fields[] = {
	{"used", "integer", sizeof(int), IOOffset(host_entry *, used)},
	{"hostname", "string", MAXHOSTNAMELEN, IOOffset(host_entry *, hostname)},
	{"runqueuelen", "integer", sizeof(int), IOOffset(host_entry *, runqueuelen)},
	{"loadavg", "integer", sizeof(int), IOOffset(host_entry *, loadavg)},
	{"freemem", "integer", sizeof(int), IOOffset(host_entry *, freemem)},
	{"netmon", "integer", sizeof(int), IOOffset(host_entry *, netmon)},
	{NULL, NULL}
};

IOField res_fields[] = {
	{"input_hosts", "host_entry[MAX_HOSTS]", sizeof(host_entry), 
		IOOffset(list_of_res *, input_hosts)},
	{NULL, NULL}
};

char *strdup(char *s)
{
  char *p;

  p = (char *)kmalloc(strlen(s)+1, GFP_KERNEL);
  if (p != NULL)
    strcpy(p,s);
  return p;
}

int my_gethostname(char *name, int len)
{
	int i, errno;
	if (len < 0)
		return -EINVAL;
	down_read(&uts_sem);
	i = 1 + strlen(system_utsname.nodename);
	if (i > len)
		i = len;
	errno = 0;
	memset(name, '\0', MAXHOSTNAMELEN-1);
	memcpy(name, system_utsname.nodename, i);
	if (memchr(name, '.', MAXHOSTNAMELEN) == NULL) {
		strcat (name, ".");
		strcat (name, system_utsname.domainname);
	}
	up_read(&uts_sem);
	return errno;
}

void handler1(void *event, int length, void *client_data, attr_list event_attrs) 
{
	struct qfabric_event qos_event;
	struct qfabric_event2 qos_event2;
	int i = 0;
	int len = length;
	char *kcode;
	int codedist = 0;

	if (length == sizeof(struct qfabric_event2)) {
		len = sizeof(struct qfabric_event2);
		memcpy((void *)&qos_event2, event, len);
		if (strcmp(qos_event2.hostname, this_node_name) == 0)
			return;
		codedist = 1;
	} else {
		memcpy((void *)&qos_event, event, length);
		if (strcmp(qos_event.hostname, this_node_name) == 0)
			return;
	}
	handler_called = 1;
	if (codedist == 1) {
	    while ((i < MAX_HOSTS) && (qos_hosts[i].used != 0) &&
		(strcmp(qos_hosts[i].hostname, qos_event2.hostname)!=0)) 
		i++;
	    if (i < MAX_HOSTS) {
	    	qos_hosts[i].runqueuelen = qos_event2.runqueuelen;
	    	qos_hosts[i].loadavg = qos_event2.loadavg;
	    	qos_hosts[i].freemem = qos_event2.freemem;
	    	qos_hosts[i].netmon = qos_event2.netmon;
	    	qos_hosts[i].uinfo1 = qos_event2.uinfo1;
	    	qos_hosts[i].uinfo2 = qos_event2.uinfo2;
	    	qos_hosts[i].uinfo3 = qos_event2.uinfo3;
	    }
	    kcode = kmalloc(length - sizeof(struct qfabric_event2), GFP_KERNEL);
	    strcpy(kcode, (char *)&qos_event2+sizeof(struct qfabric_event2));
	    
	    if (strcmp(qos_event2.hook, "q_mon") == 0) {
		q_mon_callback(qos_event2.par1, qos_event2.par2, qos_event2.par3, kcode); 
	    } else if (strcmp(qos_event2.hook, "q_con") == 0) {
		q_con_callback(qos_event2.par1, qos_event2.par2, qos_event2.par3, kcode); 
	    } else if (strcmp(qos_event2.hook, "q_com") == 0) {
		q_com_callback(qos_event2.par1, qos_event2.par2, qos_event2.par3, kcode); 
	    }
	    kfree(kcode);
	    return;
	}
	while ((i < MAX_HOSTS) && (qos_hosts[i].used != 0) &&
		(strcmp(qos_hosts[i].hostname, qos_event.hostname)!=0)) 
		i++;
	if (i == MAX_HOSTS) return;
	if (qos_hosts[i].used == 0) {
		strcpy((char *)qos_hosts[i].hostname, (char *)qos_event.hostname);
		qos_hosts[i].uinfo1 = -1;
		qos_hosts[i].uinfo2 = -1;
		qos_hosts[i].uinfo3 = -1;
		qos_hosts[i].used = 1;
	}
	if (qos_event.type == 0) { /* MON */
		qos_hosts[i].runqueuelen = qos_event.runqueuelen;
		qos_hosts[i].loadavg = qos_event.loadavg;
		qos_hosts[i].freemem = qos_event.freemem;
		qos_hosts[i].netmon = qos_event.netmon;
	    	qos_hosts[i].uinfo1 = qos_event.uinfo1;
	    	qos_hosts[i].uinfo2 = qos_event.uinfo2;
	    	qos_hosts[i].uinfo3 = qos_event.uinfo3;
	} else if (qos_event.type == 1) { /* CON */
/* TODO */
	}
}

extern int qos_register(char *app_id, char *group_id) 
{
	qos_apps pointer = head_qos_ptr;
	int unregister = 0;

	while (pointer != NULL) 
	{ 
		if (((strcmp(pointer->app_id, app_id) == 0) && 
			(strcmp(pointer->group_id, group_id) == 0) &&
			(pointer->pid = current->pid))) /* unregister */
		{
			if(pointer == head_qos_ptr)
			{
				head_qos_ptr = pointer->next;
				head_qos_ptr->prev = NULL;
			}
			else
			{
				pointer->next->prev = pointer->prev;
				pointer->prev->next = pointer->next;
			}
			kfree(pointer);
			active_qchannels--;
			unregister = 1;
			break;
		}
		pointer = pointer->next;
	}
	if (unregister == 1) return 0;

	pointer = kmalloc(sizeof(struct _qos_apps), GFP_KERNEL);
	memset(pointer, 0, sizeof(struct _qos_apps));
	
	pointer->proc = current;
	pointer->active = 1;
	pointer->pid = current->pid;
	if (app_id != NULL)
		pointer->app_id = strdup(app_id);
	else pointer->app_id = strdup("DEFAULT");
	if (group_id != NULL)
		pointer->group_id = strdup(group_id);
	else pointer->group_id = strdup("DEFAULT");
	pointer->app_lock = RW_LOCK_UNLOCKED;
	pointer->server_time = 360;
	
	pointer->next = head_qos_ptr;
	pointer->prev = NULL;
	if(head_qos_ptr != NULL)
		head_qos_ptr->prev = pointer;
	else
		pointer->prev = NULL;
	head_qos_ptr = pointer;
	qos_apps_registered++;

	active_qchannels++;
	return 0;
}

int distribute_code(char *hook, char *par1, int par2, int par3, char *code) 
{
  if (type2_event != NULL) kfree(type2_event);
  type2_event = (struct qfabric_event2 *) kmalloc(sizeof(struct qfabric_event2)+
		strlen(code), GFP_KERNEL);
  type2_event->type = 2;
  strcpy(type2_event->hook, hook);
  strcpy(type2_event->par1, par1);
  type2_event->par2 = par2;
  type2_event->par3 = 0;
  type2_event->len = strlen(code);
  strcpy((char *)(type2_event)+sizeof(struct qfabric_event2), code);

  send_type2_event = 1; 

  return 0;
}

/* q-mon-callback: */
void q_mon_callback(char *parameter1, int parameter2, int parameter3, char *code) 
{
  static qos_apps pointer;

  if (parameter3 == 1) { 
    distribute_code("q_mon", parameter1, parameter2, parameter3, code);
    return;
  }

  pointer = head_qos_ptr;
  while (pointer != NULL) {
    if (strcmp(pointer->app_id, parameter1) == 0)
      break;
    pointer = pointer->next;
  }
  if (pointer != NULL) {
    pointer->mcontext = new_ecl_parse_context();
    ecl_add_param("last_time", "unsigned long", 0, pointer->mcontext);
    pointer->mgen_code = ecl_code_gen(code, pointer->mcontext);
    pointer->filter = pointer->mgen_code->func;
  }
}

/* q-con-callback: */
void q_con_callback(char *parameter1, int parameter2, int parameter3, char *code) 
{
  static qos_apps pointer;

  if (parameter3 == 1) { 
    distribute_code("q_con", parameter1, parameter2, parameter3, code);
    return;
  }

  pointer = head_qos_ptr;
  while (pointer != NULL) {
    if (strcmp(pointer->app_id, parameter1) == 0)
      break;
    pointer = pointer->next;
  }
  if (pointer != NULL) {
    pointer->context = new_ecl_parse_context();
    ecl_add_param("id", "int", 0, pointer->context);
    ecl_add_param("cmd", "int", 0, pointer->context);
    ecl_add_struct_type("host_entry", struct_fields, pointer->context);
    ecl_add_struct_type("list_of_res", res_fields, pointer->context);
    ecl_add_param("resources", "list_of_res", 0, pointer->context);
    pointer->gen_code = ecl_code_gen(code, pointer->context);
    pointer->controller = pointer->gen_code->func;
    pointer->dynamic = 1;
  }
}

/* q-com-callback: */
void q_com_callback(char *parameter1, int parameter2, int parameter3, char *code)
{
  static qos_apps pointer;

  if (parameter3 == 1) { 
    distribute_code("q_com", parameter1, parameter2, parameter3, code);
    return;
  }

  pointer = head_qos_ptr;
  while (pointer != NULL) {
    if (strcmp(pointer->app_id, parameter1) == 0)
      break;
    pointer = pointer->next;
  }
  if (pointer != NULL) {
    ECinsert_filter(pointer->chan, code, parameter2);
  }
}

/* q-mon: resource monitor. */
int q_mon() {
	static qos_apps pointer;
	struct _local_resources *l_res;
	int cpu_load;
	int table_changed = 0;
	long (*func)(unsigned long);

	strcpy((char *)qf_event.hostname, this_node_name);
	qf_event.hostname[strlen(this_node_name)] = '\0';
 	qf_event.type = 0;

	/* Poll local resources */
	l_res = poll_resources();

	cpu_load = nr_running;	
	if (qos_hosts[0].runqueuelen != cpu_load) {
		qos_hosts[0].runqueuelen = cpu_load;
		table_changed = 1;
	}
	if (qos_hosts[0].loadavg != l_res->loadavg) {
		qos_hosts[0].loadavg = l_res->loadavg;
		table_changed = 1;
	}
	if (qos_hosts[0].freemem != l_res->freemem) {
		qos_hosts[0].freemem = l_res->freemem;
		table_changed = 1;
	}
	if (qos_hosts[0].netmon != l_res->netmon) {
		qos_hosts[0].netmon = l_res->netmon;
		table_changed = 1;
	}

	/* Update remote tables. */
	qf_event.runqueuelen = cpu_load;
	qf_event.loadavg = l_res->loadavg;
	qf_event.freemem = l_res->freemem;
	qf_event.netmon = l_res->netmon;

	pointer = head_qos_ptr;
	while (pointer != NULL) {
		if (pointer->active == 1) {
			/* New member found, join/create qchannel. */
			pointer->cm = CManager_create();
			pointer->cc = ECho_CM_init(pointer->cm);
			groups = matching_comm_groups(pointer->app_id, 
					pointer->app_id);
			if (groups->count < 1) {
				pointer->chan = EChannel_create(pointer->cc);
				channel_id = ECglobal_id(pointer->chan);
				group_id = setup_comm_group(NULL, 
					pointer->app_id,
					pointer->app_id, pointer->server_time,
					this_node_name, port_num);
				port_num++;
				if (port_num == (PORT_START+PORT_RANGE)) 
					port_num = PORT_START;
				assoc_group_info(group_id, channel_id, 
						strlen(channel_id)+1);
			} else {
				pointer->chan = EChannel_open(pointer->cc, 
						(char*)groups->list[0].info);
			}
 			pointer->handle = ECsource_subscribe(pointer->chan);
 			pointer->sink_handle = ECsink_subscribe(pointer->chan, 
					handler1, (void*)pointer);
			pointer->active = 2;
		} else if (pointer->active == 2) {
			int block = 0;
			/* Submit monitoring event. */
			if ((send_type2_event == 1) && (strcmp(type2_event->par1,
				pointer->app_id) == 0)) {
			  send_type2_event = 0;
			  strcpy(type2_event->hostname, qf_event.hostname);
			  type2_event->type = 2;
			  type2_event->runqueuelen = qf_event.runqueuelen;
			  type2_event->loadavg = qf_event.loadavg;
			  type2_event->freemem = qf_event.freemem;
			  type2_event->netmon = qf_event.netmon;
			  ECsubmit_event(pointer->handle, type2_event, 
				sizeof(struct qfabric_event2)+type2_event->len);
			  if (qf_event.type == 1) {
				ECsubmit_event(pointer->handle, &qf_event, sizeof(qf_event));
			  }
			} else {
			  if (pointer->filter != NULL) {
			 	func = (long(*)(unsigned long)) (long) pointer->filter;
				block = func(jiffies);
			  }
			  if ((block == 0) && (qf_event.type != 1)) 
				ECsubmit_event(pointer->handle, &qf_event, sizeof(struct qfabric_event));
			}
		}
		pointer = pointer->next;		
	}
	return table_changed;
}

/* q-con: resource controller. */
void q_con()
{
  static qos_apps pointer;
  long (*func)(int, int, void *);
  int retval = -1;
  int adapt = 0;
  int par1, par2, par3, par4;

  pointer = head_qos_ptr;
  while (pointer != NULL) {
    if (pointer->controller != NULL) {
      func = (long(*)(int, int, void *)) (long) pointer->controller; 
      if (pointer->dynamic == 1) {
        while (retval != -2) {
          retval = get_next_adaptor(retval);
	  if (retval == -2) break;
	  par1 = par2 = par3 = par4 = 0;
	  par1 = pointer->pid;
          adapt = func(retval, 1, &reslist);  /* 1=is resource to be adapted? */
	  if (adapt == 1) {
	    par2 = func(retval, 2, &reslist); /* 2=get par2 for adaptor call, par1=pid */
	    par3 = func(retval, 3, &reslist); /* 3=get par3 */
	    par4 = func(retval, 4, &reslist); /* 4=get par4 */
	    call_adaptor_by_id(retval, par1, par2, par3, par4);
	  }
        }
      } else {
        retval = func(0, sizeof(qos_hosts), reslist.input_hosts); 
      }
    }
    pointer = pointer->next;
  }
/* TODO: con call send control events */
}

/* TODO: make adjustable */
/* determine time q-mon needs to sleep. */
int determine_sleep_time() 
{
	return HZ;
}

/* QoS manager thread (runs q-mon and q-con). */
static int qos_mgr()
{
	int tables_changed = 0;
	static qos_apps pointer;
	int i;
	int sleep_time;

	/* Initialization. */
	init_kthread(&qfabric_thread, "qfabric");

	this_node_name = (char *) kmalloc(MAXHOSTNAMELEN, GFP_KERNEL);
	my_gethostname(this_node_name, MAXHOSTNAMELEN);

	for (i = 0; i < MAX_HOSTS; i++) {
		qos_hosts[i].used = 0;
		qos_hosts[i].uinfo1 = -1;
		qos_hosts[i].uinfo2 = -1;
		qos_hosts[i].uinfo3 = -1;
	}
	qos_hosts[0].used = 1;
	strcpy((char *)qos_hosts[0].hostname, this_node_name);

	/* Loop. */
	sleep_time = 0;
	while (loop_flag == 1) {
		if (active_qchannels > 0) {
			/* Call q-mon. */
			if (sleep_time == 0)
				tables_changed = q_mon();	
			/* Check for incoming mon-events. */
			pointer = head_qos_ptr;
			while (pointer != NULL) {
				handler_called = 0;
				if (pointer->active == 2) {
					CMpoll_network(pointer->cm);
				}
				pointer = pointer->next;
			}
			/* Call q-con. */	
			if ((tables_changed != 0) || (handler_called != 0)) {
				q_con();
				tables_changed = 0;
				handler_called = 0;
			}
			if (sleep_time <= 0) 
				sleep_time = determine_sleep_time();
			sleep_time = CMtask_sleep(sleep_time);
		} else {
			sleep_time = HZ;
			CMtask_sleep(sleep_time);
		}
	}
	exit_kthread(&qfabric_thread);
	return 0;
}

/* Handler func for user-level calls. */
void slow_handler_func(int id, unsigned long data, int len)
{
  long mydata;
  int mypid;
  struct down_info down;

  if (len == sizeof(long)) {
    memcpy(&mydata, (void *)data, sizeof(long));
    qos_hosts[0].uinfo1 = (int)mydata;
    qf_event.uinfo1 = (int)mydata;
  } else if (len == sizeof(struct down_info)) {
    memcpy(&down, (void *)data, sizeof(struct down_info));
    mypid = down.pid;
    qos_hosts[0].uinfo2 = (int)mypid;
    qf_event.uinfo2 = (int)mypid;
  }
}

int con_unregister(char *app_id, char *group_id) 
{
  static qos_apps pointer;

  pointer = head_qos_ptr;
  while (pointer != NULL) {
    if ((strcmp(pointer->app_id,app_id)==0) && (strcmp(pointer->group_id,group_id)==0)) 
      break;
    pointer = pointer->next;
  }
  if (pointer != NULL) {
    pointer->controller = NULL;
  }
  return 0;
}

int con_register(char *app_id, char *group_id, void *controller) 
{
  static qos_apps pointer;

  pointer = head_qos_ptr;
  while (pointer != NULL) {
    if ((strcmp(pointer->app_id,app_id)==0) && (strcmp(pointer->group_id,group_id)==0)) 
      break;
    pointer = pointer->next;
  }
  if (pointer != NULL) {
    pointer->controller = controller;
  }
  return 0;
}

/* Initialize module. */
int init_module(void) 
{
  original_qos_conf = sys_call_table[__NR_qos_configure];
  sys_call_table[__NR_qos_configure] = qos_register;
  if (start_kthread(qos_mgr, &qfabric_thread) < 0) {
	  printk("Not able to start Qfabric!\n");
	  return -1;
  }

  /* Register hooks */
  register_hook("q_con", "Q-Fabric Controller Function", "APP_ID",
	(void *) q_con_callback);
  register_hook("q_mon", "Q-Fabric Monitoring Function", "APP_ID", 
	(void *) q_mon_callback);
  register_hook("q_com", "Q-Fabric Event Channel", "APP_ID, type(0-in/1-out)",
	(void *) q_com_callback);

  qf_id = register_ecalls_service("QoS", NULL, slow_handler_func, IMMEDIATE);

  reslist.input_hosts = &qos_hosts[0];

  return 0;
}

/* Remove module. */
void cleanup_module(void) 
{
  static qos_apps pointer;
  /* Unregister hooks */
  unregister_hook("q_con");
  unregister_hook("q_mon");
  unregister_hook("q_com");

  unregister_ecalls_service("QoS");
 
  pointer = head_qos_ptr;
  while (pointer != NULL) {
    if (pointer->controller != NULL) {
      ecl_code_free(pointer->gen_code);
      ecl_free_parse_context(pointer->context);
    }
    pointer = pointer->next;
  }

  loop_flag = 0;
  stop_kthread(&qfabric_thread);
  sys_call_table[__NR_qos_configure] = original_qos_conf;
}
